// brm.groovy
// Backup, Restore, and Move content tool
// Copyright (c) 2009-2011 by SAS Institute Inc., Cary, NC 27513

import com.sas.solutions.brm.BRMController; 
import com.sas.solutions.brm.migrate.MainMigrator
import com.sas.solutions.brm.util.BRMUtil;

if (args.length == 0)
{
    println("BRM ${readVersion()}");
    println("Backup, Restore and Move content tool for PM Solutions");
    println("");
    println("Usage:");
    println("brm command-name [parameter ...]");
    println("");
    println("Try \"brm help\" for the list of commands.");
    System.exit(1);
}

def command = args[0]
String[] commandArgs = new String[args.length - 1];
System.arraycopy(args, 1, commandArgs, 0, commandArgs.length);

if (command.equals("groovy"))
{
    if (commandArgs.length == 0)
    {
        println("Missing groovy script name");
        System.exit(1);
    }
    else
    {
        // Run Groovy script with BRM classpath
        runGroovyScript(commandArgs);
        System.exit(0);
    }
}

// We do configuration check only if not running a Groovy script so that a
// groovy Script can be written to do configuration for us.
if (!configCheck(command))
{
    println("Configuration check failed.");
    println("Please see previous messages for potential remedies.");
    System.exit(1);
}

// Return a list of lists describing commands implemented by scripts:
// [ [command-name, usage-message], ...]. If the usage message is null
// or empty, then the command will not be listed by the "help" command.
def getScriptCommandList()
{
    return [
        ['indicators', 'brm indicators'],
        ['tar', 'brm tar tar-command options (-h for help)'],
        ];
}

// Return true iff a command is present in commandList, where commandList is
// in the form: [ [command-name, usage-message]]
def isImplemented(commandList, commandName)
{
    def result = false;
    commandList.each
    { commandEntry ->
        if (commandEntry[0].equalsIgnoreCase(commandName))
        {
            result = true;
            return;
        }
    }
    return result;
}

def printCommandList(commandList)
{
    commandList.each
    { commandEntry ->
        if (!commandEntry[1])
            return; // skip undocumented commands
        println("    ${commandEntry[1]}");
    }
}

if (command.equalsIgnoreCase("help"))
{
    println("BRM ${readVersion()}");
    println("Backup, Restore and Move content tool for PM Solutions");
    println();
    println("List of BRM sub-commands:");
    println("    brm agent - Run BRM in agent mode (not on metadata server)");
    printCommandList(BRMController.getCommandList());
    printCommandList(MainMigrator.getCommandList());
    printCommandList(getScriptCommandList());
}
else if (command.equalsIgnoreCase("agent"))
{
    com.sas.solutions.brm.BRMAgentTCPServer.main(commandArgs);
}
else if (isImplemented(BRMController.getCommandList(), command))
{
    BRMController.main(args);
}
else if (isImplemented(MainMigrator.getCommandList(), command))
{
    MainMigrator.main(args);
}
else if (isImplemented(getScriptCommandList(), command))
{
    def scriptsDir = new File(System.getProperty('BRM.install.dir'), 'scripts');
    def scriptFile = new File(scriptsDir, command.toLowerCase() + '.groovy');
    String[] scriptCommandArgs = new String[args.length];
    scriptCommandArgs[0] = scriptFile.getPath();
    System.arraycopy(args, 1, scriptCommandArgs, 1, args.length - 1);
    runGroovyScript(scriptCommandArgs);
}
else
{
    println("Unknown command: " + command);
    println('Use the "help" command to list the possible commands.');
}

// Normal exit
System.exit(0);

def readVersion()
{
    def version = BRMUtil.readVersion();
    if (!version)
    {
        println("Missing or invalid version.txt file.");
        println("BRM has not been installed properly.");
        System.exit(1);
    }
    return version;
}

// Run Groovy script with BRM classpath
def runGroovyScript(commandArgs)
{
    String groovyScriptName = commandArgs[0];
    String[] groovyArgs = new String[commandArgs.length - 1];
    System.arraycopy(commandArgs, 1, groovyArgs, 0, groovyArgs.length);
    GroovyShell gsh = new GroovyShell();
    gsh.run(new File(groovyScriptName), groovyArgs);
}

// Configuration check function
// Return true if Ok, false if we cannot proceed
// Diagnostic messages will be printed to the console because normal logging
// may not work (that's one of the configuration items that could be wrong.)
def configCheck(command)
{
    def installDir = System.getProperty('BRM.install.dir');
    if (installDir == null)
    {
        println("ERROR: Java system property BRM.install.dir is not set.");
        printInstallDirMsg();
        return false;
    }
    installDir = new File(installDir);
    if (!installDir.isDirectory())
    {
        println("ERROR: Installation directory ${installDir} does not exist.");
        printInstallDirMsg();
        return false;
    }
    def brmini = new File(installDir, 'brm.ini');
    if (!brmini.isFile())
    {
        println("ERROR: brm.ini file does not exist at ${brmini}");
        printInstallDirMsg();
        return false;
    }
    def m = null;
    try
    {
        m = readIniSection(brmini, 'properties');
    }
    catch (java.nio.charset.CharacterCodingException e)
    {
        // ignore
    }
    catch (java.io.CharConversionException e)
    {
        // ignore
    }
    if (m != null)
    {
        // Note: These error conditions cannot be detected unless the
        // regular Java launcher is somehow bypassed. Usually, an incorrect
        // MASTERPROP setting will simply result in brm.exe making a quick
        // exit with no message.
        def masterprop = m.get('MASTERPROP');
        if (masterprop == null)
        {
            println("ERROR: brm.ini is missing MASTERPROP definition");
            return false;
        }
        masterprop = new File(masterprop);
        if (!masterprop.isFile())
        {
            println("ERROR: brm.ini has invalid MASTERPROP definition");
            println("       File not found: " + masterprop);
            println("       Point MASTERPROP to a valid sassw.config file.");
            return false;
        }
    }
    def configDir = new File(installDir, 'config');
    if (!configDir.isDirectory())
    {
        println("ERROR: BRM config directory does not exist");
        println("       Directory not found: " + configDir);
        println("       Make sure BRM is installed properly.");
        return false;
    }
    def configFiles = ['brm.properties', 'brm-log4j.properties', 
                       'brmagent.properties'];
    def numMissingFiles = 0;
    configFiles.each
    { configFileName ->
        if (command.equalsIgnoreCase("agent") && 
                configFileName == 'brm.properties')
            return;
        configFile = new File(configDir, configFileName);
        if (!configFile.isFile())
        {
            println("ERROR: Missing BRM configuration file: " + configFileName);
            println("       Make sure that BRM is configured properly, and");
            println("       that all .template files have been copied to");
            println("       their new names.");
            numMissingFiles += 1;
        }
    }
    if (numMissingFiles > 0)
        return false;
    return true;
}

def printInstallDirMsg()
{
    println("       Check that brm.ini exists in the same directory as");
    println("       the launcher file brm.exe (brm.sh on Unix) and that");
    println("       the BRM.install.dir property is set correctly to ");
    println("       <LAUNCHERDIR>.");
}

def readIniSection(file, section)
{
    if (section == null)
        return [:];
    // Don't use a formatted string here, equals() won't work
    section = '[' + section.toLowerCase() + ']';
    def r = new BufferedReader(
                new InputStreamReader(
                    new FileInputStream(file)));
    def line = null;
    try
    {
        while (true)
        {
            line = r. readLine();
            if (line == null)
                return [:];
            line = line.trim().toLowerCase();
            if (line.equals(section))
                break;
        }
        def map = [:];
        while (true)
        {
            line = r. readLine();
            if (line == null)
                return map;
            line = line.trim();
            if (line.startsWith('[') && line.endsWith(']'))
                return map;
            if (line.length() == 0)
                continue;
            if (line.startsWith(';'))
                continue;
            def i = line.indexOf('=');
            if (i < 0)
                continue;
            def key = line.substring(0, i);
            def value = line.substring(i+1);
            map[key] = value;
        }
    }
    finally
    {
        r.close();
    }
}

// end-of-file
