/*
 * Copyright (c) 2006-2010 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm;

import java.io.File;
import java.net.URL;
import java.util.Collection;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Properties;

import org.apache.log4j.Logger;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.helper.ProjectHelper2;
import org.apache.tools.ant.taskdefs.MacroInstance;

import com.sas.solutions.brm.errors.BRMAntError;
import com.sas.solutions.brm.errors.BRMConfigurationException;
import com.sas.solutions.brm.errors.BRMInternalError;
import com.sas.solutions.brm.errors.BRMUserError;
import com.sas.solutions.brm.logging.BRMBuildListener;
import com.sas.solutions.brm.logging.MonitoredLogger;
import com.sas.solutions.brm.omr.OMRConnection;
import com.sas.solutions.brm.storedProcess.StoredProcDirFinder;
import com.sas.solutions.brm.util.BRMUtil;

public class BRMAgent implements BRMAgentInterface
{
    // Logger for messages that are sent only to BRM-agent.log
    private static final Logger LOG = Logger.getLogger("agent." + 
        BRMAgent.class.getName());
    
    // Logger for messages that are sent while running Ant commands or
    // related methods. These messages go either to the local BRM.log
    // or over a TCP connection to the BRM controller and its BRM.log.
    private Object antLogger = null;

    private Properties localConfig = null;
    private Properties controlConfig = null;
    private Properties agentConfig = null;

    private String machineName = "";
    private String component = "agent";
    private String timeStampStr = "";

    /**
     * @param antLogger  The Logger object to use for messages generated
     * while running Ant targets. If null is passed, then setAntLogger()
     * must be called with a valid Logger object before runTarget() is
     * called.
     */
    public BRMAgent(Object antLogger) throws BRMConfigurationException
    {
        setAntLogger(antLogger);

        // Setting a default timeStampStr value here,
        // but it will be re-set from runTarget().
        this.timeStampStr = BRMUtil.generateTimeStamp();

        init();
    }

    // Load local config files and set up for first use
    public void init() throws BRMConfigurationException
    {
        this.machineName = BRMUtil.getLocalHostName();
        this.component = "agent";
        renameThread();
        this.agentConfig = new Properties();
        BRMUtil.loadPropFileAsResource(Constants.AGENT_CONFIG_FILE,
                this.agentConfig, LOG);
        loadLocalConfig();
    }

    // This is called by BRMAgentTCPServer
    public String getAgentConfigProp(String propName)
    {
        if (propName == null)
            return null;
        if (this.agentConfig == null)
            return null;
        return this.agentConfig.getProperty(propName);
    }

    // This is called by test scripts
    public void setAgentConfigProp(String propName, String propVal)
    {
        this.agentConfig.setProperty(propName, propVal);
    }

    // This is called by test scripts and MainMigrator
    public Properties getLocalConfig()
    {
        return this.localConfig;
    }

    /**
     * @return The current Logger instance used for messages generated while
     * running Ant targets.
     */
    public Object getAntLogger()
    {
        return this.antLogger;
    }

    /**
     * Set the Logger used for messages generated while executing Ant
     * targets. If the logger is not a MonitoredLogger, it will be wrapped
     * with a MonitoredLogger so that the number of errors and warnings can
     * be tracked.
     */
    public setAntLogger(Logger logger)
    {
        if (logger != null && !(logger instanceof MonitoredLogger))
        {
            logger = new MonitoredLogger(null, logger);
        }
        this.antLogger = logger;
    }

    public setControlConfig(Properties controlConfig)
    {
        this.controlConfig = controlConfig;
    }

    public String getConfigDir()
    {
        String configDir = null;
        if (this.agentConfig != null)
        {
            configDir = this.agentConfig.getProperty("BRM.config.dir");
            if (configDir != null && configDir.length() > 0)
                return configDir;
        }
        if (this.controlConfig != null)
        {
            configDir = this.controlConfig.getProperty("BRM.config.dir");
            if (configDir != null && configDir.length() > 0)
                return configDir;
        }
        return configDir;
    }

    public boolean loadLocalConfig()
    {
        if (this.localConfig != null)
        {
            LOG.debug("loadLocalConfig: configuration is already loaded");
            return true;
        }
        String configDir = getConfigDir();
        if (configDir == null)
        {
            LOG.debug("loadLocalConfig: BRM.config.dir property not set yet");
            return false;
        }
        File utilitiesDir = new File(configDir, "Utilities");
        if (!utilitiesDir.isDirectory())
        {
            throw new BRMConfigurationException(
                "Utilities directory doesn't exist: " + utilitiesDir);
        }
        this.localConfig = new Properties();
        // Load additional properties files if they exist
        def additionalPropFiles = [
            "metadatasrv.configuration.properties",
            "mysqldbs.configuration.properties", 
            "objectspawnr.configuration.properties", 
            "weblogic.configuration.properties", 
            "tableserver.configuration.properties",
            "cfgwizard.configuration.properties",
            "javaportal.configuration.properties",
            "bidashmid.configuration.properties",
            "fmsmidtier.configuration.properties",
            "hcmmidtier.configuration.properties",
            "spmmidtier.configuration.properties",
            "levconfig.configuration.properties",
            "webspherend.configuration.properties",
           ];
        additionalPropFiles.each { propFile ->
            def fullPropPath = new File(utilitiesDir, propFile);
            if (fullPropPath.isFile()) { 
                BRMUtil.loadPropFile(fullPropPath, this.localConfig, LOG);  
            }
        }
        // Load catch-all combined prop file to get anything that the list
        // of files above did not define.
        File configPropFile = new File(utilitiesDir, 
                                            "configuration.properties");
        if (!configPropFile.isFile())
        {
            throw new BRMConfigurationException(
                "Configuration file doesn't exist: " + configPropFile);
        }
        BRMUtil.loadPropFile(configPropFile, this.localConfig, LOG);

        return true;
    }

    //////////////////////////////////////////////////////////////////////////
    // Begin BRMAgentInterface implementation

    public Set getTargetSet()
    {
        Project p = getAntProject();
        Set targetSet = new HashSet();
        targetSet.addAll(p.getTargets().keySet());
        return targetSet;
    }

    // Attempt to run an Ant target.
    // Status and error messages are returned via Log4j logging.
    // Throws BRMAntError if executing the target produces error messages,
    // or if the target doesn't exist in the ant file.
    public void runTarget(String component, String targetName,
                                Properties controlConfig)
    {
        this.component = component;
        this.controlConfig = controlConfig.clone();
        this.controlConfig.setProperty("BRM.component.name", component);
        this.controlConfig.setProperty("BRM.target.name", targetName);
        String timeStampStr = controlConfig.getProperty("BRM.time.stamp");
        if (timeStampStr != null)
        {
            this.timeStampStr = timeStampStr;
        }
        else
        {
            // Use the default timestamp from when BRMAgent was created
            this.controlConfig.setProperty("BRM.time.stamp", this.timeStampStr);
        }
        this.machineName = controlConfig.getProperty("BRM.agent.host");
        if (this.machineName == null)
        {
            this.machineName = BRMUtil.getLocalHostName();
            this.controlConfig.setProperty("BRM.agent.host", this.machineName);
        }
        Object logger = getAntLogger();
        String msg;
        
        // Get the output directory name if it is set
        // Get output directory name from Agent first
        String outputDirName = this.agentConfig.getProperty("BRM.output.dir");
        
        if (outputDirName == null)
        {
            // Get output directory from Controller since it is not in Agent
            outputDirName = controlConfig.getProperty("BRM.output.dir");  
        }

        // Replace ${stamp} token with date- and time-based code
        if (outputDirName != null)
        {
            outputDirName = outputDirName.replace('$' + '{stamp}', 
                                                    this.timeStampStr);
        }

        // We must have an output directory if we are doing a backup or
        // restore.
        if (targetName.startsWith('backup_') ||
            targetName.endsWith('_backup') || 
            targetName.indexOf('_backup_') >= 0 ||
            targetName.startsWith('restore_') ||
            targetName.endsWith('_restore') ||
            targetName.indexOf('_restore_') >= 0)
        {

            if (outputDirName == null)
            {
                msg = "Missing property: BRM.output.dir";
                LOG.error(msg); // Local BRM-agent.log
                logger.error(msg); // Remote BRM.log
                throw new BRMAntError(msg);
            }

            if (targetName.startsWith('backup_') ||
                targetName.endsWith('_backup') || 
                targetName.indexOf('_backup_') >= 0)
            {
                // Create the output directory if it does not exist
                (new File(outputDirName)).mkdirs();    
            }
                
            if (!(new File(outputDirName)).exists())
            {
                msg = "Directory for BRM.output.dir (" + outputDirName + ") does not exist";
                LOG.error(msg); // Local BRM-agent.log
                logger.error(msg); // Remote BRM.log
                throw new BRMAntError(msg);
            }  
        }

        // Set thread name based on control properties
        String prevThreadName = renameThread();

        int startNumErrors = logger.getNumErrorMessages();

        boolean result = executeTarget(targetName, getAntProject());
        if (!result)
        {
            msg = "BRMAgent.runTarget: Can't execute target " + targetName;
            LOG.error(msg); // Local BRM-agent.log
            logger.error(msg); // Remote BRM.log
            throw new BRMAntError(msg);
        }

        int endNumErrors = logger.getNumErrorMessages();

        // Reset thread name
        Thread.currentThread().setName(prevThreadName);

        if (endNumErrors > startNumErrors)
        {
            msg = (
                "Errors encountered when executing Ant target \""+targetName+ 
                "\" on " + this.machineName + 
                " .  Please check the BRM log file for details.");
            LOG.error(msg); // Local BRM-agent.log
            // Error messages already sent to remote BRM.log
            throw new BRMAntError(msg);
        }
    }

    // This is a signal that the high-level BRM command has finished.
    // No more methods should be called on this object after close() is
    // called. (The TCP server may exit at this point.)
    public void close()
    {
        def msg = "BRMAgent received signal to close.";
        LOG.info(msg);
        Object logger = getAntLogger();
        if (logger != null)
        {
            logger.info(msg);
        }
    }

    // End BRMAgentInterface implementation
    //////////////////////////////////////////////////////////////////////////

    //////////////////////////////////////////////////////////////////////
    // Private utility methods

    private boolean executeTarget(String targetName, Project p)
    {
        return executeTarget(targetName, p, false);
    }

    // Return true iff the Ant target exists and execution was attempted
    // (may have had an error during execution)
    private boolean executeTarget(String targetName, Project p, 
            boolean okIfNotDefined)
    {
        Hashtable targetTable = p.getTargets();
        if (targetTable.get(targetName) == null)
        {
            if (okIfNotDefined)
            {
                getAntLogger().debug("Skipping Ant target \"" + 
                        targetName + "\"");
            }
            else
            {
                getAntLogger().error("Ant target \"" + 
                        targetName + "\" is not defined");
            }
            return false;
        }
        try
        {
            LOG.info("Beginning target: " + targetName);
            p.executeTarget(targetName);
            LOG.info("Completed target: " + targetName);
        }
        catch (Exception e)
        {
            getAntLogger().error("Error executing target: " + targetName, e);
        }
        return true;
    }
    
    private void loadPropertiesIntoProject(Project proj, Properties props)
    {
        if (proj == null)
        {
            getAntLogger().error("Project was null.");
            return;
        }
        if (props == null)
        {
            getAntLogger().error("Properties were null.");
            return;
        }
        
        for (Iterator i = props.keySet().iterator(); i.hasNext(); )
        {
            String key = (String)i.next();
            String propVal = props.getProperty(key);
            if (propVal.length() == 0)
                continue;
            // Substitute timestamp in output directory
            if (key.equals("BRM.output.dir"))
            {
                propVal = BRMUtil.replaceTimeStampVar(propVal,
                                                      this.timeStampStr);
            }
            proj.setProperty(key, propVal);
        }
    }
    
    private String renameThread()
    {
        String newThreadName = "";
        if (this.component != null)
            newThreadName = this.machineName + "[" + this.component + "]";
        else
            newThreadName = this.machineName;
        Thread currentThread = Thread.currentThread();
        String prevThreadName = currentThread.getName();
        Thread.currentThread().setName(newThreadName);
        return prevThreadName;
    }

    private Project getAntProject()
    {
        // Create ant Project from ant XML file
        ////////////////////////////////////////////////////
        Project ant = new Project();
        ant.init();

        // Set up the project helper
        ProjectHelper2 helper = new ProjectHelper2();
        ant.addReference("ant.projectHelper", helper);

        // Create the build listener
        if (getAntLogger() != null)
        {
            BRMBuildListener logger = new BRMBuildListener(this.component,
                                                            getAntLogger());
            ant.addBuildListener(logger);
        }
        else
        {
            LOG.error(
                "Ant Logger has not been set, messages might be misdirected");
        }

        // Load the agent Ant command file
        URL uri = ClassLoader.getSystemResource(Constants.AGENT_COMMAND_FILE);
        if (uri == null)
        {
            throw new BRMConfigurationException("Missing config file: " +
                    Constants.AGENT_COMMAND_FILE);
        }
        File agentAnt = new File(uri.getFile());
        helper.parse(ant, agentAnt);
        
        // Setup Properties in Project
        ////////////////////////////////////////////////////
        loadLocalConfig();
        if (this.localConfig != null)
            loadPropertiesIntoProject(ant, this.localConfig);
        if (this.controlConfig != null)
            loadPropertiesIntoProject(ant, this.controlConfig);
        if (this.agentConfig != null)
            loadPropertiesIntoProject(ant, this.agentConfig);

        return ant;
    }
}
