/*
 * Copyright (c) 2010 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketAddress;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import com.sas.solutions.brm.errors.BRMAntError;
import com.sas.solutions.brm.logging.RemoteLoggingServer;

public class BRMAgentRemoteProxy implements BRMAgentInterface
{
    private static transient final Logger LOG = Logger.getLogger(
            BRMAgentRemoteProxy.class);

    private String hostname;
    private int port;
    private Object antLogger;

    /**
     * @param hostname  Name of remote host running BRM agent
     * @param port      TCP port number on which remote BRM agent is running
     * @param antLogger An instance of MonitoredLogger (with logEvent method)
     */
    public BRMAgentRemoteProxy(String hostname, int port, Object antLogger)
    {
        this.hostname = hostname;
        this.port = port;
        this.antLogger = antLogger;
    }

    ////////////////////////////////////////////////////////////////////////
    // Begin implementation of BRMAgentInterface

    public Set getTargetSet()
    {
        Socket sock = getAgentConnection();
        try
        {
            sendCommand(sock, "getTargetSet", null);
            Set result = (Set)readObject(sock);
            return result;
        }
        finally
        {
            closeSocket(sock);
        }
    }

    public void runTarget(String component, String targetName, 
                          Properties props)
    {
        Socket sock = getAgentConnection();
        String error = null;
        try
        {
            Map params = new HashMap();
            params.put("component", component);
            params.put("targetName", targetName);
            params.put("logLevel", this.antLogger.getEffectiveLevel());
            params.put("props", props);
            sendCommand(sock, "runTarget", params);

            RemoteLoggingServer rls = new RemoteLoggingServer(sock, 
                                                              this.antLogger);
            rls.run();

            Map result = (Map)rls.getResult();
            if (result != null)
            {
                error = result.get("error");
            }
            else
            {
                error = "Abnormal disconnect: " + rls.getNetworkError();
            }
        }
        finally
        {
            closeSocket(sock);
        }
        if (error != null)
            throw new BRMAntError(error);
    }

    public void close()
    {
        Socket sock = getAgentConnection();
        try
        {
            Map params = new HashMap();
            params.put("logLevel", this.antLogger.getEffectiveLevel());
            sendCommand(sock, "close", params);
            RemoteLoggingServer rls = new RemoteLoggingServer(sock, 
                                                              this.antLogger);
            rls.run();
            Object error = rls.getNetworkError();
            if (error != null)
            {
                LOG.warn("Abnormal disconnect: " + error);
            }
        }
        finally
        {
            closeSocket(sock);
        }
    }

    // End implementation of BRMAgentInterface
    ////////////////////////////////////////////////////////////////////////

    public Socket getAgentConnection()
    {
        String hostname = this.hostname.toLowerCase().trim();
        int port = this.port;
        LOG.info("Connecting to agent at " + hostname + ":" + port);
        Socket sock = null;
        try
        {
            InetAddress addr = InetAddress.getByName(hostname);
            SocketAddress sockaddr = new InetSocketAddress(addr, port);

            // Create an unbound socket
            sock = new Socket();

            // This method will block no more than timeoutMs.
            // If the timeout occurs, SocketTimeoutException is thrown.
            int timeoutMs = 2000;   // 2 seconds
            sock.connect(sockaddr, timeoutMs);
        }
        catch (UnknownHostException e)
        {
            throw new RuntimeException("Error connecting to server: " + e);
        }
        catch (SocketTimeoutException e)
        {
            throw new RuntimeException("Server " + hostname + 
                    " is not responding on port " + port,
                    e);
        }
        catch (IOException e)
        {
            throw new RuntimeException("Error connecting to " + hostname + 
                    " on port " + port + ": " + e);
        }
        return sock;
    }

    /**
     * Send a command to a BRM agent over a TCP socket.
     * @param command  String containing name of command to run on agent
     * @param params  Map of command parameters, may be null for empty params
     */
    public void sendCommand(Socket sock, String command, Map params)
    {
        if (params == null)
            params = new HashMap();
        try
        {
            ObjectOutputStream oos = new ObjectOutputStream(
                                                sock.getOutputStream());
            oos.writeObject(command);
            oos.writeObject(params);
            oos.flush();
        }
        catch (IOException e)
        {
            throw new RuntimeException(e);
        }
    }

    public readObject(Socket sock)
    {
        Object obj = null;
        try
        {
            ObjectInputStream ois = new ObjectInputStream(
                                            sock.getInputStream());
            obj = ois.readObject();
        }
        catch (IOException e)
        {
            throw new RuntimeException(e);
        }
        return obj;
    }

    public void closeSocket(Socket sock)
    {
        try
        {
            sock.close();
        }
        catch (IOException e)
        {
        }
    }
}

// end-of-file
