/*
 * Copyright (c) 2010 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.Socket;
import java.net.ServerSocket;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import com.sas.solutions.brm.errors.BRMAntError;
import com.sas.solutions.brm.logging.MonitoredLogger;
import com.sas.solutions.brm.logging.ObjectOutputStreamLogger;
import com.sas.solutions.brm.util.BRMUtil;

public class BRMAgentTCPServer
{
    private static final Logger LOG = Logger.getLogger("agent." + 
            BRMAgentTCPServer.class.getName());
    private static transient final Logger PROGRESS = Logger.getLogger(
            "progress.agent.brm");

    private BRMAgentInterface agent = null;
    private int port = 0;
    private boolean shutdownAllowed = false;
    private boolean argsOk = false;
    private boolean keepRunning = true;
    
    private Socket sock;    // Current client socket
    private ObjectInputStream inputStream; // Current input stream
    private ObjectOutputStream outputStream; // Current output stream

    public BRMAgentTCPServer(String[] args)
    {
        this.agent = null; // This is created on demand
        this.argsOk = parseArguments(args);
        this.sock = null;
        this.inputStream = null;
        this.outputStream = null;
    }
    
    /**
      * Parse an array of command-line arguments as BRMAgent-server-specific
      * options. Sets attributes as appropriate. If invalid or incomplete
      * options are encountered, or the "help" option is given, a message is 
      * printed and this method returns false.
      * @return true if arguments are Ok for BRM agent to run,
      *         false otherwise.
      */
    public boolean parseArguments(String[] args)
    {
        LOG.info("BRMAgent started with command-line options: " +
                 BRMUtil.stringArrayToString(args));
        for (int i =0; i<args.length; i++)
        {
            if (args[i].equalsIgnoreCase("-p"))
            {
                if (i == args.length - 1)
                {
                    System.err.println(
                      "BRMAgentTCPServer: missing port number after -p option");
                    return false;
                }
                else
                {
                    try
                    {
                        this.port = Integer.parseInt(args[i+1]);
                    }
                    catch (NumberFormatException e)
                    {
                        System.err.println(args[i+1] + 
                            " is not a valid port number.");
                        return false;
                    }
                    // consuming the arg after -p, so advance one
                    i++;
                }
            }
            else if (args[i].equalsIgnoreCase("--allow-shutdown"))
            {
                this.shutdownAllowed = true;
            }
            else if (args[i].equalsIgnoreCase("help") ||
                     args[i].equalsIgnoreCase("--help") ||
                     args[i].equalsIgnoreCase("-h") ||
                     args[i].equals("/?"))
            {
                System.out.println("BRM Agent - executes BRM commands");
                System.out.println("Usage: brm agent [options]");
                System.out.println("Options:");
                System.out.println("--help  Print this message and exit");
                System.out.println("-p port-number  Change port number");
                System.out.println("--allow-shutdown  Exit after BRM command");
                return false;
            }
            else
            {
                System.err.println(
                    "BRMAgentTCPServer: Unknown option: " + args[i]);
                System.err.println("Try -h for help");
                return false;
            }
        }
        return true;
    }

    public BRMAgentInterface getAgent()
    {
        if (this.agent == null)
        {
            // Set agent up with default logger
            // Some commands will override this logger
            this.agent = new BRMAgent(new MonitoredLogger(PROGRESS, LOG));
        }
        return this.agent;
    }

    /**
     * @return The TCP port number on which BRM agent TCP server has been
     * configured to listen, or the default port number if no port option
     * was given.
     */
    public int getPort()
    {
        int port = this.port;
        if (port == 0)
        {
            String s = getAgent().getAgentConfigProp(
                                        Constants.AGENT_PORT_PROPERTY);
            if (s != null && s.length() > 0)
                port = Integer.parseInt(s);
            if (port == 0)
                port = Constants.AGENT_DEFAULT_PORT;
        }
        return port;
    }

    // Run BRM Agent listening on TCP socket
    // Run one remote command per client connection then disconnect
    public void run()
    {
        if (!this.argsOk)
        {
            LOG.info("BRMAgent quitting because of command-line options.");
            return;
        }
        int port = getPort();
        PROGRESS.info("BRMAgent is listening for requests on port " + port);
        LOG.info("BRMAgent is listening for requests on port " + port);
        PROGRESS.info("Press Ctrl-C to stop.");
        ServerSocket srv = null;
        try
        {
            srv = new ServerSocket(port);
        }
        catch (IOException e)
        {
            String msg = "Error listening on TCP port " + port;
            PROGRESS.fatal(msg + ": " + e);
            LOG.fatal(msg, e);
            return;
        }
        this.sock = null;
        while (this.keepRunning)
        {
            try
            {
                // Wait for connection from client.
                this.sock = srv.accept();
            }
            catch (IOException e)
            {
                String msg = "Error accepting connection on TCP socket";
                PROGRESS.fatal(msg + ": " + e);
                LOG.fatal(msg, e);
                break;
            }
            LOG.info("New connection received from " + 
                        this.sock.getInetAddress());

            try
            {
                this.inputStream = new ObjectInputStream(
                                                this.sock.getInputStream());
                this.outputStream = new ObjectOutputStream(
                                                this.sock.getOutputStream());
                String command = (String)this.inputStream.readObject();
                Map params = (Map)this.inputStream.readObject();
                handleRemoteCommand(command, params);
            }
            catch (IOException e)
            {
                String msg = "Error while handling controller connection";
                PROGRESS.error(msg + ": " + e);
                msg += " (socket closed: " + this.sock.isClosed() + ")";
                LOG.error(msg, e);
            }
            catch (Throwable e)
            {
                String msg = "Error executing command from controller";
                PROGRESS.fatal(msg + ": " + e);
                LOG.fatal(msg, e);
                break;
            }
            finally
            {
                LOG.info("Disconnecting from " + 
                            this.sock.getInetAddress());
                if (this.inputStream != null)
                {
                    try { this.inputStream.close() } catch (IOException e) { }
                    this.inputStream = null;
                }
                if (this.outputStream != null)
                {
                    try { this.outputStream.close() } catch (IOException e) { }
                    this.outputStream = null;
                }
                try { this.sock.close(); } catch (IOException e) { }
                this.sock = null;
            }
        }
        try
        {
            srv.close();
        }
        catch (IOException e)
        {
        }
        LOG.info("BRMAgentTCPServer: Normal exit.");
    }

    public void handleRemoteCommand(String command, Map params)
    {
        if (command.equals("getTargetSet"))
        {
            handleRemoteGetTargetSet(params);
        }
        else if (command.equals("runTarget"))
        {
            handleRemoteRunTarget(params);
        }
        else if (command.equals("close"))
        {
            handleRemoteClose(params);
        }
        else
        {
            LOG.error("Agent received unknown command: " + command);
        }
    }

    public void handleRemoteGetTargetSet(Map params)
    {
        sendObject(getAgent().getTargetSet());
    }

    public void handleRemoteRunTarget(Map params)
    {
        String component = params.get("component");
        String targetName = params.get("targetName");
        Level logLevel = params.get("logLevel");
        Properties props = params.get("props");

        LOG.info("Received " + props.size() + " properties from controller");
        if (LOG.isDebugEnabled())
        {
            Enumeration propNames = props.propertyNames();
            while (propNames.hasMoreElements())
            {
                String propName = (String)propNames.nextElement();
                String propValue = props.getProperty(propName);
                LOG.debug(propName + "=" + propValue);
            }
        }

        // Set up agent to send Ant log events back over this TCP
        // socket.
        Logger antLogger = new MonitoredLogger(null,
                                new ObjectOutputStreamLogger("agentAnt",
                                                            this.outputStream));
        if (logLevel != null)
        {
            antLogger.setLevel(logLevel);
        }
        getAgent().setAntLogger(antLogger)

        String error = null;
        try
        {
            getAgent().runTarget(component, targetName, props);
        }
        catch (BRMAntError e)
        {
            error = e.getMessage();
            if (error == null)
                error = e.toString();
        }

        // Send result back
        Map result = new HashMap();
        result.put("error", error);
        sendObject(result);

        // Disconnect the Ant logger
        getAgent().setAntLogger(null);
    }

    public void handleRemoteClose(Map params)
    {
        Level logLevel = params.get("logLevel");

        // Send log events back over this TCP socket.
        Logger agentLogger = new MonitoredLogger(null,
                                new ObjectOutputStreamLogger("agentAnt",
                                                            this.outputStream));
        if (logLevel != null)
        {
            agentLogger.setLevel(logLevel);
        }
        getAgent().setAntLogger(agentLogger)

        getAgent().close();
        this.agent = null; // Can't call methods on this agent after close()

        if (this.shutdownAllowed)
        {
            String msg = (
                "Agent shutting down because of --allow-shutdown option");
            LOG.info(msg);
            PROGRESS.info(msg);
            agentLogger.info(msg);
            this.keepRunning = false;
        }
    }

    public void sendObject(Object obj)
    {
        this.outputStream.writeObject(obj);
        this.outputStream.flush();
    }

    //////////////////////////////////////////////////////////////////////
    // Main method

    public static void main(String[] args)
    throws Exception
    {
        Thread.currentThread().setName("agent");
        com.sas.solutions.brm.logging.LogConfigurator.configure();
        args = BRMUtil.filterCmdLineProps(args);
        BRMAgentTCPServer server = new BRMAgentTCPServer(args);
        server.run();
    }
}

// end-of-file
