/*
 * Created on Apr 28, 2006
 * Copyright (c) 2006 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm;

import java.io.IOException;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.Set;

import javax.swing.JOptionPane;

import org.apache.log4j.Logger;

import com.sas.solutions.brm.Constants;
import com.sas.solutions.brm.errors.BRMConfigurationException;
import com.sas.solutions.brm.errors.BRMAntError;
import com.sas.solutions.brm.errors.BRMError;
import com.sas.solutions.brm.errors.BRMInternalError;
import com.sas.solutions.brm.errors.BRMUserError;
import com.sas.solutions.brm.errors.DefaultConfigUsedException;
import com.sas.solutions.brm.logging.MonitoredLogger;
import com.sas.solutions.brm.omr.OMRConnection;
import com.sas.solutions.brm.util.BRMUtil;
import com.sas.solutions.brm.xml.StatusXMLParser;

public class BRMController
{
    private static transient final Logger LOG = Logger.getLogger(
            BRMController.class);
    private static transient final Logger PROGRESS = Logger.getLogger(
            "progress.controller.brm");
    
    private Properties controlConfig = null;
    private Properties agentConfig = null; // used only for the TCP port number
    private String timeStampStr = null;
    List componentList = null;
    Map hostAgentMap = new HashMap();
    
    public BRMController() throws BRMConfigurationException, IOException
    {
        String version = BRMUtil.readVersion();
        String buildtype = BRMUtil.readBuildtype();
        String builddatetime = BRMUtil.readBuilddatetime();
        LOG.info("BRM version ${version}-${buildtype} built ${builddatetime}");
        LOG.info("BRMController starting");
        this.timeStampStr = BRMUtil.generateTimeStamp();
        readProperties();
    }

    ///////////////////////////////////////////////////////////////////////
    // Public interface

    /**
      * Clean up and disconnect from all resources used by the Controller.
      */
    public void close()
    {
        LOG.info("Closing BRM controller and releasing its resources.");
        this.controlConfig = null;
        this.agentConfig = null;
        this.timeStampStr = null;
        this.componentList = null;
        this.hostAgentMap.each
        { host, agent ->
            try
            {
                agent.close();
            }
            catch (Throwable e)
            {
                LOG.warn("Exception while closing agent at ${host}: ${e}");
            }
        }
        this.hostAgentMap.clear();
    }
    
    public void doBackup()
        throws BRMConfigurationException
    {
        LOG.info("Starting backup");
        def outputDir = new File(getOutputDir());
        LOG.info("BRM output directory on controller machine: " + outputDir);
        // Create the BRM output directory if necessary
        outputDir.mkdirs();
        if (!outputDir.isDirectory())
        {
            // Directory creation failed somehow
            def msg = "BRM output directory (" + outputDir +") does not exist.";
            throw new BRMUserError(msg);
        }  
        backupControlConfig();
        
        runTargets("backup");

        LOG.info("Finished backup");
    }
    
    public void doRestore()
        throws BRMConfigurationException
    {
        LOG.info("Starting restore");
        def outputDir = new File(getOutputDir());
        LOG.info("BRM output directory on controller machine: " + outputDir);

        if (!outputDir.isDirectory())
        {
            // Directory must exist
            def msg = "BRM output directory (" + outputDir +") does not exist.";
            throw new BRMUserError(msg);
        }  
         
        if (!restorePrompt())
            return;
        LOG.info("User elected to continue with BRM Restore.");

        // If temp directory exists, copy config directory contents to temp
        // directory to be picked up later by "brm loadconfig" command -
        // this is so it can access backed-up status.xml files.
        def installDir = System.getProperty('BRM.install.dir');
        def tempDir = new File(installDir, 'temp');
        if (tempDir.isDirectory())
        {
            def backupConfigDir = new File(getOutputDir(), 'config');
            def tempConfigDir = new File(tempDir, 'config');
            tempConfigDir.mkdirs();
            LOG.info("Copying backed up status.xml files from \"${backupConfigDir}\" to \"${tempConfigDir}\"");
            backupConfigDir.list().each
            { name ->
                def subDir = new File(backupConfigDir, name);
                if (!subDir.isDirectory())
                    return;
                def statusXMLFile = new File(subDir, 'status.xml');
                if (!statusXMLFile.isFile())
                    return;
                def destSubDir = new File(tempConfigDir, name);
                destSubDir.mkdirs();
                def destStatusXMLFile = new File(destSubDir, 'status.xml');
                LOG.info("Copying \"${statusXMLFile}\" to \"${destStatusXMLFile}\"");
                BRMUtil.copyFile(statusXMLFile, destStatusXMLFile);
            }
        }

        // Run restore targets on every host
        runTargets("restore");

        def restoreStatusPropname = 'BRM.restore.statusxml';
        def restoreStatusPropval = System.getProperty(restoreStatusPropname);
        if (restoreStatusPropval != null)
            LOG.info("System property ${restoreStatusPropname}=${restoreStatusPropval}");
        else
            LOG.info("System property ${restoreStatusPropname} is not set");
        if (BRMUtil.isTrueStr(restoreStatusPropval))
        {
            getConfiguredHostList().each
            { host ->
                runTarget(host, "restore", "restorestatusxml");
            }
        }

        LOG.info("Finished restore");
    }

    ///////////////////////////////////////////////////////////////////////
    // Public "getter" methods

    public int getAgentPort()
    {
        String s = this.agentConfig.getProperty(
                Constants.AGENT_PORT_PROPERTY);
        if (s == null || s.length() == 0)
        {
            return Constants.AGENT_DEFAULT_PORT;
        }
        return Integer.parseInt(s);
    }

    // This used to be called "getProperty", but apparently this name
    // is special to the Groovy programming language, and that caused
    // problems.
    public String getConfigProp(String propName)
    {
        return this.controlConfig.get(propName);
    }

    public void setProperty(String propName, String propVal)
    {
        this.controlConfig.setProperty(propName, propVal);
    }

    ///////////////////////////////////////////////////////////////////////
    // Helper methods

    // Check the version of BRM that produced the backup.
    // Return empty string if all is well, or warning message if there is a
    // mismatch. Messages are also sent to the LOG.
    def checkBackupVersion()
    {
        def outputDir = getOutputDir();
        def backupConfigDir = new File(outputDir, 'config');
        def backupVersionFile = new File(backupConfigDir, 'version.txt');
        def msg;
        if (!backupVersionFile.isFile())
        {
            msg = ("Missing version file in backup directory: " + 
                    backupVersionFile);
            throw new BRMConfigurationException(msg);
        }
        def installDir = new File(System.getProperty('BRM.install.dir'));
        def currentVersionFile = new File(installDir, 'version.txt');
        if (!currentVersionFile.isFile())
        {
            msg = ("Missing version file in BRM installation directory: " +
                    currentVersionFile);
            throw new BRMConfigurationException(msg);
        }
        def backupVersion = BRMUtil.readFileAsString(backupVersionFile);
        backupVersion = backupVersion.trim();
        LOG.info("Backup was created with BRM version ${backupVersion}");
        def currentVersion = BRMUtil.readFileAsString(currentVersionFile);
        currentVersion = currentVersion.trim();
        LOG.info("Current BRM version is ${currentVersion}");
        if (!backupVersion.equals(currentVersion))
        {
            msg = (
"""
Restoring a backup created with BRM version ${backupVersion}
Current BRM version is ${currentVersion}
Potentially incompatible backup, proceed at your own risk!
"""
);
            LOG.warn("Backup compatibility warning:" + msg);
            PROGRESS.warn(msg);
            return msg;
        }
        return '';
    }

    def getRestorePromptMessage()
    {
        return (
"""You are about to initiate a BRM restore on this system, which will replace
all existing data on this system with data from the source system. All existing
metadata and other data will be overwritten.
""" + checkBackupVersion() +
"""
Please enter "yes" or "y" to continue.
Enter "no" or just press Enter to cancel.
""");
    }

    // Prompt the user that he/she is about to do a BRM restore, which will
    // wipe out all current content. Return true if user says yes, false
    // otherwise.
    boolean restorePrompt()
    {
        def skipRestorePromptPropName = 'BRM.skip.restore.prompt';
        def skipRestorePrompt = System.getProperty(skipRestorePromptPropName);
        LOG.debug("${skipRestorePromptPropName}=${skipRestorePrompt}");
        if (BRMUtil.isTrueStr(skipRestorePrompt))
        {
            LOG.info(
                "Skipping restore prompt: " +
                "${skipRestorePromptPropName}=${skipRestorePrompt}");
            return true;
        }
        def readFromStandardInput = false;
        if (!System.getProperty("os.name").startsWith('Win'))
        {
            // Not Windows, therefore assuming Unix
            def DISPLAY = System.getenv("DISPLAY");
            if (DISPLAY == null || DISPLAY.trim().length() == 0)
            {
                LOG.info("DISPLAY environment variable not set.");
                LOG.info("Reading prompt from standard input.");
                readFromStandardInput = true;
            }
        }

        String inputValue = null;
        if (readFromStandardInput)
        {
            // Print prompt and read from standard input
            print(getRestorePromptMessage());
            // Note: Due to the way that the Java launcher was implemented,
            // reading from standard input doesn't work on Windows.
            def reader = new BufferedReader(new InputStreamReader(System.in));
            inputValue = reader.readLine();
        }
        else
        {
            // Prompt using graphical interface
            // Prompt the user... do you want to continue with BRM Restore?    
            inputValue = JOptionPane.showInputDialog( null, 
                getRestorePromptMessage(), "BRM",
                JOptionPane.INFORMATION_MESSAGE, null, null, null); 
        }
        LOG.info("At BRM restore prompt, user entered: ${inputValue}");

        // User cancelled out 
        if (inputValue == null)
        {
            LOG.info("User cancelled out of BRM Restore at the prompt.");
            return false;
        }

        if (!(inputValue.equalsIgnoreCase('yes') || 
              inputValue.equalsIgnoreCase('y')))
        {
            LOG.info("User elected not to continue with BRM Restore.");
            return false;
        }

        return true;
    }
    
    private void runTargets(String command)
    {
        // Do per-host processing first
        List hosts = getConfiguredHostList();
        String msg;
        if (hosts.size() == 0)
        {
            msg = "No configured hosts, no " + command + " targets ran.";
            LOG.warn(msg);
            PROGRESS.warn(msg);
            return;
        }

        // Compile set of "hostname:targetName"
        // Host names were already normalized by getConfiguredHostList()
        Set hostTargetSet = new HashSet();
        Iterator hostIterator = hosts.iterator();
        String host = null;
        while (hostIterator.hasNext())
        {
            host = (String)hostIterator.next();
            Set targetSet = getTargetSet(host);
            Iterator targetSetIterator = targetSet.iterator();
            while (targetSetIterator.hasNext())
            {
                hostTargetSet.add(host + ":" +
                                    (String)targetSetIterator.next());
            }
        }

        // Run "_before" targets on each machine
        String targetName = command + "_before";
        msg = "Running " + targetName + " targets on each machine...";
        LOG.info(msg);
        hostIterator = hosts.iterator();
        while (hostIterator.hasNext())
        {
            host = (String)hostIterator.next();
            if (!hostTargetSet.contains(host + ":" + targetName))
                continue;
            LOG.info("Processing host: " + host);
            runTarget(host, "", targetName);
        }

        if (command == 'backup')
            checkStatusXMLFiles();

        // Per-component processing
        // S0800772: If a component fails, still try to run the targets for
        // the other components. For example, if the webapp_backup_before
        // target is run successfully, we must try to run the
        // webapp_backup_after target, even if the mysql_backup target
        // fails, or else the web applications may be left in a bad state.
        Throwable firstError = null;
        def okComponents = new LinkedHashSet(this.componentList);
        ["_"+command+"_before",  "_"+ command,  "_"+command+"_after"].each
        { suffix ->
            LOG.info("Running \"" + suffix + "\" tasks for each component...");
            this.componentList.each
            { component ->
                host = null;
                try
                {
                    host = getHostForComponentOrDie(component);
                }
                catch (BRMConfigurationException e1)
                {
                    if (okComponents.contains(component))
                    {
                        LOG.error(e1.toString(), e1);
                        PROGRESS.error(e1.toString());
                        okComponents.remove(component);
                    }
                }
                targetName = component + suffix;
                if (host != null &&
                    !hostTargetSet.contains(host + ":" + targetName))
                {
                    LOG.debug("Skipping target " + targetName + 
                        " because it does not exist on host " + host);
                    return;
                }
                if (!okComponents.contains(component))
                {
                    msg = ("Skipping target " + targetName + 
                           " because of previous errors for component \"" +
                           component + "\"");
                    LOG.warn(msg);
                    PROGRESS.warn(msg);
                    return;
                }
                LOG.info("Processing component: " + component);
                try
                {
                    runTarget(host, component, targetName);
                }
                catch (Throwable e2)
                {
                    if (firstError == null)
                        firstError = e2;
                    msg = ("Target " + targetName + " on " + host + ": " +
                           e2.toString());
                    if (!(e2 instanceof BRMAntError))
                        LOG.error(msg, e2);
                    PROGRESS.error(msg);
                    okComponents.remove(component);
                }
            }
        }

        // Run "_after" targets on each machine
        targetName = command + "_after";
        msg = "Running " + targetName + " targets on each machine...";
        LOG.info(msg);
        hostIterator = hosts.iterator();
        while (hostIterator.hasNext())
        {
            host = (String)hostIterator.next();
            if (!hostTargetSet.contains(host + ":" + targetName))
                continue;
            LOG.info("Processing host: " + host);
            try
            {
                runTarget(host, "", targetName);
            }
            catch (Throwable e3)
            {
                if (firstError == null)
                    firstError = e3;
                msg = ("Target " + targetName + " on " + host + ": " +
                       e3.toString());
                if (!(e3 instanceof BRMAntError))
                    LOG.error(msg, e3);
                PROGRESS.error(msg);
            }
        }

        if (firstError != null)
        {
            throw new BRMAntError(
                "One or more targets failed. See previous error messages.", 
                firstError);
        }
    }

    def checkStatusXMLFiles()
    {
        LOG.info("Checking the status.xml files");
        def conn = getNewOMRConnection();
        def outputDir = new File(getOutputDir());
        def outputConfigDir = new File(outputDir, 'config');
        def msg;
        getConfiguredHostList().each
        { host ->
            def outputHostDir = new File(outputConfigDir, host);
            def statusXMLFile = new File(outputHostDir, 'status.xml');
            if (!statusXMLFile.isFile())
            {
                msg = "Missing backed up file: ${statusXMLFile}\nMaybe the output directory is not shared on all machines?";
                LOG.warn(msg);
                PROGRESS.warn(msg);
                return;
            }
            LOG.info("Checking the status.xml file from machine ${host}: ${statusXMLFile}");
            def numOkStatusXMLEntries = 0;
            def numBadStatusXMLEntries = 0;
            def parser = StatusXMLParser.parseXML(statusXMLFile);
            parser.configuredObjects.each
            { record ->
                def fqid = record.fqid;
                if (!fqid)
                    return;
                def p = ~/\b[0-9A-Z]{8}\.[0-9A-Z]{8}\b/;
                def m = p.matcher(fqid);
                if (!m)
                    return;
                def obj = conn.getObject('Root', fqid, false);
                if (obj == null)
                {
                    numBadStatusXMLEntries += 1;
                    if (numBadStatusXMLEntries == 1)
                    {
                        msg = "Invalid fqid in status.xml file from machine ${host}: ${statusXMLFile}";
                        LOG.warn(msg);
                        PROGRESS.warn(msg);
                    }
                    msg = "Invalid status.xml line: code=\"${record.code}\" fqid=\"${fqid}\" order=\"${record.order}\"";
                    LOG.warn(msg);
                    PROGRESS.warn(msg);
                }
                else
                {
                    numOkStatusXMLEntries += 1;
                }
            }
            LOG.info("${numOkStatusXMLEntries} FQIDs were Ok");
            LOG.info("${numBadStatusXMLEntries} FQIDs could not be found");
        }
        conn.close();
    }

    private void backupControlConfig()
    {
        String installDirStr = System.getProperty("BRM.install.dir");
        if (installDirStr == null || installDirStr.length() == 0)
        {
            throw new BRMConfigurationException("Missing system property: " +
                "BRM.install.dir");
        }
        File installDir = new File(installDirStr);
        if (!installDir.isDirectory())
        {
            throw new BRMConfigurationException(
                "System property BRM.install.dir " + 
                "points to invalid directory: " + installDir);
        }
        File configSrc = new File(installDir, "config");
        File configDst = new File(getOutputDir(), "config");
        if (!configDst.isDirectory())
            configDst.mkdirs();
        BRMUtil.copyFile(new File(installDir, "builddatetime.txt"),
            new File(configDst, "builddatetime.txt"));
        BRMUtil.copyFile(new File(installDir, "buildtype.txt"),
            new File(configDst, "buildtype.txt"));
        BRMUtil.copyFile(new File(installDir, "version.txt"),
            new File(configDst, "version.txt"));
        BRMUtil.copyFile(new File(configSrc, "brm.properties"),
            new File(configDst, "brm.properties"));
        BRMUtil.copyFile(new File(configSrc, "brm-log4j.properties"),
            new File(configDst, "brm-log4j.properties"));
        BRMUtil.storeBackupInfo(this.controlConfig, configDst.getPath());
    }

    private String getOutputDir()
    {
        String outputDir = this.controlConfig.getProperty("BRM.output.dir");
        if (outputDir == null || outputDir.trim().length() == 0)
        {
            throw new BRMConfigurationException(
                    "Missing property in brm.properties: BRM.output.dir");
        }
        return BRMUtil.replaceTimeStampVar(outputDir, this.timeStampStr);
    }

    private OMRConnection getNewOMRConnection() 
        throws BRMConfigurationException
    {
        return BRMUtil.getNewOMRConnection(this.controlConfig, "admin");
    }
    
    private void readProperties() 
        throws BRMConfigurationException, IOException
    {
        LOG.debug("Reading property files");
        this.controlConfig = new Properties();
        BRMUtil.loadPropFileAsResource(Constants.BRM_CONFIG_FILE,
                this.controlConfig, LOG);
        this.controlConfig.setProperty("BRM.controller.host", 
                                        BRMUtil.getLocalHostName());
        this.controlConfig.setProperty("BRM.controller.host.fqdn", 
                                        BRMUtil.getLocalFQDN());
        this.agentConfig = new Properties();
        BRMUtil.loadPropFileAsResource(Constants.AGENT_CONFIG_FILE,
                this.agentConfig, LOG);

        // Set up the list of components to process, in the correct order
        this.componentList = new ArrayList();
        for (int i=0; i<100; ++i)
        {
            String componentKey;
            if (i >= 10)
                componentKey = "BRM.component." + i;
            else
                componentKey = "BRM.component.0" + i;
            String component = this.controlConfig.getProperty(componentKey);
            if (component == null && i < 10)
            {
                // Maybe the component order number is missing a leading 0
                componentKey = "BRM.component." + i;
                component = this.controlConfig.getProperty(componentKey);
            }
            if (component != null)
                component = component.trim();
            if (component != null)
            {
                String hostname = getHostForComponentOrDie(component);
                if (hostname.equals("default.host.com"))
                {
                    throw new DefaultConfigUsedException(
                            "Detected \"default.host.com\"... " +
                            "Did you edit your " + Constants.BRM_CONFIG_FILE
                            + " file?");
                }
                LOG.info("Component #" + i + ": " + component);
                this.componentList.add(component);
            }
        }
    }

    /**
     * Return a list of all the hosts which have configured components in
     * the BRM configuration file.
     * @return A List of the host names in the order in which their
     *         corresponding components appear.
     */
    public List getConfiguredHostList()
    {
        List hosts = new ArrayList();
        Set hostSet = new HashSet(); // avoid listing hosts twice
        Iterator componentIterator = this.componentList.iterator();
        while (componentIterator.hasNext())
        {
            String component = (String)componentIterator.next();
            String host = getHostForComponent(component);
            if (host != null && !hostSet.contains(host))
            {
                hosts.add(host);
                hostSet.add(host);
            }
        }
        return hosts;
    }

    private String getHostKeyForComponent(String component)
    {
        if (component == null || component.trim().length() == 0)
            return null;
        return "BRM.host." + component.trim();
    }

    private String getHostForComponent(String component)
    {
        String hostKey = getHostKeyForComponent(component);
        if (hostKey == null)
            return null;
        String hostname = this.controlConfig.getProperty(hostKey);
        if (hostname == null || hostname.trim().length() == 0)
            return null;
        hostname = hostname.toLowerCase();
        hostname = hostname.trim();
        return hostname;
    }

    private String getHostForComponentOrDie(String component)
    {
        String hostname = getHostForComponent(component);
        if (hostname == null)
        {
            // Missing BRM.host.<component> property
            throw new BRMConfigurationException(
                "File " + Constants.BRM_CONFIG_FILE +
                " is missing property " +
                getHostKeyForComponent(component));
        }
        return hostname;
    }

    public BRMAgentInterface getAgent(String host)
    {
        BRMAgentInterface agent = this.hostAgentMap.get(host);
        if (agent != null)
            return agent;
        boolean inProc = BRMUtil.isLocalHost(host);
        String remoteAgent = this.controlConfig.getProperty("BRM.remote.agent");
        boolean alwaysRemoteAgent = (remoteAgent != null &&
                                        remoteAgent.equals("always"));
        Object logger = getLocalAntLogger();
        if (inProc && !alwaysRemoteAgent)
        {
            agent = new BRMAgent(logger);
        }
        else
        {
            agent = new BRMAgentRemoteProxy(host, getAgentPort(), logger);
        }
        this.hostAgentMap.put(host, agent);
        return agent;
    }

    // Return a logging object that puts messages in BRM.log and also copies
    // urgent messages to the console.
    public Object getLocalAntLogger()
    {
        def antLogger = Logger.getLogger('agentAnt');
        def monitoredLogger = new MonitoredLogger(PROGRESS, antLogger);
        monitoredLogger.setLevel(antLogger.getEffectiveLevel());
        return monitoredLogger;
    }

    // Get the set of Ant targets implemented by an agent on a particular
    // host machine.
    public Set getTargetSet(String host)
    {
        return getAgent(host).getTargetSet();
    }

    /**
     * Run an Ant target on the given machine.
     * @throws BRMAntError if the Ant target generates any ERROR or FATAL
     * log messages, or if the target does not exist on that host.
     */
    public void runTarget(String host, String component, String targetName)
    {
        String msg = "";
        if (component != null && component.length() > 0)
            msg = "(" + component + ") ";
        msg += "Running " + targetName + " on " + host;
        LOG.info(msg);
        PROGRESS.info(msg);
        Properties props = this.controlConfig.clone();
        props.setProperty("BRM.time.stamp", this.timeStampStr);
        props.setProperty("BRM.agent.host", host);
        getAgent(host).runTarget(component, targetName, props);
    }

    final static String BACKUP = "backup";
    final static String RESTORE = "restore";

    // Return a list of lists describing commands implemented by this class:
    // [ [command-name, usage-message], ...]. If the usage message is null
    // or empty, then the command will not be listed by the "help" command.
    def static getCommandList()
    {
        return [
            [BACKUP, 'brm backup [backup-location]'],
            [RESTORE, 'brm restore backup-location'],
            ['runtarget', 
             'brm runtarget hostname targetname [backup-location]'],
            ];
    }

    /**
      * Set the BRM output directory from the command line.
      * @param  commandArray  Command-line parameters
      * @param  n  Number of parameters expected before the output directory
      * @return true if there was an output directory supplied, false otherwise
      */
    private boolean setOutputDirFromCmdLine(String[] commandArray, int n)
    {
        if (commandArray.length > n)
        {
            String backupLocation = commandArray[n];
            LOG.info("Setting BRM.output.dir from command-line: " + 
                    backupLocation);
            this.controlConfig.setProperty("BRM.output.dir", 
                    backupLocation);
            return true;
        }
        return false;
    }

    /**
     * Run a BRM backup, restore, or runtarget command.
     * @throws Exception if the command fails.
     */
    public void run(String[] commandArray) throws Exception
    {
        if (commandArray.length < 1)
        {
            StringBuffer msg = new StringBuffer();
            msg.append("Command not specified\n");
            msg.append('Use the "help" command to get the list of commands.');
            throw new BRMUserError(msg.toString());
        }

        this.timeStampStr = BRMUtil.generateTimeStamp();
        String command = commandArray[0];
        command = command.trim();
        command = command.toLowerCase();
        LOG.debug("About to execute command: " + command);
        PROGRESS.info("Running command: " + command);
        
        if (BACKUP.equals(command))
        {
            setOutputDirFromCmdLine(commandArray, 1);
            doBackup();
        }
        else if (RESTORE.equals(command))
        {
            if (!setOutputDirFromCmdLine(commandArray, 1))
            {
                // Must set output directory on the command line when doing
                // a restore. This avoids accidently picking up the backup
                // directory from brm.properties and destroying the previous
                // backup or getting an error message.
                def msg = "Missing backup directory location. The backup directory location must be specified on the command-line when running restore.";
                throw new BRMUserError(msg);
            }
            doRestore();
        }
        else if ("runtarget".equals(command))
        {
            if (commandArray.length < 3)
            {
                throw new BRMUserError(
                    "Missing hostname and target-name parameters");
            }
            String hostname = commandArray[1];
            String targetName = commandArray[2];
            setOutputDirFromCmdLine(commandArray, 3);
            runTarget(hostname, "", targetName);
            PROGRESS.info("Target " + targetName + " was executed");
        }
        else
        {
            StringBuffer msg = new StringBuffer();
            msg.append("Unknown command: ");
            msg.append(command);
            msg.append((char)'\n');
            msg.append('Use the "help" command to get the list of commands.');
            throw new BRMUserError(msg.toString());
        }
    }

    public static void main(String[] args)
        throws Exception
    {
        Thread.currentThread().setName("controller");
        com.sas.solutions.brm.logging.LogConfigurator.configure();
        args = BRMUtil.filterCmdLineProps(args);
        BRMController controller = null;
        int resultCode = 0;
        try
        {
            controller = new BRMController();
            controller.run(args);
        }
        catch (Throwable e)
        {
            if (e instanceof DefaultConfigUsedException)
                resultCode = 1;
            else if (e instanceof BRMUserError)
                resultCode = 2;
            else if (e instanceof BRMAntError)
                resultCode = 3;
            else
                resultCode = 4;
            if (resultCode < 4)
            {
                String msg = e.getMessage();
                if (msg == null)
                    msg = e.toString()
                LOG.error(msg);
                PROGRESS.error(msg);
            }
            else
            {
                LOG.error("Error while running BRM Controller", e);
                PROGRESS.error("${e}\nSee BRM.log for details");
            }
        }
        finally
        {
            if (controller != null)
                controller.close();
            controller = null;
        }
        System.exit(resultCode);
    }
}
