/*
 * Copyright (c) 2006-2010 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm;

import com.sas.solutions.brm.omr.GenericMetadataObject;
import com.sas.solutions.brm.omr.OMRConnection;
import com.sas.solutions.brm.util.BRMUtil;
import com.sas.solutions.brm.xml.StatusXMLParser;
 
public class StatusXMLUpdater
{
    private StatusXMLUpdater()
    {
        // a class with nothing but static utility methods
    }

    /**
      * Read a status.xml file the parser object containing the parsed
      * information from the status.xml file.
      * @param statusxmlfile  File object pointing to status.xml file
      */
    static def readStatusXML(File statusxmlfile)
    {
        return StatusXMLParser.parseXML(statusxmlfile);
    }

    static def getCodeToFQIDsMap(StatusXMLParser parser)
    {
        def codeFQIDsMap = [:];
        parser.configuredObjects.each
        { comp ->
            def code = comp.get('code');
            def fqid = comp.get('fqid');
            if (code == null && fqid == null && code == fqid)
                return;
            def fqidList = codeFQIDsMap.get(code);
            if (fqidList == null)
            {
                fqidList = [];
                codeFQIDsMap[code] = fqidList;
            }
            fqidList.add(fqid);
        }
        return codeFQIDsMap;
    }

    /**
     * Read the component:fqid info from the backed-up status.xml files.
     *
     * Each component code maps to a list of FQIDs. There will be more than
     * one fqid in the list if the same component name appears anywhere in
     * the backed-up status.xml files with different FQIDs.
     * The fqid list of each code found is stored in the BRMController
     * properties in a property named like this:
     *      BRM.code-fqids-map.[code]
     * where [code] is the product code found in the status.xml file, and
     * the square brackets are not part of the property name. The property
     * value is a comma-separated list of fqids. (Fortunately, an FQID
     * doesn't contain a comma, only letters, digits, and periods.)
     *
     * This is called by MainMigrator and from agentAnt.xml.
     *
     * @param backupConfigDir A File object pointing to a directory that has
     *      sub-directories, each of which contains a status.xml file. This
     *      directory must exist.
     * @param propSetter An object that has a setProperty(propName, propVal)
     *      method. Either BRMController or the Ant project object will do.
     * @param log A logging object with debug(msg), info(msg), warn(msg) and
     *      error(msg) methods.
     */
    static def getBackedUpComponentFQIDs(backupConfigDir, propSetter, log)
    {
        if (!backupConfigDir.isDirectory())
        {
            throw new Exception(
                        "Missing backup config directory: ${backupConfigDir}");
        }
        if (log != null)
            log.info(
              "Scanning backed-up status.xml files under: ${backupConfigDir}");
        def componentFQIDsMap = getBackedUpComponentFQIDsMap(backupConfigDir);
        componentFQIDsMap.each
        { code, fqids ->
            def propname = 'BRM.code-fqids-map.' + code;
            propSetter.setProperty(propname, fqids.join(','));
        }
    }

    /**
     * @param backupConfigDir a File object pointing to a directory
     * containing sub-directories, each of which holds a status.xml file.
     * @return A Map of [code: [fqid, ...]] as found in the status.xml files
     */
    static def getBackedUpComponentFQIDsMap(backupConfigDir)
    {
        def result = [:];
        String[] names = backupConfigDir.list();
        Arrays.sort(names);
        names.each
        { hostname ->
            def hostDir = new File(backupConfigDir, hostname);
            if (!hostDir.isDirectory())
                return;
            def statusXMLFile = new File(hostDir, 'status.xml');
            if (!statusXMLFile.isFile())
                return;
            def parser = readStatusXML(statusXMLFile);
            def codeFQIDsMap = getCodeToFQIDsMap(parser);
            codeFQIDsMap.each
            { code, fqids ->
                def prevList = result.get(code);
                if (prevList == null)
                {
                    prevList = [];
                    result[code] = prevList;
                }
                fqids.each
                { fqid ->
                    if (!(fqid in prevList))
                        prevList.add(fqid);
                }
            }
        }
        return result;
    }

    /**
     * Read the properties previously set by getBackedUpComponentFQIDs().
     * @return a Map of [code: [fqid, ...]]
     */
    static def getBackedUpComponentFQIDsFromProps(props)
    {
        def result = [:];
        props.each
        { propname, propval ->
            if (!propname.startsWith('BRM.code-fqids-map.'))
                return;
            def code = propname.substring('BRM.code-fqids-map.'.length());
            def fqids = propval.split(',', -1);
            result[code] = Arrays.asList(fqids);
        }
        return result;
    }

    // Groovy coding note: Due to a bug in Groovy 1.7.1, you must assign the
    // result of objs.findAll to a variable and then return it. You cannot
    // directly return objs.findAll {}

    // Filter out any objects not of type LogicalServer
    static def choose_LogicalServer = { conn, objs, log ->
        def result = objs.findAll
        { obj ->
            return (obj.getTypeName() == 'LogicalServer');
        }
        return result;
    }

    static def choose_ObjWithProductNamePrefix(prefix)
    {
        def closure = { conn, objs, log ->
            def result = objs.findAll
            { obj ->
                def productname = obj.getAttribute('ProductName');
                return (productname != null && 
                    productname.startsWith(prefix));
            }
            return result;
        }
        return closure;
    }

    /**
     * Choose the object with associated Public Configuration Properties
     * TextStore with os.localhost.host.name that matches the hostname
     * parameter. If hostname is null, match the local host.
     */
    static def choose_ObjOnHost(hostname)
    {
        def closure = { conn, objs, log ->
            def hostname_msg = hostname;
            if (hostname_msg == null)
                hostname_msg = '(localhost)';
            log.debug("Choosing object(s) linked with hostname: ${hostname_msg}");
            def result = objs.find
            { obj ->
                def notes = obj.getAssociation('Notes');
                if (notes == null)
                {
                    log.warn("Object has no Notes: ${obj}");
                    return false;
                }
                def pubconfig = notes.find
                {
                    return (it.getName() == 'Public Configuration Properties');
                }
                if (pubconfig == null)
                {
                    log.warn("Object has no Public Configuration Properties: ${obj}");
                    return false;
                }
                pubconfig = conn.getObject(pubconfig);
                def storedtext = pubconfig.getAttribute('StoredText');
                if (storedtext == null)
                {
                    log.warn("Public Configuration Properties has no StoredText: ${pubconfig}");
                    return false;
                }
                def props = new Properties();
                // Assuming that the properties text is in ISO 8859-1 encoding
                // as per http://download.oracle.com/javase/1.5.0/docs/api/java/util/Properties.html#load%28java.io.InputStream%29
                def bytes = storedtext.getBytes('latin1');
                def input = new ByteArrayInputStream(bytes);
                props.load(input);
                def props_hostname = props.get('os.localhost.host.name');
                log.debug("    ${obj.getTypeName()} Id=\"${obj.getId()}\" os.localhost.host.name=${props_hostname}");
                def props_hostfqdn = props.get('os.localhost.fqdn.host.name');
                log.debug("    ${obj.getTypeName()} Id=\"${obj.getId()}\" os.localhost.fqdn.host.name=${props_hostfqdn}");
                if (hostname != null)
                {
                    log.debug("    Matching with hostname=${hostname}");
                    if (props_hostname != null &&
                          BRMUtil.areHostsEquivalent(hostname, props_hostname))
                    {
                        log.debug("    Matches short name: ${props_hostname}");
                        return true;
                    }
                    if (props_hostfqdn != null &&
                          BRMUtil.areHostsEquivalent(hostname, props_hostfqdn))
                    {
                        log.debug("    Matches fqdn: ${props_hostfqdn}");
                        return true;
                    }
                    log.debug("    Does not match hostname or hostname fqdn");
                    return false;
                }
                log.debug("    Matching with local host");
                if (props_hostname != null &&
                        BRMUtil.isLocalHost(props_hostname))
                {
                    log.debug("    Host short name matches local host");
                    return true;
                }
                if (props_hostfqdn != null &&
                        BRMUtil.isLocalHost(props_hostfqdn))
                {
                    log.debug("    Host fqdn matches local host");
                    return true;
                }
                log.debug("    Does not match any local host name");
                return false;
            }
            return result;
        }
        return closure;
    }

    /**
     * Return a map of status.xml code to spec to find its
     * SoftwareComponent(s). The spec consists of a list:
     *
     *   [attribute_name, attribute_value, closure1, closure2, ...]
     *
     * The closures are optional. They are used successively to filter the
     * list returned by the search for objects matching the query
     * "@attribute_name=attribute_value".
     */
    static def getCodeLookupMap(properties)
    {
        // This is always null when not testing
        // null means this local machine
        def this_host = properties.get("BRM.test.host");

        def codeLookupMap = [
        // levconfig: special case: more than 1 per status.xml
        'levconfig': ['ClassIdentifier', 'SAS Application Server'],
        // workspacesrv: special case: more than 1 per status.xml
        'workspacesrv': ['ClassIdentifier', 
                '440196D4-90F0-11D0-9F41-00A024BB830C',
                choose_LogicalServer],
        // levdirroot: special case: 1 per status.xml, match by machine name
        'levdirroot': ['ClassIdentifier', 'levdirroot',
                choose_ObjOnHost(this_host)],
        'metadatasrv': ['ClassIdentifier', 
                '0217E202-B560-11DB-AD91-001083FF6836', 
                choose_LogicalServer],
        'globalroot': ['ClassIdentifier', 'globalroot'],
        'objectspawnr': ['ClassIdentifier', 'IOM Spawner'],
        'mgmtcondata': ['ClassIdentifier', 'mgmtcondata', 
                choose_ObjWithProductNamePrefix('Mgmt Console Server Data')],
        // 1 of these per tier?
        'dproserver': ['Name', 'Deployment Tester Server'],
        'dsbatchsrv': ['ClassIdentifier', 
                'e20e54b3-e5f4-47ed-b21d-03109bf9a4b8',
                choose_LogicalServer],
        'msofficedata': ['ClassIdentifier', 'msofficedata',
                choose_ObjWithProductNamePrefix('Add-In for Office Srvr Data')],
        'olapcubesrv': ['ClassIdentifier', 
                'f3f46472-1e31-11d5-87c2-00c04f38f9f6',
                choose_LogicalServer],
        'workspacpool': ['ClassIdentifier', 
                '620963ee-32bf-4128-bf5f-4b0df8ff90eb',
                choose_LogicalServer],
        'storprocsrv': ['ClassIdentifier', 
                '15931E31-667F-11D5-8804-00C04F35AC8C',
                choose_LogicalServer],
        'tableserver': ['ClassIdentifier', 
                'E803E04B-AAF0-4C74-86A6-F7008AA702DC',
                choose_LogicalServer],
        'connectsrv': ['ClassIdentifier', 'Connect Spawner'],
        'sharesrv': ['ClassIdentifier', 'SHARE'],
        'ossrvsched': ['ClassIdentifier', 
                'b379c9c1-89e0-4e72-ba43-6a416899b5dd',
                choose_LogicalServer],
        'biportlets': ['PublicType', 'Application',
                choose_ObjWithProductNamePrefix('BI Portlets')],
        // Enterprise Miner Client Midtier
        'dmineclntcfg': ['ClassIdentifier', 'dmineclient'],
        ];

        return codeLookupMap;
    }

    /**
     * Find a SoftwareComponent given its code from status.xml on this
     * machine. In the case of any ambiguity at all (multiple components in
     * metadata or in status.xml) return null. Also return null if no
     * matching component found.
     */
    static def findSingleComponentByCode(conn, code, log, properties)
    {
        log.info("    findSingleComponentByCode: Searching for SoftwareComponent with code=\"${code}\"");
        def codeFQIDMap = getBackedUpComponentFQIDsFromProps(properties);
        def fqids = codeFQIDMap.get(code);
        def obj = null;
        log.info("    findSingleComponentByCode: From backup: ${code} -> ${BRMUtil.stringArrayToString((String[])fqids)}");
        if (fqids != null && fqids.size() == 1)
        {
            def fqid = fqids[0];
            obj = conn.getObject('SoftwareComponent', fqid, false);
            if (obj != null)
            {
                log.info("    findSingleComponentByCode: Found ${obj.getTypeName()} with fqid=\"${fqid}\" in metadata");
                return obj;
            }
        }
        def objs = findComponentsByCode(conn, code, log, properties);
        if (objs.size() == 0)
        {
            log.info("    findSingleComponentByCode: No matches found in metadata");
            return null;
        }
        if (objs.size() > 1)
        {
            log.info("    findSingleComponentByCode: ${objs.size()} matches found in metadata, deferring handling of multiple results till later");
            return null;
        }
        obj = objs[0];
        log.info("    findSingleComponentByCode: Found ${obj.getTypeName()} with Id=\"${obj.getId()}\" in metadata");
        return obj;
    }

    /**
     * Find the software components in metadata that match this product
     * code (and this machine, for levdirroot).
     */
    static def findComponentsByCode(conn, code, log, properties)
    {
        log.info("    findComponentsByCode: Searching for SoftwareComponent with code=\"${code}\"");
        def codeLookupMap = getCodeLookupMap(properties);
        def spec = codeLookupMap.get(code);
        if (spec == null)
        {
            log.info("    findComponentsByCode: code: ${code}: not in the table");
            spec = ['ClassIdentifier', code];
        }
        def attrname = spec[0];
        def attrval = spec[1];
        def query = "@${attrname}='${attrval}'";
        def objs = [];
        conn.getRepositories().each
        { repos ->
            log.info("    findComponentsByCode: Searching ${repos.getName()} repository for SoftwareComponent objects with query: ${query}");
            conn.setRepository(repos.getName());
            objs.addAll(conn.doOMRQuery('SoftwareComponent', query, false));
        }
        conn.setFoundationRepository();
        if (objs.size() == 0)
        {
            log.info("    findComponentsByCode: No objects found matching query");
            return [];
        }
        // Filter the returned objects to make sure we get the right ones
        for (def i=2; i<spec.size(); ++i)
        {
            def closureNum = i - 1;
            def closureTot = spec.size() - 2;
            log.debug("    findComponentsByCode: filtering ${objs.size()} objects with closure #${closureNum} of ${closureTot}");
            def closure = spec[i];
            def result = closure(conn, objs, log);
            if (result == null)
            {
                log.debug("    findComponentsByCode: closure returned null");
                return [];
            }
            if (result instanceof GenericMetadataObject)
            {
                log.debug("    findComponentsByCode: closure returned a single object");
                objs = [result];
            }
            else
            {
                // result must be list of objects filtered by closure
                objs = result;
            }
            log.debug("    findComponentsByCode: closure #${closureNum} of ${closureTot} returned ${objs.size()} object(s)");
            if (objs.size() == 0)
                break
        }
        return objs;
    }

    /**
     * Synchronize the status.xml file with metadata, by attempting to look up
     * the new fqid given the product code.
     @deprecated syncStatusXMLByFQID() is the better method now.
     * @param properties  Result of calling Ant project.getProperties()
     * @param log         Result of calling BRMBuildListener.getLogger(project)
     * @param srcfile     File object, location of status.xml to read,
     *                    pass null for standard
     *                    ${BRM.config.dir}/Utilities/status.xml
     * @param destfile    File object, location to write modified status.xml,
     *                    pass null to modify srcfile in-place. Makes a
     *                    backup using BRMUtil.backupInSameDir().
     */
    static def syncStatusXMLByCode(properties, log, srcfile, destfile)
    {
        log.info("syncStatusXML: Synchronizing status.xml with metadata");

        def getpropOrDie = { propname ->
            def propval = properties.get(propname);
            if (propval == null)
                throw new Exception("Missing property: " + propname);
            return propval;
        }
        def host = getpropOrDie("BRM.host.metadata");
        def port = getpropOrDie("BRM.port.metadata");
        def user = getpropOrDie("BRM.admin.user");
        def password = getpropOrDie("BRM.admin.password");

        log.info("syncStatusXML: Connecting to metadata server at ${host}:${port} as ${user}");
        def conn = new OMRConnection(host, port, user, password);
        conn.connect();
        conn.setFoundationRepository();

        // Read file into memory
        if (srcfile == null)
        {
            def outputDir = new File(getpropOrDie("BRM.config.dir"));
            def utilitiesDir = new File(outputDir, "Utilities");
            srcfile = new File(utilitiesDir, 'status.xml');
        }
        log.info("syncStatusXML: Reading \"${srcfile}\"");
        def parser = readStatusXML(srcfile);

        def codeFQIDsMap = getCodeToFQIDsMap(parser);
        def duplicated_components = new LinkedHashSet();
        codeFQIDsMap.each
        { code, fqids ->
            if (fqids.size() > 1)
                duplicated_components.add(code);
        }

        // Return true if object has been fixed already, false otherwise
        // Side effect: save objid/code in map
        // True return value could mean error was logged
        def fixed_objid_code_map = [:];
        def checkForPrevFixedObjId = { obj, prev_fqid, code, order ->
            if (!fixed_objid_code_map.containsKey(obj.getId()))
            {
                fixed_objid_code_map[obj.getId()] = code;
                return false;
            }
            if (prev_fqid == null)
                prev_fqid_str = '';
            else
                prev_fqid_str = "original fqid=\"${prev_fqid}\"";
            def prev_code = fixed_objid_code_map[obj.getId()];
            if (prev_code == code)
            {
                log.info("syncStatusXML: Already fixed component fqid: code=\"${code}\" ${prev_fqid_str} order=\"${order}\" new fqid=${obj.getId()}");
            }
            else
            {
                log.error("syncStatusXML: Same output fqid mapped to two different codes: previous code=\"${prev_code}\" this code=\"${code}\" ${prev_fqid_str} order=\"${order}\" new fqid=${obj.getId()}");
            }
            return true;
        }

        def duplicated_component_orders = [:];
        def notFoundCount = 0;
        def component_fix_list = [];
        for (Iterator it=parser.configuredObjects.iterator(); it.hasNext(); )
        {
            def comp = it.next();
            def code = comp.get('code');
            def fqid = comp.get('fqid');
            def order = comp.get('order');
            if (code == null)
            {
                log.debug("Skipping row with missing code");
                continue;
            }
            if (fqid == null)
            {
                log.debug("Skipping row with code=\"${code}\", missing fqid");
                continue;
            }
            if (code == fqid)
            {
                log.info("syncStatusXML: Skipping row as \"${fqid}\" is not a real fqid");
                continue;
            }
            log.info("syncStatusXML: Synchronizing row in status.xml: code=\"${code}\" original fqid=\"${fqid}\" order=\"${order}\"");
            if (code in duplicated_components)
            {
                // Remove duplicate components now, add them back later
                log.info("syncStatusXML: Component code appears multiple times in status.xml: ${code}");
                if (order != null && 
                    !duplicated_component_orders.containsKey(code))
                {
                    duplicated_component_orders[code] = order;
                }
                it.remove();
                continue;
            }
            // Try simple method of looking up component
            def obj = findSingleComponentByCode(conn, code, log, properties);
            if (obj == null)
            {
                component_fix_list.add(comp);
                it.remove();
                continue;
            }
            // Update this component's fqid
            checkForPrevFixedObjId(obj, fqid, code, order);
            comp['fqid'] = obj.getId();
            log.info("syncStatusXML: Synchronized component fqid: code=\"${code}\" original fqid=\"${fqid}\" order=\"${order}\" new fqid=${obj.getId()}");
        }

        def addComponentToStatusXML = { obj, code, order ->
            def comp = new LinkedHashMap();
            def fqid = obj.getId();
            comp['code'] = code;
            comp['fqid'] = fqid;
            if (order != null)
                comp['order'] = order;
            if (checkForPrevFixedObjId(obj, null, code, order))
                return;
            parser.configuredObjects.add(comp);
        }

        duplicated_components.each
        { code ->
            log.info("syncStatusXML: Resolving code that apears multiple times: ${code}");
            def objs = findComponentsByCode(conn, code, log, properties);
            if (objs.size() == 0)
            {
                log.error("syncStatusXML: Could not find component in metadata when there were multiple in status.xml file: code=\"${code}\"");
                notFoundCount += 1;
                return;
            }
            def order = duplicated_component_orders.get(code);
            objs.each
            { obj ->
                addComponentToStatusXML(obj, code, order);
                log.info("syncStatusXML: Fixed component fqid: code=\"${code}\" order=\"${order}\" new fqid=${obj.getId()}");
            }
        }

        component_fix_list.each
        { comp ->
            def code = comp.get('code');
            def fqid = comp.get('fqid');
            def order = comp.get('order');
            log.info("syncStatusXML: Fixing component that could not be found earlier: code=\"${code}\" original fqid=\"${fqid}\" order=\"${order}\"");
            def objs = findComponentsByCode(conn, code, log, properties);
            if (objs.size() == 0)
            {
                log.error("syncStatusXML: Could not find component in metadata: code=\"${code}\" original fqid=\"${fqid}\" order=\"${order}\"");
                notFoundCount += 1;
                parser.configuredObjects.add(comp);
                return;
            }
            objs.each
            { obj ->
                addComponentToStatusXML(obj, code, order);
                log.info("syncStatusXML: Fixed component fqid: code=\"${code}\" original fqid=\"${fqid}\" order=\"${order}\" new fqid=${obj.getId()}");
            }
        }

        if (notFoundCount > 0)
            log.error("syncStatusXML: ${notFoundCount} component(s) not found");
        log.info("syncStatusXML: Done with metadata connection.");
        conn.close();

        // Write new status.xml
        if (destfile == null)
        {
            destfile = srcfile;
        }
        if (destfile.isFile())
        {
            def backfile = BRMUtil.backupInSameDir(destfile);
            log.info("syncStatusXML: Backed up \"${destfile}\" to \"${backfile}\"");
        }
        log.info("syncStatusXML: Writing results out to \"${destfile}\"");
        parser.writeXML(destfile);
    }

    /**
     * Read the old-fqid:new-fqid info from the objidmap.txt file generated
     * by ConfigObjLoader.load().
     *
     * The fqid mapping is stored in the BRMController properties like this:
     *      BRM.fqids-map.[old-fqid] = new-fqid
     * where [old-fqid] is fqid (metadata ID) of a configuration object as
     * stored by saveconfig, and new-fqid is the metadata ID of that same
     * object as re-created by loadconfig. The square brackets are not part of
     * the property name.
     *
     * This is called by MainMigrator and from agentAnt.xml.
     *
     * @param tempDir A File object pointing to a directory in which saveconfig
     *      and loadconfig stored information.
     * @param propSetter An object that has a setProperty(propName, propVal)
     *      method. Either BRMController or the Ant project object will do.
     * @param log A logging object with debug(msg), info(msg), warn(msg) and
     *      error(msg) methods.
     */
    static def getConfigObjIdMap(tempDir, propSetter, log)
    {
        def tempConfigDir = new File(tempDir, 'config');
        def configObjsDir = new File(tempConfigDir, 'configobjs');
        def objidmapFile = new File(configObjsDir, 'objidmap.txt');
        def objidmapReader = objidmapFile.newReader('UTF-8');
        def lineNum = 0;
        while (true)
        {
            def line = objidmapReader.readLine();
            if (line == null)
                break;
            lineNum += 1;
            def i = line.indexOf('=');
            if (i <= 0 || i >= line.length() - 1)
            {
                throw new Error("Error in format of line ${lineNum} of ${objidmapFile}");
            }
            def oldId = line.substring(0, i);
            def newId = line.substring(i + 1);
            def propname = 'BRM.fqids-map.' + oldId;
            propSetter.setProperty(propname, newId);
        }
    }

    /**
     * Read the properties previously set by getConfigObjIdMap().
     * @return a Map of [oldId: newId]
     */
    static def getObjIdMapFromProps(props)
    {
        def result = [:];
        def prefix = 'BRM.fqids-map.';
        props.each
        { propname, propval ->
            if (!propname.startsWith(prefix))
                return;
            def oldId = propname.substring(prefix.length());
            def newId = propval;
            result[oldId] = newId;
        }
        return result;
    }

    /**
     * Synchronize the status.xml file with metadata, using the object ID map
     * generated by loadconfig.
     * @param properties  Result of calling Ant project.getProperties()
     * @param log         Result of calling BRMBuildListener.getLogger(project)
     * @param srcfile     File object, location of status.xml to read,
     *                    pass null for standard
     *                    ${BRM.config.dir}/Utilities/status.xml
     * @param destfile    File object, location to write modified status.xml,
     *                    pass null to modify srcfile in-place. Makes a
     *                    backup using BRMUtil.backupInSameDir().
     */
    static def syncStatusXMLByFQID(properties, log, srcfile, destfile)
    {
        log.info("syncStatusXML: Synchronizing status.xml with metadata");

        def getpropOrDie = { propname ->
            def propval = properties.get(propname);
            if (propval == null)
                throw new Exception("Missing property: " + propname);
            return propval;
        }

        // Read file into memory
        if (srcfile == null)
        {
            def outputDir = new File(getpropOrDie("BRM.config.dir"));
            def utilitiesDir = new File(outputDir, "Utilities");
            srcfile = new File(utilitiesDir, 'status.xml');
        }
        log.info("syncStatusXML: Reading \"${srcfile}\"");
        def parser = readStatusXML(srcfile);

        def objIdMap = getObjIdMapFromProps(properties);

        def notFoundCount = 0;
        def foundCount = 0;
        for (Iterator it=parser.configuredObjects.iterator(); it.hasNext(); )
        {
            def comp = it.next();
            def code = comp.get('code');
            def fqid = comp.get('fqid');
            def order = comp.get('order');
            if (code == null)
            {
                log.debug("Skipping row with missing code");
                continue;
            }
            if (fqid == null)
            {
                log.debug("Skipping row with code=\"${code}\", missing fqid");
                continue;
            }
            if (code == fqid)
            {
                log.info("syncStatusXML: Skipping row as \"${fqid}\" is not a real fqid");
                continue;
            }

            def newId = objIdMap.get(fqid);
            if (newId == null)
            {
                notFoundCount += 1;
                log.warn("Synchronizing row in status.xml: code=\"${code}\" original fqid=\"${fqid}\" order=\"${order}\": no matching ID found for fqid in object ID map");
                continue;
            }
            foundCount += 1;

            // Update this component's fqid
            comp['fqid'] = newId;
            log.info("syncStatusXML: Synchronized component fqid: code=\"${code}\" original fqid=\"${fqid}\" order=\"${order}\" new fqid=${newId}");
        }

        if (notFoundCount > 0)
        {
            def msg = "syncStatusXML: ${notFoundCount} fqid(s) not found";
            if (notFoundCount >= foundCount)
                log.error(msg);
            else
                log.warn(msg);
        }
        else
        {
            log.info("Updated ${foundCount} FQIDs successfully.");
        }

        // Write new status.xml
        if (destfile == null)
        {
            destfile = srcfile;
        }
        if (destfile.isFile())
        {
            def backfile = BRMUtil.backupInSameDir(destfile);
            log.info("syncStatusXML: Backed up \"${destfile}\" to \"${backfile}\"");
        }
        log.info("syncStatusXML: Writing results out to \"${destfile}\"");
        parser.writeXML(destfile);
    }

    // Note the different in capitalization between the name of this method
    // and the getpropOrDie closure used in some other methods.
    static def getPropOrDie(properties, propname)
    {
        def propval = properties.get(propname);
        if (propval == null)
            throw new Exception("Missing property: " + propname);
        return propval;
    }

    /**
     * Restore the status.xml file from a BRM backup directory.
         This only works if we are restoring on the same machines on which
         the backup was taken, and the host names in the brm.properties file
         are correct and match the host names used when the backup was
         taken.
     * @param properties  Result of calling Ant project.getProperties()
     * @param log         Result of calling BRMBuildListener.getLogger(project)
     * @return true if the file was restored Ok.
     */
    static def restoreStatusXML(properties, log)
    {
        log.info("restoreStatusXML: Restoring backed up status.xml file");

        def getpropOrDie = { propname ->
        }
        def brmBackupDir = getPropOrDie(properties, 'BRM.output.dir');
        def brmAgentHost = getPropOrDie(properties, 'BRM.agent.host');
        def backupConfigDir = new File(brmBackupDir, 'config');
        def backupHostDir = new File(backupConfigDir, brmAgentHost);
        if (!backupHostDir.isDirectory())
        {
            log.warn("restoreStatusXML: Can't restore status.xml file on host ${brmAgentHost}: Missing backup directory: ${backupHostDir}");
            return false;
        }
        def bakStatusXMLFile = new File(backupHostDir, 'status.xml');
        if (!bakStatusXMLFile.isFile())
        {
            log.warn("restoreStatusXML: Can't restore status.xml file on host ${brmAgentHost}: Missing backed-up status.xml file: ${bakStatusXMLFile}");
            return false;
        }
        def statusXMLFile = getStatusXMLFile(properties);
        if (!statusXMLFile.isFile())
        {
            log.warn("restoreStatusXML: Can't restore status.xml file on host ${brmAgentHost}: No status.xml file deployed: ${statusXMLFile}");
            return false;
        }
        def bakBakStatusXMLFile = BRMUtil.backupInSameDir(statusXMLFile);
        log.info("restoreStatusXML: Copied \"${statusXMLFile}\" to \"${bakBakStatusXMLFile}\"");
        log.info("restoreStatusXML: Copying \"${bakStatusXMLFile}\" to \"${statusXMLFile}\"");
        statusXMLFile.setBytes(bakStatusXMLFile.readBytes());
        return true;
    }

    static def getStatusXMLFile(properties)
    {
        // e.g. C:/SAS/Config/Lev1
        def configDir = getPropOrDie(properties, 'BRM.config.dir');
        def levUtilitiesDir = new File(configDir, 'Utilities');
        return new File(levUtilitiesDir, 'status.xml');
    }
}

// end-of-file
