package com.sas.solutions.brm.dav;

import java.util.Iterator;
import java.util.List;

import com.sas.metadata.remote.MdOMIUtil;
import com.sas.metadata.remote.MetadataObjects;
import com.sas.report.repository.AdapterFactory;
import com.sas.report.repository.EntryInterface;
import com.sas.services.information.Filter;
import com.sas.services.information.metadata.MetadataInterface;
import com.sas.services.information.RepositoryInterface;
import com.sas.services.user.UserContextInterface;
import com.sas.solutions.brm.errors.BRMConfigurationException;
import com.sas.solutions.brm.omr.OMRConnection;

/**
 * Iterate over all of the BI Dashboard indicators and update those that
 * contain SAS Strategy Management providers.
 */
class UpdateStMBIDProviderUrl
{
    boolean DEBUG = true;

    def log = null;
    def props = null;
    def newhost = null;
    
    public UpdateStMBIDProviderUrl(log, props)
    {
        this.log = log;
        this.props = props;
        this.newhost = props.get('BRM.host.webapp');
        if (this.newhost == null)
        {
            throw new Exception("Missing property: BRM.host.webapp");
        }
        this.log.info(
            "The new main web application host is \"${this.newhost}\"");
    }

    def getOMRCredentials()
    {
        def getprop = { propname ->
            def propval = props.get(propname);
            if (propval == null || propval.length() == 0)
            {
                throw new BRMConfigurationException(
                    "Missing property: " + propname);
            }
            return propval;
        }
        def host = getprop("BRM.host.metadata");
        def port = getprop("BRM.port.metadata");
        def user = getprop("BRM.admin.user");
        def password = getprop("BRM.admin.password");
        return ['host': host, 'port': port, 'user': user, 
                'password': password];
    }

    
    public void updateAll()
    {
        log.info("Updating host names in DAV for BI Dashboard providers");
        def creds = getOMRCredentials();

        // Get the name of the Foundation repository
        def omr = new OMRConnection(creds.host, creds.port, creds.user, 
                                    creds.password);
        log.debug("Connecting to ${creds.host}:${creds.port} as ${creds.user}");
        omr.connect();
        omr.setFoundationRepository();
        String repositoryName = omr.getRepositoryName();
        omr.close();

        // Get a usercontext.
        SDWPlatformServices services = new SDWPlatformServices(this.log);
        services.startLocalPlatformServices(creds.host, creds.port);
        UserContextInterface userContext = services.newUser(creds.user, 
                                                            creds.password);

        if (userContext == null)
        { 
            def msg = "Unable to migrate documents. Problem with user context.";
            logError(msg);
            throw new RuntimeException(msg);
        }

        // locate the providers in metadata.
        Filter filter = new Filter();
        filter.setType(MetadataObjects.TRANSFORMATION);
        filter.setOMRFlags(MdOMIUtil.OMI_MATCH_CASE);
        //filter.setXMLSelect("<XMLSELECT search=\"Transformation[@PublicType='IndicatorData']\"/>");
        filter.setExplicitSearchString("omi", "<XMLSELECT search=\"Transformation[@PublicType='IndicatorData']\"/>");
        RepositoryInterface repositoryInterface = userContext.getAuthServer().getRepository(repositoryName);
        List transformations = repositoryInterface.search(filter); 
        if (transformations == null || transformations.size() == 0)
        {
            log.info("No Transformations found");
            return;
        }
        log.info("Found ${transformations.size()} transformation(s)");
        
        // Go through the indicators from metadata.
        Iterator iter = transformations.iterator();
        while (iter.hasNext())
        {
            // Do mysterious BID stuff to get an input stream.
            MetadataInterface transformation = (MetadataInterface)iter.next();
            log.info("Found transformation: ${transformation}");
            log.debug("Transformation class: ${transformation.getClass()}");
            log.debug("transformation.getType()=${transformation.getType()}");
            Object serviceObject = transformation.newServiceObject();
            log.debug("serviceObject=${serviceObject}");
            if (serviceObject == null)
            {
                log.warn("Unable to get ServiceObject for transformation \"${transformation}\": Missing jar files?");
                continue;
            }
            EntryInterface entry = AdapterFactory.adaptEntry(serviceObject);
            log.debug("entry=${entry}");
            InputStream istr = entry.getInputStream();

            // Get the string value from the object
            byte[] readData = new byte[istr.available()];
            istr.read(readData);
            istr.close();
            log.debug("Transformation content: ${readData.length} bytes");
            String origContent = new String(readData, "UTF-8");
            if (origContent.length() == 0)
            {
                log.info("Transformation content is empty");
                continue;
            }
            
            // Now operate on the string making necessary changes to the server
            // url.  Modify the server name if it's an StM provider.
            String modContent = fixUpXML(origContent);

            // Persist the change if there was a change.
            if (!origContent.equals(modContent))
            {
                info("modifying provider: " + transformation.getPathUrl());
                break; // XXX TESTING Do not modify DAV
                OutputStream ostr = entry.getOutputStream();
                byte[] writeData = modContent.getBytes("UTF-8");
                ostr.write(writeData);
                ostr.close();
            }
        }
    }

    private String fixUpString(String content, String oldUrl, String newUrl) {

    // fix up the string.
    String newContent = content;
    int p = content.indexOf("<ModelProvider>");
    if (p >= 0) {
        int pp = content.indexOf("</", p);
        String providerType = content.substring(p+15, pp);
        if (providerType.equalsIgnoreCase("spm2")) {
            // this is an StM provider, check url.
            p = content.indexOf("authenticationServiceURL=\"");
            if (p >= 0) {
                pp = content.indexOf("\"", p+26);
                String url = content.substring(p+26, pp);

                if (url != null && url.equals(oldUrl)) {
                    newContent = content.substring(0, p+26) + newUrl + content.substring(pp); 
                }
            }
        }
    }
    return newContent;
   }

    private String fixUpXML(String content) {

    // fix up the hostname in the string.
    String newContent = content;
    int p = content.indexOf("<ModelProvider>");
    if (p >= 0) {
        log.debug("fixUpXML: found ModelProvider");
        int pp = content.indexOf("</", p);
        String providerType = content.substring(p+15, pp);
        if (providerType.equalsIgnoreCase("spm2")) {
            log.debug("fixUpXML: found spm2 provider");
            // this is an StM provider, check url.
            p = content.indexOf("authenticationServiceURL=\"");
            if (p >= 0) {
                log.debug("fixUpXML: found authenticationServiceURL");
                pp = content.indexOf("\"", p+26);
                String url = content.substring(p+26, pp);
                log.debug("fixUpXML: found URL: ${url}");
                String newUrl = fixUpURL(url);
                if (url != null && !url.equals(newUrl)) {
                    newContent = content.substring(0, p+26) + newUrl + content.substring(pp); 
                }
            }
        }
    }
    return newContent;
   }

    // Expecting URL in the form: 
    // http://default.host.com:7201/SASStrategyManagement
    def fixUpURL(url)
    {
        if (url == null)
            return null;
        def p = url.indexOf('//');
        if (p < 0)
        {
            // Malformed URL
            return url;
        }
        p += 2;
        // Find colon or slash, whichever is first
        def q1 = url.indexOf(':', p);
        def q2 = url.indexOf('/', p);
        def q = 0;
        if (q1 >= 0 && (q2 < 0 || q1 < q2))
        {
            q = q1;
        }
        else if (q2 >= 0)
        {
            q = q2;
        }
        else
        {
            q = url.length();
        }
        def oldhost = url.substring(p, q);
        log.debug("fixUpURL: found host in URL: ${oldhost}");
        return url; // XXX temporary - avoid changing DAV while debugging
        log.info(
            "Changing host \"${oldhost}\" to \"${this.newhost}\" in URL \"${url}\"");
        return url.substring(0, p) + this.newhost + url.substring(q);
    }
    
    
   private void debug(String s) {
    this.log.debug(s);
   }
     
   private void info(String s) {
        this.log.info(s);
   }
   
   private void warn(String s) {
        this.log.warn(s);
   }
   
   private void error(String s) {
        this.log.error(s);
   }

}

// end-of-file
