/*
 * Created Fri 04 Jun 2010
 * Copyright (c) 2010 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm.logging;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.Priority;
import org.apache.log4j.spi.LoggingEvent;

/**
  * Logger that logs to console and to file, counting errors and warnings.
  * INFO and lower priority messages are not echoed to the console.
  * Stack traces are also not printed to the console.
  */
public class MonitoredLogger extends Logger
{
    def consoleLogger = null;
    def fileLogger = null;
    int messages = 0;
    int warnings = 0;
    int errors = 0;

    /**
     * @param consoleLogger Log to which WARN and higher messages will be
     * sent, can be null
     * @param fileLogger Log to which all messages will be sent. Cannot be
     * null.
     */
    public MonitoredLogger(consoleLogger, fileLogger)
    {
        super("");

        // Override protected repository field so that
        // respository.isDisabled() always returns false
        this.repository = new BRMLoggerRepository();

        this.consoleLogger = consoleLogger;
        this.fileLogger = fileLogger;
    }

    /**
     * Delegate setLevel(logLevel) to the fileLogger. The consoleLogger
     * level is left alone.
     */
    public void setLevel(Level logLevel)
    {
        this.fileLogger.setLevel(logLevel);
    }

    /**
     * Delegate getEffectiveLevel() to the fileLogger.
     */
    public Level getEffectiveLevel()
    {
        return this.fileLogger.getEffectiveLevel();
    }

    /**
     * Delegate isEnabledFor(logLevel) to the fileLogger.
     */
    public boolean isEnabledFor(Level logLevel)
    {
        return this.fileLogger.isEnabledFor(logLevel);
    }

    public int getNumErrorMessages()
    {
        return this.errors;
    }

    public int getNumWarningMessages()
    {
        return this.warnings;
    }

    public int getNumMessages()
    {
        return this.messages;
    }

    public void trace(Object msg)
    {
        log(Level.TRACE, msg);
    }

    public void trace(Object msg, Throwable e)
    {
        log(Level.TRACE, msg, e);
    }

    public void debug(Object msg)
    {
        log(Level.DEBUG, msg);
    }

    public void debug(Object msg, Throwable e)
    {
        log(Level.DEBUG, msg, e);
    }

    public void info(Object msg)
    {
        log(Level.INFO, msg);
    }

    public void info(Object msg, Throwable e)
    {
        log(Level.INFO, msg, e);
    }

    public void warn(Object msg)
    {
        log(Level.WARN, msg);
    }

    public void warn(Object msg, Throwable e)
    {
        log(Level.WARN, msg, e);
    }

    public void error(Object msg)
    {
        log(Level.ERROR, msg);
    }

    public void error(Object msg, Throwable e)
    {
        log(Level.ERROR, msg, e);
    }

    public void fatal(Object msg)
    {
        log(Level.FATAL, msg);
    }

    public void fatal(Object msg, Throwable e)
    {
        log(Level.FATAL, msg, e);
    }

    public void log(Priority priority, Object message)
    {
        log(priority, message, null);
    }

    public void log(Priority priority, Object message, Throwable t)
    {
        this.fileLogger.log(priority, message, t);
        this.messages += 1;
        if (priority.isGreaterOrEqual(Level.WARN))
        {
            if (this.consoleLogger != null)
                this.consoleLogger.log(priority, message, null);
            if (priority.isGreaterOrEqual(Level.ERROR))
            {
                this.errors += 1;
            }
            else
            {
                this.warnings += 1;
            }
        }
    }

    /**
     * Pass a LoggingEvent on to the console and file loggers, in a similar
     * manner to the log(Priority, Object, Throwable) method.
     *
     * This is called by RemoteLoggingServer, because it has a LoggingEvent
     * object that has just been deserialized from an ObjectInputStream. Any
     * attempt to call the regular log methods on the fileLogger would lose
     * serialized exception information.
     */
    public void logEvent(LoggingEvent event)
    {
        Level eventLevel = event.getLevel();
        if (this.fileLogger.isEnabledFor(eventLevel))
            this.fileLogger.callAppenders(event);
        this.messages += 1;
        if (eventLevel.isGreaterOrEqual(Level.WARN))
        {
            if (this.consoleLogger != null)
            {
                // We call log(Level, Object) on the console logger because
                // we don't want exception stack traces to be printed to the
                // console.
                this.consoleLogger.log(eventLevel, event.getMessage());
            }
            if (eventLevel.isGreaterOrEqual(Level.ERROR))
            {
                this.errors += 1;
            }
            else
            {
                this.warnings += 1;
            }
        }
    }

    /**
     * Override Category.callAppenders().
     * If this method is called, something went wrong with overriding the
     * base class.
     */
    public void callAppenders(LoggingEvent event)
    {
        assert false;
    }
}
