/*
 * Created on July 21, 2011
 * Copyright (c) 2011 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm.migrate;

import com.sas.solutions.brm.errors.BRMConfigurationException;
import com.sas.solutions.brm.omr.GenericMetadataObject;
import com.sas.solutions.brm.omr.OMRUtil;

class ConfigObjLoader extends ConfigObjVisitorBase
{
    def configFileObj = null; // MigrateConfig(loadconfig.ini)
    def objIdMap = new HashMap(); // Map XML obj ID -> new obj ID
    def reverseObjIdMap = new HashMap(); // Map new obj ID -> XML obj ID
    def loadParentStack = new LinkedList(); // [ XML obj ID, ... ]

    // ['objTypeName:Name': [xmlObjId, ...] ]
    def nativeNameIdsMap = new LinkedHashMap();

    // [nativeReposPrefix: nativeReposName]
    def nativePrefixReposMap = new LinkedHashMap();

    // [nativeReposName: nativeBIPServiceID]
    def nativeReposBIPServiceIDMap = new LinkedHashMap();

    // ['objTypeName:Name': [foreignObjId, ...] ]
    def foreignNameIdsMap = new LinkedHashMap();

    // [foreignReposPrefix: foreignReposName]
    def foreignPrefixReposMap = new LinkedHashMap();

    // [nativeBIPServiceID: foreignBIPServiceID]
    def bipServiceIDMap = new LinkedHashMap();

    // These associations are required to create objects that have them
    def requiredAssocs = [
        'NamedComponent',
        'OwningObject',
        'OwningPrototype',
        'Source',
        ];

    // If true, keep going despite warnings in checkObjectNameMatches()
    boolean force = false;

    public ConfigObjLoader(omr, tempDir, logger, force)
    {
        super(omr, tempDir, logger);
        this.force = force;
    }

    def loadConfigIni()
    {
        def loadconfigIniFile = getLoadconfigIniFile();
        if (loadconfigIniFile.isFile())
        {
            this.logger.info("Loading object name mapping file: ${loadconfigIniFile}");
            this.configFileObj = new MigrateConfig(loadconfigIniFile);
        }
        else
        {
            this.logger.info("No object name mapping file, looking up objects by name as they are: ${loadconfigIniFile}");
        }
    }

    def loadNativePrefixReposMap()
    {
        def tempConfigDir = new File(this.tempDir, 'config');
        def reposPrefixMapFile = new File(tempConfigDir, 
                                          'repos-prefix-map.properties');
        this.logger.info("Loading repository prefix map from file: ${reposPrefixMapFile}");
        def reposPrefixMapIn = new FileInputStream(reposPrefixMapFile);
        def reposPrefixMap = new Properties();
        reposPrefixMap.load(reposPrefixMapIn);
        reposPrefixMapIn.close();
        reposPrefixMap.each
        { prefix, name ->
            this.nativePrefixReposMap[prefix] = name;
        }
    }

    def loadBIPServiceIDs()
    {
        def bipserviceidsFile = new File(this.configObjsDir, 
                                         'bipserviceids.txt');
        if (!bipserviceidsFile.isFile())
        {
            this.logger.warn("Missing BIP Service IDs file: ${bipserviceidsFile}");
            return;
        }
        this.logger.info("Loading native BIP Service IDs from ${bipserviceidsFile}");
        bipserviceidsFile.eachLine('UTF-8')
        { line ->
            def nativeBIPServiceID = line.trim();
            def obj = newMinimalObj('SoftwareComponent', nativeBIPServiceID);
            def prefix = obj.getIdPrefix();
            def reposName = this.nativePrefixReposMap.get(prefix);
            this.bipServiceIDMap[nativeBIPServiceID] = null;
            this.nativeReposBIPServiceIDMap[reposName] = nativeBIPServiceID;
        }
    }

    def setupForeignPrefixReposMap()
    {
        this.omr.getRepositories().each
        { repos ->
            // Note: The ID suffix of the repository object is the same as the
            // ID prefix of ordinary objects. This is the repository
            // identifier.
            this.foreignPrefixReposMap[repos.getIdSuffix()] = repos.getName();
        }
    }

    int countXMLFiles()
    {
        if (!this.configObjsDir.isDirectory())
            return 0;
        int numXMLFiles = 0;
        this.configObjsDir.eachDir { subDir ->
            subDir.eachFile(groovy.io.FileType.FILES) { curFile ->
                if (!curFile.getName().endsWith('.xml'))
                    return;
                numXMLFiles += 1;
            }
        }
        return numXMLFiles;
    }

    boolean checkDumpIntegrity()
    {
        def dumpCompleteFile = new File(this.configObjsDir, 'dumpcomplete.txt');
        if (!dumpCompleteFile.exists())
        {
            this.logger.error("Can't load configuration objects; saveconfig did not complete successfully.");
            return false;
        }
        def configobjtypesFile = new File(this.configObjsDir, 
                                          'configobjtypes.txt');
        if (!configobjtypesFile.exists())
        {
            this.logger.error("Can't load configuration objects; saveconfig did not save configuration object types.");
            return false;
        }
        def configobjpathsFile = new File(this.configObjsDir, 
                                          'configobjpaths.txt');
        if (!configobjpathsFile.exists())
        {
            this.logger.error("Can't load configuration objects; saveconfig did not save configuration object paths.");
            return false;
        }
        int numXMLFiles = countXMLFiles();
        this.logger.info("Found ${numXMLFiles} dumped XML files in directory ${this.configObjsDir}");
        def expectedMinNumXMLFiles = 100;
        if (numXMLFiles < expectedMinNumXMLFiles)
        {
            this.logger.error("Less than ${expectedMinNumXMLFiles} XML files produced by saveconfig. Aborting loading objects.");
            return false;
        }
        def objidmapFile = new File(this.configObjsDir, 'objidmap.txt');
        if (objidmapFile.isFile())
        {
            this.logger.error("brm loadconfig has already been run. You must re-run brm saveconfig before running brm loadconfig again.");
            return false;
        }
        return true;
    }

    def splitTypeNameKey(String key)
    {
        int i = key.indexOf(':');
        if (i < 0 || i >= key.length() + 1)
            throw new IllegalArgumentException('Bad key: ' + key);
        return [
            'typename': key.substring(0, i),
            'name': key.substring(i + 1),
            ];
    }

    def makeTypeNameKey(String objTypeName, String name)
    {
        return objTypeName + ':' + name;
    }

    def getLoadconfigIniFile()
    {
        def installDir = System.getProperty('BRM.install.dir');
        if (!installDir)
        {
            throw new BRMConfigurationException('Missing system property: BRM.install.dir');
        }
        def configDir = new File(installDir, 'config');
        return new File(configDir, 'loadconfig.ini');
    }

    def splitNameValue(String line)
    {
        int i = line.indexOf('=');
        if (i < 0)
            return null;
        return [
            'name': line.substring(0, i).trim(),
            'value': line.substring(i + 1).trim(),
            ];
    }

    def scanForeignConfigObjs()
    {
        this.logger.info("Scanning foreign configuration-related metadata objects");
        def deleteMap = new LinkedHashMap(); // [ objid: objTypeName ]

        def scheduleObjToBeDeleted = { obj ->
            def objTypeName = obj.getTypeName();
            if (isReusableObject(obj))
            {
                this.logger.debug("Not deleting objects like this: ${objTypeName} Id=${obj.getId()} Name=\"${obj.getName()}\"");
                return;
            }
            if (hasContentAssoc(obj))
            {
                this.logger.debug("Foreign object with content association: ${objTypeName} \"${obj.getName()}\" Id=${obj.getId()}");
                def key = makeTypeNameKey(objTypeName, obj.getName());
                def foreignObjIds = this.foreignNameIdsMap.get(key);
                if (foreignObjIds == null)
                {
                    foreignObjIds = [obj.getId()];
                    this.foreignNameIdsMap[key] = foreignObjIds;
                }
                else
                {
                    foreignObjIds.add(obj.getId());
                }
                return;
            }
            deleteMap.put(obj.getId(), objTypeName);
        }

        def visitFolder = { obj, path ->
            this.logger.debug("Not deleting folder Tree: Id=${obj.getId()} Name=\"${obj.getName()}\" path=\"${path}\"");
            return;
        }

        def handleBIPService = { obj ->
            this.logger.info("Saving ID of foreign BIP Service SoftwareComponent, Id=${obj.getId()}");
            def reposName = this.foreignPrefixReposMap.get(obj.getIdPrefix());
            def nativeBIPServiceID = this.nativeReposBIPServiceIDMap.get(
                                                                    reposName);
            if (nativeBIPServiceID)
            {
                this.logger.info("Mapping native BIP Service SoftwareComponent Id=${nativeBIPServiceID} to foreign BIP Service SoftwareComponent Id=${obj.getId()}");
                this.bipServiceIDMap[nativeBIPServiceID] = obj.getId();
            }
            def foundation = this.omr.getFoundationRepository();
            if (foundation.getIdSuffix() == obj.getIdPrefix())
            {
                // This is the BIP Service SoftwareComponent in the
                // Foundation repository - use as default mapping
                this.logger.info("Foreign BIP Service SoftwareComponent Id=${obj.getId()} is in Foundation repository");
                // Must copy keys before iterating because map will be changed
                this.bipServiceIDMap.keySet().toArray().each
                {
                    nativeBIPServiceID = it;
                    if (!this.bipServiceIDMap[nativeBIPServiceID])
                    {
                        this.logger.debug("Mapping native BIP Service SoftwareComponent Id=${nativeBIPServiceID} to default foreign BIP Service SoftwareComponent in Foundation repository");
                        this.bipServiceIDMap[nativeBIPServiceID] = obj.getId();
                    }
                }
            }
        }

        visitConfigObjs([
            'visitConfigObj': scheduleObjToBeDeleted,
            'visitFolder': visitFolder,
            'visitBIPService': handleBIPService,
            ]);
        return deleteMap;
    }

    /**
    @return [ 'objTypeName:Name': [objid, ...] ]
    */
    def loadConfigObjsByNameFile()
    {
        File configobjsbynameFile = null;
        // Keep both names for backwards compatibility
        def filenames = ['configobjsbyname.txt', 'contentobjsbyname.txt'];
        def found = filenames.find
        { name ->
            configobjsbynameFile = new File(this.configObjsDir, name);
            if (configobjsbynameFile.isFile())
                return true;
        };
        if (!found)
        {
            throw new Exception("Missing name index file: " +
                new File(this.configObjsDir, filenames[0]));
        }
        def reader = configobjsbynameFile.newReader('UTF-8');
        def lineNum = 0;
        def result = new LinkedHashMap();
        while (true)
        {
            def line = reader.readLine();
            if (line == null)
                break;
            lineNum += 1;
            def record = splitNameValue(line);
            if (record == null)
            {
                throw new BRMConfigurationException(
                    'Format error in file ' + configobjsbynameFile +
                    'on line ' + lineNum);
            }
            def objids = null;
            if (result.containsKey(record.name))
            {
                objids = result[record.name];
            }
            else
            {
                objids = [];
                result[record.name] = objids;
            }
            objids.add(record.value);
        }
        reader.close();
        return result;
    }

    def createLoadconfigIniFile(List unmatchedContentObjs)
    {
        File loadconfigIniFile = getLoadconfigIniFile();
        def typeMap = new LinkedHashMap();
        unmatchedContentObjs.each
        { record ->
            def objTypeName = record.typename;
            def nameList = typeMap.get(objTypeName);
            if (nameList == null)
            {
                nameList = new LinkedHashSet();
                typeMap.put(objTypeName, nameList);
            }
            nameList.add(record.name);
        }
        def writer = loadconfigIniFile.newWriter('UTF-8');
        def out = new PrintWriter(writer);
        try
        {
            out.println("; loadconfig.ini");
            out.println("; This file helps brm loadconfig look up objects by name");
            out.println("; when the names on the source system are different than on the local system.");
            out.println(";");
            out.println("; File format:");
            out.println("; [ObjectTypeName]");
            out.println("; Object name from source machine=Local object name");
            out.println(";");
            out.println("; For each line below that ends with '=', append the name of the matching ");
            out.println("; object from the local machine.");
            out.println();
            typeMap.each
            { objTypeName, nameList ->
                out.println("[${objTypeName}]");
                nameList.each
                { name ->
                    out.println("${name}=");
                }
                out.println();
            }
        }
        finally
        {
            out.close();
        }
        return loadconfigIniFile;
    }

    def translateForeignToNativeName(objTypeName, foreignName)
    {
        if (this.configFileObj == null)
            return foreignName;
        def nativeName = this.configFileObj.getValue(objTypeName, foreignName);
        if (nativeName == null || nativeName.trim().length() == 0)
            return foreignName;
        if (foreignName != nativeName)
        {
            this.logger.info("Translating ${objTypeName} name from foreign \"${foreignName}\" to native \"${nativeName}\"");
        }
        return nativeName;
    }

    /**
    * Scan the objects with special associations and make sure they can be
    * looked up by name. This includes loading and/or creating the
    * loadconfig.ini file.
    * @return true if all names are matched successfully, false otherwise.
    */
    def checkObjectNameMatches()
    {
        this.nativeNameIdsMap = loadConfigObjsByNameFile();
        def unmatchedContentObjs = [];
        this.foreignNameIdsMap.each
        { foreignKey, foreignObjIds ->
            def keyobj = splitTypeNameKey(foreignKey);
            keyobj.foreignObjIds = foreignObjIds;
            def objTypeName = keyobj.typename;
            def foreignName = keyobj.name;
            def nativeName = translateForeignToNativeName(objTypeName, 
                                                           foreignName);
            def nativeKey = makeTypeNameKey(objTypeName, nativeName);
            def xmlObjIds = this.nativeNameIdsMap.get(nativeKey);
            if (xmlObjIds)
            {
                this.logger.debug("checkObjectNameMatches: nativeNameIdsMap: \"${nativeKey}\" -> ${xmlObjIds}, foreignNameIdsMap: \"${foreignKey}\" -> ${foreignObjIds}");
            }
            else
            {
                unmatchedContentObjs.add(keyobj);
            }
        }
        if (unmatchedContentObjs)
        {
            unmatchedContentObjs.each
            { record ->
                this.logger.warn("Unmatched object name: ${record.typename} \"${record.name}\" Ids=${record.foreignObjIds}");
            }
            this.logger.warn("Could not match ${unmatchedContentObjs.size()} object(s) by name.");
            if (this.force)
            {
                this.logger.warn("Continuing loading configuration objects despite warnings, because \"force\" option was specified.");
                return true;
            }
            this.logger.warn("To ignore these warnings (not recommended), re-run with the force option: brm loadconfig force");
            if (this.configFileObj == null)
            {
                def loadconfigIniFile = createLoadconfigIniFile(
                                                        unmatchedContentObjs);
                this.logger.warn("Created new template file at: ${loadconfigIniFile}");
                this.logger.warn("Edit this file and then re-run brm loadconfig to enable BRM to find the needed objects by name.");
            }
            else
            {
                this.logger.warn("To enable BRM to find unmatched objects, update this file and re-run brm loadconfig: ${this.configFileObj.iniFile}");
            }
            return false;
        }
        return true;
    }

    /**
    Delete a list of metadata objects, handling exceptions.
    @param deleteMap  [ objid: objTypeName ]
    */
    def deleteObjects(deleteMap)
    {
        def deletedobjsFile = new File(this.configObjsDir, 'deletedobjs.txt');
        // Open file in append mode
        def writer = deletedobjsFile.newWriter('UTF-8', true);
        def deleteLog = new PrintWriter(writer);
        def deleteObject = { obj ->
            obj = this.omr.getObject(obj);
            if (obj == null)
                return true; // Object was already gone
            def updateAssocs = false;
            obj.getAssociationNames().each
            { assocname ->
                if (assocname in this.requiredAssocs)
                    return;
                if (obj.getAssociation(assocname))
                {
                    obj.setAssociation(assocname, [])
                    updateAssocs = true;
                }
            }
            if (updateAssocs)
            {
                try
                {
                    this.omr.store(obj);
                }
                catch (Throwable e1)
                {
                    this.logger.warn("Error while updating associations on ${obj.getTypeName} \"${obj.getName()}\" Id=${obj.getId()} before deleting it: ${e1}");
                }
            }
            try
            {
                this.omr.delete(obj);
                return true;
            }
            catch (Throwable e2)
            {
                this.logger.warn("Could not delete object: ${obj}: ${e2}");
            }
            return false;
        }
        deleteMap.each
        { objid, objTypeName ->
            def obj = newMinimalObj(objTypeName, objid);
            if (deleteObject(obj))
                deleteLog.println("${obj.getTypeName()}:${obj.getId()}");
        }
        deleteLog.println(); // Blank line to separate batches
        deleteLog.close();
        this.logger.info("Deleted ${deleteMap.size()} objects");
    }

    /**
    * Find TextStore objects named "Public Configuration Properties" or
    * "Private Configuration Properties" that have no associations.
    * @return a map of all such orphan TextStore objects found:
    *       [objid: objTypeName] (objTypeName will be 'TextStore')
    */
    def findOrphanTextStores()
    {
        this.logger.info("Finding orphan TextStore objects named \"Public Configuration Properties\" or \"Private Configuration Properties\"");
        def deleteMap = new LinkedHashMap();
        def textstores = this.omr.doOMRQuery('TextStore', "*[@Name='Public Configuration Properties' OR @Name='Private Configuration Properties']", true);
        textstores.each
        { textstore ->
            def numassocs = 0;
            textstore.getAssociationNames().each
            { assocname ->
                def objs = textstore.getAssociation(assocname);
                numassocs += objs.size();
            }
            if (numassocs > 0)
                return;
            deleteMap.put(textstore.getId(), textstore.getTypeName());
        }
        return deleteMap;
    }

    // Load map of XML object IDs to paths of folder Tree objects
    def loadFolderPaths()
    {
        def configobjpathsFile = new File(this.configObjsDir, 
                                          'configobjpaths.txt');
        def r = configobjpathsFile.newReader('UTF-8');
        try
        {
            int lineNum = 0;
            while (true)
            {
                def line = r.readLine();
                if (line == null)
                    break;
                lineNum += 1;
                def record = splitNameValue(line);
                if (record == null)
                {
                    throw new BRMConfigurationException(
                        'Format error in file ' + configobjpathsFile +
                        'on line ' + lineNum);
                }
                def objid = record.name;
                def path = record.value;
                def obj = findTreeByPath(path);
                if (obj == null)
                    obj = createTreeWithPath(path);
                this.objIdMap.put(objid, obj.getId());
            }
        }
        finally
        {
            r.close();
        }
    }

    def getObjFile(String objTypeName, String objid)
    {
        if (!objid)
        {
            throw new IllegalArgumentException("Invalid object ID: " + objid);
        }
        def objTypeDir = new File(this.configObjsDir, objTypeName);
        return new File(objTypeDir, objid + '.xml');
    }

    def getObjIdFromFile(File xmlfile)
    {
        def name = xmlfile.getName();
        def i = name.lastIndexOf('.');
        if (i < 0)
        {
            this.logger.error("File name is missing extension: ${xmlfile}");
            return null;
        }
        if (name.substring(i).toLowerCase() != '.xml')
        {
            this.logger.error("File name does not end in .xml: ${xmlfile}");
            return null;
        }
        return name.substring(0, i);
    }

    // Given a list of objects, return the ones that are "foreign" objects, or
    // an empty list if there are none. In this context, a foreign object is
    // one whose ID is not a key in reverseObjIdMap.
    def filterForForeignObjs(objs)
    {
        if (!objs)
            return [];
        def result = [];
        objs.each
        { obj ->
            if (!this.reverseObjIdMap.containsKey(obj.getId()))
            {
                result.add(obj);
            }
        }
        return result;
    }

    // Find a "foreign" object whose type is a key in reusableObjectTypes,
    // using as search criteria the attribute names from reusableObjectTypes
    // and the attribute values from obj.
    //
    // Basically, this searches for objects of type Identity, Person,
    // Permission, etc. which we do not want to create, but we expect to find
    // in the content moved from the foreign system.
    def getMatchingForeignObj(obj)
    {
        def attrnames = this.reusableObjectTypes.get(obj.getTypeName());
        if (!attrnames)
        {
            this.logger.warn("Wrong type of object to be looked up by attributes: ${obj.getTypeName()} ID=${obj.getId()} \"${obj.getName()}\"");
            return null;
        }
        int i = 0;
        def query = new StringBuffer();
        attrnames.each
        { attrname ->
            if (i > 0)
                query.append(' AND ');
            def attrval = obj.getAttribute(attrname);
            if (attrval == null)
                return;
            query.append('@' + attrname + "='" + attrval + "'");
            i += 1;
        }
        query = query.toString();
        if (!query)
        {
            this.logger.warn("No attributes to look up this object: ${obj.getTypeName()} ID=${obj.getId()} \"${obj.getName()}\"");
            return null;
        }
        def objs = this.omr.doOMRQuery(obj.getTypeName(), query, true);
        def result = filterForForeignObjs(objs);
        if (!result)
        {
            // Put the warning message here so we can log the query string
            this.logger.warn("No foreign object with type ${obj.getTypeName()} matching criteria \"${query}\"");
            return null;
        }
        if (result.size() > 1)
            this.logger.warn("Multiple (${result.size()}) foreign objects with type ${obj.getTypeName()} matching criteria \"${query}\"");
        return result[0];
    }

    /**
    Put the associations on a new metadata object.
    @param nativeObj  The metadata object that was loaded from an XML file
    @param newObj The new metadata object on which we are putting associations
    from nativeObj.
    @param assocs  List of association names to copy from nativeObj to newObj
    */
    def findAndSetAssocs(nativeObj, newObj, assocnames)
    {
        if (!assocnames)
            return;
        assocnames.each
        { assocname ->
            if (this.doNotFollowAssocs.contains(assocname))
            {
                // This association will be set from the other end
                return;
            }
            def newassocobjs = findAssocObjs(nativeObj, assocname);
            newObj.setAssociation(assocname, newassocobjs);
        }
    }

    /**
    @param nativeObj  The metadata object that was loaded from an XML file
    @param assocname  Association name
    @return List of loaded or located objects in the association
    */
    def findAssocObjs(nativeObj, assocname)
    {
        if (isContentAssoc(nativeObj, assocname))
        {
            // Handle this association in moveContentObjs()
            return [];
        }
        def assocobjs = nativeObj.getAssociation(assocname);
        def newassocobjs = assocobjs.collect
        { assocobj ->
            def assocObjId = assocobj.getId();
            def newAssocObjId = this.objIdMap.get(assocObjId);
            if (newAssocObjId != null)
                return newMinimalObj(assocobj.getTypeName(), newAssocObjId);
            if (assocObjId in this.loadParentStack)
            {
                //this.logger.debug("Not putting object with XML Id=${assocObjId} in association ${assocname} because it is a parent object currently being loaded.");
                return null;
            }
            if (this.bipServiceIDMap.containsKey(assocObjId))
            {
                newAssocObjId = this.bipServiceIDMap[assocObjId];
                if (newAssocObjId != null)
                {
                    this.objIdMap.put(assocObjId, newAssocObjId);
                    return newMinimalObj('SoftwareComponent', newAssocObjId);
                }
            }
            def assocObjFile = getObjFile(assocobj.getTypeName(), assocObjId);
            def newAssocObj = loadObjFromFile(nativeObj.getId(), assocObjFile);
            newAssocObjId = newAssocObj.getId();
            if (newAssocObj != null)
                this.logger.debug("Loading associated object with XML Id=${assocObjId} from file returned ${newAssocObj.getTypeName()} \"${newAssocObj.getName()}\" Id=${newAssocObjId}");
            else
                this.logger.debug("Loading associated object with XML Id=${assocObjId} from file returned null");
            return newAssocObj;
        }
        // Due to a bug in Groovy, I can't directly return newassocobjs.findAll
        newassocobjs = newassocobjs.findAll { it != null; };
        return newassocobjs;
    }

    /**
    / Load a metadata object from an XML file.
    // If the object has not been loaded before, create it in metadata
    //   and recursively load or find its associated objects, and
    //   return the newly-created metadata object.
    // If the object has been loaded before, then return a
    //   GenericMetadataObject with the correct type and ID.
    @param parentId  The "native" (from XML file) ID of the object that is
    loading this object via an association.
    @param xmlfile   The XML file to load.
    */
    def loadObjFromFile(String parentId, File xmlfile)
    {
        this.logger.debug("Loading metadata object from XML file: ${xmlfile}");
        this.loadParentStack.add(parentId);
        try
        {
            if (!xmlfile.isFile())
            {
                this.logger.warn("No such metadata object file: ${xmlfile}");
                return null;
            }
            def nativeObj = OMRUtil.loadObjectFromFile(xmlfile);
            def nativeID = nativeObj.getId();
            def newID = this.objIdMap.get(nativeID);
            if (newID != null)
            {
                return newMinimalObj(nativeObj.getTypeName(), newID);
            }
            def newObj = null;
            if (isReusableObject(nativeObj))
            {
                newObj = getMatchingForeignObj(nativeObj);
                if (newObj == null)
                {
                    // Log message already emitted by getMatchingForeignObj()
                    return null;
                }
                newID = newObj.getId();
                this.objIdMap.put(nativeID, newID);
                return newObj;
            }
            newObj = OMRUtil.stripAssocs(nativeObj);
            findAndSetAssocs(nativeObj, newObj, this.requiredAssocs);
            try
            {
                newObj = this.omr.create(newObj);
            }
            catch (Throwable e)
            {
                this.logger.error("Could not create new object: ${newObj} from XML object: ${nativeObj}");
                throw e;
            }
            newID = newObj.getId();
            this.objIdMap.put(nativeID, newID);
            this.reverseObjIdMap.put(newID, nativeID);
            findAndSetAssocs(nativeObj, newObj, 
                             nativeObj.getAssociationNames());
            // Update associations
            if (newObj.isModified())
                this.omr.store(newObj);
            return newObj;
        }
        finally
        {
            this.loadParentStack.removeLast();
        }
    }

    def loadNativeConfigObjs()
    {
        this.logger.info("Loading configuration-related metadata objects from directory ${this.configObjsDir}");
        loadFolderPaths();
        def objidmapFile = new File(this.configObjsDir, 'objidmap.txt');
        def objidmapPrinter = objidmapFile.newPrintWriter('UTF-8');
        this.configObjTypes.each
        { objTypeName ->
            def objTypeDir = new File(this.configObjsDir, objTypeName);
            if (!objTypeDir.exists())
                return;
            objTypeDir.eachFile(groovy.io.FileType.FILES) { curFile ->
                def objid = getObjIdFromFile(curFile);
                def xmlfile = getObjFile(objTypeName, objid);
                def obj = loadObjFromFile(null, xmlfile);
                if (obj != null)
                {
                    objidmapPrinter.println("${objid}=${obj.getId()}");
                }
            }
        }
        objidmapPrinter.close();
    }

    /**
    * Move all objects in content-related associations from the foreign
    * configuration objects to the native ones.
    * @return a map of foreign objects to be deleted, no longer needed because
    * their associated objects have been transferred. [ objid: objTypeName ]
    */
    def moveContentObjs()
    {
        def deleteMap = new LinkedHashMap(); // [ objid: objTypeName ]
        this.foreignNameIdsMap.each
        { foreignKey, foreignObjIds ->
            def keyobj = splitTypeNameKey(foreignKey);
            keyobj.foreignObjIds = foreignObjIds;
            def objTypeName = keyobj.typename;
            def foreignName = keyobj.name;
            def nativeName = translateForeignToNativeName(objTypeName,
                                                          foreignName);
            def nativeKey = makeTypeNameKey(objTypeName, nativeName);
            def xmlObjIds = this.nativeNameIdsMap.get(nativeKey);
            if (xmlObjIds == null)
                xmlObjIds = [];
            if (foreignObjIds.size() == 0 && xmlObjIds.size() == 0)
                return;
            if (foreignObjIds.size() == xmlObjIds.size())
            {
                deleteMap.putAll(
                    moveContentBetweenObjLists(objTypeName, 
                                        foreignName, nativeName,
                                        foreignObjIds, xmlObjIds));
                return;
            }
            def msg = ("Can't map ${foreignObjIds.size()} foreign ${objTypeName} object(s) named \"${foreignName}\" to ${xmlObjIds.size()} native ${objTypeName} object(s) named \"${nativeName}\". Some metadata objects will not be transferred.");
            if (this.force)
                this.logger.warn(msg);
            else
                this.logger.error(msg);
        }
        return deleteMap;
    }

    /**
    * Utility method used by moveContentObjs().
    * @return a map of foreign objects to be deleted, no longer needed because
    * their associated objects have been transferred. [ objid: objTypeName ]
    */
    def moveContentBetweenObjLists(objTypeName, foreignName, nativeName,
                               foreignObjIds, xmlObjIds)
    {
        assert (foreignObjIds.size() == xmlObjIds.size());
        assert (foreignObjIds.size() > 0);
        if (foreignObjIds.size() > 1)
        {
            this.logger.warn("Moving associated objects from ${foreignObjIds.size()} foreign ${objTypeName} objects named \"${foreignName}\" to ${xmlObjIds.size()} native ${objTypeName} objects named \"${nativeName}\". Because the mapping is not precise, object destinations will be selected arbitrarily.");
        }
        def deleteMap = new LinkedHashMap(); // [ objid: objTypeName ]
        for (def i=0; i<foreignObjIds.size(); ++i)
        {
            def foreignObjId = foreignObjIds[i];
            def xmlObjId = xmlObjIds[i];
            moveContentBetween2Objs(objTypeName, foreignName, nativeName,
                                   foreignObjId, xmlObjId);
            deleteMap.put(foreignObjId, objTypeName);
        }
        return deleteMap;
    }

    /**
    * Utility method used by moveContentBetweenObjLists().
    */
    def moveContentBetween2Objs(objTypeName, foreignName, nativeName,
                                foreignObjId, xmlObjId)
    {
        def nativeObjId = this.objIdMap.get(xmlObjId);
        if (nativeObjId == null)
        {
            this.logger.error("moveContentBetween2Objs: Missing native ${objTypeName} \"${nativeName}\" XML Id=${xmlObjId}");
            return;
        }
        def foreignObj = this.omr.getObject(objTypeName, foreignObjId, true);
        if (foreignObj == null)
        {
            this.logger.error("moveContentBetween2Objs: Missing foreign ${objTypeName} \"${foreignName}\" Id=${foreignObjId}");
            return;
        }
        def nativeObj = this.omr.getObject(objTypeName, nativeObjId, true);
        if (nativeObj == null)
        {
            this.logger.error("moveContentBetween2Objs: Missing native ${objTypeName} \"${nativeName}\" Id=${nativeObjId}");
            return;
        }
        foreignObj.getAssociationNames().each
        { assocName ->
            if (! isContentAssoc(foreignObj, assocName))
                return;
            def assocObjs = foreignObj.getAssociation(assocName);
            this.logger.info("Moving ${assocObjs.size()} object(s) in association ${assocName} from foreign ${objTypeName} \"${foreignName}\" Id=${foreignObjId} to native ${objTypeName} \"${nativeName}\" Id=${nativeObjId}");
            foreignObj.setAssociation(assocName, []);
            nativeObj.setAssociation(assocName, assocObjs);
        }
        // Must empty the associations on the foreign object first or
        // setting the associations on the native object might fail.
        this.omr.store(foreignObj);
        this.omr.store(nativeObj);
    }

    def fixIdsInTextStores()
    {
        this.logger.info("Finding and fixing metadata IDs in TextStores");

        int numChangedLines = 0;
        int numChangedObjs = 0;
        def processLine = 
        { obj, line ->
            def metadataIds = getMetadataIDsFromLine(line);
            if (!metadataIds)
                return line;
            def i = line.indexOf('=');
            if (i < 0)
            {
                if (obj.getName() != 'SourceCode')
                {
                    this.logger.warn("Metadata ID(s) ${metadataIds} found in ${obj.getTypeName()} \"${obj.getName()}\" (${obj.getId()}) but no property name found in line: ${line}");
                }
                return line;
            }
            def propname = line.substring(0, i);
            def propval = line.substring(i + 1);
            def newline;
            if (propname == 'spmprefxpoll.dependson.webinfpltfm.fqid')
            {
                def target = getWebInfPlatformObj();
                if (target == null)
                {
                    // getWebInfPlatformObj() already reported reason not found
                    this.logger.warn("Could not check or fix ID in ${obj.getTypeName()} \"${obj.getName()}\" (${obj.getId()}): Could not find object for line: ${line}");
                    return line;
                }
                if (propval == target.getId())
                {
                    this.logger.info("${obj.getTypeName()} \"${obj.getName()}\" (${obj.getId()}): No change needed to line: ${line}");
                    return line;
                }
                newline = (propname + '=' + target.getId());
                this.logger.info("${obj.getTypeName()} \"${obj.getName()}\" (${obj.getId()}): changed line \"${line}\" to \"${newline}\"");
                numChangedLines += 1;
                return newline;
            }
            // Generic case: Look up target ID
            def nativeID = propval.trim();
            def newID = this.objIdMap.get(nativeID);
            if (newID != null && 
                metadataIds.size() == 1 && metadataIds[0] == nativeID)
            {
                newline = (propname + '=' + newID);
                this.logger.info("${obj.getTypeName()} \"${obj.getName()}\" (${obj.getId()}): changed line \"${line}\" to \"${newline}\"");
                numChangedLines += 1;
                return newline;
            }
            if (!(obj.getName() in ['SourceCode', 'WRS_Hints']))
            {
                this.logger.warn("Metadata ID(s) ${metadataIds} found in ${obj.getTypeName()} \"${obj.getName()}\" (${obj.getId()}) but property name ${propname} is unknown: ${line}");
            }
            return line;
        }

        this.omr.setFoundationRepository();
        def objs = this.omr.doOMRQuery('TextStore', "", true);
        objs.each
        { obj ->
            def storedtext = obj.getAttribute('StoredText');
            if (!storedtext)
                return;
            def lines = storedtext.split('\\n', -1);
            def result = [];
            lines.each
            { line ->
                result.add(processLine(obj, line));
            }
            def storedtextAfter = result.join('\n');
            if ((storedtextAfter != storedtext))
            {
                numChangedObjs += 1;
                obj.setAttribute('StoredText', storedtextAfter);
                omr.store(obj);
            }
        }
        this.logger.info("${numChangedLines} changed line(s)");
        this.logger.info("${numChangedObjs} changed object(s)");
    }

    def static getMetadataIDsFromLine(line)
    {
        def p = ~/\b[0-9A-Z]{8}\.[0-9A-Z]{8}\b/;
        def m = p.matcher(line);
        def result = [];
        while (m.find())
        {
            result.add(m.group());
        }
        return result;
    }

    def getWebInfPlatformObj()
    {
        def query = "@Name=:'Web Infrastructure Platform'";
        def objs = this.omr.doOMRQuery('SoftwareComponent', query, true);
        if (!objs)
        {
            this.logger.warn("Could not find Web Infrastructure Platform Software Component with query: ${query}");
            return null;
        }
        if (objs.size() > 1)
        {
            this.logger.warn("Found ${objs.size()} Infrastructure Platform Software Components with query ${query}, expected only one");
            return null;
        }
        return objs[0];
    }

    /**
    @return true if brm loadconfig should continue, false otherwise.
    */
    boolean load()
    {
        this.omr.setFoundationRepository();
        loadConfigIni();
        loadNativePrefixReposMap();
        loadBIPServiceIDs();
        setupForeignPrefixReposMap();
        if (!checkDumpIntegrity())
            return false;
        def deleteMap = scanForeignConfigObjs();
        if (!checkObjectNameMatches())
            return false;
        deleteMap.putAll( findOrphanTextStores() );
        loadNativeConfigObjs();
        deleteMap.putAll( moveContentObjs() );
        deleteObjects(deleteMap);
        fixIdsInTextStores();
        return true;
    }
}
