/*
 * Created on July 8, 2011
 * Copyright (c) 2011 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm.migrate;

import com.sas.solutions.brm.omr.GenericMetadataObject;

class ConfigObjVisitorBase extends OMRVisitorBase
{
    def tempDir = null;
    def logger = null;
    //////////
    def configObjsDir = null;

    List configObjTypes = [
        'SoftwareComponent',
        'DeployedComponent',
        'ConfiguredComponent',
        'ServerComponent',
        'ServiceComponent',
        'LogicalServer',
        'ServerContext',
        'Machine',
        ];

    Map reusableObjectTypes = [
        'AuthenticationDomain': ['Name'],
        'Identity': ['Name'],
        'IdentityGroup': ['Name'],
        'Person': ['Name'],
        'Login': ['UserID'],
        'ResponsibleParty': ['Name'],
        'AccessControlTemplate': ['Name'], 
        'Permission': ['Name', 'Type'],
        'ContentType': ['Name'],
        'Prototype': ['Name'],
        ];

    Set contentObjectTypes = new LinkedHashSet([
        'Column',
        'DatabaseSchema',
        'PhysicalTable',
        'SASLibrary',
        'Transformation',
        'UniqueKey',
        ]);

    Set contentAssocs = new LinkedHashSet([
        'DataPackages', // to PhysicalTable, Directory, etc.
        'Libraries',    // to SASLibrary
        'ComputeTasks', // to ClassifierMap
        'SpecSourceTransformations', // to JFJob
        'SpecTargetTransformations', // to JFJob
        'InitProcesses', // from ServerComponent to TransformationStep
        'ImplementedObjects', // from DeployedComponent to ClassifierMap
        'SourceTransformations', // From PropertySet to Transformation
        ]);

    Set doNotFollowAssocs = new LinkedHashSet([
        'TypedProperties',
        'TypedAttProperties',
        ]);

    Set contentBearingConfigObjTypes = new LinkedHashSet([
        'SoftwareComponent',
        'DeployedComponent',
        'ConfiguredComponent',
        'ServiceComponent',
        'ServerComponent',
        'LogicalServer',
        'ServerContext',
        'SASClientConnection',
        ]);

    Set scannedObjSet = new HashSet(); // set of OBJID
    int numVisitedObjs = 0;
    int numVisitedFolders = 0;

    public ConfigObjVisitorBase(omr, tempDir, logger)
    {
        super(omr);
        this.tempDir = tempDir;
        this.logger = logger;
        //////////
        def tempConfigDir = new File(this.tempDir, 'config');
        this.configObjsDir = new File(tempConfigDir, 'configobjs');
    }

    def static newMinimalObj(String typeName, String objid)
    {
        def newObj = new GenericMetadataObject();
        newObj.setTypeName(typeName);
        newObj.setId(objid);
        return newObj;
    }

    /**
    A "reusable object" is an object that we expect to find also in the 
    foreign metadata. When dumping resuable objects, associations are not
    followed.  When loading reusable objects, the attribute names in the
    reusableObjectTypes map are used to search for an equivalent object 
    already in metadata. Note: Not all objects whose types are in the
    reusableObjectTypes map are considered terminal objects. For example, 
    some Prototype objects should be excluded.
    @param obj The object to check. The associations should be populated so
    they can be checked.
    @return true iff this should be considered a "reusable" object.
    */
    def isReusableObject(obj)
    {
        def objTypeName = obj.getTypeName();
        if (!this.reusableObjectTypes.containsKey(objTypeName))
        {
            return false;
        }
        if (objTypeName == 'Prototype')
        {
            if (obj.getAssociation('PrimaryPropertyGroup'))
            {
                this.logger.debug("This Prototype object is NOT terminal: Id=${obj.getId()} \"${obj.getName()}\"");
                return false;
            }
            else
            {
                this.logger.debug("This Prototype object is terminal: Id=${obj.getId()} \"${obj.getName()}\"");
                return true;
            }
        }
        return true;
    }

    def hasContentAssoc(obj)
    {
        def contentAssoc = obj.getAssociationNames().find
        { assocname ->
            return isContentAssoc(obj, assocname);
        }
        return (contentAssoc != null);
    }

    /**
    * Determine if an association on an object is a content association.
    * @param obj The metadata object. This object must have a valid type
    * name and associations, but need not have a valid Id.
    * @param assocname The association name
    * @return true iff this association on this object contains
    * content-related objects (objects that can be moved between
    * installations).
    */
    def isContentAssoc(obj, assocname)
    {
        if (! (assocname in this.contentAssocs))
            return false;
        def assocobjs = obj.getAssociation(assocname);
        if (!assocobjs)
            return false;
        // If the associated objects are all DataSourceName objects, then
        // it is configuration, not content.
        def dsns = assocobjs.findAll
            { return (it.getTypeName() == 'DataSourceName'); };
        if (dsns.size() == assocobjs.size())
            return false;
        return true;
    }

    /**
    * Scan and visit configuration-related objects recursively. "Visit" means
    * to call a closure with the object as the parameter.
    * Exceptions:
    *   Don't visit a SoftwareComponent named "BIP Service" or with
    *       PublicType='RootFolder'.
    *   Dont visit objects whose types are in contentObjectTypes
    *   Don't follow associations on a true folder Tree object
    *   Don't follow associations from objects where isReusableObject() 
    *       returns true
    *   Don't follow associations that are not content associations.
    *       visited BIP Service folder Tree objects.
    * @param handlerMap Map of name to callback closure objects. At most one
    *       handler will be called per object, depending on the object:
    *       visitConfigObj(obj): Called for regular configuration-related
    *                            objects
    *       visitFolder(obj, path): Called for BIP Service folder Tree objects
    *       visitBIPService(obj): Called for the "BIP Service" 
    *                             SoftwareComponent
    */
    def visitConfigObjs(Map handlerMap)
    {
        this.logger.info("Visiting configuration-related metadata objects");
        this.scannedObjSet.clear();
        numVisitedObjs = 0;
        def queue = new LinkedList(); // [ [parent, assocName, obj], ... ]
        def scanObj = { parent, assocName, obj ->
            if (this.scannedObjSet.contains(obj.getId()))
                return;
            this.scannedObjSet.add(obj.getId());
            def objTypeName = obj.getTypeName();
            if (contentObjectTypes.contains(objTypeName))
            {
                def parentInfo;
                if (parent != null)
                    parentInfo = "${parent.getTypeName()} Id=${parent.getId()}";
                else
                    parentInfo = "(none)";
                this.logger.debug("Skipping content-related object: Parent: ${parentInfo} assocName=${assocName} ${objTypeName} Id=${obj.getId()}");
                return;
            }
            // Fetch all of the attributes and associations
            def objid = obj.getId(); // getObject() might lose the object ID
            obj = getObject(obj);
            if (obj == null)
            {
                this.logger.warn("getObject() returned null for ${objTypeName} ID=${objid}");
                return;
            }
            if (objTypeName == 'SoftwareComponent' &&
                (obj.getName() == 'BIP Service' ||
                 obj.getAttribute('PublicType') == 'RootFolder'))
            {
                handlerMap.visitBIPService(obj);
                return;
            }
            def path = null;
            if (objTypeName == 'Tree')
            {
                path = getFolderPathOfTree(obj);
            }
            if (path == null)
            {
                this.logger.debug("Visiting ${obj.getTypeName()} Id=${obj.getId()} Name=\"${obj.getName()}\"");
                handlerMap.visitConfigObj(obj);
                numVisitedObjs += 1;
            }
            else
            {
                this.logger.debug("Visiting folder ${obj.getTypeName()} Id=${obj.getId()} Name=\"${obj.getName()}\" path=\"${path}\"");
                handlerMap.visitFolder(obj, path);
                numVisitedFolders += 1;
                // Don't follow associations on folder Tree objects
                return;
            }
            if (isReusableObject(obj))
            {
                // Don't follow associations on these objects
                return;
            }
            obj.getAssociationNames().each
            { childAssocName ->
                if (isContentAssoc(obj, childAssocName))
                    return;
                if (this.doNotFollowAssocs.contains(childAssocName))
                    return;
                obj.getAssociation(childAssocName).each
                { assocobj ->
                    queue.add([newMinimalObj(objTypeName, obj.getId()), 
                               childAssocName, assocobj]);
                }
            }
        }
        this.configObjTypes.each
        { objTypeName ->
            this.omr.getRepositories().each
            { repos ->
                this.logger.info(
                    "Scanning objects of type ${objTypeName} in repository ${repos.getName()}");
                this.omr.setRepository(repos.getName());
                def objs = this.omr.doOMRQuery(objTypeName, '', true);
                this.logger.info("Found ${objs.size()} objects");
                objs.each
                { obj ->
                    def classIdentifier = obj.getAttribute('ClassIdentifier');
                    if (classIdentifier && 
                        classIdentifier.startsWith('com.sas.'))
                    {
                        // Objects with ClassIdentifier that is a fully
                        // qualified Java class name is more content than
                        // configuration.
                        return;
                    }
                    scanObj(null, null, obj);
                }
            }
            this.omr.setFoundationRepository();
        }
        this.logger.info("Scanning related objects");
        while (queue.size() > 0)
        {
            def params = queue.poll();
            scanObj(params[0], params[1], params[2]);
        }
        this.logger.info("Scanned ${this.scannedObjSet.size()} objects, visited ${numVisitedObjs} configuration-related metadata objects and ${numVisitedFolders} associated folders.");
    }
}
