/*
 * Created on Nov. 11, 2010
 * Copyright (c) 2010 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm.migrate;

import java.util.regex.Pattern;

import org.w3c.dom.Node;

import com.sas.solutions.brm.omr.OMRConnection;

class HostFinder extends OMRVisitorBase
{
    MigrateConfig components = null;
    Properties brmprops = null;

    def hostPattern = null;
    def useridPattern = null;

    def searchedObjectsCount = 0;
    def numHostsInThisText = 0;

    public HostFinder(OMRConnection omr, MigrateConfig components,
                        Properties brmprops)
    {
        super(omr);
        this.components = components;
        this.brmprops = brmprops;

        // Compile regular expressions
        def hostPropLinesStr = [
            /\.host=/,
            /\.host\.name=/,
            /\.host\.short\.name=/,
            /\.url=/,
            /^configobj.name=.* - SAS SHARE Server - .*/,
            /^configobj.name=Deployment Properties - WebLogic - .*/,
            /^configobj.name=Deployment Tester Server - .*/,
            /^server.sharesrv.servercomponent.name=.* - SAS SHARE Server - .*/,
            /.msg=rebuild the SAS Information Delivery Portal EAR file on .* and redeploy/,
            /.msg=copy PAR files from \$\{par.dest.dir\} to .* on .*/,
            ].join('|');
        this.hostPattern = Pattern.compile(hostPropLinesStr, 
                                            Pattern.CASE_INSENSITIVE);
        this.useridPattern = Pattern.compile(/\.userid=/);
    }

    public static boolean equalHosts(String host1, String host2)
    {
        if (host1 == null && host2 == null)
            return true;
        if (host1 == null && host1 != null)
            return false;
        if (host1 != null && host2 == null)
            return false;
        return host1.equalsIgnoreCase(host2);
    }

    public static String getDomainPrefix(String name)
    {
        if (name == null)
            return null;
        int i = name.indexOf('\\');
        if (i < 0)
            return null;
        return name.substring(0, i);
    }

    public static String quoteNonNullStr(String s)
    {
        if (s == null)
            return 'null';
        return "\"${s}\"";
    }

    // Searches the component table and BRM config properties for the host name
    // that should go with the given ClassIdentifier. Return the host name, or
    // null if no match found.
    def getHostForClassId(classId)
    {
        if (classId == null)
            return null;
        def propname = this.components.getValue('components', classId);
        if (propname == null)
            return null;
        return this.brmprops.getProperty(propname);
    }

    // Searches the pseudo-components table and BRM config properties for the
    // host name that should go with a "pseudo-componet", and object that does
    // not have a ClassIdentifier. The table is searched for the type name and
    // name of the obj parameter. Returns the host name, or null of no match
    // found.
    def getHostForPseudoComponent(obj)
    {
        if (obj == null)
            return null;
        def key = "${obj.getTypeName()}:${obj.getName()}";
        def propname = this.components.getValue('pseudo-components', key);
        if (propname == null)
            return null;
        return this.brmprops.getProperty(propname);
    }

    // Searches the services table and BRM config properties for the host name
    // that should go with the given service name. Return the host name, or
    // null if no match found.
    def getHostForService(service)
    {
        if (service == null)
            return null;
        def propname = this.components.getValue('services', service);
        if (propname == null)
            return null;
        return this.brmprops.getProperty(propname);
    }

    // Determine the new value for a "host" attribute in XML, givin a
    // "protocol" attribute value
    def getHostForProtocol(protocol)
    {
        if (protocol == null)
            return null;
        if (protocol == 'omi')
            return getMetadataHost();
        if (protocol == 'dav')
            return getDAVHost();
        return null;
    }

    def getMetadataHost()
    {
        return this.brmprops.getProperty('BRM.host.metadata');
    }

    def getMetadataPort()
    {
        return this.brmprops.getProperty('BRM.port.metadata');
    }

    def getDAVHost()
    {
        return this.brmprops.getProperty('BRM.host.dav');
    }

    def getDAVPort()
    {
        // TODO: DAV Property not yet implemented in BRM
        return this.brmprops.getProperty('BRM.port.dav');
    }

    def visitMachines()
    {
        def typename = 'Machine';
        def searchParam = '';
        def matchExactType = true;
        omr.setFoundationRepository();
        omr.doOMRQuery(typename, searchParam, matchExactType).each
        { machine ->
            handleMachine(machine);
        }
    }

    def handleMachine(machine)
    {
        this.searchedObjectsCount += 1;
    }

    def visitDocuments()
    {
        def typename = 'Document';
        def searchParam = '';
        def matchExactType = false;
        omr.getRepositories().each
        { repos ->
            def reposname = repos.getName();
            newRepository(reposname);
            omr.setRepository(reposname);
            omr.doOMRQuery(typename, searchParam, matchExactType).each
            { document ->
                def uri = document.getAttribute('URI');
                if (uri == null || uri.length() == 0)
                    return;
                document = getObject(document);
                def objs = document.getAssociation('Objects');
                if (objs == null)
                    objs = [];
                def pseudoCompHost = getHostForPseudoComponent(document);
                if (objs.size() == 0 && pseudoCompHost == null)
                    return;
                handleDocument(document, objs);
            }
        }
    }

    def newRepository(reposname)
    {
    }

    def handleDocument(document, objs)
    {
        this.searchedObjectsCount += 1;
    }

    def visitLogins()
    {
        def typename = 'Login';
        def searchParam = '';
        def matchExactType = true;
        omr.setFoundationRepository();
        omr.doOMRQuery(typename, searchParam, matchExactType).each
        { login ->
            def name = login.getName();
            def nameDomain = getDomainPrefix(name);
            def userID = login.getAttribute('UserID');
            def userIDDomain = getDomainPrefix(userID);
            if (nameDomain == null && userIDDomain == null)
                return;
            handleLogin(login);
        }
    }

    def handleLogin(login)
    {
        this.searchedObjectsCount += 1;
    }
    
    def visitProperties()
    {
        def typename = 'Property';
        def matchExactType = true;
        omr.setFoundationRepository();
        [
            'Connection.Property.HostName.Name.xmlKey.txt',
            'Connection.MySQL.Property.SERVER.Name.xmlKey.txt',
            'omr_host',
            'Metadata.Host',
            'bid.Midtier.trusted.URLs',
            'SPMProvider.Midtier.trusted.URLs',
            'soltnsmid.jdbc.connection.userid',
            'Host Name',
        ].each
        { propname ->
            def searchParam = "*[@Name='${propname}']";
            omr.doOMRQuery(typename, searchParam, matchExactType).each
            { property ->
                handleProperty(property);
            }
        }
    }

    def handleProperty(property)
    {
        this.searchedObjectsCount += 1;
    }

    def visitSASClientConnections()
    {
        def typename = 'SASClientConnection';
        def matchExactType = true;
        omr.setFoundationRepository();
        [
            // Connection: SASMeta - Metadata Server (Client)
            "*[Source/ServerComponent[@ClassIdentifier='0217E202-B560-11DB-AD91-001083FF6836']]",
            // Connection: SASTS SASClientConnection
            "*[Source/ServerComponent[@ClassIdentifier='E803E04B-AAF0-4C74-86A6-F7008AA702DC']]",
            // Connection: SASApp - Connect Server
            "*[Source/ServerComponent[@ClassIdentifier='028e4060-d545-11d5-880d-aa0004006d06']]",
            // Connection: Share Server - HOSTSHORTNAME
            "*[Source/ServerComponent[@ClassIdentifier='SHARE']]",
        ].each
        { searchParam ->
            omr.doOMRQuery(typename, searchParam, matchExactType).each
            { connection ->
                handleSASClientConnection(connection);
            }
        }
    }

    def handleSASClientConnection(connection)
    {
        this.searchedObjectsCount += 1;
    }

    def visitServerComponents()
    {
        def typename = 'ServerComponent';
        def matchExactType = true;
        omr.setFoundationRepository();
        [
            // Object Spawner - HOSTSHORTNAME
            "*[@ClassIdentifier='IOM Spawner']",
            // Connection: Share Server - HOSTSHORTNAME
            "*[@ClassIdentifier='SHARE']",
        ].each
        { searchParam ->
            omr.doOMRQuery(typename, searchParam, matchExactType).each
            { component ->
                handleServerComponent(component);
            }
        }
    }

    def handleServerComponent(component)
    {
        this.searchedObjectsCount += 1;
    }

    def visitSoftwareComponents()
    {
        def typename = 'SoftwareComponent';
        def matchExactType = true;
        omr.setFoundationRepository();
        [
            // MySQL Database Server VERSION - HOSTSHORTNAME
            "*[@ClassIdentifier='mysqldbs']",
            // Deployment Properties - WebLogic - HOSTNAME
            "*[@ClassIdentifier='webappsrv']",
        ].each
        { searchParam ->
            omr.doOMRQuery(typename, searchParam, matchExactType).each
            { component ->
                handleSoftwareComponent(component);
            }
        }
    }

    def handleSoftwareComponent(component)
    {
        this.searchedObjectsCount += 1;
    }

    def visitTCPIPConnections()
    {
        def typename = 'TCPIPConnection';
        def searchParam = '';
        def matchExactType = true;
        omr.getRepositories().each
        { repos ->
            def reposname = repos.getName();
            newRepository(reposname);
            omr.setRepository(reposname);
            omr.doOMRQuery(typename, searchParam, matchExactType).each
            { connection ->
                def hostName = connection.getAttribute('HostName');
                if (hostName == null)
                    return;
                handleTCPIPConnection(connection);
            }
        }
    }

    def handleTCPIPConnection(connection)
    {
        this.searchedObjectsCount += 1;
    }

    def visitTrees()
    {
        def typename = 'Tree';
        def matchExactType = true;
        omr.setFoundationRepository();
        [
            // MySQL Database Server VERSION - SHORTHOSTNAME
            "*[Members/SoftwareComponent[@ClassIdentifier='mysqldbs']]",
        ].each
        { searchParam ->
            omr.doOMRQuery(typename, searchParam, matchExactType).each
            { tree ->
                handleTree(tree);
            }
        }
    }

    def handleTree(tree)
    {
        this.searchedObjectsCount += 1;
    }

    def visitTextStoreXMLObjs()
    {
        def typename = 'TextStore';
        def searchParam = "*[@Name='Initialization Data']";
        def matchExactType = true;
        omr.setFoundationRepository();
        omr.doOMRQuery(typename, searchParam, matchExactType).each
        { obj ->
            def text = obj.getAttribute('StoredText');
            if (text == null)
                return;
            if (!isXML(text))
                return;
            if (isTemplateTextStore(obj))
                return;
            def doc = buildDOMTree(text);
            if (doc == null)
                return; // Not really XML
            this.numHostsInThisText = 0;
            handleXML(obj, doc);
        }
    }

    // Process an XML text object
    def handleXML(obj, doc)
    {
        this.searchedObjectsCount += 1;
        def nodeValueCallback = 
        { parent, node, value ->
            if (node.getNodeType() != Node.ATTRIBUTE_NODE)
                return false;
            if (!(node.getNodeName() in ['host', 'url']))
                return false;
            if (!(parent.getNodeName() in ['ReposDef', 'Profile']))
                return false;
            this.numHostsInThisText += 1;
            handleXMLValue(obj, parent, node, value);
            return true;
        }
        traverseDOMNodes(doc, nodeValueCallback);
    }

    def handleXMLValue(obj, parent, node, value)
    {
    }

    def visitTextStorePropertiesObjs()
    {
        def typename = 'TextStore';
        def searchParam = "*[@Name='Public Configuration Properties' OR @Name='Private Configuration Properties']";
        def matchExactType = true;
        omr.getRepositories().each
        { repos ->
            def reposname = repos.getName();
            newRepository(reposname);
            omr.setRepository(reposname);
            omr.doOMRQuery(typename, searchParam, matchExactType).each
            { obj ->
                def text = obj.getAttribute('StoredText');
                if (text == null)
                    return;
                this.numHostsInThisText = 0;
                handlePropertiesText(obj, text);
            }
        }
    }

    // Process a Java Properties text object
    def handlePropertiesText(obj, text)
    {
        def result = [];
        def lines = text.split('\n', -1);
        lines.each
        { line ->
            if (this.hostPattern.matcher(line).find() ||
                (this.useridPattern.matcher(line).find() &&
                 line.indexOf('\\') >= 0))
            {
                this.numHostsInThisText += 1;
                if (this.numHostsInThisText == 1)
                    this.searchedObjectsCount += 1;
                def resultLine = handlePropertiesTextLine(obj, line);
                if (resultLine != null)
                    line = resultLine;
            }
            result.add(line);
        }
        return result.join('\n');
    }

    def handlePropertiesTextLine(obj, line)
    {
    }

    def visitPromptGroups()
    {
        def typename = 'PromptGroup';
        def searchParam = '';
        def matchExactType = true;
        omr.getRepositories().each
        { repos ->
            def reposname = repos.getName();
            newRepository(reposname);
            omr.setRepository(repos.getName());
            omr.doOMRQuery(typename, searchParam, matchExactType).each
            { promptGroup ->
                def groupInfo = promptGroup.getAttribute('GroupInfo');
                if (groupInfo == null)
                    return;
                if (!isXML(groupInfo))
                    return;
                def doc = buildDOMTree(groupInfo);
                if (doc == null)
                    return; // Not really XML
                this.numHostsInThisText = 0;
                handlePromptGroupXML(promptGroup, doc);
            }
        }
        omr.setFoundationRepository();
    }

    def handlePromptGroupXML(obj, doc)
    {
        this.searchedObjectsCount += 1;
        def nodeValueCallback = 
        { parent, node, value ->
            if (!(parent.getNodeName() == 'TextDefinition'))
                return false;
            if (node.getNodeType() != Node.ATTRIBUTE_NODE)
                return false;
            if (!(node.getNodeName() == 'name'))
                return false;
            if (!(node.getNodeValue() == '_collection_url'))
                return false;
            node = parent.getElementsByTagName('DefaultValue').item(0);
            if (node == null)
                return false;
            parent = node;
            node = parent.getElementsByTagName('String').item(0);
            if (node == null)
                return false;
            parent = node;
            node = parent.getAttributeNode('value');
            if (node == null)
                return false;
            this.numHostsInThisText += 1;
            handlePromptGroupXMLValue(obj, parent, node, node.getValue());
            return true;
        }
        traverseDOMNodes(doc, nodeValueCallback);
    }

    def handlePromptGroupXMLValue(obj, parent, node, value)
    {
    }

    def visitHosts()
    {
        visitMachines();
        visitDocuments();
        visitLogins();
        visitProperties();
        visitSASClientConnections();
        visitServerComponents();
        visitSoftwareComponents();
        visitTCPIPConnections();
        visitTrees();
        visitTextStoreXMLObjs();
        visitTextStorePropertiesObjs();
        visitPromptGroups();
    }
}
