/*
 * Created on May 31, 2006
 * Copyright (c) 2006 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm.migrate;

import org.apache.log4j.Logger;

import com.sas.solutions.brm.BRMAgent;
import com.sas.solutions.brm.Constants;
import com.sas.solutions.brm.StatusXMLUpdater;
import com.sas.solutions.brm.errors.BRMConfigurationException;
import com.sas.solutions.brm.errors.BRMAntError;
import com.sas.solutions.brm.errors.BRMUserError;
import com.sas.solutions.brm.logging.MonitoredLogger;
import com.sas.solutions.brm.omr.OMRConnection;
import com.sas.solutions.brm.omr.OMRUtil;
import com.sas.solutions.brm.util.BRMUtil;
import com.sas.solutions.brm.webservice.WSMWrapper;

public class MainMigrator
{
    Properties controlProps = null;
    Properties localConfig = null;
    OMRConnection omr = null;
    private static final Logger LOG = Logger.getLogger(MainMigrator.class);
    private static transient final Logger PROGRESS = Logger.getLogger(
            "progress.controller.brm");

    public MainMigrator()
    {
        String version = BRMUtil.readVersion();
        String buildtype = BRMUtil.readBuildtype();
        String builddatetime = BRMUtil.readBuilddatetime();
        LOG.info("BRM version ${version}-${buildtype} built ${builddatetime}");
        LOG.info("MainMigrator starting");
        this.controlProps = new Properties();
        BRMUtil.loadPropFileAsResource(Constants.BRM_CONFIG_FILE, 
                                        this.controlProps, LOG);
        def agent = new BRMAgent(null);
        agent.setControlConfig(this.controlProps);
        agent.loadLocalConfig();
        this.localConfig = agent.getLocalConfig();
        init(BRMUtil.getNewOMRConnection(controlProps, 'admin'));
        def metadataserver = this.controlProps.getProperty('BRM.host.metadata');
        def msg = "Connected to metadata server on ${metadataserver}";
        LOG.info(msg);
        PROGRESS.info(msg);
    }

    public MainMigrator(OMRConnection omr)
    {
        init(omr);
    }

    private void init(OMRConnection omr)
    {
        this.omr = omr;
    }

    def close()
    {
        if (this.omr != null)
        {
            this.omr.close();
            this.omr = null;
        }
    }

    def getTempDir()
    {
        def installDir = System.getProperty('BRM.install.dir');
        def tempDir = new File(installDir, 'temp');
        return tempDir;
    }

    def getFoundationServicesEnvObj(logger)
    {
        this.omr.setFoundationRepository();
        def query = "@Name =: 'Foundation Services'";
        def objs = this.omr.doOMRQuery('SoftwareComponent', query, true);
        if (objs.size() == 0)
        {
            if (logger != null)
                logger.warn("Could not find SoftwareComponent with ${query}");
            return null;
        }
        if (objs.size() > 1)
        {
            if (logger != null)
                logger.warn("More than one SoftwareComponent with ${query}");
            return null;
        }
        def obj = objs[0];
        obj = obj.getAssociatedObject('PropertySets', 'Environment.Properties');
        if (obj == null)
        {
            if (logger != null)
                logger.warn("Foundation Services object doesn't have Environment.Properties");
            return null;
        }
        def env = this.omr.getObject(obj); // Get full object info
        return env
    }

    def getFoundationServicesProps(logger)
    {
        if (logger != null)
        {
            logger.info(
                "Getting the Foundation Services environment properties");
        }
        def env = getFoundationServicesEnvObj(logger);
        if (env == null)
        {
            // Error has already been logged
            return null;
        }
        def objs = env.getAssociation('SetProperties');
        if (objs == null)
        {
            if (logger != null)
            {
                logger.error(
                    "No Foundation Services environment properties defined");
            }
            return null;
        }
        def props = new Properties();
        objs.each
        { obj ->
            obj = this.omr.getObject(obj); // Get full object info
            def propName = obj.getName();
            def propValue = obj.getAttribute('DefaultValue');
            if (logger != null)
                logger.debug("${propName}=${propValue}");
            if (propValue == null)
            {
                logger.debug("Changing null value for property ${propName} to empty string");
                propValue = '';
            }
            props.setProperty(propName, propValue);
        }
        return props;
    }

    // Save Foundation Services environment properties to disk file
    def saveFoundationServicesProps(logger)
    {
        def props = getFoundationServicesProps(logger);
        if (props == null)
        {
            // Error message was already logged
            return;
        }
        def tempDir = getTempDir();
        def tempConfigDir = new File(tempDir, 'config');
        tempConfigDir.mkdirs();
        def envPropFile = new File(tempConfigDir, 
                                   'foundation-services-env.properties');
        if (logger != null)
        {
            logger.info("Saving Foundation Services environment properties to: ${envPropFile}");
        }
        def fos = new FileOutputStream(envPropFile);
        props.store(fos, "Foundation Services environment");
        fos.close();
    }

    def setFoundationServicesProps(props, logger)
    {
        if (logger != null)
        {
            logger.info(
                "Setting the Foundation Services environment properties");
        }
        def env = getFoundationServicesEnvObj(logger);
        if (env == null)
        {
            // Error has already been logged
            return null;
        }
        props.each
        { propName, propValue ->
            def obj = env.getAssociatedObject('SetProperties', propName);
            if (obj == null)
            {
                if (logger != null)
                {
                    logger.warn("Foundation Services is missing property found in backed-up property file: ${propName}=${propValue}");
                }
                return;
            }
            obj = this.omr.getObject(obj); // Get full object info
            obj.setAttribute('DefaultValue', propValue);
            if (logger != null)
            {
                logger.debug("Setting property object ${obj.getId()} ${obj.getName()} DefaultValue=\"${propValue}\"");
            }
            this.omr.store(obj);
        }
    }

    // Load Foundation Services environment properties from disk file
    def loadFoundationServicesProps(logger)
    {
        def tempDir = getTempDir();
        def tempConfigDir = new File(tempDir, 'config');
        def envPropFile = new File(tempConfigDir, 
                                   'foundation-services-env.properties');
        if (!envPropFile.isFile())
        {
            logger.warn("Missing foundation services props file: ${envPropFile}");
            return;
        }
        def props = new Properties();
        BRMUtil.loadPropFile(envPropFile, props, logger);
        setFoundationServicesProps(props, logger);
    }

    def getOMRCredentials()
    {
        def getprop = { propname ->
            def propval = this.controlProps.get(propname);
            if (propval == null || propval.length() == 0)
            {
                throw new BRMConfigurationException(
                    "Missing property: " + propname);
            }
            return propval;
        }
        def host = getprop("BRM.host.metadata");
        def port = getprop("BRM.port.metadata");
        def user = getprop("BRM.admin.user");
        def password = getprop("BRM.admin.password");
        return ['host': host, 'port': port, 'user': user, 
                'password': password];
    }

    // Look up and save the WebServiceMaker URL
    // Part of the fix for defect S0714201.
    def saveWebServiceMakerURL(logger)
    {
        def wsmUrl = WSMWrapper.getWebServiceMakerURL(logger, 
            this.omr.getHost(), this.omr.getPort(), this.omr.getUser(),
            this.omr.getPassword());
        if (wsmUrl != null)
        {
            logger.info("Saving WebServicesMaker URL: ${wsmUrl}");
            def servicesDir = new File(tempDir, 'services');
            servicesDir.mkdirs();
            def wsmUrlFile = new File(servicesDir, 'webservicemaker.url.txt');
            logger.info("Writing ${wsmUrlFile}");
            def autoFlush = true;
            def out = new PrintWriter(new FileWriter(wsmUrlFile), autoFlush);
            out.println(wsmUrl);
            out.close();
        }
        else
        {
            logger.warn("Unable to find WebServiceMaker URL in metadata");
        }
    }

    // Re-create the System web services
    // Fixes defect S0714201
    def recreateSystemWebServices(logger)
    {
        def webservices = querySystemWebServicesInMetadata(logger);
        if (webservices.size() == 0)
        {
            logger.warn("No web services in /System/Services metadata folder");
            return;
        }
        logger.info("Found ${webservices.size()} web services in /System/Services");
        def creds = getOMRCredentials();
        def tempDir = getTempDir();
        def servicesDir = new File(tempDir, 'services');
        def spkfile = BRMUtil.exportPackage(
                logger, this.localConfig, creds, servicesDir,
                'Service.SoapGenerated', '/System/Services', 'WebServices',
                []);
        if (spkfile == null || !spkfile.isFile())
        {
            logger.error("Failed to export web services.");
            return;
        }

        // Delete the web services
        def wsm = null;
        def wsmUrl = null;
        def wsmUrlFile = new File(servicesDir, 'webservicemaker.url.txt');
        if (wsmUrlFile.isFile())
        {
            def reader = new BufferedReader(new FileReader(wsmUrlFile));
            wsmUrl = reader.readLine();
            reader.close();
            if (wsmUrl != null)
                wsmUrl = wsmUrl.trim();
        }
        if (wsmUrl != null && wsmUrl.length() > 0)
        {
            wsm = new WSMWrapper(logger, wsmUrl);
        }
        else
        {
            wsm = new WSMWrapper(logger, creds.host, creds.port, creds.user,
                                         creds.password);
        }
        webservices.each
        { obj ->
            def name = obj.getName();
            logger.info("Deleting web service: ${name}");
            wsm.removeWebService(name, creds.user, creds.password);
        }

        // Re-import the web services
        BRMUtil.importPackage(logger, this.localConfig, creds, 
                              '/System/Services', spkfile, []);
    }

    def querySystemWebServicesInMetadata(logger)
    {
        def typename = "Tree";
        def query = "@PublicType='Folder' AND @TreeType='System' AND @Name='System'";
        omr.setFoundationRepository();
        def trees = omr.doOMRQuery(typename, query, true);
        if (trees == null || trees.size() == 0)
        {
            logger.warn("Could not find /System metadata folder");
            return [];
        }
        def tree = findTreeWithNoParent(trees);
        if (tree == null)
        {
            logger.warn("Could not find root /System metadata folder");
            return [];
        }
        tree = omr.getObject(tree);
        tree = tree.getAssociatedObject('SubTrees', 'Services');
        if (tree == null)
        {
            logger.warn("Could not find /System/Services metadata folder");
            return [];
        }
        tree = omr.getObject(tree);
        def objs = tree.getAssociation('Members');
        if (objs == null || objs.size() == 0)
        {
            logger.warn("/System/Services metadata folder is empty");
            return [];
        }
        def webservices = [];
        logger.debug("Contents of /System/Services folder:");
        objs.each
        { obj ->
            obj = omr.getObject(obj);
            def publicType = obj.getAttribute('PublicType');
            logger.debug("${obj.getTypeName()} \"${obj.getName()}\" PublicType=${publicType}");
            if (obj.getTypeName() == 'DeployedComponent' &&
                publicType == 'Service.SoapGenerated')
            {
                webservices.add(obj);
            }
        }
        return webservices;
    }

    def findTreeWithNoParent(trees)
    {
        def tree = null;
        trees.each
        {
            def parentAssoc = it.getAssociation('ParentTree');
            if (parentAssoc == null || parentAssoc.size() == 0)
            {
                tree = it;
                return;
            }
        }
        return tree;
    }

    // Look up and save the DAV content mapping info
    // Part of the fix for defect S0719081
    def saveDAVContentMapping(logger)
    {
        logger.info("Saving the DAV Content Mapping");
        def mapping = findDAVContentMapping(logger);
        if (mapping == null)
        {
            logger.warn("Could not find DAV Content Mapping");
            return;
        }
        def uri = mapping.document.getAttribute('URI');
        logger.info("DAV Content Mapping URI: ${uri}");
        def tempDir = getTempDir();
        def tempDavDir = new File(tempDir, 'dav');
        tempDavDir.mkdirs();
        def documentFile = new File(tempDavDir, 'Document.xml');
        logger.info("Writing DAV Content Mapping Document to: ${documentFile}");
        BRMUtil.writeStringToFile(documentFile, mapping.document.getXML());
        def textstoreFile = new File(tempDavDir, 'TextStore.xml');
        logger.info(
                "Writing DAV Content mapping TextStore to: ${textstoreFile}");
        BRMUtil.writeStringToFile(textstoreFile, mapping.textstore.getXML());
    }

    def findDAVContentMapping(logger)
    {
        this.omr.setFoundationRepository();
        // The Document and TextStore have the same name
        def name = 'Report Repository Content Mapping';
        def query = "*[@Name='${name}'][Notes/TextStore[@Name='${name}']]";
        logger.info("Searching for DAV Content Mapping with query: ${query}");
        def objs = this.omr.doOMRQuery('Document', query, true);
        if (objs == null || objs.size() == 0)
        {
            logger.warn("No DAV Content Mapping Document in metadata");
            return null;
        }
        if (objs.size() > 1)
        {
            // This should happen rarely if ever.
            logger.warn("Found multiple DAV Content Mapping Document objects");
        }
        def document = objs.get(0);
        def textstore = document.getAssociatedObject('Notes', name);
        if (textstore == null)
        {
            logger.warn("No DAV Content Mapping TextStore in metadata");
            return null;
        }
        textstore = this.omr.getObject(textstore); // Get full attributes, etc.
        return ['document': document, 'textstore': textstore];
    }

    // Restore the previously-saved DAV content mapping info
    // Part of the fix for defect S0719081
    def loadDAVContentMapping(logger)
    {
        logger.info("Restoring the DAV Content Mapping");
        def tempDir = getTempDir();
        def tempDavDir = new File(tempDir, 'dav');
        if (!tempDavDir.isDirectory())
        {
            logger.warn(
                "Missing directory for DAV Content Mapping: ${tempDavDir}");
            return;
        }
        def documentFile = new File(tempDavDir, 'Document.xml');
        def saved_document = OMRUtil.loadObjectFromFile(documentFile);
        if (saved_document == null)
        {
            logger.warn(
                "Missing file for DAV Content Mapping: ${documentFile}");
            return;
        }
        def textstoreFile = new File(tempDavDir, 'TextStore.xml');
        def saved_textstore = OMRUtil.loadObjectFromFile(textstoreFile);
        if (saved_textstore == null)
        {
            logger.warn(
                "Missing file for DAV Content Mapping: ${textstoreFile}");
            return;
        }
        def mapping = findDAVContentMapping(logger);
        if (mapping == null)
        {
            logger.warn("Could not find DAV Content Mapping");
            return;
        }
        logger.info("Updating DAV Content Mapping Document attributes");
        OMRUtil.copyObjectAttributes(saved_document, mapping.document);
        this.omr.store(mapping.document);
        logger.info("Updating DAV Content Mapping TextStore attributes");
        OMRUtil.copyObjectAttributes(saved_textstore, mapping.textstore);
        this.omr.store(mapping.textstore);
        logger.info("Done restoring the DAV Content Mapping");
    }

    def findSetinit(logger)
    {
        def query = "*[@Name='Setinit text'][Objects/Document[@Name='SAS Setinit Location' AND @TextRole='setinit']]";
        def objs = this.omr.doOMRQuery('TextStore', query, true);
        if (objs.size() == 0)
        {
            if (logger != null)
                logger.warn("Could not find TextStore with ${query}");
            return null;
        }
        if (objs.size() > 1)
        {
            if (logger != null)
                logger.warn("More than one TextStore matching query: ${query}");
            return null;
        }
        return objs[0];
    }

    def saveSetinit(logger)
    {
        def setinitTextStore = findSetinit(logger);
        if (setinitTextStore == null)
            return;
        def tempDir = getTempDir();
        def tempConfigDir = new File(tempDir, 'config');
        tempConfigDir.mkdirs();
        def setinitFile = new File(tempConfigDir, 'Setinit text.xml');
        logger.info("Saving XML file: ${setinitFile}");
        setinitFile.write(setinitTextStore.getXML(), 'UTF-8');
    }

    def loadSetinit(logger)
    {
        def setinitTextStore = findSetinit(logger);
        if (setinitTextStore == null)
            return;
        def tempDir = getTempDir();
        def tempConfigDir = new File(tempDir, 'config');
        def setinitFile = new File(tempConfigDir, 'Setinit text.xml');
        if (!setinitFile.isFile())
        {
            logger.warn("Missing XML file: ${setinitFile}");
            return;
        }
        logger.info("Loading XML file: ${setinitFile}");
        def savedSetinitTextStore = OMRUtil.loadObjectFromFile(setinitFile);
        OMRUtil.copyObjectAttributes(savedSetinitTextStore, setinitTextStore);
        logger.debug("Restoring Setinit text in TextStore ID=${setinitTextStore.getId()}");
        this.omr.store(setinitTextStore);
    }

    def saveconfig()
    {
        def tempDir = getTempDir();
        def logger = new MonitoredLogger(PROGRESS, LOG);
        PROGRESS.info("Saving configuration info from metadata into directory: ${tempDir}");
        if (tempDir.isDirectory())
        {
            if (tempDir.list().length > 0)
            {
                PROGRESS.info("Deleting existing contents of directory: ${tempDir}");
                BRMUtil.makeDirectoryEmpty(tempDir);
            }
        }
        else
        {
            PROGRESS.info("Creating directory: ${tempDir}");
            tempDir.mkdirs();
        }

        def dumper = new ConfigObjDumper(this.omr, tempDir, logger);
        dumper.dump();

        // Save the Foundation Services environment properties
        saveFoundationServicesProps(logger);

        // Look up and save the DAV content mapping info
        // Part of the fix for defect S0719081
        saveDAVContentMapping(logger);

        // Look up and save the WebServiceMaker URL
        // Part of the fix for defect S0714201.
        saveWebServiceMakerURL(logger);

        // Save the local SETINIT info
        saveSetinit(logger);

        // Save config items from HCM database
        def controller = new com.sas.solutions.brm.BRMController();
        def dbhost = controller.getConfigProp('BRM.host.mysql');
        if (dbhost == null)
        {
            logger.error("Missing BRM property BRM.host.mysql, cannot save config items from database");
        }
        else
        {
            try
            {
                controller.runTarget(dbhost, '', 'hcm_db_saveconfig');
            }
            catch (BRMAntError e)
            {
                logger.error(e.getMessage());
            }
        }

        if (logger.errors == 1)
            throw new BRMUserError("An error was encountered while saving the configuration info. See the BRM log file for details.");
        if (logger.errors > 1)
            throw new BRMUserError("${logger.errors} errors were encountered while saving the configuration info. See the BRM log file for details.");
        PROGRESS.info("Finished saving configuration info. See the BRM log file for details.");
    }

    def loadconfig(boolean force)
    {
        def tempDir = getTempDir();
        def logger = new MonitoredLogger(PROGRESS, LOG);
        PROGRESS.info("Loading configuration info into metadata from directory: ${tempDir}");

        // Create BRM Controller object to run remote commands
        def controller = new com.sas.solutions.brm.BRMController();

        def loader = new ConfigObjLoader(this.omr, tempDir, logger, force);
        def msg = null;
        if (!loader.load())
        {
            logger.error("Aborted loading configuration objects.");
            return;
        }

        // Synchronize status.xml file on each host with fixed-up metadata
        // Using the temp/config/configobjs/objidmap.txt file created by 
        // ConfigObjLoader.load()
        StatusXMLUpdater.getConfigObjIdMap(tempDir, controller, logger);
        controller.getConfiguredHostList().each
        { host ->
            try
            {
                controller.runTarget(host, '', 'syncstatusxml');
            }
            catch (BRMAntError e1)
            {
                logger.error(e1.getMessage());
            }
        }

        // Restore the Foundation Services environment properties
        loadFoundationServicesProps(logger);

        // Restore the previously-saved DAV content mapping info
        // Part of the fix for defect S0719081
        loadDAVContentMapping(logger);

        // Restore the local SETINIT info
        loadSetinit(logger);

        // Load config items into HCM database
        def dbhost = controller.getConfigProp('BRM.host.mysql');
        if (dbhost == null)
        {
            logger.error("Missing BRM property BRM.host.mysql, cannot load config items into database");
        }
        else
        {
            try
            {
                controller.runTarget(dbhost, '', 'hcm_db_loadconfig');
            }
            catch (BRMAntError e2)
            {
                logger.error(e2.getMessage());
            }
        }

        if (logger.errors == 1)
            throw new BRMUserError("An error was encountered while loading the configuration info. See the BRM log file for details.");
        if (logger.errors > 1)
            throw new BRMUserError("${logger.errors} errors were encountered while loading the configuration info. See the BRM log file for details.");
        PROGRESS.info("Finished loading configuration info. See the BRM log file for details.");
    }

    public void dumppaths(String filename)
    {
        def dumper = new PathDumper(omr, filename);
        dumper.visitPaths();
        dumper.out.println();
        dumper.out.println('------------------------------------------------');
        dumper.out.println("${sprintf('%5d', dumper.searchedObjectsCount)} objects searched");
        dumper.out.println("${sprintf('%5d', dumper.objectsWithPathCount)} objects with file or directory paths found");
        dumper.close();
    }

    public void fixpaths(boolean nochange)
    {
        def msg = "Modifying file and directory paths in metadata";
        LOG.info(msg);
        PROGRESS.info(msg);
        this.omr.setReadOnly(nochange);
        if (nochange)
        {
            msg = "Because 'nochange' option was used, no metadata will be actually modified";
            LOG.info(msg);
            PROGRESS.info(msg);
        }
        def config = new MigrateConfig();
        msg = "Loaded migration configuration from ${config.iniFile}";
        LOG.info(msg);
        PROGRESS.info(msg);
        def fixer = new PathFixer(omr, config, LOG);
        fixer.visitPaths();
        msg = "${sprintf('%5d', fixer.searchedObjectsCount)} objects searched";
        LOG.info(msg);
        PROGRESS.info(msg);
        msg = "${sprintf('%5d', fixer.objectsWithPathCount)} objects with file or directory paths found";
        LOG.info(msg);
        PROGRESS.info(msg);
        msg = "${sprintf('%5d', fixer.modifiedObjCount)} objects modified";
        LOG.info(msg);
        PROGRESS.info(msg);
        if (nochange)
        {
            msg = "Because 'nochange' option was used, no metadata was actually modified";
            LOG.info(msg);
            PROGRESS.info(msg);
        }
        PROGRESS.info("See the BRM.log file for details");
    }

    MigrateConfig loadComponentsFile()
    {
        def installDir = System.getProperty('BRM.install.dir');
        if (installDir == null)
            installDir = '.';
        def configDir = new File(installDir, 'config');
        def componentsFile = new File(configDir, 'components.ini');
        def components = new MigrateConfig(componentsFile);
        def msg = "Loaded components info from ${components.iniFile}";
        LOG.info(msg);
        PROGRESS.info(msg);
        return components;
    }

    public void dumphosts(String filename)
    {
        def components = loadComponentsFile();
        def dumper = new HostDumper(this.omr, components, this.controlProps,
                                        filename);
        dumper.visitHosts();
    }

    public void fixhosts(boolean nochange)
    {
        def msg = "Modifying host information in metadata";
        LOG.info(msg);
        PROGRESS.info(msg);
        this.omr.setReadOnly(nochange);
        this.omr.initStats();
        if (nochange)
        {
            msg = "Because 'nochange' option was used, no metadata will be actually modified";
            LOG.info(msg);
            PROGRESS.info(msg);
        }
        def config = new MigrateConfig();
        msg = "Loaded migration configuration from ${config.iniFile}";
        LOG.info(msg);
        PROGRESS.info(msg);
        def components = loadComponentsFile();
        def fixer = new HostFixer(this.omr, components, this.controlProps, 
                                    config);
        fixer.visitHosts();
        msg = "${sprintf('%5d', fixer.searchedObjectsCount)} objects with potential host names";
        LOG.info(msg);
        PROGRESS.info(msg);
        if (nochange)
        {
            msg = "Because 'nochange' option was used, no metadata was actually modified";
            LOG.info(msg);
            PROGRESS.info(msg);
        }
        else
        {
            msg = "${sprintf('%5d', this.omr.getNumUpdates())} objects modified";
            LOG.info(msg);
            PROGRESS.info(msg);
            msg = "${sprintf('%5d', this.omr.getNumCreates())} objects created";
            LOG.info(msg);
            PROGRESS.info(msg);
            msg = "${sprintf('%5d', this.omr.getNumDeletes())} objects deleted";
            LOG.info(msg);
            PROGRESS.info(msg);
        }
        PROGRESS.info("See the BRM.log file for details");
    }

    // Use the WebServiceMaker API to list the registered web services.
    // This is just for reference and diagnostics purposes.
    // Not all of the web services listed by this method need to be exported
    // and re-imported during migration; for example, I don't think we need
    // to export and re-import the WebServiceMaker web service itself.
    public void listwebservices()
    {
        def creds = getOMRCredentials();
        def wsm = new WSMWrapper(LOG, creds.host, creds.port, creds.user,
                                      creds.password);
        def webServices = wsm.listWebServices(creds.user, creds.password);
        webServices.each
        { webService ->
            println(webService);
        }
    }

    public void refreshservices()
    {
        def logger = new MonitoredLogger(PROGRESS, LOG);
        // Re-create the System web services
        // Part of the fix for defect S0714201.
        recreateSystemWebServices(logger);
    }

    public static void printHelp()
    {
        PROGRESS.info(BRMUtil.trimMultiLine(
        '''
        BRM "migration" commands: fixup after transporting metadata

        Command usage:
            brm saveconfig
            brm loadconfig [force]
            brm dumppaths [filename] 
            brm fixpaths (nochange | commit)
            brm dumphosts [filename]
            brm fixhosts (nochange | commit)

        Run "brm saveconfig" right before doing a BRM restore. This backs up
        the installation-specific configuration information in metadata.

        Run "brm loadconfig" right after doing a BRM restore. This restores
        the installation-specific configuration information in metadata, so
        that later SAS Deployment Manager operations will work properly.

        The dumppaths command finds and prints all of the file and
        directory paths it can find in metadata. The optional filename
        parameter is the name of a file to which output will be sent,
        instead of to the console.

        The fixpaths command finds file and directory paths in metadata and
        fixes them according to the settings in the migrate.ini file. See
        the BRM tool documentation for more info. Using the "nochange"
        keyword after fixpaths causes BRM to just report the changes that
        would be made, without actually changing metadata.  Use the
        "commit" keyword to actually make the changes to metadata.
        
        The dumphosts command finds and prints all hostnames in metadata.
        The optional filename parameter is the name of a file to which
        output will be sent.

        The fixhosts command fixes the machine and host name references in
        metadata according to the settings in brm.properties, components.ini,
        and migrate.ini. See the BRM tool documentation for more info.
        '''
        ));
    }

    // Return a list of lists describing commands implemented by this class:
    // [ [command-name, usage-message], ...]. If the usage message is null
    // or empty, then the command will not be listed by the "help" command.
    def static getCommandList()
    {
        return [
            ['mighelp',
             'brm mighelp - more information about migration-related commands'],
            ['saveconfig', 'brm saveconfig'],
            ['loadconfig', 'brm loadconfig [force]'],
            ['dumppaths', 'brm dumppaths [filename]'],
            ['fixpaths', 'brm fixpaths (nochange | commit)'],
            ['dumphosts', 'brm dumphosts [filename]'],
            ['fixhosts', 'brm fixhosts (nochange | commit)'],
            ['listwebservices', 'brm listwebservices'],
            ['refreshservices', 'brm refreshservices'],
            ];
    }

    public void run(String[] args)
    {
        if (args.length == 0)
        {
            PROGRESS.info('Missing migration command. Use "mighelp" for info.');
            return;
        }
        def command = args[0];
        def msg = "Executing migration command: ${command}";
        LOG.info(msg);
        PROGRESS.info(msg);
        if (command.toLowerCase() in ['mighelp', 'help', '-h', '--help'])
        {
            printHelp();
            return;
        }
        else if (command.toLowerCase() == 'saveconfig')
        {
            saveconfig();
        }
        else if (command.toLowerCase() == 'loadconfig')
        {
            def force = false;
            if (args.length >= 2)
                force = (args[1].toLowerCase() == 'force');
            loadconfig(force);
        }
        else if (command.toLowerCase() == 'dumppaths')
        {
            def filename = null;
            if (args.length >= 2)
                filename = args[1];
            dumppaths(filename);
        }
        else if (command.toLowerCase() == 'fixpaths')
        {
            if (args.length < 2)
            {
                msg = ('Missing "nochange" or "commit" keyword. ' +
                       'Run "brm fixpaths help" for more info.');
                throw new BRMUserError(msg);
            }
            def keyword = args[1];
            def nochange = false;
            if (keyword.toLowerCase() == 'help')
            {
                printHelp();
                return;
            }
            else if (keyword.toLowerCase() == 'nochange')
            {
                nochange = true;
            }
            else if (keyword.toLowerCase() == 'commit')
            {
                nochange = false;
            }
            else
            {
                msg = ("Invalid keyword after 'fixpaths': ${keyword}. " +
                       'Run "brm fixpaths help" for more info.');
                throw new BRMUserError(msg);
            }
            fixpaths(nochange);
        }
        else if (command.toLowerCase() == 'dumphosts')
        {
            def filename = null;
            if (args.length >= 2)
                filename = args[1];
            dumphosts(filename);
        }
        else if (command.toLowerCase() == 'fixhosts')
        {
            if (args.length < 2)
            {
                msg = ('Missing "nochange" or "commit" keyword. ' +
                       'Run "brm fixhosts help" for more info.');
                throw new BRMUserError(msg);
            }
            def keyword = args[1];
            def nochange = false;
            if (keyword.toLowerCase() == 'help')
            {
                printHelp();
                return;
            }
            else if (keyword.toLowerCase() == 'nochange')
            {
                nochange = true;
            }
            else if (keyword.toLowerCase() == 'commit')
            {
                nochange = false;
            }
            else
            {
                msg = ("Invalid keyword after 'fixhosts': ${keyword}. " +
                       'Run "brm fixhosts help" for more info.');
                throw new BRMUserError(msg);
            }
            fixhosts(nochange);
        }
        else if (command.toLowerCase() == 'listwebservices')
        {
            listwebservices();
        }
        else if (command.toLowerCase() == 'refreshservices')
        {
            refreshservices();
        }
        else
        {
            throw new BRMUserError("Invalid command: ${command}");
        }
    }

    public static void main(String[] args)
    {
        Thread.currentThread().setName("migrate");
        com.sas.solutions.brm.logging.LogConfigurator.configure();
        MainMigrator mm = null;
        int resultCode = 0;
        try
        {
            mm = new MainMigrator();
            try
            {
                mm.run(args);
            }
            finally
            {
                mm.close();
            }
        }
        catch (BRMUserError e)
        {
            LOG.error(e.getMessage());
            PROGRESS.error(e.getMessage());
            resultCode = 2;
        }
        catch (Exception e)
        {
            LOG.error("Exception while running MainMigrator", e);
            PROGRESS.error(e.getMessage());
            resultCode = 4;
        }
        catch (Throwable e)
        {
            LOG.error("Error while running MainMigrator", e);
            PROGRESS.error(e.getMessage());
            resultCode = 5;
        }
        System.exit(resultCode);
    }
}
