/*
 * Created on June 30, 2010
 * Copyright (c) 2010 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm.migrate;

import com.sas.solutions.brm.omr.GenericMetadataObject;
import com.sas.solutions.brm.omr.OMRUtil;

class OMRVisitorBase extends VisitorBase
{
    def omr = null;

    public OMRVisitorBase(omr)
    {
        this.omr = omr;
    }

    def getRepositoryNameForObjId(String objid)
    {
        return this.omr.getRepositoryNameForObjId(objid);
    }

    // Fetch an object by object ID
    // This is helpful for an object retrieved from an association, to get
    // all of its attributes and associations.
    GenericMetadataObject getObject(String typename, String objid)
    {
        return this.omr.getObject(typename, objid, true);
    }

    GenericMetadataObject getObject(GenericMetadataObject obj)
    {
        return this.omr.getObject(obj);
    }

    /**
    * Get the folder path of a Tree object.
    * @return a String containing the path, or null if the Tree is not a
    * genuine folder in the BIP Tree hierarchy.
    */
    String getFolderPathOfTree(GenericMetadataObject obj)
    {
        def result = new LinkedList();
        while (true)
        {
            if ((obj.getTypeName() != 'Tree') ||
                (obj.getAttribute('PublicType') != 'Folder'))
            {
                return null;
            }
            result.addFirst(obj.getName());
            def parent = obj.getAssociatedObject('ParentTree');
            if (parent == null)
            {
                // Note: Folder in non-Foundation repository will have
                // two associated SoftwareComponents. Both should be
                // named "BIP Service" with PublicType='RootFolder'.
                def softwareComponent = obj.getAssociatedObject(
                    'SoftwareComponents');
                if (softwareComponent == null)
                {
                    //println("getFolderPathOfTree: ${obj.getTypeName()} ID=${obj.getId()} \"${obj.getName()}\": Missing SoftwareComponent");
                    return null;
                }
                // Get all attributes of SoftwareComponent
                softwareComponent = getObject(softwareComponent);
                def publicType = softwareComponent.getAttribute('PublicType');
                if (publicType != 'RootFolder')
                {
                    //println("getFolderPathOfTree: ${obj.getTypeName()} ID=${obj.getId()} \"${obj.getName()}\": Not a RootFolder SoftwareComponent: ${softwareComponent.getTypeName()} ID=${softwareComponent.getId()} PublicType=\"${publicType}\" Name=${softwareComponent.getName()}");
                    return null;
                }
                break;
            }
            obj = getObject(parent);
        }
        def sb = new StringBuffer();
        result.each
        { part ->
            sb.append('/');
            sb.append(part);
        }
        return sb.toString();
    }

    // Find a folder Tree object by its path string
    // This works even for folders in repositories outside of Foundation
    // (due to the way that getObject() is implemented).
    GenericMetadataObject findTreeByPath(String path)
    {
        def reposName = this.omr.getRepositoryName();
        this.omr.setFoundationRepository();
        try
        {
            def objs = this.omr.doOMRQuery('SoftwareComponent', 
                                           "@PublicType='RootFolder'", true);
            if (!objs)
            {
                throw new RuntimeException("Missing RootFolder (BIP Service) SoftwareComponent while searching for Tree with path=" + path);
            }
            def root = objs[0];
            def obj = null
            int j = 0;
            for (int i = 0; i <= path.length(); i = j + 1)
            {
                j = path.indexOf('/', i);
                if (j < 0)
                    j = path.length();
                def part = path.substring(i, j);
                if (!part)
                    continue;
                if (obj == null)
                    obj = root.getAssociatedObject('SoftwareTrees', part);
                else
                    obj = obj.getAssociatedObject('SubTrees', part);
                if (obj == null)
                    break;
                obj = getObject(obj);
            }
            return obj;
        }
        finally
        {
            if (reposName != null)
                this.omr.setRepository(reposName);
            else
                this.omr.setFoundationRepository();
        }
    }

    GenericMetadataObject createTreeWithPath(String path)
    {
        def reposName = this.omr.getRepositoryName();
        this.omr.setFoundationRepository();
        try
        {
            def objs = this.omr.doOMRQuery('SoftwareComponent', 
                                           "@PublicType='RootFolder'", true);
            if (!objs)
            {
                throw new RuntimeException("Missing RootFolder (BIP Service) SoftwareComponent while searching for Tree with path=" + path);
            }
            def root = objs[0];
            def parent = root;
            def assocname = 'SoftwareTrees';
            def obj = null
            int j = 0;
            for (int i = 0; i <= path.length(); i = j + 1)
            {
                j = path.indexOf('/', i);
                if (j < 0)
                    j = path.length();
                def part = path.substring(i, j);
                if (!part)
                    continue;
                if (obj != null)
                {
                    assocname = 'SubTrees';
                    parent = obj;
                }
                obj = parent.getAssociatedObject(assocname, part);
                if (obj != null)
                {
                    obj = getObject(obj);
                    continue;
                }
                // Switch to the parent's repository if necessary
                def parentReposName = this.omr.getRepositoryNameForObjId(
                                                               parent.getId());
                if (parentReposName != this.omr.getRepositoryName())
                    this.omr.setRepository(parentReposName);
                // Create the Tree object
                obj = new GenericMetadataObject();
                obj.setTypeName('Tree');
                obj.setName(part);
                obj.setAttribute('PublicType', 'Folder');
                obj.setAttribute('TreeType', 'BIP Folder');
                def usageVersion = parent.getAttribute('UsageVersion');
                if (usageVersion)
                    obj.setAttribute('UsageVersion', usageVersion);
                obj = this.omr.create(obj);
                // Associate it with its parent
                def assocobjs = parent.getAssociation(assocname);
                if (assocobjs == null)
                    assocobjs = [];
                assocobjs.add(obj);
                parent.setAssociation(assocname, assocobjs);
                this.omr.store(parent);
            }
            return obj;
        }
        finally
        {
            if (reposName != null)
                this.omr.setRepository(reposName);
            else
                this.omr.setFoundationRepository();
        }
    }

    List queryAllRepositories(String objTypeName, String query, 
                              boolean matchExactType)
    {
        def result = [];
        def reposName = this.omr.getRepositoryName();
        try
        {
            this.omr.getRepositories().each
            { repos ->
                this.omr.setRepository(repos.getName());
                def objs = this.omr.doOMRQuery(objTypeName, query, 
                                               matchExactType);
                result.addAll(objs);
            }
        }
        finally
        {
            if (reposName != null)
                this.omr.setRepository(reposName);
            else
                this.omr.setFoundationRepository();
        }
        return result;
    }
}
