/*
 * Created on July 1, 2010
 * Copyright (c) 2010 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm.migrate;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.log4j.Logger;

import com.sas.solutions.brm.util.BRMUtil;

class PathFixer extends PathVisitorBase
{
    // Regular expressions to detect script with arguments
    def unx_cmd = /^([^ \t\r\n\-]\S+).*?\s+(?:-{1,2}\S+|"[^"]*").*$/; // "
    def unx_abs_cmd = /^(\/\S+)\s+\S+.*$/;
    def quoted_cmd = /^("[^"]+")\s+\S+.*$/; // "
    // Script path beginning with drive letter followed by something
    // that looks like a command-line switch or quoted argument
    def drive_letter_cmd = /^([A-Za-z]:[\\\/]\S+).*?\s+(?:-{1,2}\S+|"[^"]*").*$/; // "
    def cmd_and_args_pattern = Pattern.compile("${unx_cmd}|${unx_abs_cmd}|${quoted_cmd}|${drive_letter_cmd}");

    // Regular expression to separate arguments from eachother
    def arg_pattern = Pattern.compile(/\s*("[^"]*"|\S+)/);

    // Regular expression to detect just arguments, no command or script
    def args_pattern = Pattern.compile(/^\s*(?:-{1,2}\S+)(?:\s+(?:"[^"]*"|\S+))+/);

    // Substitution Table
    // List of lists: [from_string, regex_patterns, to_string]
    // Sorted in order of length of from_string, longest first
    def substTable = [];

    def logger = null;

    // Flag, true if the current StoredText object has been modified.
    // This is needed because parts of stored text (Nodes XML DOM object or
    // lines in multi-line text) could have been changed during different
    // method invocations.
    def storedTextModified = false;

    def modifiedObjCount = 0;
    def scannedObjectLogged = false;
    def objectChangeLogged = false;

    def PathFixer(omr, MigrateConfig config, Logger logger)
    {
        super(omr);
        this.logger = logger;
        def pathMap = config.getSection('file_and_directory_paths');
        pathMap.each
        { key, value ->
            def regexes = compilePatternFromPath(key);
            this.substTable.add([key, regexes, value]);
        }
        // Get the substitution table into the correct order:
        // Sorted by decreasing length of from_string and then by value of
        // from_string
        // Note: you have to do the sort operations in opposite order of
        // the final order you want.
        this.substTable.sort { it[0] };
        this.substTable.sort { a, b -> b[0].length() <=> a[0].length() };
    }

    def static compilePatternFromPath(String path)
    {
        def origPath = path;
        def isWindowsPath = BRMUtil.isWindowsPath(origPath);
        path = removeTrailingSlashes(path);
        // Replace backslashes with forward slashes
        path = path.replaceAll('\\\\', '/');
        def pattern = BRMUtil.escapeRegex(path);
        if (isWindowsPath)
        {
            // Pattern must match forward or backward slashes on Windows
            // Yes, all 8 backslashes are necessary
            pattern = pattern.replaceAll('/', '[/\\\\\\\\]');
        }
        int flags = 0;
        def slash = '[/]';
        if (isWindowsPath)
        {
            flags = Pattern.CASE_INSENSITIVE;
            slash = '[/\\\\]';
        }
        def createPatternVariants =
        { pat ->
            def patternVariants = [
                '^(' + pat + ')$',
                '^(' + pat + ')' + slash,
                '"(' + pat + ')"',
                '"(' + pat + ')' + slash + '[^"]*"',
                'file:(' + pat + ')' + slash,
                'file:(' + pat + ')\\s+',
                'file:(' + pat + ')$',
                'file:///(' + pat + ')$',
                'file:///(' + pat + ')' + slash,
                '=(' + pat + ')$',
                '=(' + pat + ')' + slash,
                ];
            if (pat.indexOf(' ') < 0)
            {
                // No space in path pattern -- allow space as a delimiter
                patternVariants.add('file:///(' + pat + ')\\s');
                patternVariants.add('^(' + pat + ')\\s');
                patternVariants.add('\\s(' + pat + ')' + slash);
                patternVariants.add('\\s(' + pat + ')\\s');
                patternVariants.add('=(' + pat +')\\s');
            }
            return patternVariants.join('|');
        }
        def p1 = Pattern.compile(createPatternVariants(pattern), flags);
        def p2 = null;
        path = removeTrailingSlashes(origPath);
        if (path.indexOf('\\') >= 0)
        {
            // Create another set of pattern variants, this time with each
            // backslash doubled
            path = doubleBackslashes(path);
            pattern = BRMUtil.escapeRegex(path);
            p2 = Pattern.compile(createPatternVariants(pattern), flags);
        }
        return [p1, p2];
    }

    def static removeTrailingSlashes(String path)
    {
        while (path.endsWith('/') || path.endsWith('\\'))
            path = path.substring(0, path.length() - 1);
        return path;
    }

    def static doubleBackslashes(String path)
    {
        return path.replaceAll('\\\\', '\\\\\\\\');
    }

    // Return [cmd, argsstr] if val is a script with command-line arguments
    // Return [null, null] otherwise
    def detectCmdLine(val)
    {
        def m = cmd_and_args_pattern.matcher(val);
        if (!m.matches())
            return [null, null];
        def cmd = null;
        for (i in 1 .. m.groupCount())
        {
            cmd = m.group(i);
            if (cmd != null)
                break;
        }
        def argsstr = val.substring(cmd.length());
        return [cmd, argsstr];
    }

    def processCmdLine(cmd, argsstr)
    {
        def parts = [];
        if (cmd != null)
            parts.add(fixPath(cmd));
        def m = arg_pattern.matcher(argsstr);
        def i = 0;
        while (m.find())
        {
            parts.add(argsstr.substring(i, m.start(1)));
            parts.add(fixPath(m.group(1)));
            i = m.end(1);
        }
        if (i < argsstr.length())
            parts.add(argsstr.substring(i));
        return parts.join('');
    }

    // Return true if value is a string containing command-line args
    // (but no script or command in front)
    def detectCmdArgs(val)
    {
        def m = args_pattern.matcher(val);
        return (m.matches());
    }

    // Take an string containing a path or a command line with arguments
    // containing multiple paths.  return the fixed version with substitutions
    // made according to the migrate.ini file settings. If no matches are found
    // in the substitution table, the original path string is returned
    // unchanged.
    def fixPaths(val)
    {
        def cmd_and_args = detectCmdLine(val);
        if (cmd_and_args != [null, null])
        {
            def cmd = cmd_and_args[0];
            def argsstr = cmd_and_args[1];
            return processCmdLine(cmd, argsstr);
        }
        if (detectCmdArgs(val))
        {
            return processCmdLine(null, val);
        }
        return fixPath(val);
    }

    // Take an original path or string containing at most one path,
    // return the fixed version with substitutions made according to the
    // migrate.ini file settings. If no matches are found in the
    // substitution table, the original path string is returned unchanged.
    def fixPath(origPath)
    {
        def tryPattern =
        { p, subst ->
            Matcher m = p.matcher(origPath);
            if (m.find())
            {
                for (def i in 1 .. m.groupCount())
                {
                    int start = m.start(i);
                    int end = m.end(i);
                    if (start < 0 || end < 0)
                        continue;
                    return (origPath.substring(0, start) +
                        subst + 
                        origPath.substring(end));
                }
            }
            return null;
        }
        def result = null;
        this.substTable.find
        {
            def regexes = it[1];
            def p1 = regexes[0]; // regular pattern
            def p2 = regexes[1]; // pattern with backslashes doubled
            def subst = it[2];
            // Look at doubled backslash pattern first
            if (p2 != null)
            {
                result = tryPattern(p2, doubleBackslashes(subst));
                if (result != null)
                    return true;
            }
            result = tryPattern(p1, subst);
            if (result != null)
                return true;
            return false;
        }
        if (result == null)
        {
            return origPath;
        }
        return result;
    }

    def logObjectsWithPaths(obj)
    {
        if (this.scannedObjectLogged)
            return;
        this.scannedObjectLogged = true;
        if (this.logger == null)
            return;
        this.logger.debug("Scanning Object: ${this.omr.getRepositoryName()}:${obj.getTypeName()}: (${obj.getId()}) '${obj.getName()}'");
    }

    def logObjectChange(obj)
    {
        this.modifiedObjCount += 1;
        this.objectChangeLogged = true;
        if (this.logger == null)
            return;
        this.logger.info("Object Changed: ${this.omr.getRepositoryName()}:${obj.getTypeName()}: (${obj.getId()}) '${obj.getName()}'");
    }

    def logAttrChange(obj, name, attrname, attrval, fixedAttrVal)
    {
        if (!this.objectChangeLogged)
            logObjectChange(obj);
        if (this.logger == null)
            return;
        this.logger.info("Attribute Changed: Name: ${attrname}  Old value: '${attrval}'  New Value: '${fixedAttrVal}'");
    }

    def logXMLChange(obj, parent, node, value, fixedValue)
    {
        if (!this.objectChangeLogged)
            logObjectChange(obj);
        if (this.logger == null)
            return;
        this.logger.info("XML Node Changed: Node: ${node}  New Value: '${fixedValue}'");
    }

    def logLineChange(obj, line, fixedLine)
    {
        if (!this.objectChangeLogged)
            logObjectChange(obj);
        if (this.logger == null)
            return;
        this.logger.info("Text Line Changed: Line: '${line}'  New Value: '${fixedLine}'");
    }
 
    /////////////////////////////////////////////////////////////////////
    // Begin overridable methods

    def handleRepository(repos)
    {
        def reposname = repos.getName();
        if (this.logger != null)
        {
            this.logger.info("Scanning repository: ${reposname}");
        }
    }

    def handleType(typename)
    {
        if (this.logger != null)
        {
            this.logger.info("Scanning objects of type: ${typename}");
        }
    }

    def handleObj(obj, attrnames)
    {
        this.scannedObjectLogged = false;
        this.objectChangeLogged = false;
        super.handleObj(obj, attrnames);
        if (obj.isModified())
        {
            if (!this.omr.isReadOnly())
            {
                this.omr.store(obj);
            }
        }
    }

    def handleAttrVal(obj, name, attrname, attrval)
    {
        logObjectsWithPaths(obj);
        def fixedAttrVal = '';
        if (name == 'bid.DashboardConfig' && attrname == 'DefaultValue')
        {
            // Special case for Property bid.DashboardConfig
            // which can have a semicolon-separated list of directories
            // Don't use String.split() here, it loses information
            def parts = BRMUtil.split(attrval, ';');
            for (def i=0; i<parts.size(); ++i)
            {
                parts[i] = fixPath(parts[i]);
            }
            fixedAttrVal = parts.join(';');
        }
        else
        {
            fixedAttrVal = fixPaths(attrval);
        }
        if (fixedAttrVal != attrval)
        {
            logAttrChange(obj, name, attrname, attrval, fixedAttrVal);
            obj.setAttribute(attrname, fixedAttrVal);
        }
    }

    def handleStoredText(obj)
    {
        this.storedTextModified = false;
        super.handleStoredText(obj);
    }

    def handleXML(obj, text)
    {
        def doc = super.handleXML(obj, text);
        if (this.storedTextModified)
        {
            text = domTreeToString(doc);
            obj.setAttribute('StoredText', text);
        }
    }

    def handleXMLValue(obj, parent, node, value)
    {
        logObjectsWithPaths(obj);
        def fixedValue = fixPaths(value);
        if (fixedValue != value)
        {
            this.storedTextModified = true;
            logXMLChange(obj, parent, node, value, fixedValue);
            node.setNodeValue(fixedValue);
        }
    }

    def handleMultiLineText(obj, text)
    {
        def lines = super.handleMultiLineText(obj, text);
        if (this.storedTextModified)
        {
            text = lines.join('\n');
            obj.setAttribute('StoredText', text);
        }
    }

    def handleTextLine(obj, line)
    {
        logObjectsWithPaths(obj);
        def fixedLine = fixPaths(line);
        if (fixedLine != line)
        {
            this.storedTextModified = true;
            logLineChange(obj, line, fixedLine);
            return fixedLine;
        }
        return line;
    }

    // End overridable methods
    /////////////////////////////////////////////////////////////////////
}
