package com.sas.solutions.brm.migrate;

import com.sas.solutions.brm.omr.GenericMetadataObject;
import com.sas.solutions.brm.omr.ObjectInfoHandler;
import com.sas.solutions.brm.omr.OMRUtil;

public class TextStorePropLoader extends OMRVisitorBase
{
    def tempDir = null;
    def logger = null;

    public TextStorePropLoader(omr, tempDir, logger)
    {
        super(omr);
        this.tempDir = tempDir;
        this.logger = logger;
    }

    // Load TextStores associated with components (other than levdirroot)
    def loadTextStoresWithComponents()
    {
        def componentsDir = new File(tempDir, 'components-with-textstores');
        this.logger.info("Loading Public and Private Configuration TextStore properties associated with SoftwareComponent objects from directory: ${componentsDir}");

        // Map source component ID to destination component ID
        def sourceDestMap = [:];

        // Map destination component ID to source component ID
        def destSourceMap = [:];

        def handleComponent1 = { component, publicConfig, privateConfig ->
            this.logger.info("Searching for equivalent object, phase 1: ${component.getTypeName()}: (${component.getId()}) \"${component.getName()}\"");
            def name = component.getName();
            def classId = component.getAttribute('ClassIdentifier');
            def searchClause = "@Name='${name}'";
            if (classId != null)
                searchClause = "${searchClause} AND @ClassIdentifier='${classId}'";
            def objs = omr.doOMRQuery(component.getTypeName(), searchClause, 
                                        true);
            def obj = null;
            if (objs.size() == 0)
            {
                // No name match, quietly defer to phase 2
                return;
            }
            else if (objs.size == 1)
            {
                this.logger.info("--> Found 1 match for ${searchClause}");
                obj = objs[0];
                this.logger.info("    Matching object: ${obj.getTypeName()}: (${obj.getId()}) \"${obj.getName()}\"");
            }
            else
            {
                // Name matches multiple objects, quietly defer to phase 2
                return;
            }
            if (obj != null)
            {
                def prevSourceId = destSourceMap.get(obj.getId());
                if (prevSourceId != null)
                {
                    this.logger.error("Destination ${obj.getTypeName()} with ID ${obj.getId()} mapped to source objects with IDs ${prevSourceId} and ${component.getId()}, can't load associated properties");
                }
                else
                {
                    this.logger.info("Loading properties associated with source object: ${component.getTypeName()}: (${component.getId()}) \"${component.getName()}\"");
                    sourceDestMap[component.getId()] = obj.getId();
                    destSourceMap[obj.getId()] = component.getId();
                    loadPropsForComponent(obj, publicConfig, privateConfig);
                }
            }
        }

        def handleComponent2 = { component, publicConfig, privateConfig ->
            def destId = sourceDestMap.get(component.getId());
            if (destId != null)
            {
                // This component was already successfully loaded in phase 1
                return;
            }
            this.logger.info("Searching for equivalent object, phase 2: ${component.getTypeName()}: (${component.getId()}) \"${component.getName()}\"");
            def classId = component.getAttribute('ClassIdentifier');
            def publicType = component.getAttribute('PublicType');
            def productName = component.getAttribute('ProductName');
            def searchClause = '';
            if (classId != null)
                searchClause = "@ClassIdentifier='${classId}'";
            if (publicType != null)
            {
                if (searchClause != '')
                    searchClause += ' AND ';
                searchClause += "@PublicType='${publicType}'";
            }
            if (classId == null && productName != null)
            {
                if (searchClause != '')
                    searchClause += ' AND ';
                searchClause += "@ProductName='${productName}'";
            }
            if (classId == null && publicType == null && productName == null)
            {
                searchClause = "@Name='${component.getName()}'";
            }
            def searchParam = "";
            if (searchClause != '')
                searchParam = "*[${searchClause}]";
            def objs = omr.doOMRQuery(component.getTypeName(), searchParam, 
                                        true);
            objs = objs.findAll
            {
                return destSourceMap.get(it.getId()) == null;
            }
            def obj = null;
            if (objs.size() == 0)
            {
                this.logger.error("No component match for ${searchClause}");
            }
            else if (objs.size == 1)
            {
                this.logger.info("--> Found 1 match for ${searchClause}");
                obj = objs[0];
                this.logger.info("    Matching object: ${obj.getTypeName()}: (${obj.getId()}) \"${obj.getName()}\"");
            }
            else
            {
                this.logger.info("--> Found ${objs.size()} matches for ${searchClause}");
                obj = findBestMatchByPropNames(component, publicConfig, 
                                                        privateConfig, objs);
                if (obj == null)
                    this.logger.error("        No suitable match found among the ${objs.size()} alternatives");
                else
                    this.logger.info("        Best match: ${obj.getTypeName()}: (${obj.getId()}) \"${obj.getName()}\"");
            }
            if (obj != null)
            {
                def prevSourceId = destSourceMap.get(obj.getId());
                if (prevSourceId != null)
                {
                    this.logger.error("Destination ${obj.getTypeName()} with ID ${obj.getId()} mapped to source objects with IDs ${prevSourceId} and ${component.getId()}, can't load associated properties");
                }
                else
                {
                    this.logger.info("Loading properties associated with source object: ${component.getTypeName()}: (${component.getId()}) \"${component.getName()}\"");
                    sourceDestMap[component.getId()] = obj.getId();
                    destSourceMap[obj.getId()] = component.getId();
                    loadPropsForComponent(obj, publicConfig, privateConfig);
                }
            }
        }

        iterateOverStoredComponents(componentsDir, handleComponent1);
        iterateOverStoredComponents(componentsDir, handleComponent2);

        def foundobjidsFile = new File(componentsDir, 'foundobjids.txt');
        def foundobjidsOut = new PrintWriter(
                                    new FileWriter(foundobjidsFile), true);
        destSourceMap.keySet().each
        { objid ->
            foundobjidsOut.println(objid);
        }
        foundobjidsOut.close();
    }

    def iterateOverStoredComponents(File componentsDir, Closure callback)
    {
        if (!componentsDir.isDirectory())
        {
            this.logger.warn("Directory \"${componentsDir}\" does not exist, can not load TextStore properties associated with SoftwareComponent objects.");
            return;
        }
        componentsDir.list().each
        { componentDir ->
            def componentId = componentDir;
            componentDir = new File(componentsDir, componentDir);
            if (!componentDir.isDirectory())
                return;
            def objFile = new File(componentDir, componentId + '.xml');
            def component = OMRUtil.loadObjectFromFile(objFile);
            if (component == null)
                return;
            def configs = loadPublicPrivateConfigs(componentDir);
            def publicConfig = configs[0];
            def privateConfig = configs[1];
            callback(component, publicConfig, privateConfig);
        }
    }

    def loadPublicPrivateConfigs(componentDir)
    {
        def publicConfigFile = new File(componentDir, 
                                    'Public Configuration Properties.xml');
        def publicConfig = OMRUtil.loadObjectFromFile(publicConfigFile);
        def privateConfigFile = new File(componentDir,
                                    'Private Configuration Properties.xml');
        def privateConfig = OMRUtil.loadObjectFromFile(privateConfigFile);
        return [publicConfig, privateConfig];
    }

    def loadPropsForComponent(obj, publicConfig, privateConfig)
    {
        def configs = getPublicAndPrivateConfigs(obj);
        def prevPubConfig = configs[0];
        def prevPrivConfig = configs[1];
        def newnotes = [];
        loadPropsIntoTextStore(publicConfig, prevPubConfig, newnotes);
        loadPropsIntoTextStore(privateConfig, prevPrivConfig, newnotes);
        obj.setAssociation('Notes', newnotes);
        this.omr.store(obj);
    }

    def loadPropsIntoTextStore(newConfigTextStore, prevConfigTextStore, 
                                newnotes)
    {
        if (newConfigTextStore == null && prevConfigTextStore == null)
            return;
        if (newConfigTextStore == null)
        {
            this.logger.info("    Deleting previous ${prevConfigTextStore.getTypeName()} (${prevConfigTextStore.getId()}) \"${prevConfigTextStore.getName()}\"");
            this.omr.delete(prevConfigTextStore);
            return;
        }
        if (prevConfigTextStore == null)
        {
            // Create new TextStore if necessary
            prevConfigTextStore = cloneObject(newConfigTextStore);
        }
        def proptext = newConfigTextStore.getAttribute('StoredText');
        if (proptext == null)
            proptext = '';
        this.logger.info("    Loading StoredText from ${newConfigTextStore.getTypeName()} (${newConfigTextStore.getId()}) \"${newConfigTextStore.getName()}\" into ${prevConfigTextStore.getTypeName()} (${prevConfigTextStore.getId()}) \"${prevConfigTextStore.getName()}\"");

        def prevproptext = prevConfigTextStore.getAttribute('StoredText');
        if (prevproptext == null)
            prevproptext = '';

        proptext = keepPrevFQIDs(prevproptext, proptext);

        prevConfigTextStore.setAttribute('StoredText', proptext);
        this.omr.store(prevConfigTextStore);
        newnotes.add(prevConfigTextStore);
    }

    // The only FQID property found in StoredText so far is this:
    // spmprefxpoll.dependson.webinfpltfm.fqid=A5NHT64K.AI00000L
    //
    // prevproptext: The StoredText from the "previous" (migrated) metadata
    // proptext: The StoredText from the saved file, about to be applied
    //
    def keepPrevFQIDs(prevproptext, proptext)
    {
        if (proptext.indexOf('.fqid=') < 0)
            return proptext;
        def prevPropMap = propMapFromPropText(prevproptext);
        def lines = proptext.split(/\n/, -1);
        for (int i=0; i<lines.size(); ++i)
        {
            def line = lines[i];
            def k = line.indexOf('=');
            if (k < 0)
                continue;
            def propname = line.substring(0, k);
            if (propname.endsWith('.fqid'))
            {
                this.logger.info("Found FQID property ${propname} in metadata");
                def prevfqid = prevPropMap.get(propname);
                if (prevfqid == null)
                {
                    this.logger.warn("Property ${propname} exists in migrated metadata but wasn't originally on destination system");
                    continue;
                }
                this.logger.info("Keeping FQID property value: ${propname}=${prevfqid}");
                lines[i] = propname + '=' + prevfqid;
            }
        }
        return lines.join('\n');
    }

    def propMapFromPropText(proptext)
    {
        def result = [:];
        def lines = proptext.split(/\n/, -1);
        lines.each
        { line ->
            def i = line.indexOf('=');
            if (i < 0)
                return;
            def propname = line.substring(0, i);
            def propvalue = line.substring(i + 1);
            result[propname] = propvalue;
        }
        return result;
    }

    def findBestMatchByPropNames(component, publicConfig, privateConfig, objs)
    {
        this.logger.debug("        findBestMatchByPropNames");
        def bestScore = -1.0;
        def indexWithBestScore = -1;
        def pubPropNames1 = getPropNamesFromTextStore(publicConfig);
        def privPropNames1 = getPropNamesFromTextStore(privateConfig);
        def pubsize1 = pubPropNames1.size();
        def privsize1 = privPropNames1.size();
        if ((pubsize1 + privsize1) == 0)
        {
            this.logger.debug("        No property names to match");
            return null;
        }
        for (def i=0; i<objs.size(); ++i)
        {
            def obj = objs[i];
            def configs = getPublicAndPrivateConfigs(obj);
            def pubConfig2 = configs[0];
            def privConfig2 = configs[1];
            def pubPropNames2 = getPropNamesFromTextStore(pubConfig2);
            def pubsize2 = pubPropNames2.size();
            def privPropNames2 = getPropNamesFromTextStore(privConfig2);
            def privsize2 = privPropNames2.size();
            def numPubMatches = compareLists(pubPropNames1, pubPropNames2);
            def numPrivMatches = compareLists(privPropNames1, privPropNames2);
            def pubscore = ( ((double)numPubMatches) /
                             ((double)(pubsize1 + pubsize2 - numPubMatches)) );
            def privscore = ( ((double)numPrivMatches) /
                           ((double)(privsize1 + privsize2 - numPrivMatches)) );
            def score = pubscore;
            if (privscore < score)
                score = privscore;
            if (score > 0.0 && score > bestScore)
            {
                bestScore = score;
                indexWithBestScore = i;
            }
            this.logger.debug("        ${obj.getTypeName()}: (${obj.getId()}) \"${obj.getName()}\": score=${score}");
        }
        if (indexWithBestScore >= 0)
        {
            this.logger.debug("        Best score: ${bestScore}");
            return objs[indexWithBestScore];
        }
        this.logger.debug("        No acceptable match found");
        return null;
    }

    // Compare two lists and return the size of the intersection
    static compareLists(list1, list2)
    {
        def s = new java.util.HashSet();
        // Swap the lists if list2 is shorter
        if (list2.size() < list1.size())
        {
            def temp = list1;
            list1 = list2;
            list2 = temp;
        }
        list1.each { s.add(it); }
        def result = 0;
        list2.each
        {
            if (s.contains(it))
                ++result;
        }
        return result;
    }

    // Return list [pubConfig, privConfig]
    def getPublicAndPrivateConfigs(obj)
    {
        def result = [null, null];
        def notes = obj.getAssociation('Notes');
        if (notes == null)
            notes = [];
        notes.each
        { note ->
            note = getObject(note);
            if (note.getName() == 'Public Configuration Properties')
            {
                result[0] = note;
                return;
            }
            if (note.getName() == 'Private Configuration Properties')
            {
                result[1] = note;
                return;
            }
        }
        return result;
    }

    def getPropNamesFromTextStore(obj)
    {
        if (obj == null)
            return [];
        def text = obj.getAttribute('StoredText');
        if (text == null)
            return [];
        def propNames = [];
        text.split('\n').each
        { line ->
            def i = line.indexOf('=');
            if (i < 0)
                return;
            propNames.add(line.substring(0, i));
        }
        propNames.sort();
        return propNames;
    }

    def loadDeploymentTesterServers()
    {
        def deploymentTesterDir = new File(tempDir, 'deployment-tester');
        this.logger.info("Loading Deployment Tester Server configuration from directory: ${deploymentTesterDir}");
        def deploymentTester = null;
        def searchParam = "@Name='Deployment Tester'";
        def objs = omr.doOMRQuery('ServiceComponent', searchParam, true);
        if (objs.size() > 0)
        {
            deploymentTester = objs[0];
            this.logger.info("Deleting foreign Deployment Tester Server TCPIPConnection objects");
            deploymentTester.getAssociation('ProviderConnections').each
            { connection ->
                this.logger.info("Deleting foreign ${connection.getTypeName()} (${connection.getId()}) \"${connection.getName()}\"");
                this.omr.delete(connection);
            }
        }
        this.logger.info("Deleting foreign Deployment Tester Server ServiceComponent objects");
        searchParam = "@Name='Deployment Tester Server'";
        omr.doOMRQuery('ServiceComponent', searchParam, true).each
        { component ->
            this.logger.info("Deleting foreign ${component.getTypeName()} (${component.getId()}) \"${component.getName()}\"");
            this.omr.delete(component);
        }
        def connections = [];
        deploymentTesterDir.list().each
        { filename ->
            def serverDir = new File(deploymentTesterDir, filename);
            if (!serverDir.isDirectory())
                return;
            def component = loadDeploymentTesterServer(serverDir);
            def conn = component.getAssociatedObject('SourceConnections');
            if (conn != null)
                connections.add(conn);
        }
        if (deploymentTester != null)
        {
            deploymentTester.setAssociation('ProviderConnections', connections);
            this.omr.store(deploymentTester);
        }
    }

    def loadDeploymentTesterServer(serverDir)
    {
        def component = OMRUtil.loadObjectFromFile(new File(serverDir, 
                                        'ServiceComponent.xml'));
        if (component == null)
            return;
        component = cloneObject(component);
        def connection = OMRUtil.loadObjectFromFile(new File(serverDir,
                                        'TCPIPConnection.xml'));
        if (connection != null)
        {
            connection = OMRUtil.stripAssocs(connection);
            connection.setAssociation('Source', [component]);
            connection = cloneObjectWithAssocs(connection);
            component = getObject(component);
        }
        def configs = loadPublicPrivateConfigs(serverDir);
        def publicConfig = configs[0];
        def privateConfig = configs[1];
        putNotesOnComponent(component, publicConfig, privateConfig);
        return component;
    }

    def putNotesOnComponent(component, publicConfig, privateConfig)
    {
        def newNotes = [];
        if (publicConfig != null)
            newNotes.add(cloneObject(publicConfig));
        if (privateConfig != null)
            newNotes.add(cloneObject(privateConfig))
        if (newNotes.size() > 0)
        {
            component.setAssociation('Notes', newNotes);
            this.omr.store(component);
        }
    }

    def loadLevDirTextStores()
    {
        def levdirrootDir = new File(tempDir, 'levdirroot');
        this.logger.info("Loading Public and Private Configuration TextStore properties associated with SoftwareComponent objects from directory: ${levdirrootDir}");
        this.logger.info("Deleting foreign LevDirRoot SoftwareComponent objects");
        def objs = this.omr.doOMRQuery('SoftwareComponent', 
                                    "*[@ClassIdentifier='levdirroot']", false);
        objs.each
        { obj ->
            this.logger.info("Deleting foreign ${obj.getTypeName()} (${obj.getId()}) \"${obj.getName()}\"");
            this.omr.delete(obj);
        }
        this.logger.info("Loading LevDirRoot SoftwareComponent objects");
        def handleComponent = { component, publicConfig, privateConfig ->
            this.logger.info("Loading properties associated with source object: ${component.getTypeName()}: (${component.getId()}) \"${component.getName()}\"");
            def newComponent = cloneObject(component);
            putNotesOnComponent(newComponent, publicConfig, privateConfig);
        }
        iterateOverStoredComponents(levdirrootDir, handleComponent);
    }

    // Create a copy of a metadata object, without associations
    def cloneObject(obj)
    {
        def newObj = new GenericMetadataObject();
        newObj.setTypeName(obj.getTypeName());
        OMRUtil.copyObjectAttributes(obj, newObj);
        newObj = this.omr.create(newObj);
        this.logger.info("Created new ${newObj.getTypeName()} (${newObj.getId()}) \"${newObj.getName()}\"");
        return newObj;
    }

    // Create a copy of a metadata object, with associations.
    // Warning: You probably want to call stripAssocs() first and then
    // put on only the associations you want, before calling this.
    def cloneObjectWithAssocs(obj)
    {
        def newObj = new GenericMetadataObject();
        newObj.setTypeName(obj.getTypeName());
        obj.getAttributeNames().each
        { attrname ->
            if (attrname in ['Id', 'MetadataCreated', 'MetadataUpdated'])
                return;
            newObj.setAttribute(attrname, obj.getAttribute(attrname));
        }
        obj.getAssociationNames().each
        { assocname ->
            newObj.setAssociation(assocname, obj.getAssociation(assocname));
        }
        try
        {
            newObj = this.omr.create(newObj);
        }
        catch (Error e)
        {
            this.logger.error("Error creating Metadata object: ${newObj}", e);
            return null;
        }
        this.logger.info("Created new ${newObj.getTypeName()} (${newObj.getId()}) \"${newObj.getName()}\"");
        return newObj;
    }

    def deleteOrphanTextStores()
    {
        def textstores = this.omr.doOMRQuery('TextStore', "*[@Name='Public Configuration Properties' OR @Name='Private Configuration Properties']", true);
        textstores.each
        { textstore ->
            def numassocs = 0;
            textstore.getAssociationNames().each
            { assocname ->
                def objs = textstore.getAssociation(assocname);
                numassocs += objs.size();
            }
            if (numassocs > 0)
                return;
            this.logger.info("Deleting orphan ${textstore.getTypeName()} (${textstore.getId()}) \"${textstore.getName()}\"");
            this.omr.delete(textstore);
        }
    }

    def loadOrphanTextStores()
    {
        this.logger.info("Re-loading \"orphan\" configuration properties TextStore objects");
        def objectsDir = new File(tempDir, 'orphan-textstores');
        def handleObject = { obj ->
            obj = cloneObject(obj);
        }
        iterateOverStoredObjects(objectsDir, handleObject);
    }

    def iterateOverStoredObjects(File objectsDir, Closure callback)
    {
        this.logger.info("Loading objects from directory: ${objectsDir}");
        if (!objectsDir.isDirectory())
        {
            this.logger.warn("Directory \"${objectsDir}\" does not exist, can not load objects.");
            return;
        }
        objectsDir.list().each
        { filename ->
            if (!filename.toLowerCase().endsWith('.xml'))
                return;
            def objFile = new File(objectsDir, filename);
            def obj = OMRUtil.loadObjectFromFile(objFile);
            if (obj == null)
                return;
            this.logger.info("Loading source object: ${obj.getTypeName()}: (${obj.getId()}) \"${obj.getName()}\"");
            callback(obj);
        }
    }

    def load()
    {
        omr.setFoundationRepository();
        loadTextStoresWithComponents();
        loadDeploymentTesterServers();
        loadLevDirTextStores();
        deleteOrphanTextStores();
        loadOrphanTextStores();
    }
}
