/*
 * Created on Jun 7, 2006
 * Copyright (c) 2006 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm.omr;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * Represents a metadata object. May or may not contain all of its attributes
 * and associations, depending on which method of OMRConnection
 * returned this object.
 * 
 * @author davhan
 *
 */
public class GenericMetadataObject
{
    public static final String TYPE_KEY = "Metadata Type";
    public static final String NAME_KEY = "Name";
    public static final String ID_KEY = "Id";
    
    private Map objInfo;
    private boolean modified;
    
    public GenericMetadataObject()
    {
        this(new HashMap());
    }

    /**
     * Create a GMO from a Map as returned by ObjectInfoHandler and 
     * ObjectListHandler.
     * 
     * @param objInfo
     */
    public GenericMetadataObject(Map objInfo)
    {
        this.objInfo = objInfo;
        this.modified = false;
    }

    public String getId()
    {
        String objid = getAttribute(ID_KEY);
        if (objid == null)
            objid = getAttribute("ObjRef");
        return objid;
    }

    /**
     * Get the prefix of the object ID.
     * The prefix of the ID is the repository identifer for ordinary objects,
     * and is "A0000001" for repository objects.
     * @return The prefix portion of the object's ID, or null if this object
     * does not have an ID or its ID does not have a "." in it.
     */
    public String getIdPrefix()
    {
        String objid = getId();
        if (objid == null)
            return null;
        int i = objid.indexOf(".");
        if (i < 0)
            return null;
        return objid.substring(0, i);
    }

    /**
     * Get the suffix of the object ID.
     * The suffix of the ID identifies the type and instance ID of ordinary
     * objects, and the is the repository identifier for repository objects.
     * @return The suffix portion of the object's ID, or null if this object
     * does not have an ID or its ID does not have a "." in it.
     */
    public String getIdSuffix()
    {
        String objid = getId();
        if (objid == null)
            return null;
        int i = objid.indexOf(".");
        if (i < 0)
            return null;
        return objid.substring(i + 1);
    }

    public void setId(String id)
    {
        setAttribute(ID_KEY, id);
    }
    
    public String getName()
    {
        return getAttribute(NAME_KEY);
    }
    
    public void setName(String name)
    {
        setAttribute(NAME_KEY, name);
    }
    
    public String getTypeName()
    {
        return (String)this.objInfo.get(TYPE_KEY);
    }
    
    public void setTypeName(String typeName)
    {
        this.objInfo.put(TYPE_KEY, typeName);
        this.modified = true;
    }
    
    /**
     * Get a simple (scalar) attribute of the metadata object.
     * If there is no such attribute, this method returns null.
     * 
     * @param attrName  The attribute name
     * @return the attribute value
     */
    public String getAttribute(String attrName)
    {
        Object attr = this.objInfo.get(attrName);
        if (attr == null || !(attr instanceof String))
            return null;
        return (String)attr;
    }
    
    public void setAttribute(String attrName, String value)
    {
        // Warning: NO check is done to verify that attrName is a valid
        // attribute name and not an association name.
        if (value == null)
        {
            throw new IllegalArgumentException("null value parameter");
        }
        String oldValue = (String)this.objInfo.get(attrName);
        if (oldValue != null && oldValue.equals(value))
            return;
        this.objInfo.put(attrName, value);
        this.modified = true;
    }
    
    /**
     * Return the list of all objects associated with this metadata object
     * by this association name. If there is no association of this name,
     * this method returns null. If there are no associated objects, this
     * method returns an empty list.
     * 
     * @param assocName  The association name
     * @return a List of GenericMetadataObject instances (may be empty), or null
     */
    public List getAssociation(String assocName)
    {
        List result = new ArrayList();
        Object assoc = this.objInfo.get(assocName);
        if (assoc == null)
            return null;
        if (!(assoc instanceof List))
            return null;
        for (Iterator assocIter=((List)assoc).iterator(); assocIter.hasNext(); )
        {
            Object assocItem = assocIter.next();
            if (assocItem instanceof Map)
            {
                result.add(new GenericMetadataObject((Map)assocItem));
            }
            else
            {
                result.add((GenericMetadataObject)assocItem);
            }
        }
        return result;
    }
    
    /**
     * Like getAssociation(), except this should only be called for
     * associations whose cardinality is "0 to 1". Returns null if there is
     * no associated object for that association name name. If there are
     * multiple associated objects, the first one is returned.
     * 
     * @param assocName
     * @return the associated object, or null
     */
    public GenericMetadataObject getAssociatedObject(String assocName)
    {
        List assocL = getAssociation(assocName);
        if (assocL == null)
            return null;
        if (assocL.size() == 0)
            return null;
        return (GenericMetadataObject)assocL.get(0);
    }

    /**
     * Get an associated object by name.
     * @param assocName The association name to search
     * @param objName The name of the associated object to find
     * @return The GenericMetadataObject of the associated object with
     *         the given name,
     *         null if there are no objects in that association, the
     *         association does not exist, or none of the objects have the
     *         given name.
     */
    public GenericMetadataObject getAssociatedObject(String assocName, 
                                                     String objName)
    {
        def objs = getAssociation(assocName);
        if (objs == null)
            return null;
        for (Iterator it=objs.iterator(); it.hasNext(); )
        {
            GenericMetadataObject obj = (GenericMetadataObject)it.next();
            if (obj.getName().equalsIgnoreCase(objName))
                return obj;
        }
        return null;
    }
    
    public void setAssociation(String assocName, List assocList)
    {
        if (assocList == null)
        {
            throw new IllegalArgumentException("null assocList parameter");
        }
        this.objInfo.put(assocName, assocList);
        this.modified = true;
    }
    
    public void setAssociation(String assocName, GenericMetadataObject gmo)
    {
        List assocList = new ArrayList();
        if (gmo != null)
            assocList.add(gmo);
        setAssociation(assocName, assocList);
    }
    
    public List getAttributeNames()
    {
        List result = new ArrayList();
        Object[] attrNames = getSortedUnfilteredAttrNames();
        for (int i=0; i<attrNames.length; ++i)
        {
            String attrName = (String)attrNames[i];
            if (TYPE_KEY.equals(attrName))
                continue; // skip special Metadata Type key
            Object attrValue = this.objInfo.get(attrName);
            if (attrValue != null && attrValue instanceof List)
                continue; // skip association
            result.add(attrName);
        }
        return result;
    }
    
    public List getAssociationNames()
    {
        List result = new ArrayList();
        Object[] attrNames = getSortedUnfilteredAttrNames();
        for (int i=0; i<attrNames.length; ++i)
        {
            String attrName = (String)attrNames[i];
            if (TYPE_KEY.equals(attrName))
                continue; // skip special Metadata Type key
            Object attrValue = this.objInfo.get(attrName);
            if (attrValue != null && attrValue instanceof List)
                result.add(attrName);
        }
        return result;
    }
    
    private Object[] getSortedUnfilteredAttrNames()
    {
        Object[] attrNames = this.objInfo.keySet().toArray();
        Arrays.sort(attrNames);
        return attrNames;
    }
    
    public String toString()
    {
        StringBuffer sb = new StringBuffer();
        String typeName = getTypeName();
        if (typeName != null)
            sb.append(typeName);
        else
            sb.append("GenericMetadataObject"); /*I18nOK:LINE*/
        sb.append("{Id=");
        sb.append(getId());
        sb.append(" Name=");
        sb.append(getName());
        for (Iterator attrNameIter=getAttributeNames().iterator();
             attrNameIter.hasNext(); )
        {
            String attrName = (String)attrNameIter.next();
            if (ID_KEY.equals(attrName) || NAME_KEY.equals(attrName))
                continue;
            sb.append(" ");
            sb.append(attrName);
            sb.append("=");
            sb.append(getAttribute(attrName));
        }
        for (Iterator assocNameIter=getAssociationNames().iterator();
             assocNameIter.hasNext(); )
        {
            sb.append(" ");
            String assocName = (String)assocNameIter.next();
            sb.append(assocName);
            sb.append("=");
            sb.append(getAssociation(assocName));
        }
        sb.append("}");
        return sb.toString();
    }
    
    public String getXML()
    {
        return getXML(null);
    }
    
    public String getXML(String function)
    {
        // Type name and attributes
        StringBuffer sb = new StringBuffer();
        sb.append("<");
        sb.append(getTypeName());
        for (Iterator attrNameIter=getAttributeNames().iterator();
             attrNameIter.hasNext(); )
        {
            String attrName = (String)attrNameIter.next();
            String attrValue = getAttribute(attrName);
            if (attrValue == null)
                continue;
            sb.append(" ");
            sb.append(attrName);
            sb.append("=\"");
            sb.append(OMRUtil.escapeXML(attrValue));
            sb.append("\"");
        }
        sb.append(">");
        
        // associations
        boolean firstAssoc = true;
        for (Iterator assocNameIter=getAssociationNames().iterator();
             assocNameIter.hasNext(); )
        {
            if (firstAssoc)
            {
                firstAssoc = false;
                sb.append("\n");
            }
            String assocName = (String)assocNameIter.next();
            sb.append("  <");
            sb.append(assocName);
            if (function != null)
            {
                sb.append(" Function=\"");
                sb.append(function);
                sb.append("\"");
            }
            sb.append(">\n");
            List assocList = getAssociation(assocName);
            for (Iterator assocObjIter=assocList.iterator(); 
                 assocObjIter.hasNext(); )
            {
                GenericMetadataObject assocObj = 
                    (GenericMetadataObject)assocObjIter.next();
                sb.append("    <");
                sb.append(assocObj.getTypeName());
                sb.append(" ObjRef=\"");
                String objId = assocObj.getId();
                if (objId == null)
                {
                    objId = assocObj.getAttribute("ObjRef");
                }
                if (objId == null)
                {
                    // Could put code here to use Id="" attribute instead,
                    // and append all other attributes, for creating new object
                    // and association at the same time. But I don't have a 
                    // need for that use case yet.
                    throw new Error(
                            "Trying to associate to non-existent object");
                }
                sb.append(objId);
                sb.append("\"/>\n");
            }
            sb.append("  </");
            sb.append(assocName);
            sb.append(">\n");
        }
        
        // closing
        sb.append("</");
        sb.append(getTypeName());
        sb.append(">");
        
        return sb.toString();
    }
    
    public boolean isModified()
    {
        return this.modified;
    }
    
    public void setModified(boolean modified)
    {
        this.modified = modified;
    }
    
    /**
     * Compare with another GenericMetadataObject for consistent sorting
     * @param gmo2  The other GenericMetadataObject
     * @return an integer similar to what String.compareTo() returns
     */
    public int compareTo(GenericMetadataObject gmo2)
    {
        GenericMetadataObject gmo1 = this;
        int result = 0;
        String[] compareKeys = [TYPE_KEY, NAME_KEY, ID_KEY];
        for (int i=0; i<compareKeys.length; ++i)
        {
            String v1 = gmo1.getAttribute(compareKeys[i]);
            String v2 = gmo2.getAttribute(compareKeys[i]);
            if (v1 == null && v2 == null)
                continue;
            // null sorts as less than any other value
            if (v1 == null)
                return -1;
            if (v2 == null)
                return 1;
            result = v1.compareTo(v2);
            if (result != 0)
                return result;
        }
        return 0;
    }
    
    public static Comparator getComparator()
    {
        return new GenericMetadataObjectComparator();
    }
}

class GenericMetadataObjectComparator implements Comparator
{
    public int compare(Object o1, Object o2)
    {
        GenericMetadataObject gmo1 = (GenericMetadataObject)o1;
        GenericMetadataObject gmo2 = (GenericMetadataObject)o2;
        return gmo1.compareTo(gmo2);
    }
};
