// MetadataMover.java
// Copyright (C) 2009 by SAS Institute Inc., Cary, NC 27513-2414
// David Handy  Tue 15 Dec 2009
package com.sas.solutions.brm.omr;

import java.nio.charset.Charset;

import org.apache.log4j.Logger;
import org.w3c.dom.Document; // result of DoRequest()

import com.sas.meta.SASOMI.IOMI;
//import com.sas.metadata.MdObjectFactory;
import com.sas.metadata.remote.CMetadata;
import com.sas.metadata.remote.Login;
import com.sas.metadata.remote.MdFactory;
import com.sas.metadata.remote.MdFactoryImpl;
import com.sas.metadata.remote.MdOMRConnection;
import com.sas.metadata.remote.MdOMIUtil;
import com.sas.workspace.Repository;

import com.sas.solutions.brm.Constants;
import com.sas.solutions.brm.util.BRMUtil;

/**
 * Move metadata from machine to machine.
 * @see com.sas.tools.webappconfig.tasks.MetadatasrvMigrate
 */
public class MetadataMover
{
    Properties configProps = null;
    Logger logger = null;

    /**
     * Construct a MetadataMover with configuration properties.
     */
    public MetadataMover(Properties props, Logger logger)
    {
        this.configProps = props;
        this.logger = logger;
    }

    /**
     * Fix entries in adminUsers.txt or trustedUsers.txt file. The file is
     * modified in-place. User IDs of the form "sourcemachine\\username" are
     * translated to "destinationmachine\\username". The name of the source
     * machine is read from the backup-info.properties file in the metadata
     * backup directory.
     *
     * @param usersFile  Users file to be fixed.
     * @param outputDir  Name of the main BRM backup directory
     * @param specialUser Name of special user to add to end of file, or
     *                      null. Include "*" prefix at front of this
     *                      parameter when fixing adminUsers.txt file.
     */
    public void fixUsersFile(File usersFile, String outputDir,
                                    String specialUser)
    {
        if (this.logger != null)
        {
            this.logger.info("Fixing user " + specialUser + " in file " + 
                                usersFile);
        }
        Properties backupInfo = getMetadataBackupInfo(outputDir);
        String sourceEncoding = backupInfo.getProperty("file.encoding");
        Charset sourceCharset = null;
        if (sourceEncoding == null)
            sourceCharset = Charset.defaultCharset();
        else
            sourceCharset = Charset.forName(sourceEncoding);
        String name = usersFile.getName();
        File tempFile = File.createTempFile(name + "-", ".tmp");
        Reader reader = new BufferedReader(
                            new InputStreamReader(
                                new FileInputStream(usersFile),
                                    sourceCharset));
        // Using file.encoding property here instead of defaultEncoding()
        // method to facilitate testing case where source and destination
        // default character sets are different. Charset.defaultEncoding()
        // does not respond to change to file.encoding property.
        Charset destCharset = Charset.forName(
                                System.getProperty("file.encoding"));
        Writer writer = new PrintWriter(
                            new BufferedWriter(
                                new OutputStreamWriter(
                                    new FileOutputStream(tempFile), 
                                        destCharset)));
        String sourceHostName = backupInfo.getProperty("hostname");
        sourceHostName = sourceHostName.trim().toLowerCase();
        boolean specialUserFound = false;
        while (true)
        {
            String line = reader.readLine();
            if (line == null)
                break;
            String trimLine = line.trim();
            if (trimLine.startsWith(";"))
            {
                // Leave comment line unchanged
                writer.println(line);
                continue;
            }
            String prefix = "";
            if (trimLine.startsWith("*"))
            {
                prefix = "*";
                trimLine = trimLine.substring(1);
            }
            if (trimLine.toLowerCase().startsWith(sourceHostName + "\\"))
            {
                String destHostName = BRMUtil.getLocalHostName();
                int sepIndex = trimLine.indexOf("\\");
                trimLine = (destHostName + trimLine.substring(sepIndex));
            }
            trimLine = prefix + trimLine;
            if (specialUser != null &&
                trimLine.equalsIgnoreCase(specialUser))
            {
                specialUserFound = true;
            }
            writer.println(trimLine);
        }
        if (specialUser != null && !specialUserFound)
        {
            writer.println(specialUser);
        }
        writer.close();
        reader.close();
        BRMUtil.copyFile(tempFile, usersFile);
        tempFile.delete();
    }

    public static Properties getMetadataBackupInfo(String outputDir)
    {
        File backupDir = new File(outputDir, "metadata");
        def agentHostName = BRMUtil.getAgentHostName(backupDir);
        def configDir = new File(outputDir, "config");
        def agentConfigDir = new File(configDir, agentHostName);
        return BRMUtil.getBackupInfo(agentConfigDir);
    }

    /**
     * Fix Login User IDs in metadata.  User IDs of the form
     * "sourcemachine\\username" are translated to
     * "destinationmachine\\username". The name of the source machine is
     * read from the backup-info.properties file in the metadata backup
     * directory. The local machine is assumed to be the destination host.
     *
     * @param outputDir  Name of the main BRM backup directory
     * @return A List of Map objects containing the keys "before" and
     *         "after", describing the User IDs that were changed.
     */
    public List fixMetadataLogins(String outputDir)
    {
        if (this.logger != null)
        {
            this.logger.info("Fixing Login User IDs in metadata");
        }
        Properties backupInfo = getMetadataBackupInfo(outputDir);
        String sourceHostName = backupInfo.getProperty("hostname");
        sourceHostName = sourceHostName.trim().toLowerCase();
        String destHostName = BRMUtil.getLocalHostName();
        Properties props = this.configProps;
        String host = props.getProperty("BRM.host.metadata");
        String port = props.getProperty("BRM.port.metadata");
        String user = props.getProperty("BRM.admin.user");
        String password = props.getProperty("BRM.admin.password");

        JomaSupport joma = new JomaSupport();
        joma.connect(host, port, user, password);
        List result = new ArrayList();
        List logins = joma.getMetadataObjects(null, "Login", "");
        Iterator loginIt = logins.iterator();
        while (loginIt.hasNext())
        {
            Login login = (Login)loginIt.next();
            String userID = login.getUserID();
            if (userID.toLowerCase().startsWith(sourceHostName + "\\"))
            {
                int sepIndex = userID.indexOf("\\");
                String newUserID = (destHostName + userID.substring(sepIndex));
                Map record = new HashMap();
                record.put("before", userID);
                record.put("after", newUserID);
                result.add(record);
                login.setUserID(newUserID);
                login.updateMetadataAll();
            }
        }
        joma.disconnect();

        return result;
    }

    /**
     * Fix the metadata admin credentials in metaparms.sas so that
     * MetadataServer backup and restore commands will work again.
     */
    public void fixMetaParms()
    {
        Properties props = this.configProps;
        String sasConfigDir = props.getProperty("BRM.config.dir");
        File SASMetaDir = new File(sasConfigDir, "SASMeta");
        File MetadataServerDir = new File(SASMetaDir, "MetadataServer");
        File metaParms = new File(MetadataServerDir, "metaparms.sas");
        String user = props.getProperty("BRM.admin.user");
        String password = props.getProperty("BRM.admin.password");
        if (this.logger)
        {
            this.logger.info("Setting admin user ID to " + user + " in file " +
                                metaParms + " and fixing the password also.");
        }

        // create the backup    
        File metaParmsBak = new File(MetadataServerDir, "metaparms.sas.bak");
        BRMUtil.copyFile(metaParms, metaParmsBak);    
 
        // string replace the metauser    
        BRMUtil.stringReplaceAllInFile(metaParms.toString(),"metauser='.*?'",BRMUtil.escapeRepl("metauser='" + user +"'"));

        // string replace the metapass    
        BRMUtil.stringReplaceAllInFile(metaParms.toString(),"metapass='.*?'",BRMUtil.escapeRepl("metapass='" + password +"'"));

    }

    public void testFixMetaParms()
    {
        println("Testing fixMetaParms()");
        fixMetaParms();
    }

    public void testFixMetadataLogins()
    {
        println("Login User IDs before:");
        printUserLogins();

        println();
        String outputDir = this.configProps.getProperty("BRM.output.dir");
        List changes = fixMetadataLogins(outputDir);
        Iterator changeIt = changes.iterator();
        while (changeIt.hasNext())
        {
            Map change = (Map)changeIt.next();
            String before = change.get("before");
            String after = change.get("after");
            println("Changed ${before} to ${after}");
        }

        println();
        println("Login User IDs after:");
        printUserLogins();
    }

    public void printUserLogins()
    {
        JomaSupport joma = new JomaSupport();
        Properties props = this.configProps;
        String host = props.getProperty("BRM.host.metadata");
        String port = props.getProperty("BRM.port.metadata");
        String user = props.getProperty("BRM.admin.user");
        String password = props.getProperty("BRM.admin.password");
        joma.connect(host, port, user, password);
        List logins = joma.getMetadataObjects(null, "Login", "");
        Iterator loginIt = logins.iterator();
        while (loginIt.hasNext())
        {
            Login login = (Login)loginIt.next();
            println(login.getUserID());
        }
        joma.disconnect();
    }
}

// end-of-file
