// MetadataServerWaiter.java
// Copyright (c) 2006 by SAS Institute Inc., Cary, NC 27513
// Created by David Handy  29 Jan 2007

package com.sas.solutions.brm.omr;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.ConnectException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.SocketChannel;
import java.util.Iterator;
import java.util.Properties;
import java.util.Set;

import org.apache.log4j.Logger;

import com.sas.solutions.brm.Constants;
import com.sas.solutions.brm.errors.BRMConfigurationException;
import com.sas.solutions.brm.util.BRMUtil;

public class MetadataServerWaiter
{
    private static final Logger LOG = Logger.getLogger(
            MetadataServerWaiter.class);
    
    private Properties controlProps;

    public MetadataServerWaiter(Properties controlProps)
    {
        this.controlProps = controlProps;
    }

    /**
     * Wait until the metadata server is no longer running.
     *
     * First, wait until the files in the repository manager directory are
     * no longer locked. This only works on Windows, because on UNIX there
     * basically is no file locking.
     *
     * Next, wait until the server.pid file is gone. This only works on
     * UNIX. (MetadataServer.sh creates server.pid on startup and deletes it
     * on shutdown.)
     * 
     * @param timeoutSeconds  Maximum time to wait
     * @return true if server shutdown was detected, false if timeout.
     */
    public boolean waitTillServerIsStopped(final int timeoutSeconds)
    {
        LOG.info("Waiting up to " + timeoutSeconds + 
                " seconds for metadata server to shut down.");
        final long timeoutMilliSeconds = timeoutSeconds * 1000L;

        // C:\\SAS\\SASSolutionsConfig\\Lev1
        String n1 = this.controlProps.getProperty("BRM.config.dir");
        if (n1 == null || n1.length() == 0)
        {
            throw new BRMConfigurationException(
                "Missing property SOLUTIONS_SAS_SERVER_DIR " +
                "in " + Constants.BRM_CONFIG_FILE);
        }
        String n2 = "SASMeta/MetadataServer/MetadataRepositories";
        
        File repositoriesDir = new File(n1, n2);
        File metadataServerDir = repositoriesDir.getParentFile();
        File rposmgrDir = new File(metadataServerDir, "rposmgr");
        
        if (!rposmgrDir.isDirectory())
        {
            throw new Error("Missing repository manager directory: " + 
                    rposmgrDir);
        }
        
        boolean atLeastOneTimeThroughLoop = false;
        boolean allFilesUnlocked = false;
        long retryIntervalMilliSeconds = 2L * 1000L;
        long t0 = System.currentTimeMillis();
        for ( ; (!atLeastOneTimeThroughLoop) || 
                 ((System.currentTimeMillis() - t0) < timeoutMilliSeconds); 
             delayMS(retryIntervalMilliSeconds))
        {
            atLeastOneTimeThroughLoop = true;
            allFilesUnlocked = true;
            String[] fileNames = rposmgrDir.list();
            for (int i=0; i<fileNames.length; ++i)
            {
                String fileName = fileNames[i];
                File file = new File(rposmgrDir, fileName);
                if (file.isDirectory())
                    continue;
                if (!file.canRead() && !file.canWrite())
                {
                    allFilesUnlocked = false;
                    break;
                }
                FileInputStream fis = null;
                try
                {
                    fis = new FileInputStream(file);
                }
                catch (FileNotFoundException e1)
                {
                    if (file.isFile())
                    {
                        // File *is* found, but it is locked
                        allFilesUnlocked = false;
                        break;
                    }
                    throw new Error(
                        "Unexpected error opening file " + file, e1);
                }
                try
                {
                    fis.close();
                }
                catch (IOException e2)
                {
                    throw new Error(
                        "Unexpected error closing file " + file, e2);
                }
            }
            if (allFilesUnlocked || timeoutMilliSeconds == 0)
                break;
        }
        if (!allFilesUnlocked)
        {
            LOG.warn("Timed out after " + timeoutSeconds + 
                    " seconds while waiting for metadata server to stop");
            return false;
        }

        LOG.info("Metadata Server files are unlocked");

        File serverDotPID = new File(metadataServerDir, "server.pid");
        for ( ; ((System.currentTimeMillis() - t0) < timeoutMilliSeconds); 
             delayMS(retryIntervalMilliSeconds))
        {
            if (!serverDotPID.exists())
                break;
        }
        if (serverDotPID.exists())
        {
            LOG.warn("Timed out after " + timeoutSeconds +
                    " seconds while wating for " + serverDotPID + 
                    " to be deleted.");
            return false;
        }

        LOG.info("Metadata Server server.pid file removed/not created");

        return true;
    }

    /**
     * Wait until the metadata server has finished starting.
     *
     * According to emails received from Michael Burns
     * (Michael.Burns@sas.com) on July 7 and 8, 2008, it is sufficient to
     * detect that the metadata server is listening on its TCP port. See
     * also Michael Burns' notes on defect S0520495. I was skeptical, but I
     * decided to trust Michael as he is a metadata server developer.
     * 
     * @param timeoutSeconds  Maximum time to wait
     * @return true if server startup was detected, false if timeout.
     */
    public boolean waitTillServerIsStarted(final int timeoutSeconds)
        throws Exception
    {
        LOG.info("Waiting up to " + timeoutSeconds + 
                " seconds for metadata server to start up.");
        final long timeoutMilliSeconds = timeoutSeconds * 1000L;

        long retryIntervalMilliSeconds = 2L * 1000L;
        String host = this.controlProps.getProperty("BRM.host.metadata");
        if (host == null || host.length() == 0)
        {
            throw new BRMConfigurationException(
                "Missing property BRM.host.metadata in " +
                Constants.BRM_CONFIG_FILE);
        }

        String port = this.controlProps.getProperty("BRM.port.metadata");
        if (port == null || port.length() == 0)
        {
            throw new BRMConfigurationException(
                "Missing property BRM.port.metadata in " +
                Constants.BRM_CONFIG_FILE);
        }
        int portNum = Integer.parseInt(port);
        // Look up host name and convert to IP address
        InetAddress hostaddr = InetAddress.getByName(host);
        InetSocketAddress sockaddr = new InetSocketAddress(hostaddr, portNum);
        // Try making TCP connection to host at port, with timeout
        long t0 = System.currentTimeMillis();
        for ( ; ((System.currentTimeMillis() - t0) < timeoutMilliSeconds); 
             delayMS(retryIntervalMilliSeconds))
        {
            if (trySockConnect(sockaddr, retryIntervalMilliSeconds))
                return true;
        }
        return false;
    }

    public static boolean trySockConnect(InetSocketAddress sockaddr, 
            long timeoutMilliSeconds)
    {
        // Try making TCP connection to sockaddr, with timeout
        try
        {
            SocketChannel channel = SocketChannel.open();
            channel.configureBlocking(false);
            channel.connect(sockaddr);
            Selector selector = Selector.open();
            SelectionKey key = channel.register(selector, 
                    SelectionKey.OP_CONNECT);
            try {
                int numKeys = selector.select(timeoutMilliSeconds);
                Set selectedKeys = selector.selectedKeys();
                Iterator it = selectedKeys.iterator();
                while (it.hasNext())
                {
                    // We can assume this is the connecting channel key
                    SelectionKey k = (SelectionKey)it.next();
                    SocketChannel c = (SocketChannel)k.channel();
                    try {
                        return c.finishConnect();
                    } catch (ConnectException e) {
                        return false;
                    }
                }
            }
            finally {
                selector.close();
                channel.close();
            }
        }
        catch (java.io.IOException e)
        {
            return false;
        }
        return false;
    }
    
    private static void delayMS(long delayMilliSeconds)
    {
        try
        {
            Thread.sleep(delayMilliSeconds);
        }
        catch (InterruptedException e)
        {
            ; // ignore
        }
    }

    public static final String[] USAGE_MESSAGE = [
        "MetadataServerWaiter",
        "Utility to wait for the SAS Metadata Server to finish starting or stopping",
        "",
        "Use:",
        "java com.sas.solutions.brm.omr.MetadataServerWaiter start|stop",
        "",
        "Note: This does not actually start or stop the server,",
        "it just waits for the start or stop operation to complete.",
    ];

    public static void printUsage()
    {
        for (int i=0; i<USAGE_MESSAGE.length; ++i)
        {
            System.out.println(USAGE_MESSAGE[i]);
        }
    }

    public static void main(String[] args) throws Exception
    {
        com.sas.solutions.brm.logging.LogConfigurator.configure();
        Properties controlProps = new Properties();
        BRMUtil.loadPropFileAsResource(Constants.BRM_CONFIG_FILE, controlProps,
                                            LOG);
        int timeoutSeconds = 30 * 60;
        MetadataServerWaiter msw = new MetadataServerWaiter(controlProps);
        String waitFor = "stop";
        if (args.length > 0)
            waitFor = args[0];
        if ("stop".equalsIgnoreCase(waitFor))
            msw.waitTillServerIsStopped(timeoutSeconds);
        else if ("start".equalsIgnoreCase(waitFor))
            msw.waitTillServerIsStarted(timeoutSeconds);
        else
            printUsage();
    }
}

// end-of-file
