/*
 * Created on Jul 7, 2006
 * Copyright (c) 2006 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm.omr;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

public class OMRPropertyFileLoader
{
    private static final Logger LOG = Logger.getLogger(OMRPropertyFileLoader.class);
    
    private OMRConnection omr;

    public OMRPropertyFileLoader(OMRConnection omr)
    {
        this.omr = omr;
    }
    
    /**
     * Load a Java properties file into the OMR. Each Property in the file is
     * loaded into a Property object in the Foundation repository of the OMR 
     * named groupName+"."+propName. The Property is associated with a 
     * PropertySet whose name is the groupName parameter value. Call 
     * omr.setRepository() before calling this method, to select the repository
     * that will be used.
     * 
     * Previously-existing properties of the same name will be deleted. (This 
     * is to avoid duplicate properties, and to clean up after previous installs
     * which create duplicate property objects, a problem observed with 
     * Solutions x.2)
     * 
     * @param omr
     * @param file
     * @param groupName
     * @throws IllegalArgumentException if the file is not valid.
     */
    public void loadPropertyFile(File file, String groupName)
    {
        if (!file.isFile())
        {
            throw new IllegalArgumentException(
                    MessageFormat.format("Property file {0} does not exist", 
                            (Object[])[file.getAbsolutePath()]));
        }
        try
        {
            FileInputStream fin = new FileInputStream(file);
            try
            {
                Properties props = new Properties();
                props.load(fin);
                loadProperties(props, groupName);
            }
            finally
            {
                fin.close();
            }
        }
        catch (IOException e)
        {
            // I already checked earlier in this method to see if the file
            // exists, therefore any IOException at this point is unexpected.
            throw new Error(e);
        }
    }
    
    public void loadPropertyFile(File file)
    {
        loadPropertyFile(file, groupNameFromFile(file));
    }
    
    /**
     * Like loadPropertiesFile(), except that it takes a Properties parameter
     * instead of a File parameter.
     * 
     * @param omr
     * @param props
     * @param groupName
     */
    public void loadProperties(Properties props, String groupName)
    {
        // Validate parameters
        if (omr == null)
        {
            throw new IllegalArgumentException("null omr parameter");
        }
        if (!omr.isConnected())
        {
            throw new IllegalStateException("Call omr.connect() first");
        }
        if (omr.getRepositoryName() == null || 
            omr.getRepositoryName().length() == 0)
        {
            throw new IllegalStateException("Call omr.setRepository() first");
        }
        if (props == null)
        {
            throw new IllegalArgumentException("null props parameter");
        }
        if (groupName == null || groupName.length() == 0)
        {
            throw new IllegalArgumentException(
                    "null or empty groupName parameter");
        }
        
        // Get list of all properties that look like they should belong to this
        // property group. The "=:" in the query means "starting with".
        List omrProps = omr.doOMRQuery(
                "Property", "*[@Name=:'" + groupName + ".']");
        // Create a Map to quickly index OMR Properties in this group by name.
        // At the same time, find the PropertySet and PropertyType to which
        // new Property objects in this group should be associated.
        Map omrNamePropsMap = new HashMap();
        GenericMetadataObject assocPropSet = null;
        GenericMetadataObject assocPropType = null;
        for (Iterator omrPropIter=omrProps.iterator(); omrPropIter.hasNext(); )
        {
            GenericMetadataObject gmo = 
                (GenericMetadataObject)omrPropIter.next();
            List omrPropsForName = (List)omrNamePropsMap.get(gmo.getName());
            if (omrPropsForName == null)
            {
                omrPropsForName = new ArrayList();
            }
            omrPropsForName.add(gmo);
            omrNamePropsMap.put(gmo.getName(), omrPropsForName);
            
            GenericMetadataObject curPropSet = gmo.getAssociatedObject(
                    "AssociatedPropertySet");
            if (assocPropSet == null)
            {
                assocPropSet = curPropSet;
            }
            else if (curPropSet != null)
            {
                if (! (assocPropSet.getId().equals(curPropSet.getId())))
                {
                    LOG.debug(
                            "Multiple PropertySets found for property group " + 
                            groupName);
                }
            }
            
            GenericMetadataObject curPropType = gmo.getAssociatedObject(
                    "OwningType");
            if (assocPropType == null)
            {
                assocPropType = curPropType;
            }
            else if (curPropType != null)
            {
                if (! (assocPropType.getId().equals(curPropType.getId())))
                {
                    LOG.debug(
                            "Multple PropertyTypes found for property group " + 
                            groupName);
                }
            }
        }
        
        if (assocPropSet == null)
        {
            LOG.warn("No previous PropertySet for property group " + 
                    groupName + ". Cannot load properties file into metadata");
            return;
        }
        
        if (assocPropType == null)
        {
            List stringPropTypeL = omr.doOMRQuery("PropertyType", 
                    "*[@Name='String']");
            if (stringPropTypeL == null || stringPropTypeL.size() == 0)
            {
                // Very highly unlikely
                LOG.error("No String PropertyType in the OMR");
            }
            assocPropType = (GenericMetadataObject)stringPropTypeL.get(0);
        }
        
        for (Iterator propNameIter=props.keySet().iterator(); 
             propNameIter.hasNext(); )
        {
            String key = (String)propNameIter.next();
            String propName = groupName + "." + key;
            List prevOMRProps = (List)omrNamePropsMap.get(propName);
            GenericMetadataObject omrProp = null;
            if (prevOMRProps != null && prevOMRProps.size() > 1)
            {
                for (Iterator prevOMRPropIter=prevOMRProps.iterator(); 
                     prevOMRPropIter.hasNext(); )
                {
                    GenericMetadataObject prevOMRProp = 
                        (GenericMetadataObject)prevOMRPropIter.next();
                    LOG.debug("Deleting previous OMR Property: " + prevOMRProp);
                    omr.delete(prevOMRProp);
                }
            }
            else if (prevOMRProps != null && prevOMRProps.size() == 1)
            {
                omrProp = (GenericMetadataObject)prevOMRProps.get(0);
            }
            if (omrProp == null)
            {
                omrProp = new GenericMetadataObject();
                omrProp.setTypeName("Property");
                omrProp.setName(propName);
                omrProp.setAssociation("OwningType", assocPropType);
                omrProp.setAssociation("AssociatedPropertySet", assocPropSet);
                LOG.debug("Creating OMR Property: " + omrProp);
                omrProp = omr.create(omrProp);
            }
            String prevPropVal = omrProp.getAttribute("DefaultValue");
            String propVal = props.getProperty(key);
            if (propVal != null)
            {
                // I call trim() because older Solutions versions sometimes
                // had trailing spaces after property values. FYI, OMR
                // strips trailing spaces from strings!
                propVal = propVal.trim();
            }
            if (prevPropVal == null || (!prevPropVal.equals(propVal)))
            {
                LOG.debug("Setting OMR Property: " + propName + "=" + propVal);
                omrProp.setAttribute("DefaultValue", propVal);
                omr.store(omrProp);
            }
            else
            {
                LOG.debug("Leaving OMR Property unchanged: " + propName);
            }
        }
    }
    
    public static String groupNameFromFile(File file)
    {
        String s = file.getName();
        int i = s.lastIndexOf(".");
        if (i >= 0)
        {
            s = s.substring(0, i);
        }
        return s.toLowerCase();
    }

    /**
     * @param args
     */
    public static void main(String[] args) throws Exception
    {
        System.out.println("Testing loading Properties from file into OMR");
        if (args.length < 2)
        {
            System.err.println(
                    "I need at least two parameters: OMR host and filename.");
            return;
        }
        String omrHost = args[0];
        String filename = args[1];
        com.sas.solutions.brm.logging.LogConfigurator.configure();
        Logger.getRootLogger().setLevel(Level.DEBUG);
        OMRConnection omr = new OMRConnection(
                omrHost, "8561", "sasadm", "AdminAdmin1"
                );
        omr.connect();
        omr.setFoundationRepository();
        LOG.info("Connected to OMR server: " + omr.getHost());
        if (args.length >= 3 && args[2].equalsIgnoreCase("readonly"))
        {
            LOG.info("Setting OMR connection to readonly");
            omr.setReadOnly(true);
        }
        File file = new File(filename);
        String groupName = groupNameFromFile(file);
        OMRPropertyFileLoader loader = new OMRPropertyFileLoader(omr);
        loader.loadPropertyFile(file, groupName);
        omr.close();
    }

}
