/*
 * Created Fri 26 Feb 2010
 * Copyright (c) 2010 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm.util;

import java.sql.Connection;
import java.sql.Driver;
import java.sql.DriverPropertyInfo;
import java.sql.DriverManager;
import java.sql.SQLException;

import org.codehaus.groovy.tools.RootLoader;
import org.codehaus.groovy.tools.LoaderConfiguration;

/**
 * BRM (Backup, Restore, Migration tool) database utilities.
 */
public class BRMDBUtil
{
    private static final String BAD = "(non-existent)";
    private static final String DUP = "(duplicate-definition)";

    private static driverTableInitialized = false;
    private static driverTable = [
            // driver, directory, jarname, drivername
            // If string starts with $ it is looked up in config props
            [null, '$data.jdbc.jar.dir', '$data.dbms.jdbc.jar.name',
             '$dbms.mysql.driver'],
            // dbms.mysql.driver is not defined on every system
            [null, '$data.jdbc.jar.dir', '$data.dbms.jdbc.jar.name',
             'com.mysql.jdbc.Driver'],
        ];

    private static RootLoader rootLoader = null;

    /**
     * Initialize the locations of JDBC drivers from configuration
     * properties.
     * @return the number of JDBC drivers successfully loaded
     */
    public static int initJDBCDrivers(props, log)
    {
        def debug = { msg ->
            if (log == null)
                return;
            log.debug(msg);
        }
        def info = { msg ->
            if (log == null)
                return;
            log.info(msg);
        }
        def warn = { msg ->
            if (log == null)
                return;
            log.warn(msg);
        }
        def getprop = { propname_or_value ->
            if (propname_or_value.startsWith('$'))
                return props.get(propname_or_value.substring(1));
            else
                return propname_or_value;
        }
        def driverSpec = { row ->
            def dirspec = row[1];
            def dir = getprop(dirspec);
            def jarspec = row[2];
            def jar = getprop(jarspec);
            def drivernamespec = row[3];
            def drivername = getprop(drivernamespec);
            def sb = new StringBuffer();
            if (dir != null && dirspec != dir)
                sb.append("dir: ${dirspec}=${dir}");
            else
                sb.append("dir: ${dirspec}");
            if (jar != null && jarspec != jar)
                sb.append(" jar: ${jarspec}=${jar}");
            else
                sb.append(" jar: ${jarspec}");
            if (drivername != null && drivernamespec != drivername)
                sb.append(" drivername: ${drivernamespec}=${drivername}");
            else
                sb.append(" drivername: ${drivernamespec}");
            return sb.toString();
        }
        def numDriversFound = 0;
        def driverSpecSet = new HashSet(); // key is dir:jar:drivername
        driverTable.each
        { row ->
            def driver = row[0];
            def dir = getprop(row[1]);
            def jar = getprop(row[2]);
            def drivername = getprop(row[3]);
            def driverspec = "${dir}:${jar}:${drivername}".toString();
            if (driver != null)
            {
                if (dir != null && jar != null && drivername != null)
                    driverSpecSet.add(driverspec);
                if (driver != BAD && driver != DUP)
                {
                    numDriversFound += 1;
                    debug("Already found JDBC driver with spec: ${driverSpec(row)}");
                }
                else
                {
                    debug("Already searched for JDBC driver with spec: ${driverSpec(row)}");
                }
                return;
            }
            if (driverSpecSet.contains(driverspec))
            {
                driver = DUP;
                row[0] = driver;
                debug("Skipping duplicate JDBC driver with spec: ${driverSpec(row)}");
                return;
            }
            if (dir == null || jar == null || drivername == null)
            {
                driver = BAD;
                row[0] = driver;
                debug("Skipping JDBC driver with undefined property: ${driverSpec(row)}");
                return;
            }
            driverSpecSet.add(driverspec);
            def jarFile = new File(dir, jar);
            if (!jarFile.exists())
            {
                driver = BAD;
                row[0] = driver;
                debug("Skipping JDBC driver with non-existent jar file: ${driverSpec(row)}");
                return;
            }
            debug("Attempting to load JDBC driver with spec: ${driverSpec(row)}");
            driver = loadDriver(jarFile, drivername);
            if (driver == null)
            {
                driver = BAD;
                row[0] = driver;
                debug("Skipping non-existent JDBC driver with spec: ${driverSpec(row)}");
                return;
            }
            row[0] = driver;
            numDriversFound += 1;
            info("Found JDBC driver with spec: ${driverSpec(row)}: ${driver.getClass().getName()}");
        }
        if (numDriversFound == 0)
        {
            warn("No JDBC drivers found");
        }
        else if (numDriversFound == 1)
        {
            debug("Found one JDBC driver");
        }
        else
        {
            debug("Found ${numDriversFound} JDBC drivers");
        }
        driverTableInitialized = true;
        return numDriversFound;
    }

    /**
     * Load a driver from a jar file. Try not to call this twice with the
     * same parameters as it will add duplicate jar files to the classpath.
     */
    public static Driver loadDriver(File jarFile, String drivername)
    {
        if (rootLoader == null)
        {
            def loaderConfig = new LoaderConfiguration();
            loaderConfig.setRequireMain(false);
            rootLoader = new RootLoader(loaderConfig);
        }
        def url = jarFile.toURL();
        rootLoader.addURL(url);
        try
        {
            return rootLoader.loadClass(drivername, false).newInstance();
        }
        catch (ClassNotFoundException e)
        {
            return null;
        }
    }

    def properties = null;
    def logger = null;

    public BRMDBUtil(properties, logger)
    {
        this.properties = properties;
        this.logger = logger;
    }

    public Driver getDriver(String url)
    {
        if (!driverTableInitialized)
        {
            // Initialize static driver table
            initJDBCDrivers(this.properties, this.logger);
        }
        def matchingRow = driverTable.find { row ->
            def driver = row[0];
            if (driver == BAD)
                return false;
            if (driver == DUP)
                return false;
            if (driver == null)
                return false;
            return driver.acceptsURL(url);
        }
        if (matchingRow == null)
        {
            this.logger.warn("BRMDBUtil.getDriver: Could not load driver for JDBC URL: ${url}");
            return null;
        }
        return matchingRow[0];
    }

    public Connection getConnection(String url, Properties info)
    {
        Driver driver = getDriver(url);
        if (driver == null)
            return null;
        return driver.connect(url, info);
    }

    /**
     * Get a property value from the configuration properties.
     * Throw an exception if the property is not defined or the property
     * value is empty.
     */
    def getPropOrDie(logprefix, propname)
    {
        def propval = this.properties.get(propname);
        if (propval != null && propval.trim() != '')
            return propval;
        def msg = "${logprefix}: Missing property: \"${propname}\"";
        this.logger.error(msg);
        throw new SQLException(msg);
    }

    /**
     * Get a connection to the main MySQL server as the sqladmin user.
     * Configuration properties used:
     *   MySQL server host   : BRM.host.mysql
     *   MySQL port number   : data.dbms.port
     *   MySQL admin username: BRM.mysql.admin.user
     *   MySQL admin password: BRM.mysql.admin.password
     * @param dbname The database to use
     * @return a java.sql.Connection object
     * @throws java.sql.SQLException if a connection cannot be created
     */
    def getAdminMysqlConnection(dbname)
    {
        def logprefix = "Getting MySQL admin connection";
        def host = getPropOrDie(logprefix, 'BRM.host.mysql');
        def port = getPropOrDie(logprefix, 'data.dbms.port');
        def url = "jdbc:mysql://${host}:${port}/${dbname}";
        this.logger.info("${logprefix}: url=${url}");
        def username = getPropOrDie(logprefix, 'BRM.mysql.admin.user');
        def password = getPropOrDie(logprefix, 'BRM.mysql.admin.password');
        def info = new Properties();
        info.setProperty("user", username);
        info.setProperty("password", password);
        def connection = getConnection(url, info);
        if (connection == null)
        {
            def msg = "${logprefix}: No driver for url: \"${url}\"";
            this.logger.error(msg);
            throw new SQLException(msg);
        }
        return connection;
    }

    // Part of the fix for defect S0724712
    // Get the number of "master" servers in the SAS_ODCS_SERVER table.
    def getODCSMasterServerCount()
    {
        def query = 'SELECT COUNT(*) FROM sas_odcs_server WHERE SERVER_TYPE=1';
        def connection = getAdminMysqlConnection('sassdm');
        def sql = new groovy.sql.Sql(connection);
        def result = null;
        try
        {
            sql.eachRow(query) { row ->
                result = row[0];
            }
        }
        finally
        {
            connection.close();
        }
        return result;
    }
        
    /**
     * List the names of the databases in a MySQL server. This list excludes
     * 'information_schema', 'mysql', and 'test'.
     */
    public List listMySQLDatabases()
    {
        def query = "SELECT SCHEMA_NAME FROM SCHEMATA";
        def result = new ArrayList();
        def connection = getAdminMysqlConnection('information_schema');
        def sql = new groovy.sql.Sql(connection);
        try
        {
            sql.eachRow(query) { row ->
                def dbname = row.SCHEMA_NAME;
                if (dbname.toLowerCase() in 
                        ['information_schema', 'mysql', 'test'])
                    return;
                result.add(dbname);
            }
        }
        finally
        {
            connection.close();
        }
        return result;
    }

    def mysqlDatabaseExists(dbname)
    {
        def dbfound = false;
        listMySQLDatabases().each
        { thisname ->
            if (thisname.equalsIgnoreCase(dbname))
                dbfound = true;
        }
        return dbfound;
    }

    /**
     * Get the PropFilePath property from the SAS_DEFAULT_PROPERTIES table in
     * the HCM database.
     */
    def getHCMPropFilePath()
    {
        if (!mysqlDatabaseExists('hcm'))
            return null;
        def query = "SELECT DEFAULT_VALUE FROM SAS_DEFAULT_PROPERTIES WHERE MODULE_TYPE='Config' AND PROPERTY_NM='PropFilePath'";        
        def connection = getAdminMysqlConnection('hcm');
        def sql = new groovy.sql.Sql(connection);
        def result = null;
        try
        {
            def row = sql.firstRow(query);
            if (row != null)
                result = row.DEFAULT_VALUE;
        }
        finally
        {
            connection.close();
        }
        return result;
    }

    /**
     * Set the PropFilePath property in the SAS_DEFAULT_PROPERTIES table in
     * the HCM database.
     */
    def setHCMPropFilePath(propFilePath)
    {
        if (!mysqlDatabaseExists('hcm'))
            return;
        def update = "UPDATE SAS_DEFAULT_PROPERTIES SET DEFAULT_VALUE=? WHERE MODULE_TYPE='Config' AND PROPERTY_NM='PropFilePath'";
        def connection = getAdminMysqlConnection('hcm');
        def sql = new groovy.sql.Sql(connection);
        def result = null;
        try
        {
            result = sql.executeUpdate(update, [propFilePath]);
        }
        finally
        {
            connection.close();
        }
        return result;
    }

    // I hate that terminology "slaves" but that is what the doc uses
    def static slavesPropname = 'BRM.host.odcs.slaves';

    /**
     * Get a list of information about the destination ODCS secondary ("slave")
     * servers from a BRM property.
     * Note: The SERVER_ID values must match those in the SAS_ODCS_SERVER
     * table.
     * @param logprefix Prefix indicating context for log messages
     * @return a List of records, each record is a list containing 
     *  [SERVER_ID, host, port]
     */
    def getODCSSecondaryServersFromProp(logprefix)
    {
        def slaves = this.properties.get(slavesPropname);
        if (slaves == null || slaves.trim() == '')
        {
            this.logger.info("${logprefix}: Property ${slavesPropname} not set. If any ODCS \"slave\" servers exist, they will not be updated.");
            return [];
        }
        this.logger.info("${logprefix}: ${slavesPropname}=${slaves}");

        def result = [];
        BRMUtil.split(slaves, ',').each
        { slaveInfo ->
            slaveInfo = slaveInfo.trim();
            if (slaveInfo == '')
                return;
            def i = slaveInfo.indexOf(':');
            if (i < 0)
            {
                this.logger.error("${logprefix}: Missing SERVER_ID number in \"slave\" server info: \"${slaveInfo}\"");
                return;
            }
            def slaveId = slaveInfo.substring(0, i);
            try
            {
                slaveId = Integer.valueOf(slaveId);
            }
            catch (NumberFormatException e)
            {
                this.logger.error("${logprefix}: SERVER_ID \"${slaveId}\" in \"slave\" server info \"${slaveInfo}\" is not a valid integer.");
                return;
            }
            def slaveServer = slaveInfo.substring(i + 1);
            def hostPort = BRMUtil.splitHostPort(slaveServer);
            def slaveHost = hostPort[0];
            def slavePort = hostPort[1];
            if (slavePort instanceof String)
            {
                this.logger.error("${logprefix}: Invalid ODCS \"slave\" server port number: ${slavePort}");
                return;
            }
            this.logger.debug("slaveHost=\"${slaveHost}\" slavePort=\"${slavePort}\"");
            result.add([slaveId, slaveHost, slavePort]);
        }
        return result;
    }

    /**
     * Update the SASSDM.SAS_ODCS_SERVER table with new host names and port
     * numbers.
     */
    def updateSasOdcsServerTable()
    {
        def logprefix = "Updating SAS_ODCS_SERVER table";
        this.logger.info(logprefix);
        def connection = getAdminMysqlConnection('sassdm');
        def sql = new groovy.sql.Sql(connection);
        try
        {
            def masterPropname = 'BRM.host.odcs.master';
            def master = this.properties.get(masterPropname);
            if (master != null && master.trim() != '')
            {
                this.logger.info("Got master ODCS server from ${masterPropname} property: \"${master}\"");
            }
            else
            {
                master = getPropOrDie(logprefix, 'BRM.host.webapp');
                this.logger.info("Got master ODCS server from BRM.host.webapp property: \"${master}\"");
            }
            def result = BRMUtil.splitHostPort(master);
            def masterHost = result[0];
            def masterPort = result[1];
            if (masterPort instanceof String)
            {
                this.logger.error("${logprefix}: Invalid master ODCS server port number: ${masterPort}");
                return;
            }
            this.logger.debug("masterHost=\"${masterHost}\" masterPort=\"${masterPort}\"");
            def update = "UPDATE SAS_ODCS_SERVER SET SERVER_CD=?, SERVER_HOSTNAME=?, SERVER_PORT=? WHERE SERVER_TYPE=1";
            def updateParams = [masterHost + '_MASTER', masterHost, masterPort];
            if (masterPort == null)
            {
                update = "UPDATE SAS_ODCS_SERVER SET SERVER_CD=?, SERVER_HOSTNAME=? WHERE SERVER_TYPE=1";
                updateParams = [masterHost + '_MASTER', masterHost];
            }
            int rowsUpdated = sql.executeUpdate(update, updateParams);
            if (rowsUpdated == 0)
            {
                this.logger.error("${logprefix}: No master server rows found in table, none updated");
            }
            else if (rowsUpdated > 1)
            {
                this.logger.error("${logprefix}: Updated ${rowsUpdated} rows with new ODCS master server \"${master}\", but there should only be one master server.");
            }
            else
            {
                this.logger.info("${logprefix}: Updated one row with new ODCS master server \"${master}\"");
            }

            getODCSSecondaryServersFromProp(logprefix).each
            { slaveInfo ->
                def slaveId = slaveInfo[0];
                def slaveHost = slaveInfo[1];
                def slavePort = slaveInfo[2];
                update = "UPDATE SAS_ODCS_SERVER SET SERVER_CD=?, SERVER_HOSTNAME=?, SERVER_PORT=? WHERE SERVER_TYPE=0 AND SERVER_ID=?";
                updateParams = [slaveHost+'_SLAVE', slaveHost, slavePort, slaveId];
                if (slavePort == null)
                {
                    update = "UPDATE SAS_ODCS_SERVER SET SERVER_CD=?, SERVER_HOSTNAME=? WHERE SERVER_TYPE=0 AND SERVER_ID=?";
                    updateParams = [slaveHost+'_SLAVE', slaveHost, slaveId];
                }
                rowsUpdated = sql.executeUpdate(update, updateParams);
                if (rowsUpdated == 0)
                {
                    this.logger.error("${logprefix}: No ODCS \"slave\" server found with SERVER_ID=${slaveId}");
                }
                else if (rowsUpdated > 1)
                {
                    this.logger.error("${logprefix}: Updated ${rowsUpdated} rows with new ODCS \"slave\" server info \"${slaveInfo}\", but there should only be one server with that ID.");
                }
                else
                {
                    this.logger.info("${logprefix}: Updated one row with new ODCS \"slave\" server info: \"${slaveInfo}\"");
                }
            }
        }
        finally
        {
            connection.close();
        }
    }

    /**
     * Back up a MySQL database by running mysqldump.
     * The MySQL admin credentials come from these properties:
     * Username: BRM.mysql.admin.user
     * Password: BRM.mysql.admin.password
     * Note: This method is called from Ant as well as from this file.
     * @param dbname The database name
     * @param dbfile The MySQL dump file to be created
     * @param mysql_host The MySQL server host name
     *                   If null, the property BRM.host.mysql is used.
     * @param mysql_port The MySQL server port number.
     *                   If null, the property data.dbms.port is used.
     * @param throw_on_error If true, throw an exception if the exit code from
     * mysqldump is non-zero.
     * @return The exit code from mysqldump
     */
    def backupMysqlDatabase(dbname, dbfile, mysql_host, mysql_port,
                            throw_on_error)
    {
        def logprefix = "Backing up MySQL database";
        if (dbname == null)
        {
            throw new Exception("Missing dbname attribute");
        }
        if (dbfile == null)
        {
            dbfile = getPropOrDie(logprefix, "BRM.output.dir");
            dbfile = new File(dbfile, 'mysql');
            dbfile = new File(dbfile, 'databases');
            dbfile = new File(dbfile, dbname + '.sql');
        }
        if (mysql_host == null)
            mysql_host = getPropOrDie(logprefix, "BRM.host.mysql");
        if (mysql_port == null)
            mysql_port = getPropOrDie(logprefix, "data.dbms.port");
        this.logger.info("Backing up database ${dbname} to ${dbfile}");

        def mysqldbs_install_dir = this.properties.get("mysqldbs.install.dir");
        if (mysqldbs_install_dir == null)
        {
            throw new Exception("Missing MySQL config property: mysqldbs.install.dir");
        }
        def mysqlhome_bin = new File(mysqldbs_install_dir, 'bin');
        if (!mysqlhome_bin.isDirectory())
        {
            throw new Exception("Missing MySQL home bin directory: ${mysqlhome_bin}");
        }
        def mysqldump = new File(mysqlhome_bin, 'mysqldump');

        def user = getPropOrDie(logprefix, "BRM.mysql.admin.user");
        def password = getPropOrDie(logprefix, "BRM.mysql.admin.password");
        def masked_password = "*****";

        def cmd = [
            mysqldump,
            '--user=' + user,
            '--password=' + password,
            '--host=' + mysql_host,
            '--port=' + mysql_port,
            '--flush-logs',
            '--add-locks',
            '--disable-keys',
            '--extended-insert',
            '--lock-all-tables',
            '--add-drop-table',
            '--all',
            '--quick',
            '--quote-names',
            '--hex-blob',
            '--result-file=' + dbfile,
            dbname,
            ];
        def masked_cmd = [];
        masked_cmd.addAll(cmd);
        masked_cmd[2] = '--password=' + masked_password;
        this.logger.info(BRMUtil.arrayToStr((String[])masked_cmd));

        def proc = cmd.execute();
        proc.waitFor();
        def rc = proc.exitValue();
        if (rc != 0)
        {
            def msg = "mysqldump failed with non-zero exit code: " + rc;
            if (throw_on_error)
            {
                throw new Exception(msg);
            }
            else
            {
                this.logger.warn("${logprefix}: ${msg}")
            }
        }
        return rc;
    }

    /**
     * Restore a MySQL database from a dump file.
     * The MySQL admin credentials come from these properties:
     * Username: BRM.mysql.admin.user
     * Password: BRM.mysql.admin.password
     * Note: This method is called from Ant as well as from this file.
     * @param dbname The database name
     * @param dbfile The MySQL dump file to be loaded
     * @param mysql_host The MySQL server host name
     *                   If null, the property BRM.host.mysql is used.
     * @param mysql_port The MySQL server port number
     *                   If null, the property data.dbms.port is used.
     * @param throw_on_error If true, throw an exception if the exit code from
     * mysql is non-zero.
     * @return The exit code from mysql
     */
    def restoreMysqlDatabase(dbname, dbfile, mysql_host, mysql_port, 
                             throw_on_error)
    {
        def logprefix = "Restoring MySQL database";
        def msg = null;

        if (dbname == null)
        {
            msg = "Missing dbname attribute";
            this.logger.error(msg);
            throw new Exception(msg);
        }
        if (dbfile == null)
        {
            msg = "Missing dbfile attribute";
            this.logger.error(msg);
            throw new Exception(msg);
        }
        if (dbfile instanceof File)
            dbfile = dbfile.getPath();
        if (!(new File(dbfile)).isFile())
        {
            msg = "Could not find database load file ${dbfile}";
            this.logger.error(msg);
            throw new Exception(msg);
        }
        if (mysql_host == null)
            mysql_host = getPropOrDie(logprefix, "BRM.host.mysql");
        if (mysql_port == null)
            mysql_port = getPropOrDie(logprefix, "data.dbms.port");
        this.logger.info("Dropping database ${dbname} and loading ${dbfile}");

        def mysqldbs_install_dir = getPropOrDie(logprefix, 
                                                "mysqldbs.install.dir");
        def mysqlhome_bin = new File(mysqldbs_install_dir, 'bin');
        if (!mysqlhome_bin.isDirectory())
        {
            msg = "Missing MySQL home bin directory: ${mysqlhome_bin}";
            this.logger.error(msg);
            throw new Exception(msg);
        }
        def mysql = new File(mysqlhome_bin, 'mysql');

        def user = getPropOrDie(logprefix, "BRM.mysql.admin.user");
        def password = getPropOrDie(logprefix, "BRM.mysql.admin.password");
        def masked_password = "*****";

        // Must always use forward slash (/) in path to dbfile or MySQL will
        // die, even on Windows.
        dbfile = dbfile.replace((char)'\\', (char)'/');
        def cmd = [
            mysql,
            '--user=' + user,
            '--password=' + password,
            '--host=' + mysql_host,
            '--port=' + mysql_port,
            '--batch',
            '-e',
            "DROP DATABASE IF EXISTS ${dbname}; CREATE DATABASE ${dbname}; USE ${dbname}; SOURCE ${dbfile};",
            ];
        def masked_cmd = [];
        masked_cmd.addAll(cmd);
        masked_cmd[2] = '--password=' + masked_password;
        this.logger.info(BRMUtil.arrayToStr((String[])masked_cmd));

        def proc = cmd.execute();
        proc.waitFor();
        def rc = proc.exitValue();
        if (rc != 0)
        {
            msg = "MySQL failed with non-zero exit code: " + rc;
            if (throw_on_error)
            {
                throw new Exception(msg);
            }
            else
            {
                this.logger.warn(msg);
            }
        }
        return rc;
    }
    
    /**
     * List the secondary ("slave") ODCS servers in the SAS_ODCS_SERVER table. 
     * @return a List of records, each record is a list containing 
     *  [SERVER_ID, host, port]
     */
    public List listODCSSecondaryServers()
    {
        def query = "SELECT * FROM SAS_ODCS_SERVER WHERE SERVER_TYPE=0";
        def result = new ArrayList();
        def connection = getAdminMysqlConnection('sassdm');
        def sql = new groovy.sql.Sql(connection);
        try
        {
            sql.eachRow(query)
            { row ->
                def record = [row.SERVER_ID, row.SERVER_HOSTNAME, 
                              row.SERVER_PORT];
                result.add(record);
            }
        }
        finally
        {
            connection.close();
        }
        return result;
    }
    
   /**
     * Back up the secondary ("slave") ODCS server databases. 
     * Create directory ${BRM.output.dir}/mysql/odcs
     * In this directory create a numbered subdirectory for each secondary ODCS
     * database found. The name of each subdirectory is the SERVER_ID number
     * for that server. Inside each numbered subdirectory, save the secondary
     * ODCS database in the file sassdm.sql.
     */
    def backupODCSSecondaryServers()
    {
        def logprefix = 'Backing up ODCS "slave" databases';
        this.logger.info(logprefix);
        def server_found = false;
        def odcs_dir = null;
        listODCSSecondaryServers().each
        { record ->
            def server_id = record[0];
            def host = record[1];
            if (!server_found)
            {
                server_found = true;
                def output_dir = getPropOrDie(logprefix, 'BRM.output.dir');
                def mysql_dir = new File(output_dir, 'mysql');
                odcs_dir = new File(mysql_dir, 'odcs');
                odcs_dir.mkdirs();
            }
            this.logger.info("${logprefix}: Backing up database for SERVER_ID=${server_id} on host \"${host}\"");
            def server_dir = new File(odcs_dir, "${server_id}");
            server_dir.mkdir();
            def sassdm_sql = new File(server_dir, 'sassdm.sql');
            backupMysqlDatabase('sassdm', sassdm_sql, host, null, false);
        }
    }

    /**
      * Restore the secondary ("slave") ODCS server databases.
      */
    def restoreODCSSecondaryServers()
    {
        def logprefix = 'Restoring ODCS "slave" databases';
        this.logger.info(logprefix);
        def output_dir = getPropOrDie(logprefix, 'BRM.output.dir');
        def mysql_dir = new File(output_dir, 'mysql');
        def odcs_dir = new File(mysql_dir, 'odcs');
        if (!odcs_dir.isDirectory())
        {
            this.logger.info("${logprefix}: Directory ${odcs_dir} does not exist, no ODCS \"slave\" databases to restore.");
            return;
        }

        def odcs_server_map = [:];
        getODCSSecondaryServersFromProp(logprefix).each
        { slaveInfo ->
            def slaveId = slaveInfo[0];
            def slaveHost = slaveInfo[1];
            def slavePort = slaveInfo[2];
            odcs_server_map[slaveId] = slaveHost;
        }
        if (odcs_server_map.size() == 0)
        {
            this.logger.info("${logprefix}: property ${slavesPropname} is not defined, no ODCS \"slave\" databases will be restored.");
            return;
        }

        odcs_dir.list().each
        { name ->
            def server_dir = new File(odcs_dir, name);
            if (!server_dir.isDirectory())
                return;
            def server_id;
            try
            {
                server_id = Integer.valueOf(name);
            }
            catch (NumberFormatException e)
            {
                // Not a numbered sub-directory
                return;
            }
            def sassdm_sql = new File(server_dir, 'sassdm.sql');
            if (!sassdm_sql.isFile())
            {
                this.logger.warn("${logprefix}: Missing expected file: ${sassdm_sql}");
                return;
            }
            def host = odcs_server_map.get(server_id);
            if (host == null)
            {
                this.logger.error("${logprefix}: SERVER_ID=${server_id} is not listed in value for property ${slavesPropname}");
                return;
            }
            restoreMysqlDatabase('sassdm', sassdm_sql, host, null, false);
        }
    }
}
