// BRMTarUtil.groovy
// Helper class for creating and extracting gzipped tar files.

package com.sas.solutions.brm.util;

class BRMTarUtil
{
    /**
     * Create a gzipped tar archive file from a directory tree.
     * @param logger  a Log4j-compatible Logger object.
     * @param destfile  The destination file. The file name must end with
     * .tar.gz to ensure gzipped tar format.
     * @param basedir  The root of the directory tree to be archived.
     * @param includes  Comma-separated list (no spaces) of names to include
     * @param excludes  Comma-separated list (no spaces) of names to include
     */
    def static createArchive(logger, File destfile, File basedir, 
                             String includes, String excludes)
    {
        // Example includes: "SASHumanCapitalManagement*/**"
        // Example excludes: "SASHumanCapitalManagement*/HCMConfig.*"
        logger.info("[tarAndGzip] destfile=${destfile} basedir=${basedir} includes=${includes} excludes=${excludes}");
        def tar = new com.sas.solutions.brm.tar.Tar();
        tar.setLogger(logger);
        tar.setDestFile(destfile);
        tar.setBasedir(basedir);
        tar.setIncludes(includes);
        tar.setExcludes(excludes);
        // Write the parameters to an accompanying text file
        def destdir = destfile.getAbsoluteFile().getParentFile();
        def desttxt = new File(destdir, destfile.getName() + '.txt');
        def out = new PrintWriter(
                    new OutputStreamWriter(
                        new FileOutputStream(desttxt), 'UTF-8'));
        out.println("destfile: ${destfile}");
        out.println("basedir: ${basedir}");
        out.println("includes: ${includes}");
        out.println("excludes: ${excludes}");
        out.flush()
        // Run the tar operation
        tar.run(); 
        out.println("compression: ${tar.getCompression()}");
        out.println("Number of files and directories: ${tar.getNumFiles()}");
        out.close();
    }

    /**
     * Extract a gzipped tar archive into a directory tree.
     * On Unix, if any extracted files cannot be written to disk
     * (presumeably because of permissions errors) then those files are
     * extracted again as the SAS Server User.
     * @param logger  a Log4j-compatible Logger object.
     * @param properties  a Hashtable-compatible Properties object
     * containing configuration properties.
     * @param src  The archive file. If the file name ends with .tar.gz it
     * is assumed to be in gzipped tar format.
     * @param dest  The destination directory.
     */
    def static extractArchive(logger, properties, File src, File dest)
    {
        // Delete the error list file if it exists
        def installDir = System.getProperty('BRM.install.dir');
        def tempDir = new File(installDir, 'temp');
        def scriptsDir = new File(installDir, 'scripts');
        def errorFile = new File(tempDir, 'skippedfiles.txt');
        if (errorFile.exists())
            errorFile.delete();
        // Un-tar the source file
        def task = new com.sas.solutions.brm.tar.Untar();
        task.setLogger(logger);
        task.setSrc(src);
        task.setDest(dest);
        task.setErrorfile(errorFile);
        task.run();
        // Check if an error list file was created
        def showWarnLines = { warnFile ->
            BufferedReader br = new BufferedReader(
                    new InputStreamReader(
                        new FileInputStream(warnFile), "UTF-8"));
            while (true)
            {
                String line = br.readLine();
                if (line == null)
                    break;
                line = line.trim();
                if (line.length() == 0)
                    continue;
                logger.warn("Unable to extract file: " + line);
            }
            br.close();
        }
        if (errorFile.isFile())
        {
            if (BRMUtil.isWindows())
            {
                // On Windows:
                // Log warning messages
                showWarnLines(errorFile);
            }
            else
            {
                // On Unix:
                // Re-try the extraction with sassrv credentials
                def errorFile2 = new File(tempDir, 'failedfiles.txt');
                if (errorFile2.exists())
                    errorFile2.delete();
                def cmd = [
                    (new File(installDir, 'brm.sh')).getPath(),
                    'groovy',
                    (new File(scriptsDir, 'tar.groovy')).getPath(), 
                    'xf', src.getPath(),
                    '-C', dest.getPath(),
                    '--filelist', errorFile.getPath(),
                    '--errorfile', errorFile2.getPath()];
                def runner = new RunCommandAsSASServerUser(logger, 
                                                            properties, cmd);
                runner.run();
                if (errorFile2.isFile())
                    showWarnLines(errorFile2);
            }
        }
    }
}

// end-of-file
