/*
 * Created on Jun 23, 2006
 * Copyright (c) 2006 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm.util;


import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.UnknownHostException;
import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.Random;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.log4j.Logger;

import com.sas.util.SasPasswordEncodingException;
import com.sas.util.SasPasswordString;

import com.sas.solutions.brm.Constants;
import com.sas.solutions.brm.errors.BRMConfigurationException;
import com.sas.solutions.brm.omr.OMRConnection;

public class BRMUtil
{
    private BRMUtil()
    {
        // a class with nothing but static utility methods
    }

    /**
     * Prefix the characters that are special in a regex parameter, so they
     * are not interpreted with their special meaning but instead are used
     * as literal match characters.
     * @param s
     * @return the escaped version of s
     */
    public static String escapeRegex(String s)
    {
        // Yes, we really need all of these backslashes. Honest.
        // Groovy needs one more backslash in front of "$"
        s = s.replaceAll("\\\\", "\\\\\\\\");
        s = s.replaceAll("\\[", "\\\\[");
        s = s.replaceAll("\\.", "\\\\.");
        s = s.replaceAll("\\^", "\\\\^");
        s = s.replaceAll("\\\$", "\\\\\$");
        s = s.replaceAll("\\+", "\\\\+");
        s = s.replaceAll("\\{", "\\\\{");
        s = s.replaceAll("\\(", "\\\\(");
        s = s.replaceAll("\\?", "\\\\?");
        s = s.replaceAll("\\*", "\\\\*");
        s = s.replaceAll("\\|", "\\\\|");
        return s;
    }

    /**
     * Prefix the characters that are special in a replacement parameter, so 
     * they are not interpreted with their special meaning but instead are
     * replaced literally.
     * 
     * @param s
     * @return the escaped version of s
     */
    public static String escapeRepl(String s)
    {
        // Yes, we really need all of these backslashes. Honest.
        // Groovy needs one more backslash in front of "$"
        s = s.replaceAll("\\\\", "\\\\\\\\");
        s = s.replaceAll("\\\$", "\\\\\$");
        return s;
    }

    /**
     * Escape XML special characters
     */
    public static String escapeXML(String s)
    {
        s = s.replaceAll("&", "&amp;");
        s = s.replaceAll("\"", "&quot;");
        s = s.replaceAll("<", "&lt;");
        s = s.replaceAll(">", "&gt;");
        return s;
    }

    /**
     * Guess whether the string looks like a Unix-style file or directory
     * path.
     * @param s
     * @return true if the string contains at least one forward slash and 
     *              does not look like a Windows path (see isWindowsPath()).
     */
    public static boolean isUnixPath(String s)
    {
        if (s == null)
            return false;
        boolean hasForwardSlash = (s.indexOf("/") >= 0);
        return (hasForwardSlash && !isWindowsPath(s));
    }

    /**
     * Guess whether the string looks like a Windows-style file or directory
     * path.
     * @param s
     * @return true if the string contains at least one backslash or starts
     *         with a single letter followed immediately by colon (:).
     */
    public static boolean isWindowsPath(String s)
    {
        if (s == null)
            return false;
        if (s.indexOf("\\") >= 0)
            return true;
        Pattern p = Pattern.compile("[A-Za-z]:");
        Matcher m = p.matcher(s);
        return m.lookingAt();
    }

    /**
     * @return true if the operating system in which this Java VM is running
     * is Windows(tm), false otherwise.
     */
    public static boolean isWindows()
    {
        String osname = System.getProperty('os.name');
        if (osname.toLowerCase().startsWith('windows'))
        {
            return true;
        }
        return false;
    }

    /**
     * @param origString
     * @param fromPrefix
     * @param toPrefix
     * @return
     */
    public static String replacePathPrefix(String origString, 
            String fromPrefix, String toPrefix)
    {
        char fromPrefixLastChar = fromPrefix.charAt(fromPrefix.length() - 1);
        boolean fromPrefixEndsWithSlash = (fromPrefixLastChar == (char)'/') ||
                                          (fromPrefixLastChar == (char)'\\');
        boolean toPrefixIsUnixPath = isUnixPath(toPrefix);
        char toPrefixLastChar = toPrefix.charAt(toPrefix.length() - 1);
        boolean toPrefixEndsWithSlash = (toPrefixLastChar == (char)'/') || 
                                        (toPrefixLastChar == (char)'\\');
        if (fromPrefixEndsWithSlash && !toPrefixEndsWithSlash)
        {
            // Add a slash (forward or backward, *as appropriate*) to the end
            // of toPrefix. This will fix problems with disappearing directory
            // separator characters.
            if (toPrefixIsUnixPath)
                toPrefix = toPrefix + "/";
            else
                toPrefix = toPrefix + fromPrefixLastChar;
        }
        else if (!fromPrefixEndsWithSlash && toPrefixEndsWithSlash)
        {
            // Remove a trailing slash from toPrefix to avoid creating bogus
            // doubled slashes in the resulting paths.
            toPrefix = toPrefix.substring(0, toPrefix.length() - 1);
        }
        Pattern p = Pattern.compile("\\b" + escapeRegex(fromPrefix) + "\\b");
        Matcher m = p.matcher(origString);
        String result = null;
        if (m.find())
        {
            int prefixStart = m.start();
            int prefixEnd = m.end();
            StringBuffer sb = new StringBuffer();
            sb.append(origString.substring(0, prefixStart));
            sb.append(toPrefix);
            String tail = origString.substring(prefixEnd);
            // If toPrefix contains at least one forward slash, and contains no
            // backslashes, then convert the entire result part from the end of
            // the prefix till the end of the string from backslashes to 
            // forward slashes.
            if (toPrefixIsUnixPath)
            {
                tail = tail.replace((char)'\\', (char)'/');
            }
            sb.append(tail);
            result = sb.toString();
        }
        else
        {
            result = origString;
        }
        return result;
    }

    /**
      * Like loadPropFile(), except it loads a resource instead of a file.
      * @param logger Log4j Logger object, could be null.
      */
    public static void loadPropFileAsResource(String resourceName,
            Properties properties, Logger logger)
        throws BRMConfigurationException
    {
        URL url = ClassLoader.getSystemResource(resourceName);
        if (url == null)
        {
            throw new BRMConfigurationException(
                    "Resource " + resourceName + " not found");
        }
        File propFile = new File(url.getFile());
        loadPropFile(propFile, properties, logger);
    }

    /**
      * Load a local properties file. The properties in that file are added
      * to the Properties object passed as a parameter to this method. If
      * the property already exists in the properties object, it is *not*
      * overwritten.
      *
      * @param propFile The property file. It must exist.
      * @param properties The Properties object into which the file is loaded.
      * @param logger  Log4j Logger object to track file loading, can be null.
      * @throws BRMConfigurationException if there is an error loading the
      *         file.
      */
    public static void loadPropFile(File propFile, Properties properties,
                            Logger logger)
        throws BRMConfigurationException
    {
        if (logger != null)
            logger.info("Loading " + propFile.getAbsolutePath());
        FileInputStream fis = null;
        Properties newProps = new Properties();
        try
        {
            fis = new FileInputStream(propFile);
            newProps.load(fis);
        }
        catch(IOException e)
        {
            if (logger != null)
                logger.error("Loading " + propFile, e);
            throw new BRMConfigurationException(
                "Could not load property file: " + propFile, e);
        }
        finally
        {
            try
            {
                if (fis != null)
                    fis.close();
            }
            catch (Exception e)
            {
                // Ignore errors on closing stream
            }
        }
        // Merge the new property values into the existing Properties obj
        // If the property was previously set, do not overwrite it
        for (Enumeration e=newProps.propertyNames(); e.hasMoreElements(); )
        {
            String propName = (String)e.nextElement();
            if (properties.getProperty(propName) == null)
                properties.setProperty(propName,
                        newProps.getProperty(propName));
        }
    }

    public static String decodePW(String encodedPW)
    {
        if (encodedPW == null)
            return null;
        String decodedPW = null;
        try
        {
            decodedPW = SasPasswordString.decode(encodedPW);
        }
        catch (SasPasswordEncodingException e)
        {
            throw new BRMConfigurationException("Error decoding password", e);
        }
        return decodedPW;
    }

    /**
      * Filter an array of command-line arguments, removing and executing
      * options of the form "-Dname=value".
      * @return the command-line arguments minus the property settings
      */
    public static String[] filterCmdLineProps(String[] args)
    {
        List result = new ArrayList();
        for (int i=0; i<args.length; ++i)
        {
            if (args[i].startsWith("-D"))
            {
                String name = null;
                String value = null;
                int eqIndex = args[i].indexOf("=");
                if (eqIndex >= 0)
                {
                    name = args[i].substring(2, eqIndex);
                    value = args[i].substring(eqIndex + 1);
                }
                else
                {
                    name = args[i].substring(2);
                    value = "";
                }
                if (name.length() > 0)
                {
                    System.setProperty(name, value);
                    continue;
                }
            }
            result.add(args[i]);
        }
        String[] resultArray = new String[result.size()];
        int argIndex = -1;
        for (Iterator argIter=result.iterator(); argIter.hasNext(); )
        {
            resultArray[++argIndex] = (String)argIter.next();
        }
        return resultArray;
    }

    // Convert an array of strings into a printable string
    // with the array items separated by commas, and the whole
    // thing surrounded by square brackets. This is aimed mostly at debug
    // logging; for logging command lines you probably want arrayToStr().
    public static String stringArrayToString(String[] args)
    {
        if (args == null)
            return "null";
        StringBuffer sb = new StringBuffer();
        sb.append((char)'[');
        for (int i=0; i<args.length; ++i)
        {
            if (i > 0)
            {
                sb.append(", ");
            }
            sb.append(args[i]);
        }
        sb.append((char)']');
        return sb.toString();
    }

    // Convert an array of bytes into a printable string with the array
    // items separated by commas, formatted as hexadecimal integers, and the
    // whole thing surrounded by square brackets. This is aimed mostly at
    // debug logging.
    public static String byteArrayToString(byte[] args)
    {
        if (args == null)
            return "null";
        StringBuffer sb = new StringBuffer();
        sb.append((char)'[');
        for (int i=0; i<args.length; ++i)
        {
            if (i > 0)
            {
                sb.append(", ");
            }
            sb.append(String.format("0x%02X", args[i]));
        }
        sb.append((char)']');
        return sb.toString();
    }

    /**
      * Mangle a file or directory pathname such that it can be used as a
      * valid file name (but a user can still recognize the directory names
      * as parts of the file name).
      */
    public static String fusePath(String pathName)
    {
        String fusedPath = pathName.replaceAll("[\\\\/|:]", "-");
        fusedPath = fusedPath.replaceAll("\\s", "_");
        return fusedPath;
    }

    /**
      * Change the filename portion of a file path, leaving the rest alone.
      * This method should work regardless of what kind of slashes are used
      * in the pathName parameter.
      *
      * Examples:<br/>
      *   changeFileNameInPath("a/b/c", "d") returns "a/b/d"
      *   changeFileNameInPath("a/b/", "d") returns "a/b/d"
      *   changeFileNameInPath("/a", "d") returns "/d"
      *   changeFileNameInPath("a", "d") returns "d"
      */
    public static String changeFileNameInPath(String pathName, 
            String newFileName)
    {
        int i = pathName.lastIndexOf("/");
        if (i < 0)
            i = pathName.lastIndexOf("\\");
        // If still (i < 0), this algorithm replaces the whole string,
        // which makes as much sense as anything else.
        return pathName.substring(0, i + 1) + newFileName;
    }

    /**
      * @param s A String
      * @return true iff The string parameter is non-null and equal
      * (case-insensitive) to "1", "true", "yes", "y", or "on".
      * This is compatible with the Ant definition of truth, except that Ant
      * does not recognize "1" or "y".
      */
    public static boolean isTrueStr(String s)
    {
        if (s == null)
            return false;
        s = s.trim().toLowerCase();
        if (s.length() == 0)
            return false;
        if (s.equals("1"))
            return true;
        if (s.equals("true"))
            return true;
        if (s.equals("yes"))
            return true;
        if (s.equals("y"))
            return true;
        if (s.equals("on"))
            return true;
        return false;
    }

    public static void dumpObjInfo(Object obj, PrintStream out)
    {
        try
        {
            Class contextClass = obj.getClass();
            out.println(contextClass.toString());
            Method[] methods = contextClass.getDeclaredMethods();
            for (int methodIndex=0; methodIndex<methods.length; methodIndex++)
            {
                Method m = methods[methodIndex];
                out.print(m.toString());
                if (m.getName().startsWith("get") &&
                        m.getParameterTypes().length==0 &&
                        !(m.getReturnType().equals(void.class)))
                {
                    out.print("=");
                    out.print(m.invoke(obj, (Object[])[]));
                }
                out.println();
            }
        }
        catch (Exception e)
        {
            throw new Error(e);
        }
    }

    public static Object getDeclaredFieldValue(Object obj, String fieldName)
    {
        try
        {
            Class contextClass = obj.getClass();
            Field[] fields = contextClass.getDeclaredFields();
            for (int i=0; i<fields.length; ++i)
            {
                Field field = fields[i];
                if (field.getName().equals(fieldName))
                {
                    return field.get(obj);
                }
            }
            return null;
        }
        catch (Exception e)
        {
            throw new Error(e);
        }
    }

    /**
    * @param controlConfig  BRM configuration properties
    * @param userType       "trusted" or "admin"
    */
    public static OMRConnection getNewOMRConnection(Properties controlConfig, 
                                                    String userType)
        throws BRMConfigurationException
    {
        def getprop = { propname ->
            def propval = controlConfig.getProperty(propname);
            if (propval == null || propval.length() == 0)
            {
                throw new BRMConfigurationException(
                    "Cannot connect to metadata server because " +
                    Constants.BRM_CONFIG_FILE +
                    " is missing property " +
                    propname);
            }
            return propval;
        }
        def host = getprop("BRM.host.metadata");
        def port = controlConfig.getProperty("BRM.port.metadata");
        if (port == null || port.length() == 0)
            port = "8561";
        def user = getprop("BRM." + userType + ".user");
        def password = getprop("BRM." + userType + ".password");
        OMRConnection omr = new OMRConnection(host, port, user, password);
        omr.connect();
        return omr;
    }

    /**
      * Return a file or directory path that has been converted to an
      * common format such that it can be compared with other normalized
      * paths. The normalized path is absolute and contains no "." or ".."
      * elements. On Windows, the path is converted to lower case.
      */
    public static String normPath(String pathName)
    {
        return normPath(pathName, File.separatorChar);
    }

    public static String normPath(String pathName, char separatorChar)
    {
        if (pathName == null)
            return null;
        if (separatorChar == (char)'\\')
        {
            // Special case for Windows
            try
            {
                return ((new File(pathName)).getCanonicalPath()).toLowerCase();
            }
            catch (IOException e)
            {
                // Ignore the error in getCanonicalPath(), fall back to our
                // own algorithm after normalizing the case. (We call
                // getAbsolutePath() first so that when it gets called again
                // it won't add a capitalized drive letter.)
                pathName = (new File(pathName)).getAbsolutePath();
                pathName = pathName.toLowerCase();
                // The fallback algorithm can't handle UNC paths.
                if (pathName.length() >= 2 &&
                        pathName.charAt(0) == (char)'\\' &&
                        pathName.charAt(1) == (char)'\\')
                    throw new Error(e);
            }
        }
        // We don't use File.getCanonicalPath() here because on Unix it
        // dereferences symbolic links. We want to preserve symbolic links.
        // This "if" statement is for testing separatorChar='/' on Windows.
        if (pathName.length() == 0 || pathName.charAt(0) != separatorChar)
            pathName = (new File(pathName)).getAbsolutePath();
        List parts = splitPath(pathName);
        List result = new ArrayList();
        // This "if" statement is for testing separatorChar='/' on Windows.
        if (parts.size() > 0 && ((String)parts.get(0)).length() == 0)
            result.add("");
        for (Iterator it=parts.iterator(); it.hasNext(); )
        {
            String part = (String)it.next();
            if (part.length() == 0)
                continue;
            if (part.equals("."))
                continue;
            if (part.equals(".."))
            {
                if (result.size() > 1)
                    result.remove(result.size() - 1);
                continue;
            }
            result.add(part);
        }
        if (result.size() < 2)
            result.add("");
        return joinPath(result, separatorChar);
    }

    /**
      * Split a file or directory path into a list containing its component
      * elements without the separators. Both / and \ are treated as path
      * separators. The first element of a split path beginning with / or \
      * will be an empty string (this is a feature). splitPath() and
      * joinPath() are opposites of each other.
      */
    public static List splitPath(String path)
    {
        path = path.replace((char)'\\', (char)'/');
        List parts = new ArrayList();
        int j = 0;
        for (int i = 0; i <= path.length(); i = j + 1)
        {
            j = path.indexOf('/', i);
            if (j < 0)
                j = path.length();
            parts.add(path.substring(i, j));
        }
        return parts;
    }

    /**
      * Join a list of path elements into a path string. Each element is
      * separated in the string by separatorChar.  If the list has only 0 or
      * 1 elements, then no separatorChar is added. splitPath() and
      * joinPath() are opposites of each other.
      */
    public static String joinPath(List parts, char separatorChar)
    {
        StringBuffer sb = new StringBuffer();
        boolean first = true;
        for (Iterator it=parts.iterator(); it.hasNext(); )
        {
            if (first)
                first = false;
            else
                sb.append(separatorChar);
            sb.append(it.next());
        }
        return sb.toString();
    }

    /**
      * Join a list of path elements by the default separator char.
      */
    public static String joinPath(List parts)
    {
        return joinPath(parts, File.separatorChar);
    }

    /**
    * Find a path relative to another path.
    * @param dir Path to a parent directory
    * @param file Path to a file or directory inside of dir (may be nested)
    * @return The File that consists of the file parameter relative to the
    * dir parameter. Return null if file is not inside of dir or if file is
    * equal to dir.
    */
    def static getRelativeFile(File dir, File file)
    {
        def parts1 = BRMUtil.splitPath(dir.getPath());
        def parts2 = BRMUtil.splitPath(file.getPath());
        if (parts1.size() >= parts2.size())
            return null;
        def i = 0;
        for (i=0; i<parts1.size(); ++i)
        {
            // Optimized for Windows: case-insensitive compare
            // This is just an additional integrity check, so doesn't
            // have to be 100% perfect.
            if (!parts1[i].equalsIgnoreCase(parts2[i]))
                return null;
        }
        def resultparts = new ArrayList();
        for (i=parts1.size(); i<parts2.size(); ++i)
        {
            resultparts.add(parts2[i]);
        }
        return new File(BRMUtil.joinPath(resultparts));
    }

    /**
    * Split a filename into base and extension parts. The extension is
    * everything after and including the last '.' character in the filename.
    * @return A String[] array of length two. The first item in the array is
    * the filename up to but not including the '.' in the filename
    * extension, the last item in the array is the extension including the
    * '.'.
    */
    public static String[] splitExt(File file)
    {
        String[] result = new String[2];
        String path = file.getPath();
        int i = path.lastIndexOf('.');
        if (i < 0)
        {
            result[0] = path;
            result[1] = '';
            return result;
        }
        result[0] = path.substring(0, i);
        result[1] = path.substring(i);
        return result;
    }

    /**
      * Compare two lists and return the number of first elements that are
      * common between them. This implementation works most efficiently with
      * ArrayList instances.
      */
    public static int compareListPrefixes(List list1, List list2)
    {
        int min = list1.size();
        if (list2.size() < min)
            min = list2.size();
        int i = 0;
        // "i=i" keeps Groovy happy
        for (i=i; i<min; ++i)
        {
            Object o1 = list1.get(i);
            Object o2 = list2.get(i);
            if (!o1.equals(o2))
                break;
        }
        return i;
    }

    // Convert an array of strings into a printable string
    // with the array items separated by spaces, and items containing
    // spaces surrounded by quote (") characters. This is intended mainly
    // for logging command lines. For a similar method with a different
    // output format, see stringArrayToString().
    public static String arrayToStr(array)
    {
        StringBuffer sb = new StringBuffer();
        def firstItem = true;
        array.each
        { item ->
            if (firstItem)
                firstItem = false;
            else
                sb.append(" ");
            boolean hasSpace = (item.indexOf(" ") >= 0);
            if (hasSpace)
                sb.append("\"");
            sb.append(item);
            if (hasSpace)
                sb.append("\"");
        }
        return sb.toString();
    }
    
    // For some strange reason, Ant Project.getProperties() returns a
    // Hashtable object instead of a Properties object. This method serves
    // as an adapter between Ant and a more sane environment:
    // Properties props = BRMUtil.hashToProps(project.getProperties());
    public static Properties hashToProps(Hashtable hash)
    {
        Properties props = new Properties();
        props.putAll(hash);
        return props;
    }

    // Return a string containing the number of spaces given in the indent
    // parameter
    public static String getIndentStr(int indent)
    {
        StringBuffer sb = new StringBuffer();
        for (int i=0; i<indent; ++i)
            sb.append(" ");
        return sb.toString();
    }

    public static boolean isLocalHost(String hostname)
    {
        // If hostname is empty or null then it can't be a remote host
        // therefore we return true.
        if (hostname == null)
            return true;
        hostname = hostname.trim().toLowerCase();
        if (hostname.length() == 0)
            return true;
        if (hostname.equals("localhost"))
            return true;

        // Try to get the local host name
        String localName = null;
        String localCanonicalName = null;
        try
        {
            InetAddress addr = InetAddress.getLocalHost();
            localName = addr.getHostName();
            localCanonicalName = addr.getCanonicalHostName();
        }
        catch (UnknownHostException ex)
        {
            // Can't determine local host name
        }
        if (localName != null)
        {
            localName = localName.trim().toLowerCase();
            if (hostname.equals(localName))
                return true;
        }
        if (localCanonicalName != null)
        {
            localCanonicalName = localCanonicalName.trim().toLowerCase();
            if (hostname.equals(localCanonicalName))
                return true;
        }

        InetAddress hostAddress = null;
        try
        {
            hostAddress = InetAddress.getByName(hostname);
        }
        catch (UnknownHostException ex)
        {
            // Can't be local if it doesn't exist
            return false;
        }
        if (hostAddress.isLoopbackAddress())
            return true;

        // Attempt to bind to the host address.
        try
        {
            ServerSocket srv = new ServerSocket(0, 1, hostAddress);
            srv.close();
        }
        catch (IOException e)
        {
            // Can't bind to the address if it is not local.
            return false;
        }
        // Successfully bound to this address, it must be local.
        return true;
    }

    public static String getLocalHostName()
    {
        String localName = null;
        try
        {
            InetAddress addr = InetAddress.getLocalHost();
            localName = addr.getHostName();
        }
        catch (UnknownHostException ex)
        {
            // Can't determine local host name
            return "localhost";
        }
        return localName.trim().toLowerCase();
    }

    public static String getLocalFQDN()
    {
        String localCanonicalName = null;
        try
        {
            InetAddress addr = InetAddress.getLocalHost();
            localCanonicalName = addr.getCanonicalHostName();
        }
        catch (UnknownHostException ex)
        {
            // Can't determine local host name
            return "localhost";
        }
        return localCanonicalName.trim().toLowerCase();
    }
    
    /**
      * Determine if two host names represent the same machine.
      * @return true if the host names match by case-insensitive string
      * comparison, or if they both have the same network address.
      * @throws NullPointerException of host1 or host2 are null.
      */
    public static boolean areHostsEquivalent(String host1, String host2)
    {
        if (host1.trim().equalsIgnoreCase(host2.trim()))
            return true;
        InetAddress[] hostAddrs1 = null;
        try
        {
            hostAddrs1 = InetAddress.getAllByName(host1);
        }
        catch (UnknownHostException ex)
        {
            // Can't be equal if it doesn't exist
            return false;
        }
        InetAddress[] hostAddrs2 = null;
        try
        {
            hostAddrs2 = InetAddress.getAllByName(host2);
        }
        catch (UnknownHostException ex)
        {
            // Can't be equal if it doesn't exist
            return false;
        }
        // In the vast majority of cases both lists of addresses will be of
        // length 1.  However, I'm doing this cross comparison in case some
        // future DNS server provides both IPv4 and IPv6 addresses, etc.
        // Note: In Groovy 1.7.1 I've had problems directly returning the
        // result of find() or findAll(), that's why I'm using the variables
        // match and innerMatch.
        def match = hostAddrs1.find
        { hostAddr1 ->
            def innerMatch = hostAddrs2.find
            { hostAddr2 ->
                return Arrays.equals(hostAddr1.getAddress(), 
                                        hostAddr2.getAddress());
            }
            return (innerMatch != null);
        }
        return (match != null);
    }

    /**
     * Write a properties file named backup-info.properties to the specified
     * directory. The property file contains information about BRM at the
     * time a backup is being performed (i.e. BRM version, OS info, etc.)
     * @param configProps  The properties from brm.properties or
     *                 brmagent.properties
     * @param dirName  The name of the directory in which the properties
     *                 file will be created. The directory must exist.
     */
    public static void storeBackupInfo(Hashtable configProps, String dirName)
    {
        Properties props = new Properties();
        def installDir = System.getProperty('BRM.install.dir');
        def versionFile = new File(installDir, 'version.txt');
        def version = versionFile.getText('UTF-8').trim();
        props.setProperty("BRM.version", version);
        setPropIfNotNull(props, "BRM.cross.platform", 
                            configProps.get("BRM.cross.platform"));
        setPropIfNotNull(props, "BRM.config.dir", 
                            configProps.get("BRM.config.dir"));
        setPropIfNotNull(props, "BRM.output.dir", 
                            configProps.get("BRM.output.dir"));
        setPropIfNotNull(props, "BRM.agent.host", 
                            configProps.get("BRM.agent.host"));
        setPropIfNotNull(props, "BRM.controller.host", 
                            configProps.get("BRM.controller.host"));
        setPropIfNotNull(props, "BRM.controller.host.fqdn", 
                            configProps.get("BRM.controller.host.fqdn"));
        setPropIfDirExists(props, "data.jdbc.jar.dir",
                            configProps.get("data.jdbc.jar.dir"));
        setPropIfDirExists(props, "mysqldbs.install.dir",
                            configProps.get("mysqldbs.install.dir"));
        // I think this is the "mid-tier" version of data.jdbc.jar.dir
        setPropIfDirExists(props, "mysqljdbc.install.dir",
                            configProps.get("mysqljdbc.install.dir"));
        setPropIfNotNull(props, "os.localhost.machine.type",
                            configProps.get("os.localhost.machine.type"));
        setPropIfDirExists(props, "soldatatier.install.dir",
                            configProps.get("soldatatier.install.dir"));
        setPropIfDirExists(props, "soltnsmid.install.dir",
                            configProps.get("soltnsmid.install.dir"));
        setPropIfNotNull(props, "webappsrv.container.type",
                            configProps.get("webappsrv.container.type"));
        setPropIfNotNull(props, "webappsrv.version",
                            configProps.get("webappsrv.version"));
        setPropIfDirExists(props, "webappsrv.weblogic.home.dir",
                            configProps.get("webappsrv.weblogic.home.dir"));
        setPropIfDirExists(props, "webspherend.install.dir",
                            configProps.get("webspherend.install.dir"));
        setPropIfNotNull(props, "BRM.install.dir", 
                            System.getProperty("BRM.install.dir"));
        props.setProperty("file.encoding", 
                            System.getProperty("file.encoding"));
        props.setProperty("os.arch", System.getProperty("os.arch"));
        props.setProperty("os.name", System.getProperty("os.name"));
        props.setProperty("os.version", System.getProperty("os.version"));
        props.setProperty("hostname", getLocalHostName());
        props.setProperty("hostname.fqdn", getLocalFQDN());
        def baseInstallDir = configProps.get("base.install.dir");
        def sasExecFileName = configProps.get("sas.exec.file.name");
        if (baseInstallDir != null && sasExecFileName != null)
        {
            def sasExecFile = new File(baseInstallDir, sasExecFileName);
            if (sasExecFile.isFile())
            {
                props.setProperty("BRM.sas.exec.file", sasExecFile.getPath());
            }
        }
        File backupInfo = new File(dirName, "backup-info.properties");
        OutputStream fos = new FileOutputStream(backupInfo);
        props.store(fos, "BRM backup information");
        fos.close();
    }

    /**
     * Get the information stored by storeBackupInfo().
     * @param backupDir  The directory in which the info is stored
     * @return  A Properties object containing the data from appropriate
     * backup-info.properties file.
     * @throws FileNotFoundException if a property file does not exist.
     */
    public static Properties getBackupInfo(File backupDir)
    {
        File backupInfoFile = new File(backupDir, "backup-info.properties");
        Properties backupInfo = new Properties();
        loadPropFile(backupInfoFile, backupInfo, null);
        return backupInfo;
    } 

    public static void setPropIfNotNull(Properties props, String propName,
                                            String propValue)
    {
        if (propValue != null)
            props.setProperty(propName, propValue);
    }

    public static void setPropIfDirExists(Properties props, String propName,
                                            String propValue)
    {
        if (propValue == null)
            return;
        def dir = new File(propValue);
        if (!dir.isDirectory())
            return;
        props.setProperty(propName, propValue);
    }

    public static void storeAgentHostName(Hashtable props, String dir)
    {
        def hostname = props.get("BRM.agent.host");
        if (hostname == null)
            hostname = getLocalFQDN();
        def agentHostFile = new File(dir, "agent-host-name.txt");
        def out = new PrintWriter(
                    new OutputStreamWriter(
                        new FileOutputStream(agentHostFile), "UTF-8"));
        out.println(hostname);       // Line 1 is BRM.agent.host property
        out.println(getLocalFQDN()); // Line 2 is local FQDN (may be the same)
        out.close();
    }

    /**
      * Given a BRM component backup directory, return a host name.
      * This returns the host name as was used in the original
      * brm.properties file when the BRM backup was created.
      * @param backupDir  File object pointing to a component directory
      * @return  A host name, or null if host name could not be determined
      */
    public static String getAgentHostName(File backupDir)
    {
        def agentHostFile = new File(backupDir, "agent-host-name.txt");
        if (!agentHostFile.isFile())
            return null;    // File doesn't exist
        def reader = new BufferedReader(
                        new InputStreamReader(
                            new FileInputStream(agentHostFile), "UTF-8"));
        def agentHostName = reader.readLine();
        reader.close();
        if (agentHostName == null)
            return null;    // empty file
        agentHostName = agentHostName.trim();
        return agentHostName;
    }

    /**
      * Extract a map of component to hostname from a properties object.
      */
    static def getComponentHostMap(properties)
    {
        def result = [:];
        def prefix = 'BRM.host.';
        properties.keySet().each
        { propname ->
            if (propname.startsWith(prefix))
            {
                def component = propname.substring(prefix.length());
                def host = properties.get(propname);
                host = host.trim().toLowerCase();
                if (host.length() == 0)
                    return; // empty string not a real host
                result[component] = host;
            }
        }
        return result;
    }

    /**
      * Query environment to find out if BRM restore is taking place on the
      * same machine(s) as the BRM source backup.
      * @param properties  Current BRM config/agent properties
      * @param log  Logger object with info, debug, error methods
      * @return true if on the same machine(s) as source, false otherwise
      */
    static def isRestoreOnSameSystem(properties, log)
    {
        log.debug("isRestoreOnSameSystem: determining if we are on the same machine(s) as the BRM source backup");

        def getpropOrDie = { propname ->
            def propval = properties.get(propname);
            if (propval == null)
                throw new Exception("Missing property: " + propname);
            return propval;
        }

        def outputDir = new File(getpropOrDie("BRM.output.dir"));
        log.debug("isRestoreOnSameSystem: BRM.output.dir=\"${outputDir}\"");
        def backupConfigDir = new File(outputDir, 'config');
        def backupBRMConfigFile = new File(backupConfigDir, 'brm.properties');
        if (!backupBRMConfigFile.isFile())
        {
            throw new Exception("Missing backed up config file: \"${backupBRMConfigFile}\"");
        }
        def backedUpProps = new Properties();
        def input = new FileInputStream(backupBRMConfigFile);
        backedUpProps.load(input);
        input.close();

        def backupComponentHostMap = getComponentHostMap(backedUpProps);
        def currentComponentHostMap = getComponentHostMap(properties);
        def componentSet = new HashSet(backupComponentHostMap.keySet());
        componentSet.addAll(currentComponentHostMap.keySet());
        def componentList = new ArrayList(componentSet);
        Collections.sort(componentList);

        def numMatches = 0;
        def numMismatches = 0;
        componentList.each
        { component ->
            def backupHost = backupComponentHostMap.get(component);
            def currentHost = currentComponentHostMap.get(component);
            if (backupHost == null)
            {
                log.debug("isRestoreOnSameSystem: Host for component ${component} is not in the backed-up brm.properties file");
                return;
            }
            if (currentHost == null)
            {
                log.debug("isRestoreOnSameSystem: Host for component ${component} is not in the current brm.properties file");
                return;
            }
            if (areHostsEquivalent(backupHost, currentHost))
            {
                log.debug("isRestoreOnSameSystem: Hosts are equivalent for component ${component}: \"${backupHost}\", \"${currentHost}\"");
                numMatches += 1;
            }
            else
            {
                log.debug("isRestoreOnSameSystem: Hosts not equivalent for component ${component}: \"${backupHost}\", \"${currentHost}\"");
                numMismatches += 1;
            }
        }

        boolean restoreOnSameSystem = (numMatches > 0 && numMismatches == 0);
        log.debug("isRestoreOnSameSystem: numMatches=${numMatches} numMismatches=${numMismatches}: returning ${restoreOnSameSystem}");
        return restoreOnSameSystem;
    }

    /**
     * Copy a source file to a destination file. If the destination does not
     * exist, it is created.
     */
    public static void copyFile(File srcFile, File dstFile)
    {
        InputStream inStream = new FileInputStream(srcFile);
        OutputStream outStream = new FileOutputStream(dstFile);
        byte[] buf = new byte[2048];
        while (true)
        {
            int len = inStream.read(buf);
            if (len < 0)
                break;
            outStream.write(buf, 0, len);
        }
        inStream.close();
        outStream.close();
    }

    public static String generateTimeStamp()
    {
        Calendar gc = new GregorianCalendar();
        Date timeStamp = gc.getTime();
        SimpleDateFormat f = new SimpleDateFormat("yyyyMMdd.HHmm");
        return f.format(timeStamp);
    }

    public static String replaceTimeStampVar(String propVal, 
                                             String timeStampStr)
    {
        if (propVal == null)
            return null;
        // Groovy requires extra backslash in front of "$"
        return propVal.replaceAll("\\\$\\{stamp\\}", timeStampStr);
    }

    /**
     * Return a copy of the string with leading spaces removed.
     */
    public static String lstrip(String s)
    {
        def m = (s =~ /^\s*/);
        if (!m)
            return s;
        return s.substring(m.end());
    }

    /**
     * Return a copy of the string with trailing spaces removed.
     */
    public static String rstrip(String s)
    {
        def m = (s =~ /\s*$/);
        if (!m)
            return s;
        return s.substring(0, m.start());
    }

    /**
     * Remove common leading spaces from each line of a multi-line string.
     */
    public static String trimMultiLine(String s)
    {
        if (s == null || s.length() == 0)
            return '';
        // split into a list of lines
        def lines = s.split('\n');
        def maxint = 2147483647; // some very large number
        def indent = maxint;
        lines[1..(lines.size() - 1)].each
        { line ->
            def stripped = BRMUtil.lstrip(line);
            if (stripped.length() > 0)
            {
                indent = [indent, line.length() - stripped.length()].min();
            }
        }
        // Remove indentation (first line is special)
        def trimmed = [lines[0].trim()];
        if (indent < maxint)
        {
            lines[1..(lines.size() - 1)].each
            { line ->
                if (line.size() > indent)
                    trimmed.add(BRMUtil.rstrip(line[indent..-1]));
                else
                    trimmed.add('');
            }
        }
        // Strip off trailing and leading blank lines
        def firstindex = 0;
        trimmed.find
        { line ->
            if (line.length() > 0)
                return true;
            firstindex += 1;
            return false;
        }
        def lastindex = trimmed.size() - 1;
        trimmed[-1..0].find
        { line ->
            if (line.length() > 0)
                return true;
            lastindex -= 1;
            return false;
        }
        if (firstindex >= trimmed.size() || lastindex < 0)
            return '';
        return trimmed[firstindex .. lastindex].join('\n');
    }
    
    // Split string s on delimiter d in a way that allows round-trip
    // transformation, i.e. trailing empty strings can be produced.  This is
    // similar to String.split(regex, -1), except that the delimiter is an
    // exact string to match instead of a regular expression.
    def static split(s, d)
    {
        if (d.length() == 0)
            throw new Exception("Empty delimiter");
        def i = 0;
        def j = 0;
        def a = [];
        while (i <= s.length())
        {
            j = s.indexOf(d, i);
            if (j < 0)
                j = s.length();
            a.add(s.substring(i, j));
            i = j + d.length();
        }
        return a;
    }

    /**
     * Split a string in host:port format into [host, port].
     * port is converted from a string to an integer.
     * If a port number is not given or is empty, return null for port.
     * If port is not a valid integer, leave it as a string.
     * Test it later: if (port instanceof String) ...
     */
    def static splitHostPort(s)
    {
        def i = s.indexOf(':');
        if (i < 0)
            i = s.length();
        def host = s.substring(0, i).trim();
        def port = null;
        if ((s.length() - i - 1) > 0)
        {
            port = s.substring(i + 1);
            try
            {
                port = Integer.valueOf(port);
            }
            catch (NumberFormatException e)
            {
                // Leave port as original string
            }
        }
        return [host, port];
    }

    def static makeDirectoryEmpty(File directory)
    {
        directory.list().each
        { entry ->
            def fileordir = new File(directory, entry);
            if (fileordir.isDirectory())
                makeDirectoryEmpty(fileordir);
            fileordir.delete();
        }
    }

    /*
     *  Given a file, this replaces all instances of a search pattern with a given string. It overwrites the file. 
     */ 
    public static void stringReplaceAllInFile(String filePath, String searchString, String replaceString){
    
        String newFile = readFileAsString(filePath);
        
        newFile=newFile.replaceAll("(?i)"+searchString,replaceString);
        
        writeStringToFile(filePath, newFile);
    
    }
    
    /*
     *  Writes a String out to a text file.  Does not append, it overwrites. 
     */ 
    public static void writeStringToFile(String file, String text){
        writeStringToFile(new File(file), text);
    }
    
    /*
     *  Writes a String out to a text file.  Does not append, it overwrites. 
     */ 
    public static void writeStringToFile(File file, String text){
        try {
            FileWriter outFile = new FileWriter(file);
            PrintWriter out = new PrintWriter(outFile);
                       
            // Write text to file
            out.print(text);
            
            out.close();
            
        } catch (IOException e){
            e.printStackTrace();
        }
    }
    
    /*
     *  Reads in a file to a String variable.  
     */ 
    private static String readFileAsString(File filePath)
      throws IOException{
        StringBuffer fileData = new StringBuffer(1000);
        BufferedReader reader = new BufferedReader(
                new FileReader(filePath));
        char[] buf = new char[1024];
        int numRead=0;
        
        while((numRead=reader.read(buf)) != -1){
            String readData = String.valueOf(buf, 0, numRead);
            fileData.append(readData);
            buf = new char[1024];
        }
        
        reader.close();
        
        return fileData.toString();
    }

    private static String readFileAsString(String filePath)
      throws IOException{
        return readFileAsString(new File(filePath));
    }
    

    def static dumpProps(props)
    {
        def sb = new StringBuffer();
        props.each
        { key, value ->
            sb.append("${key}=${value}");
            sb.append(File.separatorChar);
        }
        return sb.toString();
    }

    // Find location of ExportPackage/ImportPackage programs
    // C:\Program Files\SAS\SASPlatformObjectFramework\9.2\ExportPackage.exe
    // /usr/local/SAS/SASPlatformObjectFramework/9.2/ExportPackage
    def static findPlatformObjDir(props)
    {
        // From metadatasrv.configuration.properties
        def propname = 'platfmltypes.install.dir';
        def platformdir = props.get(propname);
        if (platformdir == null || platformdir.length() == 0)
        {
            throw new BRMConfigurationException(
                                "Missing property: " + propname);
        }
        return platformdir;
    }

    // Run the ExportPackage program
    // logger -- Logger object for logging messages
    // props -- configuration properties
    // creds -- dictionary with credentials for OMR server
    // output_dir -- directory into which all output will be put
    // types -- type name to export, i.e. "Tables"
    // objectpath -- Metadata folder path, i.e.
    //   "/Solutions/SAS Data Integration Studio Custom Tree/Solution Data Marts (SDM)/Metric Tables"
    // packagename - base name of .spk and .log file to generate in output_dir
    // extraopts - null, or an array of strings to add to the options
    // Return full path of generated .spk file
    def static exportPackage(logger, props, creds, output_dir,
                             types, objectpath, packagename, extraopts)
    {
        def ExportPackage = new File(findPlatformObjDir(props),
                                        "ExportPackage");
        def spkfile = new File(output_dir, packagename + '.spk');
        def logfile = new File(output_dir, packagename + '-export.log');
        def args = [];
        def trace_args = [];
        def addArg = { arg ->
            args.add(arg);
            trace_args.add(arg);
        }
        def addHiddenArg = { arg ->
            args.add(arg);
            trace_args.add('*****');
        }
        addArg(ExportPackage.getPath());
        addArg("-host");
        addArg(creds['host']);
        addArg("-port");
        addArg(creds['port']);
        addArg("-user");
        addArg(creds['user']);
        addArg("-password");
        addHiddenArg(creds['password']);
        addArg("-package");
        addArg(spkfile.getPath());
        addArg("-log");
        addArg(logfile.getPath());
        addArg("-types");
        addArg(types);
        addArg("-objects");
        addArg(objectpath);
        if (extraopts != null)
        {
            extraopts.each
            { opt ->
                addArg(opt);
            }
        }
        addArg("-debug");
        logger.info(arrayToStr(trace_args));
        def rc = execProgram(null, args);
        if (rc != 0)
        {
            throw new Exception("ExportPackage error, exit code: " + rc);
        }
        return spkfile;
    }

    // Run the ImportPackage program
    // Produces an '-import.log' file in the same directory as spkfile
    // logger -- Logger object for logging messages
    // props -- configuration properties
    // creds -- dictionary with credentials for OMR server
    // targetpath -- Destination metadata folder path, i.e.
    //   "/Products/SAS Solutions Services/SDMMET"
    // spkfile -- File object of .spk file to import
    // extraopts -- null, or an array of strings to add to the options
    def static importPackage(logger, props, creds,
                             targetpath, spkfile, extraopts)
    {
        def ImportPackage = new File(findPlatformObjDir(props),
                                        "ImportPackage");
        def output_dir = spkfile.getParentFile();
        if (output_dir == null)
            output_dir = '.';
        def packagename = spkfile.getName();
        def extIndex = packagename.toLowerCase().indexOf('.spk');
        if (extIndex >= 0)
        {
            packagename = packagename.substring(0, extIndex);
        }
        def logfile = new File(output_dir, packagename + '-import.log');
        def args = [];
        def trace_args = [];
        def addArg = { arg ->
            args.add(arg);
            trace_args.add(arg);
        }
        def addHiddenArg = { arg ->
            args.add(arg);
            trace_args.add('*****');
        }
        addArg(ImportPackage.getPath());
        addArg("-host");
        addArg(creds.host);
        addArg("-port");
        addArg(creds.port);
        addArg("-user");
        addArg(creds.user);
        addArg("-password");
        addHiddenArg(creds.password);
        addArg("-package");
        addArg(spkfile.getPath());
        addArg("-log");
        addArg(logfile.getPath());
        addArg("-target");
        addArg(targetpath);
        if (extraopts != null)
        {
            extraopts.each
            { opt ->
                addArg(opt);
            }
        }
        logger.info(arrayToStr(trace_args));
        def rc = execProgram(null, args);
        if (rc == 4)
        {
            logger.warn("ImportPackage warnings, exit code: " + rc);
        }
        else if (rc != 0)
        {
            throw new Exception("ImportPackage error, exit code: " + rc);
        }
        return rc;
    }

    // Run a program, return the exit code
    // logger -- Logger object to trace program execution, or null
    // cmd -- Array or List of program executable and arguments
    def static execProgram(logger, cmd)
    {
        if (logger != null)
            logger.info("Executing program: " + arrayToStr(cmd));
        def proc = cmd.execute();
        def sout = new StringBuffer();
        def serr = new StringBuffer();
        proc.consumeProcessOutput(sout, serr);
        proc.waitFor();
        def exitValue = proc.exitValue();
        if (logger != null)
        {
            if (sout.length() > 0)
            {
                logger.info("Standard output: " + sout);
            }
            if (serr.length() > 0)
            {
                logger.info("Standard error: " + serr);
            }
            logger.info("Exit value: " + exitValue);
        }
        return exitValue;
    }

    /**
    @return a String containing the BRM version (e.g. '5.2.2'), or null if
    there was a problem getting the version.
    */
    def static readVersion()
    {
        def installDir = System.getProperty('BRM.install.dir');
        if (installDir == null)
            installDir = '.';
        def versionFile = new File(installDir, 'version.txt');
        if (!versionFile.isFile())
        {
            return null;
        }
        return versionFile.getText('UTF-8').trim();
    }

    /**
    @return a String containing the BRM build type ('testing' or 'release') or
    null if there was a problem getting the build type.
    */
    def static readBuildtype()
    {
        def installDir = System.getProperty('BRM.install.dir');
        if (installDir == null)
            installDir = '.';
        def buildtypeFile = new File(installDir, 'buildtype.txt');
        if (!buildtypeFile.isFile())
        {
            return null;
        }
        return buildtypeFile.getText('UTF-8').trim();
    }

    /**
    @return a String containing the BRM build date and time in the format
    'YYYY-DD-MM-HHMMSS', or null if there was a problem getting the build date
    and time.
    */
    def static readBuilddatetime()
    {
        def installDir = System.getProperty('BRM.install.dir');
        if (installDir == null)
            installDir = '.';
        def builddatetimeFile = new File(installDir, 'builddatetime.txt');
        if (!builddatetimeFile.isFile())
        {
            return null;
        }
        return builddatetimeFile.getText('UTF-8').trim();
    }

    /**
    * Copy a file to file with the same name but with a unique, time-stamped
    * string appended to the name. The timestamp portion of the file name is
    * generated by generateTimeStamp().
    * @return the newly-created backup File object
    */
    def static backupInSameDir(File file)
    {
        File dir = file.getParentFile();
        if (dir == null)
            dir = new File('.');
        int sequence = 0;
        String timestamp = generateTimeStamp() ;
        def rand = new Random();
        while (true)
        {
            sequence += 1;
            def bakFile = new File(
                dir,
                file.getName() + '-' + timestamp + '-' + sequence);
            if (bakFile.createNewFile())
            {
                bakFile.setBytes(file.getBytes());
                return bakFile;
            }
            Thread.sleep(rand.nextInt(100) + 1);
        }
        throw new RuntimeException("Should never reach this point");
    }
}
