/*
 * Created on Jan 31, 2007
 * Copyright (c) 2007 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.Serializable;
import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

import com.sas.solutions.brm.errors.BRMConfigurationException;

/**
 * A class for reading configuration files in the common
 * "name=value" format.
 * 
 * IMPORTANT NOTE: This class stores keys in their original case, but retrieves
 * values case-insensitively. Therefore you cannot have two separate keys that 
 * differ only by case. If two such keys appear in the same config file, the 
 * second instance will change the value of the first.
 * 
 * Note: Leading and trailing spaces are trimmed off of both variable names 
 * and values.
 * 
 * Feature: This class remembers the order in which keys first appear in the 
 * config file. Use orderedKeyIterator() to iterate over the keys in the order
 * in which they first appeared. They keys are returned in their original case,
 * even though get() operates in a case-insensitive manner.
 * 
 * Feature: Any "${stamp}" sequence appearing in a variable value will be
 * substituted with "yyyyMMdd.HHmm", where yyyy is the year, MM is the month,
 * dd the day, HH the hour (24-hour format), and mm the minute. The time stamp
 * is created at the time the class is instantiated and does not change after
 * that. For example, if the config file contains:
 * 
 *    run.time=Started at: ${stamp}
 * 
 * and the the ConfigFile instance is created on 5 Feb 2007, 5:38 PM, then
 * config.get("run.time") will return "Started at: 20070205.1738".
 * 
 * @author davhan
 * 
 */
public class ConfigFile implements Serializable
{
    private static final Logger LOG = Logger.getLogger(ConfigFile.class);

    private Map valueMap;
    private List orderedKeys;
    private Date timeStamp;

    public ConfigFile()
    {
        this.valueMap = new HashMap();
        this.orderedKeys = new ArrayList();
        Calendar gc = new GregorianCalendar();
        this.timeStamp = gc.getTime();
    }
    
    public void loadFile(File file)
        throws IOException
    {
        LOG.info("Loading config file: " + file);
        BufferedReader br = new BufferedReader(new FileReader(file));
        while (true)
        {
            String line = br.readLine();
            if (line == null)
                break;
            String trimmedLine = line.trim();
            if (trimmedLine.length() == 0)
                continue;
            if (trimmedLine.startsWith("#"))
                continue;
            int i = trimmedLine.indexOf((char)'=');
            String key = null;
            String value = null;
            if (i >= 0)
            {
                key = trimmedLine.substring(0, i);
                key = key.trim();
                value = trimmedLine.substring(i + 1);
                value = value.trim();
            }
            else
            {
                key = trimmedLine;
                // leave value null if there is no "=" on the line
            }
            LOG.debug(key + "=" + value);
            set(key, value);
        }
        br.close();
    }
    
    public void loadFileByName(String fileName)
        throws IOException
    {
        File file = new File(fileName);
        loadFile(file);
    }
    
    public void loadResource(String resourceName)
        throws IOException
    {
        URL url = ClassLoader.getSystemResource(resourceName);
        if (url == null)
        {
            throw new BRMConfigurationException(
                    "Resource " + resourceName + " not found");
        }
        File file = new File(url.getFile());
        loadFile(file);
    }

    /**
     * Iterate over the keys from the config file in the order they appear
     * in the file.
     * @return an Iterator that traverses the keys in order
     */
    public Iterator orderedKeyIterator()
    {
        return this.orderedKeys.iterator();
    }
    
    public boolean containsKey(String key)
    {
        key = key.toLowerCase();
        return this.valueMap.containsKey(key);
    }
    
    public void set(String key, String value)
    {
        String lcKey = key.toLowerCase();
        if (!this.valueMap.containsKey(lcKey))
        {
            this.orderedKeys.add(key);
        }
        this.valueMap.put(lcKey, value);
    }
    
    public void del(String key)
    {
        key = key.toLowerCase();
        if (!this.valueMap.containsKey(key))
            return;
        this.valueMap.remove(key);
        this.orderedKeys.remove(key);
    }
    
    /**
     * @param key
     * @return The value for the variable named by key, or null if not set
     */
    public String get(String key)
    {
        String value = getRaw(key);
        if (value != null)
        {
            value = substituteVariables(value);
        }
        return value;
    }
    
    /**
     * @param key
     * @return The variable value without any ${variable} substitution
     */
    public String getRaw(String key)
    {
        key = key.toLowerCase();
        return (String)this.valueMap.get(key);
    }

    private String substituteVariables(String inputStr)
    {
        SimpleDateFormat f = new SimpleDateFormat("yyyyMMdd.HHmm");
        // Groovy requires extra backslash in front of "$"
        String s = inputStr.replaceAll("\\\$\\{stamp\\}", 
                f.format(this.timeStamp));
        return s;
    }
}
