// RunCommandAsSASServerUser.groovy
/*
 * Created on Mar 1, 2011
 * Copyright (c) 2011 by SAS Institute Inc., Cary, NC 27513
 */

package com.sas.solutions.brm.util;

import com.sas.iom.SASIOMDefs.VariableArray2dOfStringHolder;
import com.sas.meta.SASOMI.ISecurity_1_1;
import com.sas.meta.SASOMI.ISecurity_1_1Helper;
import com.sas.util.SasPasswordString;

import com.sas.solutions.brm.omr.JomaSupport;

/**
 * Run a command with login credentials fetched from metadata for the SAS
 * General Servers group (typical username "sassrv"). This only works on
 * Unix, because it uses the "su" command.
 */
public class RunCommandAsSASServerUser implements Runnable
{
    def logger;
    def props;
    def cmd;
    def numErrors = 0;

    /**
     * @param logger a Logger object used for logging status or errors
     * @param props  a Properties or Hashtable containing all the loaded
     *               BRM configuration properties.
     * @param cmd    a List containing the command name and arguments
     */
    public RunCommandAsSASServerUser(logger, props, cmd)
    {
        this.logger = logger;
        this.props = props;
        this.cmd = cmd;
    }

    def getProp(propname)
    {
        return this.props.get(propname);
    }

    def getPropOrDie(propname)
    {
        def propval = getProp(propname);
        if (propval == null)
        {
            def msg = ("Can't run command as SAS Server User: Missing property: " + propname);
            this.logger.error(msg);
            throw new Exception(msg);
        }
        return propval;
    }

    def getLoginName()
    {
        return getPropOrDie('oma.person.gensrvusr.person.name');
    }

    def getLoginCredentials()
    {
        def authdomain = getPropOrDie('oma.person.gensrvusr.login.authdomain');
        def loginname = getLoginName();
        this.logger.debug("getLoginCredentials: Getting the credentials for AuthDomain ${authdomain}, login name ${loginname}");
        def host = getPropOrDie('BRM.host.metadata');
        def port = getPropOrDie('BRM.port.metadata');
        def user = getPropOrDie('BRM.trusted.user');
        def password = getPropOrDie('BRM.trusted.password');
        this.logger.debug("getLoginCredentials: Connecting to metadata server ${host}:${port} as user ${user}");
        def joma = new JomaSupport();
        joma.connect(host, port, user, password);
        try
        {
            ISecurity_1_1 isec = ISecurity_1_1Helper.narrow(joma._factory.getConnection().getCMRHandle()._duplicate());
            VariableArray2dOfStringHolder output = new VariableArray2dOfStringHolder();
            String[][] options = new String[4][2];
            options[0][0] = "MaxListLen";
            options[0][1] = "10";
            options[1][0] = "IncludeBlankPasswords";
            options[1][1] = "NO";
            options[2][0] = "PrimaryOnly";
            options[2][1] = "NO";
            options[3][0] = "IdentityInfo";
            options[3][1] = "NO";
            isec.GetLoginForAuthDomain("", authdomain, options, output);
            String[][] value = output.value;
            def result = value.find
            { credentials ->
                def username = credentials[0];
                this.logger.debug("getLoginCredentials: Login name: ${username}");
                return (username == loginname);
            }
            if (result == null)
            {
                this.logger.warn("Could not retreive login credentials for user ${loginname}");
                return null;
            }
            this.logger.debug("getLoginCredentials: Login name matches");
            def encodedPassword = result[1];
            def decodedPassword = SasPasswordString.decode(encodedPassword);
            return ['username': loginname, 'password': decodedPassword];
        }
        finally
        {
            joma.disconnect();
        }
    }

    def getenv(varname)
    {
        def val = System.getenv(varname);
        if (val == null || val.trim().length() == 0)
        {
            def msg = ("Can't run command as SAS Server User: Missing environment variable: ${varname}, use brmrunas.sh to start BRM command.");
            this.logger.error(msg);
            throw new Exception(msg);
        }
        return val;
    }

    void run()
    {
        def loginname = getLoginName();
        def cmdStr = BRMUtil.arrayToStr(this.cmd);
        this.logger.info("RunCommandAsSASServerUser: Running command as user ${loginname}: ${cmdStr}");
        def BRM_EXEC_FIFO_IN = getenv('BRM_EXEC_FIFO_IN');
        def BRM_EXEC_FIFO_OUT = getenv('BRM_EXEC_FIFO_OUT');
        def BRM_SECURITY_TOKEN = getenv('BRM_SECURITY_TOKEN');
        this.logger.debug("RunCommandAsSASServerUser: opening: ${BRM_EXEC_FIFO_IN}");
        def writer = new PrintWriter(BRM_EXEC_FIFO_IN, 'UTF-8');
        this.logger.debug("RunCommandAsSASServerUser: sending command: ${cmdStr}");
        writer.println(BRM_SECURITY_TOKEN);
        writer.println(loginname);
        writer.println(cmdStr);
        writer.close();
        this.logger.debug("RunCommandAsSASServerUser: opening: ${BRM_EXEC_FIFO_OUT}");
        def reader = new BufferedReader(
                            new InputStreamReader(
                                new FileInputStream(BRM_EXEC_FIFO_OUT), 
                                "UTF-8"));
        this.logger.debug("RunCommandAsSASServerUser: executor opened ${BRM_EXEC_FIFO_OUT} for writing");
        def firstLine = true;
        while (true)
        {
            def line = reader.readLine();
            if (line == null)
                break;
            if (firstLine)
            {
                firstLine = false;
                try
                {
                    def exitCode = Integer.parseInt(line);
                    def msg = "RunCommandAsSASServerUser: program exit code: ${exitCode}";
                    if (exitCode == 0)
                    {
                        this.logger.info(msg);
                    }
                    else
                    {
                        ++this.numErrors;
                        this.logger.warn(msg);
                    }
                    continue;
                }
                catch (NumberFormatException e)
                {
                    this.logger.warn("RunCommandAsSASServerUser: Unable to get exit code from first output line");
                }
            }
            this.logger.info("RunCommandAsSASServerUser: program output: ${line}");
        }
        reader.close();
    }
}

// end-of-file
