// SASFileUtil.groovy
package com.sas.solutions.brm.util;

/**
 * Utility to back up and restore SAS data files in a cross-platform manner.
 */
class SASFileUtil
{
    def properties = null;
    def logger = null;
    File backupdir = null;
    File progdir = null;

    /**
     @param properties An object with get() method that retrieves
     configuration properties.
     @param logger A logger with an interface like log4j Logger.
     @param backupdir The full path to the backup directory into which
     directory trees will be backed up, or from which they will be restored.
     @param progdir The directory in which SAS program files will be
     created.
     */
    public SASFileUtil(properties, logger, File backupdir, File progdir)
    {
        this.properties = properties;
        this.logger = logger;
        this.backupdir = backupdir;
        this.progdir = progdir;
    }

    /**
     * Search for and back up all SAS datasets and catalogs found in a
     * directory tree.
     @param basename Backup base name.
        <ul>
            <li>basename is used as the name of the sub-directory of
            backupdir into which files are backed up</li>
            <li>basename + '-backup.sas' is the name of the SAS program file
            generated</li>
        </ul>
     @param basedir Full path to base of directory tree to be backed up by
     the generated SAS program.
     @return 0 if the files were backed up successfully, non-zero otherwise.
    */
    def backupSASFiles(String basename, File basedir)
    {
        File sasProg = createBackupProgram(basename, basedir);
        return execSASProgram(sasProg);
    }

    /**
     * Search a directory tree and create a SAS program to back up SAS
     * files.
     @param basename Backup base name.
        <ul>
            <li>basename is used as the name of the sub-directory of
            backupdir into which files are backed up</li>
            <li>basename + '-backup.sas' is the name of the SAS program file
            generated</li>
        </ul>
     @param basedir Full path to base of directory tree to be backed up by
     the generated SAS program.
     @return The File containing the generated SAS program.
     */
    def createBackupProgram(String basename, File basedir)
    {
        this.progdir.mkdirs();
        def progFile = new File(this.progdir, basename + '-backup.sas');
        progFile.write(''); // Create new empty file
        def numSASDirs = 0;
        def addSASDirToProgram = { dir ->
            numSASDirs += 1;
            if (numSASDirs > 1)
            {
                progFile.append("""
libname source;
filename cportout;
""");
            }
            def cportFile;
            if (dir.equals(basedir))
            {
                cportFile = new File(this.backupdir, basename + '.cport');
            }
            else
            {
                def destdir = new File(this.backupdir, basename);
                def parentdir = dir.getParentFile();
                def cportname = dir.getName() + '.cport';
                def reldir = BRMUtil.getRelativeFile(basedir, parentdir);
                if (reldir == null)
                {
                    destdir.mkdirs();
                    cportFile = new File(destdir, cportname);
                }
                else
                {
                    def cportDir = new File(destdir, reldir.getPath());
                    cportDir.mkdirs();
                    cportFile = new File(cportDir, cportname);
                }
            }
            progFile.append("""
libname source '${dir.getPath()}';
filename cportout '${cportFile.getPath()}';
proc cport library=source file=cportout;
run;
""");
        }
        findSASFiles(basedir, addSASDirToProgram);
        return progFile;
    }

    /**
     * Scan a directory tree for directories that contain SAS files.
     * For each such directory found, call callback(dirpath).
     * @param basedir Full path to base of directory tree to scan
     * @param callback Closure that will be called with one parameter, that
     * parameter will be the full path to a directory containing SAS files.
     */
    def static findSASFiles(File basedir, Closure callback)
    {
        def queue = new LinkedList();
        queue.add(basedir);
        while (queue.size() > 0)
        {
            def curdir = queue.remove();
            String[] dirlist = curdir.list();
            if (dirlist == null)
            {
                // Can't process non-existent directory
                continue;
            }
            def items = Arrays.asList(dirlist);
            Collections.sort(items, String.CASE_INSENSITIVE_ORDER);
            def sasFileFound = false;
            items.each
            { item ->
                File curFile = new File(curdir, item);
                if (curFile.isDirectory())
                {
                    queue.add(curFile);
                }
                else if (!sasFileFound && isSASFile(curFile))
                {
                    sasFileFound = true;
                    callback(curdir);
                }
            }
        }
    }

    /**
    * @return true if the given file is a SAS dataset or catalog,
    * false otherwise.
    */
    def static isSASFile(File file)
    {
        if (!file.isFile())
            return false;
        String name = file.getPath().toLowerCase();
        if (name.endsWith('.sas7bcat') || name.endsWith('.sas7bdat'))
            return true;
        return false;
    }

    /**
     * Restore SAS files from a backup subdirectory to a directory tree.
     @param basename Backup base name.
        <ul>
            <li>basename is used as the name of the sub-directory of
            backupdir from which files are restored</li>
        </ul>
     @param basedir Full path to base of directory tree into which files
     will be restored.
     @return The number of times there were errors running PROC CIMPORT. 0
     means all is good, or that there were no SAS files to restore.
     */
    def restoreSASFiles(String basename, File basedir)
    {
        // Directory for temporary SAS programs
        this.progdir.mkdirs();
        def numErrors = 0;
        def handleCPORTFile = { cportFile ->
            def sasProg = createRestoreProgram(basename, basedir, cportFile);
            def rc = execSASProgram(sasProg);
            if (rc != 0)
            {
                if (BRMUtil.isWindows())
                {
                    ++numErrors;
                }
                else
                {
                    def runner = new RunCommandAsSASServerUser(
                                        this.logger, this.properties, cmd);
                    runner.run();
                    numErrors += runner.numErrors;
                }
            }
        }
        findCPORTFiles(basename, this.backupdir, handleCPORTFile);
        return numErrors;
    }

    /** Remove any temporary SAS programs or log files from previous runs.
    */
    def deleteTempSASProgramsAndLogs()
    {
        this.logger.info("Deleting temporary files in ${this.progdir}");
        this.progdir.list().each
        { name ->
            def lowerName = name.toLowerCase();
            if (!(lowerName.startsWith('tmp') &&
                  (lowerName.endsWith('.sas') || lowerName.endsWith('.log'))))
                return;
            File curFile = new File(this.progdir, name);
            curFile.delete();
        }
    }

    /**
     * Create a SAS program that restores SAS files to a directory tree.
     * As a side-effect, destination directories into which the files will
     * be restored will be created.
     @param basename Backup base name.
        <ul>
            <li>basename is used as the name of the sub-directory of
            backupdir from which files are restored</li>
            <li>basename + '-restore.sas' is the name of the SAS program
            file generated</li>
        </ul>
     @param basedir Full path to base of directory tree into which files
     will be restored by the generated SAS program.
     @return The File containing the generated SAS program.
     */
    def createRestoreProgram(String basename, File basedir, File cportFile)
    {
        def progFile = File.createTempFile('tmp', '.sas', this.progdir);
        progFile.write(''); // Create new empty file
        def backupBase = new File(this.backupdir, basename);
        def destdir;
        if (cportFile.getName() == basename + '.cport')
        {
            destdir = basedir;
        }
        else
        {
            def cportDir = new File(BRMUtil.splitExt(cportFile)[0]);
            def relpath = BRMUtil.getRelativeFile(backupBase, cportDir);
            destdir = new File(basedir, relpath.getPath());
        }
        destdir.mkdirs();
        progFile.append("libname dest '${destdir.getPath()}';\n");
        progFile.append("filename cportout '${cportFile.getPath()}';\n");
        progFile.append("proc cimport library=dest infile=cportout;\n");
        progFile.append("run;\n");
        return progFile;
    }

    def findCPORTFiles(String basename, File basedir, Closure callback)
    {
        File topLevelCPortFile = new File(basedir, basename + '.cport');
        if (topLevelCPortFile.isFile())
        {
            callback(topLevelCPortFile);
        }
        def queue = new LinkedList();
        queue.add(new File(basedir, basename));
        while (queue.size() > 0)
        {
            def curdir = queue.remove();
            String[] dirlist = curdir.list();
            if (dirlist == null)
            {
                // Can't process non-existent directory
                continue;
            }
            def items = Arrays.asList(dirlist);
            Collections.sort(items, String.CASE_INSENSITIVE_ORDER);
            items.each
            { item ->
                File curFile = new File(curdir, item);
                if (curFile.isDirectory())
                {
                    queue.add(curFile);
                }
                else if (item.toLowerCase().endsWith('.cport'))
                {
                    callback(curFile);
                }
            }
        }
    }

    /**
    * Run a SAS program. Create a log file with the same name as the SAS
    * program, except with a .log extension.
    * Uses the following configuration properties:
    *   base.install.dir  The directory containing the SAS executable
    *   sas.exec.file.name  The SAS executable file name (sas.exe on Windows)
    * @return The exit code from sas.
    */
    def execSASProgram(File sasprog)
    {
        // <exec executable="${base.install.dir}/${sas.exec.file.name}">
        // "C:\Program Files\SAS\SASFoundation\9.2\sas.exe" -icon -nosplash -log backupsas.log -sysin backupsas.sas
        if (!sasprog.isFile())
        {
            throw new Exception("Missing SAS program: ${sasprog}");
        }
        def logfile = new File(BRMUtil.splitExt(sasprog)[0] + '.log');
        def sashome = getProp('base.install.dir');
        def sasexe_name = getProp('sas.exec.file.name');
        def sasexe = new File(sashome, sasexe_name);
        if (!sasexe.isFile())
        {
            throw new Exception("Missing SAS executable: ${sasexe}");
        }
        def cmd = [
            sasexe.getPath(),
            '-icon',
            '-nosplash',
            '-log',
            logfile.getPath(),
            '-sysin',
            sasprog.getPath(),
            ];
        return BRMUtil.execProgram(this.logger, cmd);
    }

    def getProp(propname)
    {
        def propval = this.properties.get(propname);
        if (propval == null)
        {
            throw new Exception("Missing configuration property: ${propname}");
        }
        return propval;
    }
}

// end-of-file
