// SASTableServerUtil.groovy

package com.sas.solutions.brm.util;

class SASTableServerUtil
{
    // Initialized by constructor
    def logger;
    def props;
    def backupdir;

    // Initialized by init()
    def contentDir = null;
    def vulcanBinDir = null;
    def sastback = null;
    def tsBackupDir = null;

    def SASTableServerUtil(logger, props, backupdir)
    {
        this.logger = logger;
        this.props = props;
        this.backupdir = backupdir;
    }

    /**
      * Intialize the SAS Content Server utility common fields.
      * @return true if SAS Table Server is installed, false otherwise.
      */
    def init()
    {
        contentDir = props.get("iomsrv.tableserver.content.dir");
        if (contentDir == null)
        {
            logger.warn("Missing property: iomsrv.tableserver.content.dir");
            logger.info("Assuming SAS Table Server is not installed.");
            return false;
        }
        contentDir = new File(contentDir);
        if (!contentDir.isDirectory())
        {
            logger.warn("No SAS Table Server content directory: ${contentDir}");
            logger.info("Assuming SAS Table Server is not installed.");
            return false;
        }
        logger.info("Table Server content directory: ${contentDir}");
        def baseInstallDir = props.get("base.install.dir");
        if (baseInstallDir == null)
        {
            logger.warn("Missing property: base.install.dir");
            logger.info("Assuming SAS Table Server is not installed.");
            return false;
        }
        def vulcanDir = new File(baseInstallDir, 'vulcan');
        vulcanBinDir = new File(vulcanDir, 'bin');
        sastback = new File(vulcanBinDir, 'sastback.exe');
        if (!sastback.isFile())
            sastback = new File(vulcanBinDir, 'sastback');
        if (!sastback.isFile())
        {
            logger.warn("Missing sastback program in directory: ${vulcanBinDir}");
            logger.info("Assuming SAS Table Server is not installed.");
            return false;
        }
        logger.info("sastback utility program: ${sastback}");
        tsBackupDir = new File(backupdir, 'tableServer');
        logger.info("Table Server backup directory: ${tsBackupDir}");
        return true;
    }

    def backupSASTableServer()
    {
        logger.info("Backing up SAS Table Server");
        if (!init())
            return false;
        tsBackupDir.mkdirs();
        // iterate through all the Content files in table server
        def errors = 0;
        contentDir.eachFile() { file->   
            if (file.getName().toLowerCase().endsWith(".fbd") ||
                file.getName().toLowerCase().endsWith(".fdb"))
            {
                File dataFile = new File(contentDir, file.getName());
                File backupFile = new File(tsBackupDir, file.getName()+'.fbk');
                def cmd = [
                        sastback.getPath(), 
                        '-B',
                        dataFile.getPath(),
                        '-MO',
                        'read_only',
                        backupFile.getPath()
                    ];
                def rc = BRMUtil.execProgram(logger, cmd);
                if (rc != 0)
                {
                    logger.error("Could not back up file ${dataFile}: See log for details.");
                    errors += 1;
                }

                File lockFile = new File(contentDir, file.getName() + ".lck");
                if (lockFile.exists())
                {
                    lockFile.delete();
                }
            }
        }
        if (errors > 0)
            return false;
        return true;
    }

    def restoreSASTableServer()
    {
        logger.info("Restoring SAS Table Server");
        if (!init())
            return false;
        if (!tsBackupDir.isDirectory())
        {
            logger.warn("Missing SAS Table Server backup directory: ${tsBackupDir}");
            return false;
        }
        // iterate through all the Content archives for table server
        def errors = 0;
        tsBackupDir.eachFile() { file->   
            if (file.getName().endsWith(".fbk"))
            {
                File backupFile = new File(tsBackupDir, file.getName());
                File dataFile = new File(contentDir, 
                                         file.getName().replace(".fbk", ""));
                def option = '-R';
                if (dataFile.exists())
                    option = '-REP';
                def cmd = [
                        sastback.getPath(), 
                        option,
                        backupFile.getPath(),
                        dataFile.getPath()
                  ];
                def rc = BRMUtil.execProgram(logger, cmd);
                if (rc != 0)
                {
                    logger.error("Could not restore backup file ${backupFile}: See log for details.");
                    errors += 1;
                }
                
                File lockFile = new File(contentDir, 
                                file.getName().replace(".fbk","") + ".lck");
                if (lockFile.exists())
                {
                    lockFile.delete();
                }
            }
        }       
        if (errors > 0)
            return false;
        return true;
    }
}

// end-of-file
