/*
 * Created on Aug 05, 2010
 * Copyright (c) 2010 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm.webapp;

import com.sas.solutions.brm.util.BRMUtil;

/**
 * WebLogic scripting utilities.
 */
class WLAdmin extends WebappAdminBase
{
    public WLAdmin(props, logger)
    {
        super(props, logger);
    }

    /**
     * @param command  Command to execute (i.e. start, stop, status)
     * @param serverName Server name in which application is running
     * @param appName  Name of application to control
     * @return The exit code of the command run (0 means success)
     */
    def sendCommand(command, serverName, appName)
    {
        if (!isManagedServerRunning(serverName))
        {
            this.logger.info(
                "Will not send ${command} command to webapp ${appName} in WebLogic. ${serverName} is not running.");
            return;
        }
        this.logger.info(
            "Sending ${command} command to webapp ${appName} in WebLogic server ${serverName}");
        def exitCode = -1;
        if (("-stop".equals(command.toLowerCase())) || 
            ("-start".equals(command.toLowerCase())))
        {
            // handle -stop and -start commands
            def moreParams = [];

            def maxAttempts = 1;
            if ("-start".equals(command.toLowerCase()))
            {
                // Due to WebLogic issues, try twice when restarting
                maxAttempts = 2;
            }
            else
            {
                // Put the app in admin mode when stopping
                moreParams = ['-adminmode'];
            }

            def prevExitCodePolicy = getExitCodePolicy();
            try
            {
                def attemptNum;
                for (attemptNum=1; attemptNum <= maxAttempts; ++attemptNum)
                {
                    if (attemptNum < maxAttempts)
                        setExitCodePolicy('info');
                    else
                        setExitCodePolicy(prevExitCodePolicy);
                    exitCode = weblogicDeployCommand(appName, command,
                                                                moreParams); 
                    if (exitCode == 0)
                        break; // successful, no need to retry
                }
            }
            finally
            {
                setExitCodePolicy(prevExitCodePolicy);
            }
        }
        else if ("forceshutdown".equals(command.toLowerCase()))
        {
            // handle forceshutdown of the managed server 
            exitCode = weblogicAdminCommand(serverName, command);
        }
        else
        {
            this.logger.error("Unsupported WebLogic command: ${command}");
        }
        return exitCode;
    }

    /**
     * @return a List of AppInfo objects
     */
    def getAppList()
    {
        def apps = [];
        def propertyList = [  
                "webapp.portal.archive.name",
                "webapp.bidashmid.archive.name", 
                "webapp.fmsmidtier.archive.name",
                "webapp.hcmmidtier.archive.name",
                "webapp.spmmidtier.archive.name",
                "webapp.svcs.scs.ear.archive.name", // sas.wip.scs9.2
                ];
        propertyList.each
        { appPropName ->
            def webappname = getProp(appPropName);
            if (webappname == null)
            {
                // Not installed on this server
                return;
            }
            this.logger.info("Webapp property: ${appPropName}=${webappname}");
            webappname += ".ear";

            // Managed server is server1 by default   
            String serverNum ="server1";  
            if (webappname.contains("sas.portal"))
            {
                serverNum = getPropOrDie("webappsrv.javaportal.server");
            }
            if (webappname.contains("sas.bidashboard"))
            {
                serverNum = getPropOrDie("webappsrv.bidashmid.server");
            }
            if (webappname.contains("sas.financialmanagement"))
            {
                serverNum = getPropOrDie("webappsrv.fmsmidtier.server");
            }
            if (webappname.contains("sas.humancapitalmanagement"))
            {
                serverNum = getPropOrDie("webappsrv.hcmmidtier.server");
            }
            if (webappname.contains("sas.strategymanagement"))
            {
                serverNum = getPropOrDie("webappsrv.spmmidtier.server");
            }
            if (webappname.contains("sas.wip.scs"))
            {
                serverNum = getPropOrDie("webappsrv.webinfpltfm.server");
            }
            def serverPropName = "webappsrv.${serverNum}.name";
            def managedServerName = getProp(serverPropName);
            if (managedServerName == null)
                managedServerName = getProp(serverPropName  + ".tmp");
            if (managedServerName == null)
            {
                this.logger.error(
                    "Could not find managed server name for application: ${webappname}: Missing property: ${serverPropName}");
                return;
            }
            apps.add(new AppInfo(managedServerName, webappname));
        }
        return apps;
    }

    // This is called during a restore.
    // This stops the servers, not just the web applications.
    def stopServers()
    {
        this.logger.info("Stopping managed servers");
        getServerList().each()
        { serverName ->
            stopServer(serverName);
        }
    }

    def stopServer(serverName)
    {
        this.logger.info("Stopping managed server: \"${serverName}\"");
        weblogicAdminCommand(serverName, "FORCESHUTDOWN");
    }

    // This is mainly for testing.
    // This starts the known web application servers, there are others.
    def startServers()
    {
        this.logger.info("Starting managed servers");
        getServerList().each()
        { serverName ->
            startServer(serverName);
        }
    }

    def startServer(serverName)
    {
        this.logger.info("Starting managed server: \"${serverName}\"");
        weblogicAdminCommand(serverName, "START");
    }
    
    /*
     *  weblogic.Deploy command
     */
    def weblogicDeployCommand(String webappname, String command, moreParams)
    {
        def params = ['-name', webappname, command];
        if (moreParams)
            params += moreParams;
        def result = weblogicJavaCommand('weblogic.Deployer', params, true);
        return result.exitValue;
    }
    
    /*
     *  weblogic.Admin managed server commands
     */
    def weblogicAdminCommand(String managedServerName, String command)
    {
        def params = [command, managedServerName];
        def result = weblogicJavaCommand('weblogic.Admin', params, true);
        return result.exitValue;
    }

    /*
     *  weblogic.Admin GETSTATE
     */
    public boolean isManagedServerRunning(String server)
    {
        def params = ['GETSTATE', server];
        def result = weblogicJavaCommand('weblogic.Admin', params, false);
        if (result.sout.length() > 0)
        {
            if (result.sout.indexOf("RUNNING") >= 0)
            {
                if (logger != null)
                    logger.info("Managed server (" + server + ") is RUNNING.");
                return true;
            }
        }
        // anything else means it is not running 
        return false;   
    }

    /**
     * Run a WebLogic Java tool.
     * @param className  The full name of the Java class to execute
     * @param params  Additional command-line parameters to pass
     * @param  dolog  If true, log cmd and result
     * @return a Result object with attributes sout, serr, exitValue
     */
    def weblogicJavaCommand(className, params, dolog)
    {
        def weblogic_home = getPropOrDie("webappsrv.weblogic.home.dir");
        def jar = new File(weblogic_home, "server/lib/weblogic.jar");
        def url = getPropOrDie("webappsrv.server.admin.url");
        def user = getPropOrDie("webappsrv.admin.userid");
        def password = getPropOrDie("BRM.webapp.admin.password");
        password = BRMUtil.decodePW(password);
        
        def sout = new StringBuffer();
        def serr = new StringBuffer();
        def cmd = ['java', '-cp', jar, 
            className,
            '-adminurl', url,
            '-username', user,
            '-password',
            ];
        def cmdTrace = cmd.collect { return it; }
        cmd.add(password);
        cmdTrace.add('*****');
        params.each { cmd.add(it); }
        params.each { cmdTrace.add(it); }
        if (dolog)
        {
            def trace = BRMUtil.arrayToStr((String[])cmdTrace);
            this.logger.info("Running WebLogic command: ${trace}");
        }
        def proc = cmd.execute();
        proc.consumeProcessOutput(sout, serr);
        proc.waitFor();
        def exitValue = proc.exitValue();
        def result = ['sout': sout, 'serr': serr, 'exitValue': exitValue];
        if (dolog)
        {
            logCommandResult(result);
        }
        return result;
    }
}
