/*
 * Created on Aug 04, 2010
 * Copyright (c) 2010 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm.webapp;

import com.sas.solutions.brm.util.BRMUtil;

/**
 * WebSphere scripting utilities.
 */
class WSAdmin extends WebappAdminBase
{
    // e.g. C:\Program Files\IBM\WebSphere6\AppServer
    def WAS_HOME = null;

    public WSAdmin(props, logger)
    {
        super(props, logger);
        def wsInstallDirProp = 'webspherend.install.dir';
        WAS_HOME = getPropOrDie(wsInstallDirProp);
        if (!(new File(WAS_HOME)).isDirectory())
        {
            def msg = "Missing WebSphere installation directory: ${wsInstallDirProp}=${WAS_HOME}";
            logger.error(msg);
            throw new Exception(msg);
        }
        logger.debug("WAS_HOME=${WAS_HOME}");
    }

    /**
     * @param command  Command to execute (i.e. start, stop, status)
     * @param serverName Server name in which application is running
     * @param appName  Name of application to control
     * @return The exit code of the command run (0 means success)
     */
    def sendCommand(command, serverName, appName)
    {
        // Translate WebLogic command to WebSphere command
        if (command == "-start")
            command = "start";
        else if (command == "-stop" || command == "FORCESHUTDOWN")
            command = "stop";
        def installDir = System.getProperty("BRM.install.dir");
        def scriptsDir = new File(installDir, "scripts");
        def scriptFile = new File(scriptsDir, "SASApp.jy");
        if (!scriptFile.isFile())
        {
            throw new Exception("Missing WebSphere admin helper script: ${scriptFile}");
        }
        def binDir = new File(WAS_HOME, 'bin');
        def adminScript = null;
        def cmd = [];
        def cmdTrace = []
        def addToCmd = { part ->
            cmd.add(part);
            cmdTrace.add(part);
        }
        def addPasswordToCmd = { password ->
            cmd.add(password);
            cmdTrace.add("*****");
        }
        if (System.getProperty("os.name").startsWith('Win'))
        {
            adminScript = new File(binDir, "wsadmin.bat");
            addToCmd("cmd.exe");
            addToCmd("/c");
        }
        else
        {
            adminScript = new File(binDir, "wsadmin.sh");
        }
        if (!adminScript.isFile())
        {
            throw new Exception("Missing WebSphere admin script: ${adminScript}");
        }
        addToCmd(adminScript.getPath());
        addToCmd("-lang");
        addToCmd("jython");
        // Missing user or password is Ok for unsecured server
        def username = props.get("BRM.webapp.admin.user");
        def password = props.get("BRM.webapp.admin.password");
        password = BRMUtil.decodePW(password);
        if (username != null && username.length() > 0)
        {
            addToCmd("-username");
            addToCmd(username);
        }
        if (password != null && password.length() > 0)
        {
            addToCmd("-password");
            addPasswordToCmd(password);
        }
        addToCmd("-f");
        addToCmd(scriptFile.getPath());
        addToCmd(command);
        addToCmd(serverName);
        addToCmd(appName);

        def trace = BRMUtil.arrayToStr((String[])cmdTrace);
        logger.info("Running WebSphere admin command: ${trace}");

        def sout = new StringBuffer();
        def serr = new StringBuffer();
        def proc = cmd.execute();
        proc.consumeProcessOutput(sout, serr);
        proc.waitFor();
        def exitValue = proc.exitValue();
        def result = ['sout': sout, 'serr': serr, 'exitValue': exitValue];
        logCommandResult(result);
        return result;
    }

    def getAppList()
    {
        def applist = [];
        getAppPropNames().each
        { appPropName ->
            def appName = getProp(appPropName);
            if (appName == null)
            {
                // This app must not be installed on this server, skip it.
                return;
            }
            def serverNumPropName = getServerNumPropName(appPropName);
            def serverNum = getPropOrDie(serverNumPropName);
            def serverPropName = (String)"webappsrv.${serverNum}.name";
            def serverName = getProp(serverPropName);
            applist.add(new AppInfo(serverName, appName));
        }
        return applist;
    }

    // This is called during a restore.
    def stopServers()
    {
        this.logger.info("Stopping WebSphere application servers");
        getServerList().each()
        { serverName ->
            stopServer(serverName);
        }
    }

    def stopServer(serverName)
    {
        this.logger.info("Stopping WebSphere application server: \"${serverName}\"");
        def cmd = getServerCmd('stop', serverName);
        def rc = BRMUtil.execProgram(logger, cmd);
        if (rc != 0)
        {
            logger.warn("When stopping WebSphere server ${serverName}, stop script returned exit code ${rc}. Check BRM log for details.");
        }
    }

    // This is mainly for testing.
    // This starts the known web application servers, there are others.
    def startServers()
    {
        this.logger.info("Starting WebSphere application servers");
        getServerList().each()
        { serverName ->
            startServer(serverName);
        }
    }

    def startServer(serverName)
    {
        this.logger.info("Starting WebSphere application server: \"${serverName}\"");
        def cmd = getServerCmd('start', serverName);
        def rc = BRMUtil.execProgram(logger, cmd);
        if (rc != 0)
        {
            logger.warn("When starting WebSphere server ${serverName}, start script returned exit code ${rc}. Check BRM log for details.");
        }
    }

    // Internal method
    // Return a list containing the command and parameters to stop or start
    // a server. scriptBase is 'start' or 'stop', used to form the
    // startServer or stopServer script name.
    def getServerCmd(scriptBase, serverName)
    {
        def s;
        def wsInstallDir = new File(WAS_HOME);
        def wsProfileProp = 'webappsrv.websphere.profile.name';
        def wsProfile = props.get(wsProfileProp);
        if (wsProfile == null || wsProfile.length() == 0)
        {
            s = "Can't find WebSphere profile: Missing property: ${wsProfileProp}";
            logger.error(s);
            throw new Exception(s);
        }
        logger.debug("WebSphere profile: ${wsProfile}");
        def wsProfileDir = new File(wsInstallDir, 'profiles');
        wsProfileDir = new File(wsProfileDir, wsProfile);
        if (!wsProfileDir.isDirectory())
        {
            s = "Missing WebSphere profile directory: ${wsProfileProp}=${wsProfile}: ${wsProfileDir}";
            logger.error(s);
            throw new Exception(s);
        }
        def cmd = []
        def binDir = new File(wsProfileDir, 'bin');
        def scriptFile = null;
        if (System.getProperty("os.name").startsWith('Win'))
        {
            scriptFile = new File(binDir, scriptBase + 'Server.bat');
            cmd.add("cmd.exe");
            cmd.add("/c");
        }
        else
        {
            scriptFile = new File(binDir, scriptBase + 'Server.sh');
        }
        if (!scriptFile.isFile())
        {
            s = "Missing WebSphere script file: ${scriptFile}";
            logger.error(s);
            throw new Exception(s);
        }
        cmd.add(scriptFile.getPath());
        cmd.add(serverName);
        return cmd;
    }

    def static appPropMap = [  
                "webapp.portal.app.name": "webappsrv.javaportal.server",
                "webapp.bidashmid.app.name": "webappsrv.bidashmid.server", 
                "webapp.fmsmidtier.app.name": "webappsrv.fmsmidtier.server",
                "webapp.hcmmidtier.app.name": "webappsrv.hcmmidtier.server",
                "webapp.spmmidtier.app.name": "webappsrv.spmmidtier.server",
                ];

    def getAppPropNames()
    {
        return (appPropMap.collect { key, value -> return key; } );
    }

    def getServerNumPropName(appPropName)
    {
        return appPropMap.get(appPropName);
    }
}

// end-of-file
