/*
 * Created on Aug 05, 2010
 * Copyright (c) 2010 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm.webapp;

/**
 * Web application scripting utilities.
 */
public class WebappAdminBase
{
    // Useful closures
    def errorOutput = { s ->
        logger.error(s);
    }
    def warnOutput = { s ->
        logger.warn(s);
    }
    def infoOutput = { s ->
        logger.info(s);
    }
    def debugOutput = { s ->
        logger.debug(s);
    }

    def props = null;
    def logger = null;
    def ecpolicy = 'error';
    def nonZeroExitLogMethod = errorOutput;

    /**
     * @param props  Hashtable containing all BRM properties
     * @param logger  A Log4j Logger object
     */
    public WebappAdminBase(props, logger)
    {
        this.props = props;
        this.logger = logger;
    }

    /**
     * @param ecpolicy  Code string that determines what to do in case a
     *  command generates a non-zero exit code:
     * 'error': Log an ERROR message
     * 'warn': Log a WARNING message
     * 'info': Log an INFO message
     * 'debug': Log a DEBUG message
     */
    def setExitCodePolicy(ecpolicy)
    {
        this.ecpolicy = ecpolicy.toLowerCase();
        def outputMethod = null;
        if (getExitCodePolicy().equals('error'))
            outputMethod = errorOutput;
        else if (getExitCodePolicy().equals('warn'))
            outputMethod = warnOutput;
        else if (getExitCodePolicy().equals('info'))
            outputMethod = infoOutput;
        else
            outputMethod = debugOutput;
        this.nonZeroExitLogMethod = outputMethod;
    }

    /**
     * @return the current exit code policy in effect.
     * @see setExitCodePolicy()
     */
    def getExitCodePolicy()
    {
        return this.ecpolicy;
    }

    // Internal utility method
    def logCommandResult(result)
    {
        def msg;
        if (result.sout.length() > 0)
        {
            msg = result.sout.toString();
            logger.info(msg);
        }
        if (result.exitValue != 0)
        {
            if (result.serr.length() > 0)
            {
                msg = result.serr.toString();
                nonZeroExitLogMethod(msg);
            }
            msg = 'Web server command exit code: ' + result.exitValue;
            nonZeroExitLogMethod(msg);
        }
        else
        {
            if (result.serr.length() > 0)
            {
                logger.info(result.serr.toString());
            }
            logger.info("Web server command command exit code: ${result.exitValue}");
        }
    }

    /**
     * @param command  Command to execute (i.e. start, stop, status)
     * @param serverName Server name in which application is running
     * @param appName  Name of application to control
     * @return The exit code of the command run (0 means success)
     */
    def sendCommand(command, serverName, appName)
    {
        throw new Exception("Abstract method called");
    }

    /**
     * @return a List of AppInfo objects
     */
    def getAppList()
    {
        throw new Exception("Abstract method called");
    }

    def getServerList()
    {
        def result = new LinkedHashSet();
        def appList = getAppList();
        appList.each
        { appInfo ->
            result.add(appInfo.serverName);
        }
        return result;
    }

    def stopServers()
    {
        throw new Exception("Abstract method called");
    }

    // Only supported for WebLogic managed servers
    def stopServer(serverName)
    {
        throw new Exception("Abstract method called");
    }

    // This is mainly for testing.
    def startServers()
    {
        throw new Exception("Abstract method called");
    }

    // Only supported for WebLogic managed servers
    def startServer(serverName)
    {
        throw new Exception("Abstract method called");
    }

    def sendCommandToAppList(command)
    {
        getAppList().each
        { appinfo ->
            sendCommand(command, appinfo.serverName, appinfo.appName);
        }
    }

    def getProp(propname)
    {
        // Note: The (String) cast is necessary!
        return this.props.get((String)propname);
    }

    def getPropOrDie(propname)
    {
        def propval = getProp(propname);
        if (propval == null)
        {
            throw new Exception("Missing config property: ${propname}");
        }
        return propval;
    }
}
