// scripts/executor.groovy
// Execute commands read from a pipe, sent by a BRM command.
//
// Run this script as user root to enable it to switch to other users as
// needed. Run as user sassrv if special commands only need to be run as
// sassrv. The executor enforces security by requiring the BRM command to
// send a security token, which was randomly generated by brmrunas.sh and
// passed to the executor and the BRM command in an environment variable.

// Communication protocol:
//
// The executor starts by opening the input named pipe for reading. The open
// operation blocks and waits.
//
// The BRM command opens the input named pipe for writing, which unblocks
// the executor.  The BRM command sends the following lines:
// line 1: The security token
// line 2: The username of the user that will run the command
// lines 3-N: Command lines to be executed as a batch sh script
// The BRM command then closes the input named pipe, then opens the output
// named pipe for reading.
//
// If the first command line received by the executor consists only of the
// word "quit" (without quotes), then the executor exits immediately.
// This is how brmrunas.sh forces the executor to exit, after the BRM command
// has finished.
//
// Otherwise, the executor saves the command lines in a temporary shell
// script and executes it. If the requested username is not blank and is
// different than the user that the executor is already running as, and
// the executor has root privileges, then the executor switches to the
// requested user before running the shell script:
//     "su username -c temporary-script"
//
// The executor opens the output named pipe for writing. The open operation
// should succeed quickly because the BRM command has opened it for reading.
// The executor sends the following lines:
// line 1: An integer indicating the exit status of the temporary shell script
// lines 2-N: The standard output lines from the script, followed by the
//            standard error lines if any.
//
// Finally, the executor closes the output named pipe.
// The executor then re-opens the input named pipe and waits again.


def getEffectiveUsername()
{
    // Get the current effective user name.
    // You can't get this from the USER environment variable, because
    // that variable does not change when "su -c" is run.
    def proc = ['id', '-u', '-n'].execute();
    def sout = new StringBuffer();
    def serr = new StringBuffer();
    proc.consumeProcessOutput(sout, serr);
    proc.waitFor();
    return sout.toString().trim();
}


def getenv(varname)
{
    def val = System.getenv(varname);
    if (val == null || val.trim().length() == 0)
    {
        println("Missing environment variable: ${varname}");
        System.exit(1);
    }
    return val;
}

def createTempCmdFile()
{
    def tempCmdFile = File.createTempFile('BRM-', '.sh');
    // Make the script executable.
    // Don't allow any user but us (root) to change this file.
    // This closes a potential security hole!
    def proc = ['chmod', '755', tempCmdFile.getPath()].execute();
    proc.waitFor();
    return tempCmdFile;
}

def executeCommand(user, commandFile)
{
    // Execute the command
    def cmd = null;
    def exitValue = -1;
    def sout = new StringBuffer();
    def serr = new StringBuffer();
    if (user == null || user.trim().length() == 0 || user.equals(EXEC_USER))
    {
        // No need to switch user
        cmd = "${commandFile}";
    }
    else if (EXEC_USER.equals('root'))
    {
        // Can switch to any user without entering password
        cmd = "su ${user} -c ${commandFile}";
    }
    else
    {
        // Leave cmd null
        serr.append(
            "executor: can't switch to user ${user}: not running as root\n");
    }
    if (cmd != null)
    {
        def input = { inputWriter ->
            // http://groovy.codehaus.org/groovy-jdk/java/lang/Process.html#withWriter%28groovy.lang.Closure%29
            // "Creates a new BufferedWriter as stdin for this process,
            // passes it to the closure, and ensures the stream is flushed
            // and closed after the closure returns."
            inputWriter.write('');
        }
        try
        {
            def proc = cmd.execute();
            proc.withWriter(input);
            proc.consumeProcessOutput(sout, serr);
            proc.waitFor();
            exitValue = proc.exitValue();
        }
        catch (Throwable e)
        {
            serr.append(e.toString());
            serr.append('\n');
        }
    }
    // This blocks until the other process opens the FIFO for reading
    def out = new PrintWriter(BRM_EXEC_FIFO_OUT, 'UTF-8');
    out.println(exitValue);
    out.flush();
    if (serr.length() > 0)
    {
        out.print(serr.toString());
        out.flush();
    }
    if (sout.length() > 0)
    {
        out.print(sout.toString());
        out.flush();
    }
    out.close();
}

EXEC_USER = getEffectiveUsername();
BRM_EXEC_FIFO_IN = getenv('BRM_EXEC_FIFO_IN');
BRM_EXEC_FIFO_OUT = getenv('BRM_EXEC_FIFO_OUT');
BRM_SECURITY_TOKEN = getenv('BRM_SECURITY_TOKEN');

def running = true;
while (running)
{
    // This blocks until another process opens the FIFO for writing
    def reader = new BufferedReader(new InputStreamReader(
                            new FileInputStream(BRM_EXEC_FIFO_IN), 'UTF-8'));
    def securityTokenReceived = false;
    def user = null;
    def tempCmdFile = null;
    def cmdFileWriter = null;
    while (true)
    {
        def line = reader.readLine();
        if (line == null)
        {
            if (cmdFileWriter != null)
                cmdFileWriter.close();
            break;
        }
        if (!securityTokenReceived)
        {
            if (!line.equals(BRM_SECURITY_TOKEN))
            {
                println("BRM executor: Invalid security token received");
                break;
            }
            securityTokenReceived = true;
            continue;
        }
        if (user == null)
        {
            user = line;
            continue;
        }
        if (tempCmdFile == null)
        {
            if (line.equals('quit'))
            {
                running = false;
                break;
            }
            tempCmdFile = createTempCmdFile();
            cmdFileWriter = new PrintWriter(tempCmdFile, 'UTF-8');
            cmdFileWriter.println('#!/bin/sh');
        }
        cmdFileWriter.println(line);
    }
    reader.close();
    if (tempCmdFile != null)
    {
        executeCommand(user, tempCmdFile);
        tempCmdFile.delete();
    }
}

// end-of-file
