// Fix incorrect IDs in TextStore objects that may have been created due to
// defect S0705632.
import com.sas.solutions.brm.omr.OMRConnection;

class MetadataIDFixer
{
    def conn = null;
    boolean commit = false;

    int numChangedLines = 0;
    int numChangedObjs = 0;
    int numWarnings = 0;

    public MetadataIDFixer(conn, boolean commit)
    {
        this.conn = conn;
        this.commit = commit;
    }

    def static getMetadataIDsFromLine(line)
    {
        def p = ~/\b[0-9A-Z]{8}\.[0-9A-Z]{8}\b/;
        def m = p.matcher(line);
        def result = [];
        while (m.find())
        {
            result.add(m.group());
        }
        return result;
    }

    def getWebInfPlatformObj()
    {
        def query = "@Name=:'Web Infrastructure Platform'";
        def objs = this.conn.doOMRQuery('SoftwareComponent', query, true);
        if (!objs)
        {
            warn("Could not find Web Infrastructure Platform Software Component with query: ${query}");
            return null;
        }
        if (objs.size() > 1)
        {
            warn("Found ${objs.size()} Infrastructure Platform Software Components with query ${query}, expected only one");
            return null;
        }
        return objs[0];
    }

    def processLine(obj, line)
    {
        def metadataIds = getMetadataIDsFromLine(line);
        if (!metadataIds)
            return line;
        def i = line.indexOf('=');
        if (i < 0)
        {
            if (obj.getName() != 'SourceCode')
            {
                warn("Metadata ID(s) ${metadataIds} found in ${obj.getTypeName()} \"${obj.getName()}\" (${obj.getId()}) but no property name found in line: ${line}");
            }
            return line;
        }
        def propname = line.substring(0, i);
        def propval = line.substring(i + 1);
        if (propname == 'spmprefxpoll.dependson.webinfpltfm.fqid')
        {
            def target = getWebInfPlatformObj();
            if (target == null)
            {
                // getWebInfPlatformObj() already reported reason not found
                warn("Could not check or fix ID in ${obj.getTypeName()} \"${obj.getName()}\" (${obj.getId()}): Could not find object for line: ${line}");
                return line;
            }
            if (propval == target.getId())
            {
                info("${obj.getTypeName()} \"${obj.getName()}\" (${obj.getId()}): No change needed to line: ${line}");
                return line;
            }
            def newline = (propname + '=' + target.getId());
            info("${obj.getTypeName()} \"${obj.getName()}\" (${obj.getId()}): changed line \"${line}\" to \"${newline}\"");
            this.numChangedLines += 1;
            return newline;
        }
        if (!(obj.getName() in ['SourceCode', 'WRS_Hints']))
        {
            warn("Metadata ID(s) ${metadataIds} found in ${obj.getTypeName()} \"${obj.getName()}\" (${obj.getId()}) but property name ${propname} is unknown: ${line}");
        }
        return line;
    }

    def warn(msg)
    {
        this.numWarnings += 1;
        println("WARN: ${msg}");
    }

    def info(msg)
    {
        println(msg);
    }

    def run()
    {
        this.conn.setFoundationRepository();
        def objs = this.conn.doOMRQuery('TextStore', "", true);
        objs.each
        { obj ->
            def storedtext = obj.getAttribute('StoredText');
            if (!storedtext)
                return;
            def lines = storedtext.split('\\n', -1);
            def result = [];
            lines.each
            { line ->
                result.add(processLine(obj, line));
            }
            def storedtextAfter = result.join('\n');
            if ((storedtextAfter != storedtext))
            {
                this.numChangedObjs += 1;
                if (this.commit)
                {
                    obj.setAttribute('StoredText', storedtextAfter);
                    conn.store(obj);
                }
            }
        }
        info("${this.numChangedLines} changed line(s)");
        info("${this.numChangedObjs} changed object(s)");
        if (!this.commit)
        {
            info("No objects were actually changed in metadata because the commit option was not given.");
        }
    }
}

def command = 'help';
def commit = false;
if (args.length > 0)
    command = args[0].toLowerCase();;
if (command in ['help', '/?', '-h', '--help'])
{
    println("Script to find and fix metadata IDs in TextStore objects");
    println("Usage: brm groovy fixidsinmetadata.groovy (nochange|commit)")
    println();
    println("nochange: Find and print the IDs, but don't change them");
    println("commit: Find and fix the IDs");
    System.exit(0);
}
else if (command == 'commit')
{
    println("Finding and fixing metadata IDs in TextStore objects");
    commit = true;
}
else if (command == 'nochange')
{
    println("Finding and printing metadata IDs in TextStore objects");
    commit = false;
}
else
{
    println("Unknown command: ${command}. Use the \"help\" command.");
    System.exit(1);
}

def installDir = System.getProperty('BRM.install.dir');
if (installDir == null)
    installDir = '.';
def brmConfigDir = new File(installDir, 'config');
def propFile = new File(brmConfigDir, 'brm.properties');
def props = new Properties();
println("Reading BRM config file: ${propFile}");
props.load(new FileInputStream(propFile));

def host = props.get('BRM.host.metadata');
def port = props.get('BRM.port.metadata');
def user = props.get('BRM.admin.user');
def password = props.get('BRM.admin.password');
println("Connecting to metadata server at ${host}:${port} as ${user}");

def conn = new OMRConnection(host, port, user, password);
conn.connect();

def fixer = new MetadataIDFixer(conn, commit);
fixer.run();

conn.close();
System.exit(0);

// end-of-file
