// graphobjs.groovy
// Produce graph of metadata object relationships

import com.sas.solutions.brm.omr.OMRConnection;

doc = """
Make a DOT undirected graph of metadata object relationships.
See http://www.graphviz.org/ for an explanation of DOT graph files.

Usage:

brm groovy scripts/graphobjs.groovy startobjid maxlevels filename

startobjid: starting object ID
maxlevels: number of associations to traverse from starting object
filename: Name of file to which DOT format graph is written

Example graphviz command to produce SVG file from DOT file:
"C:\\Program Files\\Graphviz 2.28\\bin\\sfdp.exe" -Tsvg mdata.dot -o mdata.svg
SVG files can be viewed using the Firefox web browser.
""";

def graphObjects(OMRConnection conn, String startobjid, int maxlevels, 
                 File outFile)
{
    def visited_objids = new HashSet();
    def visited_edges = new HashSet();
    def out = outFile.newPrintWriter('UTF-8');
    out.println('graph MetadataObjectGraph {')
    def objidref = { objid ->
        return objid.replaceAll(/\./, '_');
    }
    def edgekey = { objid1, objid2 ->
        return objid1 + ':' + objid2;
    }
    def stack = new LinkedList();
    stack.addLast(['objid': startobjid, 'level': 0]);
    while (stack.size() > 0)
    {
        def record = stack.removeLast();
        def objid = record.objid;
        def level = record.level;
        def obj = conn.getObject('Root', objid, false);
        if (obj == null)
            continue;
        if (!(objid in visited_objids))
        {
            visited_objids.add(objid);
            def style = '';
            if (level == 0)
                style=', style=filled, fillcolor=aliceblue';
            out.println(
                "    ${objidref(objid)} [overlap=false,label=\"${obj.getTypeName()}\",tooltip=\"${objid} ${obj.getName()}\"${style}];");
        }
        if (level >= maxlevels)
            continue;
        obj.getAssociationNames().each
        { assocname ->
            def assocobjs = obj.getAssociation(assocname);
            assocobjs.each
            { assocobj ->
                def assocobjid = assocobj.getId();
                if (edgekey(objid, assocobjid) in visited_edges ||
                        edgekey(assocobjid, objid) in visited_edges)
                    return;
                visited_edges.add(edgekey(objid, assocobjid));
                out.println("    ${objidref(objid)} -- ${objidref(assocobjid)};");
                stack.addLast(['objid': assocobjid, 'level': level+1]);
            }
        }
    }
    out.println('}')
    out.close();
}

if (args.length < 3 || args[0].toLowerCase() in ['help', '-h', '/?'])
{
    println(doc);
    System.exit(1);
}

def startobjid = args[0];
def maxlevels = args[1];
def filename = args[2];

try
{
    maxlevels = Integer.parseInt(maxlevels);
}
catch (NumberFormatException e1)
{
    println("Invalid number of levels: ${maxlevels}");
    println("Use \"help\" for more info.");
    System.exit(1);
}

def installDir = System.getProperty('BRM.install.dir');
def configDir = new File(installDir, 'config');
def propFile = new File(configDir, 'brm.properties');
println("Reading BRM config file: ${propFile}");
def props = new Properties();
props.load(new FileInputStream(propFile));

def host = props.get('BRM.host.metadata');
def port = props.get('BRM.port.metadata');
def user = props.get('BRM.admin.user');
def password = props.get('BRM.admin.password');
println("Connecting to metadata server at ${host}:${port} as ${user}");

def conn = new OMRConnection(host, port, user, password);
conn.connect();
conn.setFoundationRepository();

try
{
    graphObjects(conn, startobjid, maxlevels, new File(filename));
}
finally
{
    conn.close();
}

// end-of-file
