// indicators.groovy

import com.sas.solutions.brm.Constants;
import com.sas.solutions.brm.omr.OMRConnection;
import com.sas.solutions.brm.util.BRMUtil;

def printHelp()
{
    println("Find BI Dashboard indicators in metadata");
    println("Run this script on the metadata server.");
    println("Use:");
    println("brm groovy scripts/indicators.groovy");
}


def printObj(obj, indent)
{
    def attrnames = obj.getAttributeNames();
    attrnames.each
    { attrname ->
        println("${indent}${attrname}=${obj.getAttribute(attrname)}");
    }
    def assocnames = obj.getAssociationNames();
    assocnames.each
    { assocname ->
        println("${indent}Association: ${assocname}:");
        assoclist = obj.getAssociation(assocname);
        assoclist.each
        { assocobj ->
            println(
                "${indent}  ${assocobj.getTypeName()}:${assocobj.getName()}");
        }
    }
}

def printTree(omr, tree, indent)
{
    println("${indent}Tree: ${tree.getName()}");
    indent += '    ';
    println("${indent}Attributes:");
    def attrnames = ['Id', 'PublicType', 'TreeType'];
    attrnames.each
    { attrname ->
        println("${indent}  ${attrname}=${tree.getAttribute(attrname)}");
    }
    println("${indent}Members:");
    def members = tree.getAssociation('Members');
    members.each
    { memberobj ->
        println("${indent}  ${memberobj.getTypeName()}:${memberobj.getName()}");
    }
    def parent = tree.getAssociatedObject('ParentTree');
    if (parent)
    {
        parent = omr.getObject(parent);
        println("${indent}Parent:");
        printTree(omr, parent, indent + '  ');
    }
}

def getTreePath(omr, tree)
{
    def result = [];
    while (tree != null)
    {
        result.add(tree.getName());
        // Fetch associations
        tree = omr.getObject(tree);
        tree = tree.getAssociatedObject('ParentTree');
    }
    def sb = new StringBuffer();
    for (def i=result.size()-1; i>=0; --i)
    {
        sb.append('/');
        sb.append(result[i]);
    }
    return sb.toString();
}

// Return 0 if Ok, 1 otherwise
// return value goes to System.exit()
def main()
{
    if (args.length > 0)
    {
        printHelp();
        return 1;
    }

    def controlProps = new Properties();
    BRMUtil.loadPropFileAsResource(Constants.BRM_CONFIG_FILE, controlProps, 
                                   null);

    def getPropOrDie = { propname ->
        def propval = controlProps.getProperty(propname);
        if (propval == null)
        {
            throw new Exception("Missing property: ${propname}");
        }
        return propval;
    }

    def host = getPropOrDie('BRM.host.metadata');
    def port = getPropOrDie('BRM.port.metadata');
    def user = getPropOrDie('BRM.admin.user');
    def password = getPropOrDie('BRM.admin.password');

    println("Connecting to metadata server ${host}:${port} as user ${user}");
    def omr = new OMRConnection(host, port, user, password);
    omr.connect();
    omr.setFoundationRepository();

    def objs = omr.doOMRQuery('SoftwareComponent', 
                                    "@ClassIdentifier='bidashmid'");
    if (objs.size() == 0)
    {
        println("SAS BI Dashboard is not installed.");
        return 0;
    }
    def softwareComponent = objs.get(0);
    // Print version of BI Dashboard that is installed
    println(softwareComponent.getName());

    def appConfig = softwareComponent.getAssociatedObject('PropertySets', 
                                                'Application Configuration');
    if (appConfig == null)
    {
        println("SAS BI Dashboard is missing Application Configuration object");
        return 0;
    }
    // Get the full set of attributes and associations
    appConfig = omr.getObject(appConfig);
    def dashboardConfig = appConfig.getAssociatedObject('SetProperties',
                                                'bid.DashboardConfig');
    def modelConfigsDir = null;
    if (dashboardConfig != null)
    {
        dashboardConfig = omr.getObject(dashboardConfig);
        def configDir = dashboardConfig.getAttribute('DefaultValue');
        modelConfigsDir = new File(configDir, 'modelConfigs');
        //println("Configuration directory: " + configDir);
    }

    def typenames = ['File', 'Transformation'];
    def query = "@PublicType='IndicatorData'";
    typenames.each
    { typename ->
        objs = omr.doOMRQuery(typename, query, true);
        objs.each
        { obj ->
            if (obj.getTypeName() == 'File' && modelConfigsDir != null)
            {
                def filePath = new File(modelConfigsDir, obj.getName()+'.imx');
                println(filePath.getPath());
                return;
            }
            def trees = obj.getAssociation('Trees');
            if (!trees)
            {
                println(obj.getName());
                return;
            }
            trees.each
            { tree ->
                def path = getTreePath(omr, tree);
                if (!path)
                {
                    println(obj.getName());
                    return;
                }
                println("${path}/${obj.getName()}");
            }
        }
    }

    omr.close();
    return 0;
}

System.exit(main());

// end-of-file
