/*
 * Copyright (c) 2008 by SAS Institute Inc., Cary, NC 27513
 */

import java.io.File;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import com.sas.solutions.brm.logging.ConsoleLogger;
import com.sas.solutions.brm.tar.Tar;
import com.sas.solutions.brm.tar.Untar;
import com.sas.solutions.brm.tar.TarCompressionCode;

/**
  * The main class for running the BRM Ant Tar code from a script.
  */
public class ScriptMain
{
    public static final String help = (
"Create or extract \"tar\" format archive file compressed with gzip.\n" +
"Uses the BRM-modified Ant Java tar library, so is compatible with BRM.\n" +
"\n" +
"Use examples:\n" +
"\n" +
"brm groovy scripts/tar.groovy cfz sample.tar.gz -C C:/temp --includes \"sasdav/**\"\n" +
"    -> Creates gzipped tar file sample.tar.gz, containing the sasdav\n" +
"    subdirectory of the C:/temp directory, and all its contents.\n" +
"    The --includes parameter works like the \"includes\" option in Ant.\n" +
"    There is a similar --excludes option.\n" +
"    If you omit the --includes option, all the files and sub-directories \n"+
"    of C:/temp will be included.\n" +
"\n" +
"brm groovy scripts/tar.groovy xfz sample.tar.gz -C C:/temp\n" +
"    -> Extracts contents of gzipped tar file to C:/temp directory\n" +
"\n" +
"brm groovy scripts/tar.groovy tfz sample.tar.gz\n" +
"    -> Lists contents of gzipped tar file.\n" +
"       File names with non-ASCII characters will be printed with escapes\n" +
"       in the format \\uNNNN, where NNNN is the hex unicode number.\n" +
"\n" +
"This utility does not implement all of the Unix tar command options.\n" +
"Only the options demonstrated above may be relied upon.\n");

    public static void createTarGz(Map params) throws Exception
    {
        Tar task = new Tar();
        Logger logger = new ConsoleLogger();
        task.setLogger(logger);
        task.setDestFile(new File((String)params.get("destfile")));
        task.setBasedir(new File((String)params.get("basedir")));
        task.setIncludes((String)params.get("includes"));
        task.setExcludes((String)params.get("excludes"));
        if (params.containsKey("compression"))
            task.setCompression((String)params.get("compression"));
        task.setVerbose(((Boolean)params.get("verbose")).booleanValue());
        task.run();
    }

    public static void extractTarGz(Map params) throws Exception
    {
        Untar task = new Untar();
        Logger logger = new ConsoleLogger();
        task.setLogger(logger);
        task.setSrc(new File((String)params.get("tarfile")));
        task.setDest(new File((String)params.get("basedir")));
        if (params.containsKey("compression"))
            task.setCompression((String)params.get("compression"));
        task.setVerbose(((Boolean)params.get("verbose")).booleanValue());
        task.setOverwrite(true);
        if (params.containsKey("errorfile"))
            task.setErrorfile(new File((String)params.get("errorfile")));
        if (params.containsKey("filelist"))
            task.setFilelist(new File((String)params.get("filelist")));
        task.run();
    }

    public static void listTarGz(Map params) throws Exception
    {
        Untar task = new Untar();
        Logger logger = new ConsoleLogger();
        task.setLogger(logger);
        task.setSrc(new File((String)params.get("tarfile")));
        task.setDest(new File((String)params.get("basedir")));
        if (params.containsKey("compression"))
            task.setCompression((String)params.get("compression"));
        task.setVerbose(((Boolean)params.get("verbose")).booleanValue());
        task.setListmode(true);
        task.run();
    }

    public static void aliasParam(Map params, String oldParamName, 
                                  String newParamName)
    {
        if (params.containsKey(oldParamName))
            params.put(newParamName, (String)params.get(oldParamName));
    }

    public static void main(String[] args) throws Exception
    {
        if (args.length >= 1 && (
            args[0].equalsIgnoreCase("-h") ||
            args[0].equalsIgnoreCase("--help") ||
            args[0].equalsIgnoreCase("/?") ||
            args[0].equalsIgnoreCase("/h")))
        {
            System.out.println(help);
            System.exit(0);
        }
        if (args.length < 2)
        {
            System.err.println("Missing operation. Use -h for help.");
            System.exit(1);
        }

        String command = args[0];
        String tarfile = args[1];
        boolean timeit = false;

        Map paramMap = new HashMap();
        paramMap.put("command", command);
        paramMap.put("tarfile", tarfile);
        paramMap.put("destfile", tarfile);
        paramMap.put("basedir", ".");
        if (command.indexOf('v') >= 0)
            paramMap.put("verbose", new Boolean(true));
        else
            paramMap.put("verbose", new Boolean(false));
        if (command.indexOf('z') >= 0)
            paramMap.put("compression", TarCompressionCode.GZIP);
        else if (command.indexOf('j') >= 0)
            paramMap.put("compression", TarCompressionCode.BZIP2);
        else
            paramMap.put("compression", null); // guess from filename

        String[] optionsWithParams = ["-C", "--includes", "--excludes",
                                      "--errorfile", "--filelist"];
        for (int i=2; i<args.length; ++i)
        {
            if (args[i].equals("--timeit"))
            {
                timeit = true;
                continue;
            }
            boolean optionFound = false;
            for (int j=0; j<optionsWithParams.length; ++j)
            {
                if (optionsWithParams[j].equals(args[i]))
                {
                    ++i;
                    if (i >= args.length)
                    {
                        System.err.println("Missing parameter to option " +
                                optionsWithParams[j]);
                        System.exit(1);
                    }
                    paramMap.put(optionsWithParams[j], args[i]);
                    optionFound = true;
                    break;
                }
            }
            if (!optionFound)
            {
                System.err.println("Unknown option: " + args[i]);
                System.exit(1);
            }
        }
        aliasParam(paramMap, "-C", "basedir");
        aliasParam(paramMap, "--includes", "includes");
        aliasParam(paramMap, "--excludes", "excludes");
        // errorfile and filelist are for extracting only.
        // errorfile: Name of file in which to write the names of files that
        //            could not be extracted, due to errors, one per line.
        // filelist: Name of file containing list of files to be extracted,
        //            one filename per line, CASE SENSITIVE (must match case
        //            as stored in archive).
        aliasParam(paramMap, "--errorfile", "errorfile");
        aliasParam(paramMap, "--filelist", "filelist");

        try
        {
            long t1 = System.currentTimeMillis();
            if (command.startsWith("c"))
            {
                createTarGz(paramMap);
            }
            else if (command.startsWith("x"))
            {
                extractTarGz(paramMap);
            }
            else if (command.startsWith("t"))
            {
                listTarGz(paramMap);
            }
            else
            {
               System.err.println("Unknown operation:" + command);
               System.exit(1);
            }
            long t2 = System.currentTimeMillis();
            if (timeit)
            {
                double dt = ((t2 - t1) * 1.0) / 1000.0;
                NumberFormat nf = NumberFormat.getInstance(Locale.US);
                DecimalFormat df = (DecimalFormat)nf;
                df.applyPattern("0.000");
                System.out.println("Elapsed time: " + df.format(dt) + 
                        " seconds");
            }
        }
        catch (Throwable e)
        {
            e.printStackTrace();
        }
    }
}

// end-of-file
