// setup.groovy
// Set up the BRM configuration files

isWindows = false;
if (System.getProperty('os.name').toLowerCase().startsWith('windows'))
{
    isWindows = true;
}

def groovy = './groovy.sh';
def example_SASHOME = '/usr/local/SAS/SASHome';
if (isWindows)
{
    groovy = 'groovy';
    example_SASHOME = '"C:\\Program Files\\SAS"';
}


if (args.length < 1)
{
    println("BRM configuration setup tool");
    println("Usage:");
    println("");
    println("${groovy} setup.groovy SASHOME");
    println("");
    println("SASHOME is the SAS installation directory, something like:");
    println(example_SASHOME);
    System.exit(1);
}

SASHOME = new File(args[0]);

if (!SASHOME.isDirectory())
{
    println("ERROR: Missing SASHOME directory: ${SASHOME}");
    System.exit(1);
}
println("SASHOME=${SASHOME}");

def sassw_config = new File(SASHOME, 'sassw.config');
if (!sassw_config.isFile())
{
    println("ERROR: Invalid SASHOME, missing file: ${sassw_config}");
    System.exit(1);
}

// http://jira.codehaus.org/browse/GROOVY-1642
def thisURL = getClass().protectionDomain.codeSource.location.path;
if (!thisURL.toString().startsWith('file://'))
    thisURL = 'file://' + thisURL;
// http://weblogs.java.net/blog/kohsuke/archive/2007/04/how_to_convert.html
def thisURI = thisURL.toURI();
def thisFile = null;
try
{
    thisFile = new File(thisURI);
}
catch (URISyntaxException e)
{
    thisFile = new File(url.getPath());
}
def thisDir = thisFile.parent;

writeAuditRecordToSetupHistory(thisDir);

def brminitemplate = new File(thisDir, 'brm.ini.template');
def brmini = new File(thisDir, 'brm.ini');
if (!brminitemplate.isFile())
{
    println("ERROR: Missing brm.ini template file: ${brminitemplate}");
    println("       BRM installation is not correct or complete.");
    System.exit(1);
}
else if (brmini.exists())
{
    println("File already exists, not overwriting: ${brmini}");
}
else
{
    createBRMIniFromTemplate(brminitemplate, brmini);
}

def configDir = new File(thisDir, 'config');
copyFile(new File(configDir, 'brm-log4j.properties.template'),
            new File(configDir, 'brm-log4j.properties'));
copyFile(new File(configDir, 'brmagent.properties.template'),
            new File(configDir, 'brmagent.properties'));

// Normal exit
System.exit(0);

def writeAuditRecordToSetupHistory(thisDir)
{
    def version = readFileFirstLine(new File(thisDir, 'version.txt'));
    if (version == null)
        return;
    def datetime = readFileFirstLine(new File(thisDir, 'builddatetime.txt'));
    if (datetime == null)
        return;
    def buildtype = readFileFirstLine(new File(thisDir, 'buildtype.txt'));
    if (buildtype == null)
        return;
    def manifestDir = new File(thisDir, 'manifest');
    if (!manifestDir.isDirectory())
        manifestDir.mkdir();
    def setupHistoryFile = new File(manifestDir, 'setup-history.txt');
    def date = new Date();
    println("Writing audit record to ${setupHistoryFile}");
    appendLineToFile(setupHistoryFile,
        "${date} BRM-${version}-${datetime}-${buildtype} ${SASHOME}");
}

// Take a list of strings
// Keep the ones that look like version numbers in "X.Y" or "X.Y.Z" format.
// Sort them, then return the highest version string found.
// If no version string found in names, return null.
def filterSortVersionStrings(names)
{
    def numbers = [];

    names.each { name ->
        if (name.endsWith('.'))
            return;
        def name_parts = name.split('\\.');
        def valid = true;
        def number_parts = [];
        for (int i=0; i<name_parts.size(); ++i)
        {
            try
            {
                def number = Integer.parseInt(name_parts[i]);
                number_parts.add(number)
            }
            catch (NumberFormatException e)
            {
                valid = false;
                break;
            }
        }
        if (valid)
            numbers.add(number_parts);
    }

    numbers = numbers.sort { a, b ->
        def i = 0;
        for ( ; ; ++i)
        {
            if (i == a.size() || i == b.size())
            {
                if (i == a.size() && i == b.size())
                    return 0;
                if (i == a.size())
                    return -1;
                return 1
            }
            if (a[i] < b[i])
                return -1;
            if (a[i] > b[i])
                return 1;
        }
        return 0;
    }

    if (numbers.size() == 0)
        return null;
    def biggest_number = numbers[numbers.size() - 1];
    def number_str = biggest_number.join('.');
    return number_str;
}

def getHighestVersionSubdir(dir)
{
    def names = []
    dir.eachFile { file ->
        if (!file.isDirectory())
            return;
        names.add(file.getName());
    }
    return filterSortVersionStrings(names);
}

def createBRMIniFromTemplate(brminitemplate, brmini)
{
    println("Creating \"${brmini}\" from \"${brminitemplate}\"");
    def infile = new BufferedReader(
                    new InputStreamReader(
                        new FileInputStream(brminitemplate)));
    def outfile = new PrintWriter(new FileWriter(brmini));
    def section = null;
    def platformobjdir = new File(SASHOME, 'SASPlatformObjectFramework');
    if (!platformobjdir.isDirectory())
    {
        println("ERROR: Missing SAS Platform Object Framework directory: ${platformobjdir}");
        System.exit(1);
    }
    def sasversion = getHighestVersionSubdir(platformobjdir);
    if (sasversion == null)
    {
        println("ERROR: Missing version sub-directory in SAS Platform Object Framework directory: ${platformobjdir}");
        System.exit(1);
    }
    println("Detected SAS platform version ${sasversion}");
    def filterDefaultSectionLine = { line ->
        def i = line.indexOf('=');
        if (i < 0)
            return line;
        def varname = line.substring(0, i).trim();
        def value = line.substring(i + 1).trim();
        def extConfigPrefix = '-Dsas.ext.config=';
        if (value.startsWith(extConfigPrefix))
        {
            // On SAS 9.2, the sas.java.ext.config file comes from
            // the PLATFORMOBJ directory. On SAS 9.3, it comes from
            // the SASHOME directory.
            def sasExtConfig;
            if (sasversion == '9.2')
                sasExtConfig = new File('<PLATFORMOBJ>', 'sas.java.ext.config');
            else
                sasExtConfig = new File('<SASHOME>', 'sas.java.ext.config');
            def sb = new StringBuffer();
            sb.append(varname);
            sb.append('=');
            sb.append(extConfigPrefix);
            if (isWindows)
                sb.append('"');
            sb.append(sasExtConfig.getPath());
            if (isWindows)
                sb.append('"');
            return sb.toString();
        }
        return line;
    }
    while (true)
    {
        def line = infile.readLine();
        if (line == null)
            break
        def tline = line.trim();
        if (tline.length() == 0)
        {
            outfile.println(line);
            continue;
        }
        if (tline.startsWith(';'))
        {
            outfile.println(line);
            continue;
        }
        if (tline.startsWith('[') && tline.endsWith(']'))
        {
            section = tline.substring(1, tline.length()-1).trim();
            outfile.println(line);
            continue;
        }
        if ('properties'.equalsIgnoreCase(section))
        {
            line = line.replace('<SASHOME>', SASHOME.getAbsolutePath());
            line = line.replace('<SASVERSION>', sasversion);
        }
        if ('default'.equalsIgnoreCase(section))
        {
            line = filterDefaultSectionLine(line);
        }
        if (!isWindows)
        {
            line = line.replace((char)'\\', (char)'/');
            if (line.indexOf(';') >= 0)
            {
                def parts = line.split(';', -1);
                line = parts.join(':');
            }
            // Remove quote marks around things, makes Unix choke
            line = line.replace('"', '');
        }
        outfile.println(line);
    }
    outfile.close();
    infile.close();
}

// Read a file and return the first line as a String
// with leading and trailing whitespace removed
def readFileFirstLine(file)
{
    if (!file.isFile())
    {
        println("ERROR: Installation problem: Missing file: ${file}");
        return null;
    }
    def infile = new BufferedReader(
                    new InputStreamReader(
                        new FileInputStream(file)));
    try
    {
        def line = infile.readLine();
        if (line == null)
            return "";
        return line.trim();
    }
    finally
    {
        infile.close();
    }
}

// Append a string to the end of a file
// The line parameter is not expected to include the newline character;
// a newline will automatically be added.
def appendLineToFile(file, line)
{
    def outfile = new PrintWriter(new FileWriter(file, true));
    outfile.println(line);
    outfile.close();
}

def copyFile(File src, File dst)
{
    if (dst.exists())
    {
        println("File already exists, not overwriting: ${dst}");
        return false;
    }
    println("Copying \"${src}\" to \"${dst}\"");
    def instream = new FileInputStream(src);
    def outstream = new FileOutputStream(dst);
    int bufsize = 1024 * 8;
    byte[] buffer = new byte[bufsize];
    while (true)
    {
        int n = instream.read(buffer);
        if (n <= 0)
            break;
        outstream.write(buffer, 0, n);
    }
    outstream.close();
    instream.close();
    return true;
}

// end-of-file
