/*
 * Created on Nov. 11, 2010
 * Copyright (c) 2010 by SAS Institute Inc., Cary, NC 27513
 */
package com.sas.solutions.brm.migrate;

import com.sas.solutions.brm.omr.GenericMetadataObject;
import com.sas.solutions.brm.omr.OMRConnection;

class HostFixer extends HostFinder
{
    def logger = null;
    def progress = null;
    def machineMap = [:]; // [ lower-case-machine-name: Machine object ]
    def domainMap = [:]; // [lower-case-domain-name: new-domain-name ]
    def loadedTextStoreSet = new HashSet(); // Object IDs

    // Flag, true if the current StoredText object has been modified.  This is
    // needed because parts of stored text (Nodes in XML DOM object) could have
    // been changed during different method invocations.
    def storedTextModified = false;

    public HostFixer(OMRConnection omr, MigrateConfig components,
            Properties brmprops, 
            MigrateConfig migrateconfig)
    {
        super(omr, components, brmprops);
        this.logger = MainMigrator.LOG;
        this.progress = MainMigrator.PROGRESS;
        def domain_names = migrateconfig.getSection('domain_names');
        domain_names.each
        { key, value ->
            this.domainMap[key.toLowerCase()] = value;
        }
        def installDir = System.getProperty('BRM.install.dir');
        def tempDir = new File(installDir, 'temp');
        def componentsDir = new File(tempDir, 'components-with-textstores');
        def foundobjidsFile = new File(componentsDir, 'foundobjids.txt');
        if (foundobjidsFile.isFile())
        {
            this.logger.info("Reading list of component object IDs whose properties  have been successfully loaded from: " + foundobjidsFile);
            def br = new BufferedReader(new FileReader(foundobjidsFile))
            while (true)
            {
                def line = br.readLine();
                if (line == null)
                    break;
                line = line.trim();
                if (line.length() == 0)
                    continue;
                this.loadedTextStoreSet.add(line);
            }
        }
        else
        {
            this.logger.debug("Missing file: ${foundobjidsFile}");
        }
    }

    def getMachineByName(name)
    {
        if (name != null)
            name = name.toLowerCase();
        return this.machineMap.get(name);
    }

    def createMachine(name)
    {
        def machine = new GenericMetadataObject();
        machine.setTypeName('Machine');
        machine.setName(name);
        if (!this.omr.isReadOnly())
        {
            machine = this.omr.create(machine);
        }
        if (name != null)
            name = name.toLowerCase();
        this.machineMap[name] = machine;
        return machine;
    }

    def visitMachines()
    {
        this.logger.info("Scanning Machine objects");
        // Pre-scan machine objects and load machineMap
        def typename = 'Machine';
        def searchParam = '';
        def matchExactType = true;
        def oldMachines = [];
        omr.setFoundationRepository();
        omr.doOMRQuery(typename, searchParam, matchExactType).each
        { machine ->
            def name = machine.getName();
            if (name != null)
                name = name.toLowerCase();
            this.machineMap[name] = machine;
            oldMachines.add(machine);
        }

        super.visitMachines();
    }

    def handleMachine(machine)
    {
        // S0763630 Don't change names or associations of Machine objects.
        // This is now handled by the enhanced saveconfig/loadconfig.
        def name = machine.getName();
        def nameStr = quoteNonNullStr(name);
        this.logger.debug("${machine.getTypeName()}: (${machine.getId()}) ${nameStr}: this object fixed by loadconfig");
    }

    def handleDocument(document, objs)
    {
        def uri = document.getAttribute('URI');
        def u = null;
        try
        {
            u = new java.net.URI(uri);
        }
        catch (java.net.URISyntaxException e)
        {
            // Can't replace the host part if I can't parse it
            this.logger.warn("Document: (${document.getId()}) \"${document.getName()}\": Can't parse URI \"${uri}\": ${e}");
            return;
        }
        super.handleDocument(document, objs);
        def scheme = u.getScheme();
        def userInfo = u.getUserInfo();
        def host = u.getHost();
        def port = u.getPort();
        def path = u.getPath();
        def query = u.getQuery();
        def fragment = u.getFragment();
        def newhost = null;
        def newhostStr = 'null';
        if (objs.size() > 0)
        {
            def obj = objs[0];
            obj = getObject(obj.getTypeName(), obj.getId());
            def classId = obj.getAttribute('ClassIdentifier');
            def classIdStr = 'null';
            if (classId != null)
                classIdStr = "\"${classId}\"";
            newhost = getHostForClassId(classId);
            if (newhost != null)
                newhostStr = "\"${newhost}\"";
            this.logger.debug("Document: (${document.getId()}) \"${document.getName()}\": Associated ${obj.getTypeName()}: (${obj.getId()}) \"${obj.getName()}\" ClassIdentifier=${classIdStr}: maps to host ${newhostStr}");
        }
        else
        {
            newhost = getHostForPseudoComponent(document);
            if (newhost != null)
                newhostStr = "\"${newhost}\"";
            this.logger.debug("Document: (${document.getId()}) \"${document.getName()}\": document name maps to host ${newhostStr}");
        }
        if (equalHosts(host, newhost))
        {
            this.logger.debug("Document: (${document.getId()}) \"${document.getName()}\" URI=\"${uri}\": no change, host name stays as \"${host}\"");
        }
        else if (newhost != null)
        {
            u = new java.net.URI(scheme, userInfo, newhost, port, path, query, 
                                    fragment);
            def newuri = "${u}";
            this.logger.info("Document: (${document.getId()}) \"${document.getName()}\" URI=\"${uri}\": host name changes from \"${host}\" to ${newhostStr}, URI changes to \"${newuri}\"");
            document.setAttribute('URI', newuri);
            if (!this.omr.isReadOnly())
            {
                this.omr.store(document);
            }
        }
    }

    def handleLogin(login)
    {
        super.handleLogin(login);
        def name = login.getName();
        def userID = login.getAttribute('UserID');
        def nameStr = quoteNonNullStr(name);
        def userIDStr = quoteNonNullStr(userID);
        def newname = fixDomainPrefix(name);
        def newnameStr = quoteNonNullStr(newname);
        def newuserID = fixDomainPrefix(userID);
        def newuserIDStr = quoteNonNullStr(newuserID);
        if (name == newname && userID == newuserID)
        {
            this.logger.debug("Login: (${login.getId()}) Name=${nameStr} UserID=${userIDStr}: no change");
        }
        else
        {
            this.logger.info("Login (${login.getId()}) Name=${nameStr} UserID=${userIDStr}: changed: Name=${newnameStr} UserID=${newuserIDStr}");
            login.setAttribute('Name', newname);
            login.setAttribute('UserID', newuserID);
            if (!this.omr.isReadOnly())
            {
                this.omr.store(login);
            }
        }
    }

    def fixDomainPrefix(name)
    {
        if (name == null)
            return null;
        def i = name.indexOf('\\');
        if (i < 0)
            return name;
        def domain = name.substring(0, i);
        def newdomain = this.domainMap.get(domain.toLowerCase());
        if (newdomain == null)
            return name;
        if (newdomain == '')
            return name.substring(i + 1);
        return newdomain + name.substring(i);
    }

    def handleProperty(property)
    {
        def name = property.getName();
        if (name == null)
            return;
        def defaultValue = property.getAttribute('DefaultValue');
        if (defaultValue == null)
            return;
        super.handleProperty(property);
        this.logger.debug("Property: (${property.getId()}) \"${name}\" DefaultValue=\"${defaultValue}\": this property fixed by loadconfig");
        return;
    }

    def handleSASClientConnection(connection)
    {
        super.handleSASClientConnection(connection);
        def name = connection.getName();
        def nameStr = quoteNonNullStr(name);
        def remoteAddress = connection.getAttribute('RemoteAddress');
        def remoteAddressStr = '';
        if (remoteAddress != null)
            remoteAddressStr = " RemoteAddress=\"${remoteAddress}\"";
        def serverShortName = connection.getAttribute('ServerShortName');
        def serverShortNameStr = '';
        if (serverShortName != null)
            serverShortNameStr = " ServerShortName=\"${serverShortName}\"";
        this.logger.debug("${connection.getTypeName()}: (${connection.getId()}) ${nameStr}${remoteAddressStr}${serverShortNameStr}: this object fixed by loadconfig");
    }

    def handleServerComponent(component)
    {
        super.handleServerComponent(component);
        def name = component.getName();
        def nameStr = quoteNonNullStr(name);
        this.logger.debug("${component.getTypeName()}: (${component.getId()}) ${nameStr}: this object fixed by loadconfig");
    }

    def handleSoftwareComponent(component)
    {
        super.handleSoftwareComponent(component);
        def name = component.getName();
        def nameStr = quoteNonNullStr(name);
        this.logger.debug("${component.getTypeName()}: (${component.getId()}) ${nameStr}: this object fixed by loadconfig");
    }

    def handleTCPIPConnection(connection)
    {
        super.handleTCPIPConnection(connection);
        def name = connection.getName();
        def nameStr = quoteNonNullStr(name);
        this.logger.debug("${connection.getTypeName()}: (${connection.getId()}) ${nameStr}: this object fixed by loadconfig");
    }

    def handleTree(tree)
    {
        super.handleTree(tree);
        def name = tree.getName();
        def nameStr = quoteNonNullStr(name);
        this.logger.debug("${tree.getTypeName()}: (${tree.getId()}) ${nameStr}: this object fixed by loadconfig");
    }

    // Process an XML text object
    def handleXML(obj, doc)
    {
        this.storedTextModified = false;
        super.handleXML(obj, doc);
        if (this.storedTextModified)
        {
            def text = domTreeToString(doc);
            obj.setAttribute('StoredText', text);
            if (!this.omr.isReadOnly())
            {
                this.omr.store(obj);
            }
        }
    }

    def handleXMLValue(obj, parent, node, value)
    {
        def parentName = '';
        def grandParentName = '';
        if (parent != null)
        {
            parentName = parent.getNodeName();
            def grandParent = parent.getParentNode();
            if (grandParent != null)
                grandParentName = grandParent.getNodeName();
        }
        def host = null;
        def newhost = null;
        if (grandParentName == 'InformationService' && parentName == 'ReposDef')
        {
            if (node.getNodeName() != 'host')
                return;
            host = value;
            def protocol = parent.getAttribute('protocol');
            newhost = getHostForProtocol(protocol);
            if (newhost == null)
            {
                this.logger.warn("${obj.getTypeName()}: (${obj.getId()}) \"${obj.getName()}\" StoredText: ${grandParentName}/${parentName}/${node} protocol=\"${protocol}\": no host found for protocol");
                return;
            }
            if (newhost.equalsIgnoreCase(host))
            {
                this.logger.debug("${obj.getTypeName()}: (${obj.getId()}) \"${obj.getName()}\" StoredText: ${grandParentName}/${parentName}/${node} protocol=\"${protocol}\": no change to host \"${newhost}\"");
                return;
            }
            this.storedTextModified = true;
            this.logger.info("${obj.getTypeName()}: (${obj.getId()}) \"${obj.getName()}\" StoredText: ${grandParentName}/${parentName}/${node} protocol=\"${protocol}\": changing host to \"${newhost}\"");
            node.setNodeValue(newhost);
            return;
        }
        if (grandParentName == 'UserService' && parentName == 'Profile')
        {
            if (node.getNodeName() != 'url')
                return;
            def u = null;
            try
            {
                u = new java.net.URI(value);
            }
            catch (java.net.URISyntaxException e)
            {
                // Can't replace the host and port if I can't parse it
                this.logger.warn("${obj.getTypeName()}: (${obj.getId()}) \"${obj.getName()}\" StoredText: ${grandParentName}/${parentName}/${node}: can't parse url: ${e}");
                return;
            }
            def scheme = u.getScheme();
            def userInfo = u.getUserInfo();
            host = u.getHost();
            def port = u.getPort();
            def path = u.getPath();
            def query = u.getQuery();
            def fragment = u.getFragment();
            newhost = getMetadataHost();
            if (newhost == null)
            {
                this.logger.warn("${obj.getTypeName()}: (${obj.getId()}) \"${obj.getName()}\" StoredText: ${grandParentName}/${parentName}/${node}: could not determine new metadata host");
                return;
            }
            def newport = getMetadataPort();
            if (newport == null)
            {
                this.logger.warn("${obj.getTypeName()}: (${obj.getId()}) \"${obj.getName()}\" StoredText: ${grandParentName}/${parentName}/${node}: could not determine new metadata port");
                return;
            }
            try
            {
                newport = Integer.parseInt(newport);
            }
            catch (NumberFormatException e)
            {
                this.logger.error("${obj.getTypeName()}: (${obj.getId()}) \"${obj.getName()}\" StoredText: ${grandParentName}/${parentName}/${node}: new metadata port \"${newport}\" is not an integer");
                return;
            }
            if (newhost.equalsIgnoreCase(host) && "${port}" == "${newport}")
            {
                this.logger.debug("${obj.getTypeName()}: (${obj.getId()}) \"${obj.getName()}\" StoredText: ${grandParentName}/${parentName}/${node}: no change to host \"${newhost}\" or to port ${newport}");
                return;
            }
            this.storedTextModified = true;
            def newurl = new java.net.URI(scheme, userInfo, newhost, 
                                            newport, path, query, fragment);
            this.logger.info("${obj.getTypeName()}: (${obj.getId()}) \"${obj.getName()}\" StoredText: ${grandParentName}/${parentName}/${node}: changing url to ${newurl}");
            node.setNodeValue("${newurl}");
            return;
        }
    }

    def handlePropertiesText(obj, text)
    {
        def resultText = super.handlePropertiesText(obj, text);
        if (resultText != text)
        {
            obj.setAttribute('StoredText', resultText);
            if (!this.omr.isReadOnly())
            {
                this.omr.store(obj);
            }
        }
    }

    def handlePropertiesTextLine(obj, line)
    {
        def i = line.indexOf('=');
        if (i < 0)
        {
            this.logger.debug("(${obj.getId()}) \"${obj.getName()}\" StoredText (Properties): not a property line: ${line}");
            return line;
        }
        def propname = line.substring(0, i);
        if (!propname.toLowerCase().endsWith('.host'))
        {
            this.logger.debug("(${obj.getId()}) \"${obj.getName()}\" StoredText (Properties): not a .host property: ${line}");
            return line;
        }
        def parentObj = obj.getAssociatedObject('Objects');
        if (parentObj == null)
        {
            this.logger.debug("(${obj.getId()}) \"${obj.getName()}\" StoredText (Properties): TextStore has no parent object: ${line}");
            return line;
        }
        if (this.loadedTextStoreSet.contains(parentObj.getId()))
        {
            this.logger.debug("(${obj.getId()}) \"${obj.getName()}\" StoredText (Properties): Parent Object (${parentObj.getId()}): line with host name fixed by loadconfig: ${line}");
            return line;
        }
        parentObj = this.omr.getObject(parentObj);
        def classId = parentObj.getAttribute('ClassIdentifier');
        if (classId == null)
        {
            if (parentObj.getName() == 'Deployment Tester Server')
            {
                this.logger.debug("(${obj.getId()}) \"${obj.getName()}\" StoredText (Properties): Parent Object (${parentObj.getId()}): Deployment Tester Server handled by saveconfig/loadconfig");
            }
            else
            {
                this.logger.warn("(${obj.getId()}) \"${obj.getName()}\" StoredText (Properties): Parent Object (${parentObj.getId()}): no ClassIdentifier on parent object");
            }
            return line;
        }
        def host = getHostForClassId(classId);
        if (host == null)
        {
            this.logger.warn("(${obj.getId()}) \"${obj.getName()}\" StoredText (Properties): Parent Object (${parentObj.getId()}): no host for ClassIdentifier: ${classId}");
            return line;
        }
        def newLine = propname + '=' + host;
        if (newLine == line)
        {
            this.logger.debug("(${obj.getId()}) \"${obj.getName()}\" StoredText (Properties): Parent Object (${parentObj.getId()}): line unchanged: ${line}");
            return line;
        }
        this.logger.info("(${obj.getId()}) \"${obj.getName()}\" StoredText (Properties): Parent Object (${parentObj.getId()}): old line: [${line}] new line: [${newLine}]");
        return newLine;
    }

    def handlePromptGroupXML(obj, doc)
    {
        this.storedTextModified = false;
        super.handlePromptGroupXML(obj, doc);
        if (this.storedTextModified)
        {
            def text = domTreeToString(doc);
            obj.setAttribute('GroupInfo', text);
            if (!this.omr.isReadOnly())
            {
                this.omr.store(obj);
            }
        }
    }

    def handlePromptGroupXMLValue(obj, parent, node, value)
    {
        value = value.trim();
        if (value == '')
            return;
        if (value.equalsIgnoreCase('Supply_Valid_Value'))
            return;
        def parentName = '';
        def grandParentName = '';
        if (parent != null)
        {
            parentName = parent.getNodeName();
            def grandParent = parent.getParentNode();
            if (grandParent != null)
                grandParentName = grandParent.getNodeName();
        }
        def u = null;
        try
        {
            u = new java.net.URI(value);
        }
        catch (java.net.URISyntaxException e)
        {
            // Can't replace the host and port if I can't parse it
            this.logger.warn("${obj.getTypeName()}: (${obj.getId()}) \"${obj.getName()}\" GroupInfo: ${grandParentName}/${parentName}/${node}: can't parse url: ${e}");
            return;
        }
        def scheme = u.getScheme();
        def userInfo = u.getUserInfo();
        def host = u.getHost();
        def port = u.getPort();
        def path = u.getPath();
        def query = u.getQuery();
        def fragment = u.getFragment();
        def newhost = getDAVHost();
        if (newhost == null)
        {
            this.logger.warn("${obj.getTypeName()}: (${obj.getId()}) \"${obj.getName()}\" GroupInfo: ${grandParentName}/${parentName}/${node}: could not determine new DAV host");
            return;
        }
        def newport = getDAVPort();
        if (newport == null)
        {
            newport = port;
        }
        else
        {
            try
            {
                newport = Integer.parseInt(newport);
            }
            catch (NumberFormatException e)
            {
                this.logger.error("${obj.getTypeName()}: (${obj.getId()}) \"${obj.getName()}\" GroupInfo: ${grandParentName}/${parentName}/${node}: new DAV port \"${newport}\" is not an integer");
                return;
            }
        }
        if (newhost.equalsIgnoreCase(host) && "${port}" == "${newport}")
        {
            this.logger.debug("${obj.getTypeName()}: (${obj.getId()}) \"${obj.getName()}\" GroupInfo: ${grandParentName}/${parentName}/${node}: no change to host \"${newhost}\" or to port ${newport}");
            return;
        }
        this.storedTextModified = true;
        def newurl = new java.net.URI(scheme, userInfo, newhost, 
                                        newport, path, query, fragment);
        this.logger.info("${obj.getTypeName()}: (${obj.getId()}) \"${obj.getName()}\" GroupInfo: ${grandParentName}/${parentName}/${node}: changing url to ${newurl}");
        node.setNodeValue("${newurl}");
    }
}
